@@ -1,4 +1,4 @@
-/*  $Id: init.pl,v 1.139 2006/08/21 15:26:25 jan Exp $
+/*  $Id: init.pl,v 1.150 2007/06/21 14:07:06 jan Exp $
 
     Part of SWI-Prolog
 
@@ -121,6 +121,11 @@ noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 	call/4,
 	call/5,
 	call/6,
+	call/7,
+	call/8,
+	call/9,
+	call/10,
+	call/11,
 	(^)/2,
 	(not)/1,
 	(\+)/1,
@@ -164,6 +169,16 @@ call(G, A, B, C, D) :-
 	call(G, A, B, C, D).
 call(G, A, B, C, D, E) :-
 	call(G, A, B, C, D, E).
+call(G, A, B, C, D, E, F) :-
+	call(G, A, B, C, D, E, F).
+call(G, A, B, C, D, E, F, G) :-
+	call(G, A, B, C, D, E, F, G).
+call(G, A, B, C, D, E, F, G, H) :-
+	call(G, A, B, C, D, E, F, G, H).
+call(G, A, B, C, D, E, F, G, H, I) :-
+	call(G, A, B, C, D, E, F, G, H, I).
+call(G, A, B, C, D, E, F, G, H, I, J) :-
+	call(G, A, B, C, D, E, F, G, H, I, J).
 
 not(Goal) :-
 	\+ Goal.
@@ -395,6 +410,9 @@ expand_file_search_path(Spec, Expanded) :-
 	'$expand_file_search_path'(Exp0, Exp1, NN, [Alias=Exp0|Used]),
 	arg(1, Spec, Base),
 	'$make_path'(Exp1, Base, Expanded).
+'$expand_file_search_path'(A/B, Expanded, _, _) :- !,
+	'$make_path_from_slash'(A/B, Parts, []),
+	concat_atom(Parts, Expanded).
 '$expand_file_search_path'(Spec, Spec, _, _) :-
 	atomic(Spec).
 
@@ -402,15 +420,28 @@ expand_file_search_path(Spec, Expanded) :-
 	atom_concat(_, /, Dir), !,
 	atom_concat(Dir, File, Path).
 '$make_path'(Dir, File, Path) :-
-	'$concat_atom'([Dir, '/', File], Path).
+	'$make_path_from_slash'(File, Parts, []),
+	'$append'([Dir, /], Parts, AllParts),
+	'$concat_atom'(AllParts, Path).
+
+
+%%	'$make_path_from_slash'(+SlashPath, -Parts, ?Tail) is det.
+%
+%	Translate a/b/c into [a,/,b,/,c]. We cannot   use DCG as the DCG
+%	compiler is not yet defined.
+
+'$make_path_from_slash'(A/B, P0, P) :- !,
+	'$make_path_from_slash'(A, P0, [/|P1]),
+	'$make_path_from_slash'(B, P1, P).
+'$make_path_from_slash'(A, [A|T], T).
 
 
 		/********************************
 		*         FILE CHECKING         *
 		*********************************/
 
-%	absolute_file_name(+Term, +Args, -AbsoluteFile)
-%	absolute_file_name(+Term, -AbsoluteFile, +Args)
+%%	absolute_file_name(+Term, +Args, -AbsoluteFile) is nondet.
+%%	absolute_file_name(+Term, -AbsoluteFile, +Args) is nondet.
 %	
 %	Translate path-specifier into a full   path-name. This predicate
 %	originates from Quintus was introduced  in SWI-Prolog very early
@@ -431,8 +462,9 @@ absolute_file_name(Spec, Args, Path) :-
 	),
 	(   '$select'(extensions(Exts), Args, Conditions)
 	->  true
-	;   '$select'(file_type(Type), Args, Conditions)
-	->  '$file_type_extensions'(Type, Exts)
+	;   memberchk(file_type(Type), Args)
+	->  '$file_type_extensions'(Type, Exts),
+	    Conditions = Args
 	;   Conditions = Args,
 	    Exts = ['']
 	),
@@ -535,9 +567,11 @@ user:prolog_file_type(Ext,	executable) :-
 
 '$relative_to'(Conditions, Default, Dir) :-
 	(   '$member'(relative_to(FileOrDir), Conditions)
-	*-> (   exists_file(FileOrDir)
-	    ->  file_directory_name(FileOrDir, Dir)
-	    ;   Dir = FileOrDir
+	*-> (   exists_directory(FileOrDir)
+	    ->	Dir = FileOrDir
+	    ;	atom_concat(Dir, /, FileOrDir)
+	    ->	true
+	    ;	file_directory_name(FileOrDir, Dir)
 	    )
 	;   Default == cwd
 	->  working_directory(Dir, Dir)
@@ -589,7 +623,7 @@ user:prolog_file_type(Ext,	executable) :-
 '$file_condition'(file_type(directory), File) :- !,
 	exists_directory(File).
 '$file_condition'(file_type(_), File) :- !,
-	exists_file(File).
+	\+ exists_directory(File).
 '$file_condition'(access([A1|AT]), File) :- !,
 	'$file_condition'(access(A1), File),
 	'$file_condition'(access(AT), File).
@@ -860,6 +894,9 @@ load_files(Files, Options) :-
 	->  '$load_file'(Id, Module, Options)
 	;   throw(error(type_error(atom, Id), _))
 	).
+'$load_files'(X, _, _) :-
+	var(X), !,
+	throw(error(instantiation_error, _)).
 '$load_files'([], _, _) :- !.
 '$load_files'([H|T], Module, Options) :- !,
 	'$load_files'(H, Module, Options),
@@ -952,6 +989,7 @@ load_files(Files, Options) :-
 	'$get_option'(autoload(Autoload), Options, false),
 	'$get_option'(derived_from(DerivedFrom), Options, -),
 	'$get_option'(encoding(Encoding), Options, default),
+	current_prolog_flag(generate_debug_info, DebugInfo),
 
 	(   Autoload == false
 	->  flag('$autoloading', AutoLevel, AutoLevel)
@@ -1014,15 +1052,16 @@ load_files(Files, Options) :-
 	    TimeUsed is Time - OldTime,
 
 	    '$print_message'(MessageLevel,
-			   load_file(done(Level,
-					  file(File, Absolute),
-					  Action,
-					  LM,
-					  TimeUsed,
-					  HeapUsed)))
+			     load_file(done(Level,
+					    file(File, Absolute),
+					    Action,
+					    LM,
+					    TimeUsed,
+					    HeapUsed)))
 	),
 	flag('$autoloading', _, AutoLevel),
-	set_prolog_flag(verbose_load, DefVerbose).
+	set_prolog_flag(verbose_load, DefVerbose),
+	set_prolog_flag(generate_debug_info, DebugInfo).
 
 
 '$negate'(false, true).
@@ -1061,12 +1100,10 @@ load_files(Files, Options) :-
 	),
 	'$assert_load_context_module'(Id, OldModule),
 
-	current_prolog_flag(generate_debug_info, DebugInfo),
 	'$style_check'(OldStyle, OldStyle),	% Save style parameters
 	'$open_source'(Absolute, Enc, In,
 		     '$load_file'(In, Id, Import, IsModule, LM)),
 	'$style_check'(_, OldStyle),		% Restore old style
-	set_prolog_flag(generate_debug_info, DebugInfo),
 	'$set_source_module'(_, OldModule).	% Restore old module
 
 '$load_id'(stream(Id, _), Id) :- !.
@@ -1476,6 +1513,9 @@ expand_term(Term, Term).
 '$expand_clauses'([H0|T0], [H|T]) :- !,
 	'$expand_clauses'(H0, H),
 	'$expand_clauses'(T0, T).
+'$expand_clauses'('$source_location'(File, Line):Clause0,
+		  '$source_location'(File, Line):Clause) :- !,
+	'$expand_clauses'(Clause0, Clause).
 '$expand_clauses'((Head :- Body), (Head :- ExpandedBody)) :-
 	nonvar(Body), !,
 	expand_goal(Body,  ExpandedBody).
@@ -1606,9 +1646,13 @@ resulting code is simply the same), I've removed that.
             '$extend'(LP, S, SR, H)
         ).
 '$translate_rule'((LP-->RP), (H:-B)):-
-	'$t_head'(LP, S, SR, H),
-	'$t_body'(RP, S, SR, B).
-
+	'$t_head'(LP, S0, SR, H),
+	'$t_body'(RP, S0, SR, B0),
+	(   B0 = (S00=X, B),		% map a(H,T) :- H = [a,b|T], b(T)
+	    S00 == S0
+	->  S0 = X			% into a([a,b|T]) :- b(T).
+	;   B0 = B
+	).
 
 '$t_head'((LP, List), S, SR, H) :-
 	'$append'(List, SR, List2), !,
@@ -1677,10 +1721,8 @@ then the call p([a], [a]) will succeed, which is quite definitely wrong.
 	var(Var), !.
 '$t_body'([], S, SR, S=SR) :- !.		% inline lists
 '$t_body'(List, S, SR, C) :-
-	List = [X|T], !,
-	(   T == []
-	->  C = 'C'(S, X, SR)
-	;   is_list(T)
+	List = [_|_], !,
+	(   is_list(List)
 	->  '$append'(List, SR, OL),
 	    C = (S = OL)
 	;   C = '$append'(List, SR, S)	% Deals with [H|T] in body
@@ -1752,9 +1794,6 @@ then the call p([a], [a]) will succeed, which is quite definitely wrong.
 	'$copy_args'(I2, Arity, Old, New).
 '$copy_args'(_, _, _, _).
 
-
-'C'([X|S], X, S).
-
 :- module_transparent
 	phrase/2,
 	phrase/3.
