@@ -1,11 +1,11 @@
-/*  $Id: messages.pl,v 1.81 2006/08/14 11:28:11 jan Exp $
+/*  $Id: messages.pl,v 1.90 2007/06/27 19:52:39 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@scienc.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -36,8 +36,9 @@
 	  ]).
 
 :- multifile
-	prolog:message/3,
-	prolog:error_message/3.
+	prolog:message/3,		% entire message
+	prolog:error_message/3,		% 1-st argument of error term
+	prolog:message_context/3.	% Context of error messages
 :- discontiguous
 	prolog_message/3.
 
@@ -195,6 +196,9 @@ swi_message(goal_failed(Goal)) -->
 	[ 'goal (~p) failed'-[Goal] ].
 swi_message(shared_object(_Action, Message)) --> % Message = dlerror() 
 	[ '~w'-[Message] ].
+swi_message(system_error(Error)) -->
+	[ 'error in system call: ~w'-[Error]
+	].
 swi_message(system_error) -->
 	[ 'error in system call'
 	].
@@ -204,6 +208,9 @@ swi_message(timeout_error(Op, Stream)) -->
 	[ 'Timeout in ~w from ~p'-[Op, Stream] ].
 swi_message(not_implemented(Type, What)) -->
 	[ '~w `~p\' is not implemented in this version'-[Type, What] ].
+swi_message(context_error(nodirective, Goal)) -->
+	{ goal_to_predicate_indicator(Goal, PI) },
+	[ 'Wrong context: ~p can only be used in a directive'-[PI] ].
 swi_message(format_argument_type(Fmt, Arg)) -->
 	[ 'Illegal argument to format sequence ~~~w: ~p'-[Fmt, Arg] ].
 swi_message(format(Msg)) -->
@@ -239,6 +246,8 @@ swi_extra(X) -->
 	{ var(X)
 	}, !,
 	[].
+swi_extra(Context) -->
+	prolog:message_context(Context).
 swi_extra(context(_, Msg)) -->
 	{ nonvar(Msg),
 	  Msg \== ''
@@ -263,7 +272,8 @@ prolog_message(initialization_exception(E)) -->
 	[ 'Prolog initialisation failed:', nl ],
 	translate_message(E).
 prolog_message(unhandled_exception(E)) -->
-	[ 'Unhandled exception: ~p~n'-[E] ].
+	[ 'Unhandled exception: ' ],
+	translate_message(E).
 prolog_message(goal_failed(Goal, Context)) -->
 	[ 'Goal (~w) failed: ~p'-[Goal, Context] ].
 prolog_message(no_current_module(Module)) -->
@@ -281,7 +291,7 @@ prolog_message(profile_no_cpu_time) -->
 prolog_message(non_ascii(Text, Type)) -->
 	[ 'Unquoted ~w with non-portable characters: ~w'-[Type, Text] ].
 prolog_message(io_warning(Stream, Message)) -->
-	{ stream_property(Stream, Position),
+	{ stream_property(Stream, position(Position)), !,
 	  stream_position_data(line_count, Position, LineNo),
 	  stream_position_data(line_position, Position, LinePos),
 	  (   stream_property(Stream, file_name(File))
@@ -290,6 +300,8 @@ prolog_message(io_warning(Stream, Message)) -->
 	  )
 	},
 	[ '~p:~d:~d: ~w'-[Obj, LineNo, LinePos, Message] ].
+prolog_message(io_warning(Stream, Message)) -->
+	[ 'stream ~p: ~w'-[Stream, Message] ].
 
 
 		 /*******************************
@@ -314,6 +326,8 @@ prolog_message(redefined_procedure(Type, Proc)) -->
 	[ 'Redefined ~w procedure ~p'-[Type, Proc] ].
 prolog_message(declare_module(Module, abolish(Predicates))) -->
 	[ 'Loading module ~w abolished: ~p'-[Module, Predicates] ].
+prolog_message(undefined_export(Module, PI)) -->
+	[ 'Exported procedure ~q:~q is not defined'-[Module, PI] ].
 prolog_message(discontiguous((-)/2)) -->
 	prolog_message(minus_in_identifier).
 prolog_message(discontiguous(Proc)) -->
@@ -401,6 +415,17 @@ prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
 	  'gained ~D+~D in ~2f sec; used ~D+~D; free ~D+~D'-
 	  [G, T, Time, UG, UT, RG, RT]
 	].
+prolog_message(shift_stacks(start(_L,_G,_T))) -->
+	[ 'Stack-shift: ', flush ].
+prolog_message(shift_stacks(done(Time, L, G, T))) -->
+	{ LKB is L//1024,
+	  GKB is G//1024,
+	  TKB is T//1024
+	},
+	[ at_same_line,
+	  'local: ~DKB, global: ~DKB, trail: ~DKB bytes (~2f sec)'-
+	  [LKB, GKB, TKB, Time]
+	].
 prolog_message(agc(start)) -->
 	{ thread_self(Me)
 	},
@@ -449,7 +474,7 @@ prolog_message(threads) -->
 prolog_message(threads) -->
 	[].
 prolog_message(copyright) -->
-	[ 'Copyright (c) 1990-2006 University of Amsterdam.', nl,
+	[ 'Copyright (c) 1990-2007 University of Amsterdam.', nl,
 	  'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software,', nl,
 	  'and you are welcome to redistribute it under certain conditions.', nl,
 	  'Please visit http://www.swi-prolog.org for details.'
@@ -502,9 +527,13 @@ prolog_message(query(eof)) -->
 bindings([]) -->
 	[].
 bindings([Name = Value|T]) -->
-	{ current_prolog_flag(toplevel_print_options, Options)
+	{ current_prolog_flag(toplevel_print_options, Options),
+	  (   T == []
+	  ->  Cont = ''
+	  ;   Cont = (,)
+	  )
 	},
-	[ nl, '~w = ~W'-[Name, Value, Options] ],
+	[ nl, '~w = ~W~w'-[Name, Value, Options, Cont] ],
 	bindings(T).
 
 prolog_message(query(help)) -->
