@@ -1,4 +1,4 @@
-/*  $Id: prolog_clause.pl,v 1.13 2006/10/22 15:36:58 jan Exp $
+/*  $Id: prolog_clause.pl,v 1.15 2007/02/28 13:17:49 jan Exp $
 
     Part of SWI-Prolog
 
@@ -154,7 +154,7 @@ read(Handle, Module, Clause, TermPos, VarNames) :-
 	->  Syntax = system
 	;   true
 	),
-	seek(Handle, 0, current, Here),
+	stream_property(Handle, position(Here)),
 	alternate_syntax(Syntax, Module, Setup, Restore),
 	peek_char(Handle, X),
 	debug(clause_info, 'Using syntax ~w (c=~w)', [Syntax, X]),
@@ -169,30 +169,11 @@ read(Handle, Module, Clause, TermPos, VarNames) :-
 	Restore,
 	(   var(Error)
 	->  !
-	;   text_seek(Handle, Here),
+	;   set_stream_position(Handle, Here),
 	    fail
 	).
 	
 
-%%	text_seek(+Handle, +To)
-%	
-%	Re-position the stream at position `To'.  On decent systems this
-%	is easy, but on M$-Windows with their text-mode files holding an
-%	unknown number of ^M it is   very complicated. We simply restart
-%	reading the file. Not nice, but   alternatives  are really hard,
-%	especially as many files moving  between well designed operating
-%	systems and Windows do not have a consistent line delimiter.
-
-text_seek(Handle, To) :-
-	current_prolog_flag(windows, true), !,
-	seek(Handle, 0, bof, _),
-	open_null_stream(Tmp),
-	call_cleanup(copy_stream_data(Handle, Tmp, To),
-		     close(Tmp)).
-text_seek(Handle, To) :-
-	seek(Handle, To, bof, _).
-
-
 %%	make_varnames(+ReadClause, +Offsets, +Names, -Term) is det.
 %	
 %	Create a Term varnames(...) where each argument contains the name
@@ -271,7 +252,7 @@ unify_clause((Head :- Read),
 unify_clause(Read, Compiled1, TermPos0, TermPos) :-
 	Read = (_ --> List, _),
 	is_list(List),
-	expand_term(Read, Compiled2),
+	catch(expand_term(Read, Compiled2), E, expand_failed(E, Read)),
 	Compiled2 = (DH :- _),
 	functor(DH, _, Arity),
 	DArg is Arity - 1,
@@ -284,7 +265,7 @@ unify_clause(Read, Compiled1, TermPos0, TermPos) :-
 	match_module(Compiled2, Compiled1, TermPos1, TermPos).
 					% general term-expansion
 unify_clause(Read, Compiled1, TermPos0, TermPos) :-
-	expand_term(Read, Compiled2),
+	catch(expand_term(Read, Compiled2), E, expand_failed(E, Read)),
 	match_module(Compiled2, Compiled1, TermPos0, TermPos).
 					% I don't know ...
 unify_clause(_, _, _, _) :-
@@ -301,6 +282,16 @@ match_module((H1 :- B1), (H2 :- B2), Pos0, Pos) :- !,
 match_module(H1, H2, Pos, Pos) :-	% deal with facts
 	unify_clause_head(H1, H2).
 
+%%	expand_failed(+Exception, +Term)
+%
+%	When debugging, indicate that expansion of the term failed.
+
+expand_failed(E, Read) :-
+	debugging(clause_info),
+	message_to_string(E, Msg),
+	debug(clause_info, 'Term-expand ~p failed: ~w', [Read, Msg]),
+	fail.
+
 %%	unify_body(+Read, +Decompiled, +Pos0, -Pos)
 %	
 %	Deal with translations implied by the compiler.  For example,
