@@ -1,4 +1,4 @@
-/*  $Id: prolog_xref.pl,v 1.28 2006/08/09 19:18:33 jan Exp $
+/*  $Id: prolog_xref.pl,v 1.32 2007/07/09 14:51:44 jan Exp $
 
     Part of SWI-Prolog
 
@@ -43,6 +43,7 @@
 	    xref_built_in/1,		% ?Callable
 	    xref_expand/2,		% +Term, -Expanded
 	    xref_source_file/3,		% +Spec, -Path, +Source
+	    xref_source_file/4,		% +Spec, -Path, +Source, +Options
 	    xref_public_list/4,		% +Path, -Export, +Src
 	    xref_meta/2,		% +Goal, -Called
 	    xref_hook/1,		% ?Callable
@@ -56,6 +57,7 @@
 	      [pop_operators/0, push_op/3, push_operators/1]).
 :- use_module(library(shlib), [current_foreign_library/2]).
 :- use_module(library(prolog_source)).
+:- use_module(library(option)).
 
 :- dynamic
 	called/3,			% Head, Src, From
@@ -512,11 +514,17 @@ process_meta_predicate(Decl) :-
 
 meta_args(I, Arity, _, _, []) :-
 	I > Arity, !.
-meta_args(I, Arity, Decl, Head, [H|T]) :-
+meta_args(I, Arity, Decl, Head, [H|T]) :- 		% :
 	arg(I, Decl, :), !,
 	arg(I, Head, H),
 	I2 is I + 1,
 	meta_args(I2, Arity, Decl, Head, T).
+meta_args(I, Arity, Decl, Head, [H+A|T]) :-		% I --> H+I
+	arg(I, Decl, A), 
+	integer(A), A > 0, !,
+	arg(I, Head, H),
+	I2 is I + 1,
+	meta_args(I2, Arity, Decl, Head, T).
 meta_args(I, Arity, Decl, Head, Meta) :-
 	I2 is I + 1,
 	meta_args(I2, Arity, Decl, Head, Meta).
@@ -558,6 +566,7 @@ xref_meta(phrase(G, _A),	[G+2]).
 xref_meta(phrase(G, _A, _R),	[G+2]).
 xref_meta(catch(A, _, B),	[A, B]).
 xref_meta(thread_create(A,_,_), [A]).
+xref_meta(thread_signal(_,A),   [A]).
 xref_meta(thread_at_exit(A),	[A]).
 xref_meta(predsort(A,_,_),	[A+3]).
 xref_meta(call_cleanup(A, B),	[A, B]).
@@ -576,6 +585,7 @@ xref_meta(at_halt(G),		[G]).
 xref_meta(call_with_time_limit(_, G), [G]).
 xref_meta(call_with_depth_limit(G, _, _), [G]).
 xref_meta('$add_directive_wic'(G), [G]).
+xref_meta(with_output_to(_, G),	[G]).
 
 					% XPCE meta-predicates
 xref_meta(pce_global(_, new(_)), _) :- !, fail.
@@ -1216,11 +1226,15 @@ hooking can be databases, (HTTP) URIs, etc.
 	prolog:xref_source_directory/2.		% +Source, -Dir
 
 
-%%	xref_source_file(+Spec, -File, +Src)
+%%	xref_source_file(+Spec, -File, +Src) is semidet.
+%%	xref_source_file(+Spec, -File, +Src, +Options) is semidet.
 %	
 %	Find named source file from Spec, relative to Src.
 
 xref_source_file(Plain, File, Source) :-
+	xref_source_file(Plain, File, Source, []).
+
+xref_source_file(Plain, File, Source, Options) :-
 	atom(Plain),
 	\+ is_absolute_file_name(Plain),
 	(   prolog:xref_source_directory(Source, Dir)
@@ -1229,17 +1243,18 @@ xref_source_file(Plain, File, Source) :-
 	    file_directory_name(Source, Dir)
 	),
 	concat_atom([Dir, /, Plain], Spec),
-	do_xref_source_file(Spec, File), !.
-xref_source_file(Spec, File, _) :-
-	do_xref_source_file(Spec, File), !.
-xref_source_file(Spec, _, _) :-
+	do_xref_source_file(Spec, File, Options), !.
+xref_source_file(Spec, File, _, Options) :-
+	do_xref_source_file(Spec, File, Options), !.
+xref_source_file(Spec, _, _, _) :-
 	verbose,
 	print_message(warning, error(existence_error(file, Spec), _)),
 	fail.
 
-do_xref_source_file(Spec, File) :-
+do_xref_source_file(Spec, File, Options) :-
+	option(file_type(Type), Options, prolog),
 	absolute_file_name(Spec,
-			   [ file_type(prolog),
+			   [ file_type(Type),
 			     access(read),
 			     file_errors(fail)
 			   ], File), !.
