@@ -1,4 +1,4 @@
-/*  $Id: url.pl,v 1.50 2006/08/20 19:34:00 jan Exp $
+/*  $Id: url.pl,v 1.53 2007/04/12 14:53:57 jan Exp $
 
     Part of SWI-Prolog
 
@@ -37,7 +37,9 @@
 	    global_url/3,		% +Local, +Base, -Global
 	    http_location/2,		% ?Parts, ?Location
 	    www_form_encode/2,		% Value <-> Encoded
-	    parse_url_search/2		% Form-data <-> Form fields
+	    parse_url_search/2,		% Form-data <-> Form fields
+
+	    file_name_to_url/2		% ?FileName, ?URL
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(utf8)).
@@ -230,7 +232,7 @@ chost(A) -->
 	).
 
 cport(A) -->
-	(   { memberchk(port(Port), A) }
+	(   { memberchk(port(Port), A), Port \== 80 }
 	->  { number_codes(Port, Codes) },
 	    ":",
 	    www_encode(Codes, "")
@@ -258,24 +260,38 @@ csearch([Parameter|Parameters], Sep) --> !,
 	csearch(Parameters, "&"). 
 
 cparam(Name=Value) --> !, 
-	cform(Name), 
+	cname(Name), 
 	"=", 
-	cform(Value). 
+	cvalue(Value). 
 cparam(NameValue) -->			% allow to feed Name(Value)
 	{ compound(NameValue), !,
 	  NameValue =.. [Name,Value]
 	},
-	cform(Name), 
+	cname(Name), 
 	"=", 
-	cform(Value). 
+	cvalue(Value). 
 cparam(Name)--> 
-	cform(Name). 
+	cname(Name). 
 	
 
-cform(Atom) --> 
+cname(Atom) --> 
 	{ atom_codes(Atom, Codes) },
 	www_encode(Codes, "").
 
+%%	cvalue(+Value)// is det.
+%
+%	Construct a string from  Value.  Value   is  either  atomic or a
+%	code-list.
+
+cvalue(Value) -->
+	{ atomic(Value), !,
+	  atom_codes(Value, Codes)
+	},
+	www_encode(Codes, "").
+cvalue(Codes) -->
+	www_encode(Codes, "").
+
+
 %%	cfragment(+Attributes)//
 
 cfragment(A) -->
@@ -871,3 +887,22 @@ parse_url_search(Codes, Fields) :-
 parse_url_search(Codes, Fields) :-
 	phrase(csearch(Fields, ""), Codes).
 	
+
+		 /*******************************
+		 *	    FILE URLs		*
+		 *******************************/
+
+%%	file_name_to_url(+File, -URL) is det.
+%%	file_name_to_url(-File, +URL) is semidet.
+%
+%	Translate between a filename and a URL.
+%	
+%	@tbd	Deal with encoding?
+
+file_name_to_url(File, FileURL) :-
+	nonvar(File), !,
+	absolute_file_name(File, Path),
+	atom_concat('file://', Path, FileURL), !.
+file_name_to_url(File, FileURL) :-
+	atom_concat('file://', File, FileURL), !.
+
