@@ -1,4 +1,4 @@
-/*  $Id: win_menu.pl,v 1.19 2006/03/24 16:16:14 jan Exp $
+/*  $Id: win_menu.pl,v 1.21 2007/03/25 15:29:34 jan Exp $
 
     Part of SWI-Prolog
 
@@ -66,7 +66,9 @@ menu('&Debug',
        %'&Debug mode'	     = debug,
        %'&No debug mode'     = nodebug,
        '&Edit spy points ...' = user:prolog_ide(open_debug_status),
+       '&Edit exceptions ...' = user:prolog_ide(open_exceptions(@on)),
        '&Threads monitor ...' = user:prolog_ide(thread_monitor),
+       'Debug &messages ...'  = user:prolog_ide(debug_monitor),
        'Cross &referencer ...'= user:prolog_ide(xref),			     
        --,
        '&Graphical debugger' = user:guitracer
