@@ -1,4 +1,4 @@
-/*  $Id: nonblockio.c,v 1.15 2007/01/13 17:14:46 jan Exp $
+/*  $Id: nonblockio.c,v 1.28 2007/04/17 18:05:01 jan Exp $
 
     Part of SWI-Prolog
 
@@ -129,7 +129,6 @@ leave the details to this function.
 
 #ifndef __WINDOWS__
 #define closesocket(n) close((n))	/* same on Unix */
-typedef int SOCKET;
 #endif
 
 #ifndef SD_SEND
@@ -143,26 +142,43 @@ typedef int SOCKET;
 #endif
 
 #ifdef _REENTRANT
+#if __WINDOWS__
+static CRITICAL_SECTION mutex;
+static CRITICAL_SECTION mutex_free;
+
+#define LOCK() EnterCriticalSection(&mutex)
+#define UNLOCK() LeaveCriticalSection(&mutex)
+#define LOCK_FREE() EnterCriticalSection(&mutex_free)
+#define UNLOCK_FREE() LeaveCriticalSection(&mutex_free)
+#define INITLOCK() (InitializeCriticalSection(&mutex), \
+		    InitializeCriticalSection(&mutex_free))
+#else
 #include <pthread.h>
 
 static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+
 #define LOCK() pthread_mutex_lock(&mutex)
 #define UNLOCK() pthread_mutex_unlock(&mutex)
+#define LOCK_FREE()
+#define UNLOCK_FREE()
+#define INITLOCK()
+#endif
 #else
 #define LOCK()
 #define UNLOCK()
+#define LOCK_FREE()
+#define UNLOCK_FREE()
+#define INITLOCK()
 #endif
 
 #define SOCK_INSTREAM	0x01
 #define SOCK_OUTSTREAM	0x02
 #define SOCK_BIND	0x04		/* What have we done? */
 #define SOCK_LISTEN	0x08
-#define SOCK_CONNECT	0x10
-#define SOCK_ACCEPT	0x20		/* Set on accepted sockets */
+#define SOCK_CONNECT	0x10		/* Connected (client) socket */
+#define SOCK_ACCEPT	0x20		/* Accepted (server) socket */
 #define SOCK_NONBLOCK	0x40		/* Set to non-blocking mode */
 #define SOCK_DISPATCH   0x80		/* do not dispatch events */
-#define SOCK_CLOSE_SEEN	0x100		/* FD_CLOSE seen */
-#define SOCK_EOF_SEEN   0x200		/* Seen end-of-file */
 
 #define set(s, f)   ((s)->flags |= (f))
 #define clear(s, f) ((s)->flags &= ~(f))
@@ -171,9 +187,15 @@ static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
 
 #define SOCK_MAGIC 0x38da3f2c
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+NOTE: We must lock  the  structure   to  avoid  freeSocket() called from
+Prolog deleting the socket while there are   still  pending events on it
+that are concurrently executed in the socket thread.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 typedef struct _plsocket
 { int		    magic;		/* SOCK_MAGIC */
-  struct _plsocket *next;		/* next in list */
+  nbio_sock_t	    id;			/* Integer id */
   SOCKET	    socket;		/* The OS socket */
   int		    flags;		/* Misc flags */
   IOSTREAM *	    input;		/* input stream */
@@ -188,7 +210,7 @@ typedef struct _plsocket
   { struct
     { struct sockaddr_in addr;		/* accepted address */
       int addrlen;			/* address length */
-      SOCKET slave;			/* descriptor of slave */
+      nbio_sock_t slave;		/* descriptor of slave */
     } accept;
     struct
     { struct sockaddr_in addr;		/* accepted address */
@@ -202,16 +224,17 @@ typedef struct _plsocket
     struct
     { int bytes;			/* byte count */
       char *buffer;			/* the buffer */
-      int written;
+      size_t written;
       size_t size;			/* buffer size */
     } write;
   } rdata;
 #endif
 } plsocket;
 
-static plsocket *lookupSocket(SOCKET socket);
+static plsocket *allocSocket(SOCKET socket);
+static plsocket *nbio_to_plsocket(nbio_sock_t socket);
 #ifdef __WINDOWS__
-static plsocket   *lookupExistingSocket(SOCKET socket);
+static plsocket *lookupOSSocket(SOCKET socket);
 static const char *WinSockError(unsigned long eno);
 #endif
 
@@ -247,6 +270,12 @@ nbio_debug(int level)
 
 #endif
 
+void					/* allow debugger breakpoint */
+tcp_debug()
+{ Sdprintf("Trapping debugger\n");
+}
+
+
 		 /*******************************
 		 *	  COMPATIBILITY		*
 		 *******************************/
@@ -257,12 +286,50 @@ static UINT WM_REQUEST  = WM_APP+21;
 static UINT WM_READY	= WM_APP+22;
 static UINT WM_DONE	= WM_APP+23;
 
+#if O_DEBUG
+static char *
+request_name(nbio_request request)
+{ switch(request)
+  { case REQ_NONE:    return "req_none";
+    case REQ_ACCEPT:  return "req_accept";
+    case REQ_CONNECT: return "req_connect";
+    case REQ_READ:    return "req_read";
+    case REQ_WRITE:   return "req_write";
+    default:	      return "req_???";
+  }
+}
+
+static char *
+event_name(int ev)
+{ char buf[256];
+  char *o = buf;
+
+  o[0] = '\0';
+  if ( (ev & FD_READ) )    strcat(o, "|FD_READ");
+  if ( (ev & FD_WRITE) )   strcat(o, "|FD_WRITE");
+  if ( (ev & FD_ACCEPT) )  strcat(o, "|FD_ACCEPT");
+  if ( (ev & FD_CONNECT) ) strcat(o, "|FD_CONNECT");
+  if ( (ev & FD_CLOSE) )   strcat(o, "|FD_CLOSE");
+  if ( (ev & FD_OOB) )	   strcat(o, "|FD_OOB");
+  if ( (ev & ~(FD_READ|FD_WRITE|FD_ACCEPT|FD_CONNECT|FD_CLOSE)) )
+    strcat(o, "|FD_???");
+
+  return strdup(buf);
+}
+
+#endif
+
 #define F_SETFL		0
 #define O_NONBLOCK	0
 
 static int
-nbio_fcntl(int fd, int op, int arg)
-{ switch(op)
+nbio_fcntl(nbio_sock_t socket, int op, int arg)
+{ plsocket *s;
+
+  if ( !(s=nbio_to_plsocket(socket)) )
+    return -1;
+
+  switch(op)
   { case F_SETFL:
       switch(arg)
       { case O_NONBLOCK:
@@ -270,12 +337,13 @@ nbio_fcntl(int fd, int op, int arg)
 	  int non_block;
 
 	  non_block = 1;
-	  rval = ioctlsocket(fd, FIONBIO, &non_block);
+	  rval = ioctlsocket(s->socket, FIONBIO, &non_block);
 	  if ( rval )
-	  { lookupSocket(fd)->flags |= SOCK_NONBLOCK;
+	  { s->flags |= SOCK_NONBLOCK;
 	    return 0;
-	  } else
-	    return -1;
+	  }
+
+	  return -1;
 	}
 	default:
 	  return -1;
@@ -298,9 +366,16 @@ static local_state nbio_state;
 
 static int
 doneRequest(plsocket *s)
-{ s->done = TRUE;
+{ SOCKET sock;
+
+  s->done = TRUE;
   s->request = REQ_NONE;
 
+  if ( (s->w32_flags & FD_CLOSE) && (sock=s->socket) >= 0 )
+  { s->socket = -1;
+    closesocket(sock);
+  }
+
   if ( s->thread )
   { DEBUG(2, Sdprintf("doneRequest(): posting %d\n", s->thread));
     PostThreadMessage(s->thread, WM_DONE, 0, (WPARAM)s);
@@ -312,8 +387,11 @@ doneRequest(plsocket *s)
 
 static int
 waitRequest(plsocket *s)
-{ DEBUG(2, Sdprintf("[%d] (%ld): Waiting ...",
-		    PL_thread_self(), s->thread));
+{ assert(s->magic == SOCK_MAGIC);
+
+  DEBUG(2, Sdprintf("[%d] (%ld): Waiting for %s on %d ...",
+		    PL_thread_self(), s->thread,
+		    request_name(s->request), (int)s->socket));
 
   for(;;)
   { MSG msg;
@@ -342,8 +420,11 @@ waitRequest(plsocket *s)
 
 
 int
-nbio_wait(int socket, nbio_request request)
-{ plsocket *s = lookupSocket(socket);
+nbio_wait(nbio_sock_t socket, nbio_request request)
+{ plsocket *s;
+
+  if ( !(s=nbio_to_plsocket(socket)) )
+    return -1;
 
   s->flags  |= SOCK_WAITING;
   s->done    = FALSE;
@@ -382,6 +463,14 @@ nbio_wait(int socket, nbio_request request)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+nbio_select() selects using a set of socket streams. 
+
+NOTE: The Windows versions uses our   nbio_sock_t abstraction, while the
+other version uses  the  raw  Unix   file  descriptors  referencing  the
+underlying socket.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 nbio_select(int n,
 	    fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
@@ -400,14 +489,18 @@ nbio_select(int n,
   if ( readfds )
   { for(i=0; i<n; i++)
     { if ( FD_ISSET(i, readfds) )
-      { plsocket *s = lookupSocket(i);
+      { plsocket *s = nbio_to_plsocket(i);
 
-	s->flags  |= SOCK_WAITING;
-	s->done    = FALSE;
-	s->error   = 0;
-	s->thread  = GetCurrentThreadId();
-	s->request = (s->flags & SOCK_LISTEN) ? REQ_ACCEPT : REQ_READ;
-	sockets[i] = s;
+	if ( s )
+	{ s->flags  |= SOCK_WAITING;
+	  s->done    = FALSE;
+	  s->error   = 0;
+	  s->thread  = GetCurrentThreadId();
+	  s->request = (s->flags & SOCK_LISTEN) ? REQ_ACCEPT : REQ_READ;
+	  sockets[i] = s;
+	} else
+	{ DEBUG(2, Sdprintf("nbio_select(): no socket for %d\n", i));
+	}
       }
     }
   }
@@ -481,22 +574,29 @@ nbio_select(int n,
 
 static int
 placeRequest(plsocket *s, nbio_request request)
-{ s->error   = 0;
+{ if ( s->magic != SOCK_MAGIC )
+    Sdprintf("placeRequest: %p has bad magic\n", s);
+
+  s->error   = 0;
   s->done    = FALSE;
   s->thread  = GetCurrentThreadId();
   s->request = request;
   clear(s, SOCK_WAITING);
 
   SendMessage(State()->hwnd, WM_REQUEST, 1, (LPARAM)&s);
-  DEBUG(2, Sdprintf("%d (%ld): Placed request %d\n",
-		    PL_thread_self(), s->thread, request));
+  DEBUG(2, Sdprintf("%d (%ld): Placed %s request for %d\n",
+		    PL_thread_self(), s->thread,
+		    request_name(request), (int)s->socket));
 
   return TRUE; 
 }
 
 static int
 doRequest(plsocket *s)
-{ switch(s->request)
+{ if ( s->magic != SOCK_MAGIC )
+    Sdprintf("doRequest: %p has bad magic\n", s);
+
+  switch(s->request)
   { case REQ_NONE:
       break;
     case REQ_CONNECT:
@@ -533,31 +633,43 @@ doRequest(plsocket *s)
       break;
     case REQ_ACCEPT:
       if ( s->w32_flags & FD_ACCEPT )
-      { s->w32_flags &= ~FD_ACCEPT;
+      { SOCKET slave;
 
+	s->w32_flags &= ~FD_ACCEPT;
 	if ( true(s, SOCK_WAITING) )
 	{ doneRequest(s);
 	  break;
 	}
 
-	s->rdata.accept.slave = accept(s->socket,
-				       (struct sockaddr*)&s->rdata.accept.addr,
-				       &s->rdata.accept.addrlen);
+	slave = accept(s->socket,
+		       (struct sockaddr*)&s->rdata.accept.addr,
+		       &s->rdata.accept.addrlen);
 
-	DEBUG(2, Sdprintf("Accept() --> %d\n", s->rdata.accept.slave));
-
-	if ( s->rdata.accept.slave == SOCKET_ERROR )
+	if ( slave == SOCKET_ERROR )
 	{ s->error = WSAGetLastError();
 	  
-	  if ( s->error != WSAEWOULDBLOCK )
-	    doneRequest(s);
-	} else
-	{ plsocket *s2;
+	  DEBUG(2, Sdprintf("Accept(%d): %s\n",
+			    (int)s->socket, WinSockError(s->error)));
 
-	  s2 = lookupSocket(s->rdata.accept.slave);
-	  s2->flags |= SOCK_ACCEPT;
-	  s->error = 0;
-	  doneRequest(s);
+	  if ( s->error != WSAEWOULDBLOCK )
+	  { s->rdata.accept.slave = (SOCKET)-1;
+	    doneRequest(s);
+	  }
+	} else
+	{ plsocket *pls;
+
+	  DEBUG(2, Sdprintf("Accept(%d) --> %d\n",
+			    (int)s->socket, (int)slave));
+	  if ( (pls = allocSocket(slave)) )
+	  { pls->flags |= SOCK_ACCEPT;	/* requests */
+
+	    s->rdata.accept.slave = pls->id;
+	    s->error = 0;
+	    doneRequest(s);
+	  } else
+	  { DEBUG(1, Sdprintf("Socket %d already registered; "
+			      "considering bogus\n", (int)slave));
+	  }
 	}
       }
       break;
@@ -583,12 +695,15 @@ doRequest(plsocket *s)
 	    doneRequest(s);
 	  }
 	} else
-	  doneRequest(s);
+	{ doneRequest(s);
+	}
       }
       break;
     case REQ_WRITE:
       if ( s->w32_flags & FD_WRITE )
       { int n;
+	int len = (int)(s->rdata.write.size - s->rdata.write.written);
+
 	s->w32_flags &= ~FD_WRITE;
 
 	if ( true(s, SOCK_WAITING) )
@@ -599,21 +714,23 @@ doRequest(plsocket *s)
 	DEBUG(2, Sdprintf("send() %d bytes\n", s->rdata.write.size));
 	n = send(s->socket,
 		 s->rdata.write.buffer + s->rdata.write.written,
-		 (int)(s->rdata.write.size - s->rdata.write.written),
-		 0);
+		 len, 0);
 	DEBUG(2, Sdprintf("Wrote %d bytes\n", n));
 	if ( n < 0 )
 	{ s->error = WSAGetLastError();
 	  if ( s->error == WSAEWOULDBLOCK )
 	    break;
 	  s->rdata.write.bytes = n;
+	  DEBUG(1, Sdprintf("[%d]: send(%d, %d bytes): %s\n",
+			    PL_thread_self(), s->socket, len,
+			    WinSockError(s->error)));
 	  doneRequest(s);
 	} else
 	  s->error = 0;
 	  
 	s->rdata.write.written += n;
 	if ( s->rdata.write.written >= s->rdata.write.size )
-	{ s->rdata.write.bytes = s->rdata.write.written;
+	{ s->rdata.write.bytes = (int)s->rdata.write.written;
 	  doneRequest(s);
 	}
       }
@@ -652,13 +769,47 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
   { SOCKET sock = (SOCKET) wParam;
     int err     = WSAGETSELECTERROR(lParam);
     int evt     = WSAGETSELECTEVENT(lParam);
-    plsocket *s = lookupExistingSocket(sock);
+    plsocket *s;
+
+    if ( evt&FD_CLOSE )
+    { DEBUG(1,
+	    { char *nm = event_name(evt);
+	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n", 
+		       (int)sock, nm, err ? WinSockError(err) : "none");
+	      free(nm);
+	    });
+    } else
+    { DEBUG(3,
+	    { char *nm = event_name(evt);
+	      Sdprintf("WM_SOCKET on %d: ev=(%s); err=%s\n", 
+		       (int)sock, nm, err ? WinSockError(err) : "none");
+	      free(nm);
+	    });
+    }
+
+    LOCK_FREE();
+    s = lookupOSSocket(sock);
 
     if ( s )
-    { s->w32_flags |= evt;
+    { if ( (s->w32_flags & FD_CLOSE) )
+      { DEBUG(1,
+	      { char *nm = event_name(evt);
+		Sdprintf("Got event %s (err=%s) on closed socket %d=%d\n",
+			 nm, WinSockError(err), s->id, sock);
+		free(nm);
+	      })
+      }
+
+      s->w32_flags |= evt;
       if ( err )
-      { s->error = err;
-	
+      { SOCKET sock = s->socket;
+
+	s->error = err;
+	if ( sock )
+	{ s->socket = -1;
+	  closesocket(sock);
+	}
+
 	switch(s->request)
 	{ case REQ_CONNECT:
 	    break;
@@ -673,11 +824,17 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	    break;
 	}
 	doneRequest(s);
-      } else
+      } else if ( s->socket >= 0 )
       { doRequest(s);
+      } else
+      { doneRequest(s);
       }
     } else
-      DEBUG(2, Sdprintf("Socket %d is gone\n", sock));
+    { DEBUG(1, Sdprintf("Socket %d is gone (error=%s)\n",
+			sock, WinSockError(err)));
+    }
+
+    UNLOCK_FREE();
   }
 
   return DefWindowProc(hwnd, message, wParam, lParam);
@@ -790,9 +947,11 @@ without dispatching if no input is available.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-wait_socket(plsocket *s, int fd)
+wait_socket(plsocket *s)
 { if ( true(s, SOCK_DISPATCH) )
-  { if ( true(s, SOCK_NONBLOCK) && !PL_dispatch(fd, PL_DISPATCH_INSTALLED) )
+  { int fd = s->socket;
+
+    if ( true(s, SOCK_NONBLOCK) && !PL_dispatch(fd, PL_DISPATCH_INSTALLED) )
     { fd_set rfds;
       struct timeval tv;
 
@@ -813,26 +972,42 @@ wait_socket(plsocket *s, int fd)
 
 
 int
-nbio_wait(int socket, nbio_request request)
-{ plsocket *s = lookupSocket(socket);
+nbio_wait(nbio_sock_t socket, nbio_request request)
+{ plsocket *s;
 
-  return wait_socket(s, socket) ? 0 : -1;
+  if ( !(s=nbio_to_plsocket(socket)) )
+    return -1;
+
+  return wait_socket(s) ? 0 : -1;
 }
 
 
 static int
-nbio_fcntl(int fd, int op, int arg)
-{ int rc = fcntl(fd, op, arg);
+nbio_fcntl(nbio_sock_t socket, int op, int arg)
+{ plsocket *s;
+  int rc;
+
+  if ( !(s=nbio_to_plsocket(socket)) )
+    return -1;
+
+  rc = fcntl(s->socket, op, arg);
 
   if ( rc == 0 )
   { if ( op == F_SETFL && arg == O_NONBLOCK )
-      lookupSocket(fd)->flags |= SOCK_NONBLOCK;
+      s->flags |= SOCK_NONBLOCK;
   } else
     nbio_error(errno, TCP_ERRNO);
 
   return rc;
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+NOTE: when called througb wait_for_input/3, the  descriptors in the sets
+are real underlying Unix socket descriptors   and  *not* the nbio_sock_t
+psuedo descriptors.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 nbio_select(int n,
 	    fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
@@ -854,45 +1029,141 @@ static atom_t ATOM_any;
 static atom_t ATOM_broadcast;
 static atom_t ATOM_loopback;
 
-static plsocket *sockets;
+static plsocket **sockets = NULL;	/* id --> plsocket* */
+static size_t	socks_count = 0;	/* #registered sockets */
+static size_t	socks_allocated = 0;	/* #allocated entries */
 static int initialised = FALSE;		/* Windows only */
 
 #ifdef __WINDOWS__
+
 static plsocket *
-lookupExistingSocket(SOCKET socket)
+lookupOSSocket(SOCKET socket)
 { plsocket *p;
+  size_t i;
 
   LOCK();
-  for(p=sockets; p; p = p->next)
-  { if ( p->socket == socket )
+  for(i=0; i<socks_allocated; i++)
+  { if ( (p=sockets[i]) && p->socket == socket )
     { UNLOCK();
+
+      if ( p->magic != SOCK_MAGIC )
+      { errno = EINVAL;
+	DEBUG(1, Sdprintf("Invalid OS socket: %d\n", socket));
+	return NULL;
+      }
+
       return p;
     }
   }
   UNLOCK();
+
   return NULL;
 }
-#endif
+
+#endif /*__WINDOWS__*/
 
 
 static plsocket *
-lookupSocket(SOCKET socket)
+nbio_to_plsocket_raw(nbio_sock_t socket)
 { plsocket *p;
 
-  LOCK();
-  for(p=sockets; p; p = p->next)
-  { if ( p->socket == socket )
-    { UNLOCK();
-      return p;
-    }
-  }
-
-  if ( !(p = PL_malloc(sizeof(plsocket))) )
-  { pl_error(NULL, 0, NULL, ERR_ERRNO);
-    UNLOCK();
+  if ( socket < 0 || (size_t)socket >= socks_allocated )
+  { errno = EINVAL;
     return NULL;
   }
 
+  LOCK();
+  p = sockets[socket];
+  UNLOCK();
+
+  if ( !p || p->magic != SOCK_MAGIC )
+  { DEBUG(1, Sdprintf("Invalid NBIO socket: %d\n", socket));
+    errno = EINVAL;
+    return NULL;
+  }
+
+  return p;
+}
+
+
+static plsocket *
+nbio_to_plsocket(nbio_sock_t socket)
+{ plsocket *p;
+
+  if ( !(p=nbio_to_plsocket_raw(socket)) )
+    return NULL;
+
+#ifdef __WINDOWS__
+  if ( p->socket < 0 )
+  { p->error = WSAECONNRESET;
+    return NULL;
+  }
+#endif
+
+  return p;
+}
+
+
+SOCKET
+nbio_fd(nbio_sock_t socket)
+{ plsocket *p;
+
+  if ( !(p=nbio_to_plsocket(socket)) )
+    return -1;
+
+  return p->socket;
+}
+
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Allocate a wrapper for an OS  socket.   The  wrapper  is allocated in an
+array of pointers, to keep small  integer   identifiers  we can use with
+FD_SET, etc. for implementing a compatible nbio_select().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static plsocket *
+allocSocket(SOCKET socket)
+{ plsocket *p = NULL;
+  size_t i;
+
+#ifdef __WINDOWS__
+  if ( (p=lookupOSSocket(socket)) )
+  { DEBUG(1, Sdprintf("WinSock %d already registered on %d\n",
+		      (int)socket, p->id));
+    p->socket = (SOCKET)-1;
+  }
+#endif
+
+  LOCK();
+  if ( socks_count+1 > socks_allocated )
+  { if ( socks_allocated )
+    { size_t newa = socks_allocated*2;
+
+      sockets = PL_realloc(sockets, sizeof(plsocket*)*newa);
+      for(i=socks_allocated; i<newa; i++)
+	sockets[i] = NULL;
+      socks_allocated = newa;
+    } else
+    { socks_allocated = 32;
+      sockets = PL_malloc(sizeof(plsocket*)*socks_allocated);
+      memset(sockets, 0, sizeof(plsocket*)*socks_allocated);
+    }
+  }
+
+  for(i=0; i<socks_allocated; i++)
+  { if ( sockets[i] == NULL )
+    { sockets[i] = p = PL_malloc(sizeof(*p));
+      socks_count++;
+      break;
+    }
+  }
+  UNLOCK();
+
+  assert(i<socks_allocated);
+  
+  memset(p, 0, sizeof(*p));
+  p->id     = (int)i;			/* place in the array */
   p->socket = socket;
   p->flags  = SOCK_DISPATCH;		/* by default, dispatch */
   p->magic  = SOCK_MAGIC;
@@ -901,45 +1172,49 @@ lookupSocket(SOCKET socket)
   p->request   = REQ_NONE;
 #endif
   p->input = p->output = (IOSTREAM*)NULL;
-  p->next   = sockets;
-  sockets   = p;
 
-  DEBUG(2, Sdprintf("[%d]: lookupSocket(%d): bound to %p\n",
+  DEBUG(2, Sdprintf("[%d]: allocSocket(%d): bound to %p\n",
 		    PL_thread_self(), socket, p));
 
-  UNLOCK();
   return p;
 }
 
 
 static int
-freeSocket(int socket)
-{ plsocket **p;
-  int rval;
+freeSocket(plsocket *s)
+{ int rval;
+  nbio_sock_t socket;
+  SOCKET sock;
 
-  DEBUG(2, Sdprintf("Closing %d\n", socket));
-
-  LOCK();
-  p = &sockets;
-
-  for( ; *p; p = &(*p)->next)
-  { if ( (*p)->socket == socket )
-    { plsocket *s = *p;
-      
-      *p = s->next;
-      PL_free(s);
-      break;
-    }
+  DEBUG(2, Sdprintf("Closing %d\n", s->id));
+  if ( !s || s->magic != SOCK_MAGIC )
+  { errno = EINVAL;
+    return -1;
   }
+
+  LOCK_FREE();
+  LOCK();
+  sockets[s->id] = NULL;
+  socks_count--;
   UNLOCK();
 
-  again:
-  if ( (rval=closesocket(socket)) == SOCKET_ERROR )
-  { if ( errno == EINTR )
-      goto again;
-  }
+  sock = s->socket;
+  socket = s->id;
+  s->magic = 0;
+  PL_free(s);
+  UNLOCK_FREE();
 
-  DEBUG(2, Sdprintf("freeSocket(%d) returned %d\n", socket, rval));
+  if ( sock >= 0 )
+  { again:
+    if ( (rval=closesocket(sock)) == SOCKET_ERROR )
+    { if ( errno == EINTR )
+	goto again;
+    }
+    DEBUG(2, Sdprintf("freeSocket(%d=%d) returned %d\n",
+		      socket, (int)sock, rval));
+  } else
+  { rval = 0;				/* already closed.  Use s->error? */
+  }
 
   return rval;
 }
@@ -1048,7 +1323,7 @@ WinSockError(unsigned long error)
   }
 
   sprintf(tmp, "Unknown error %ld", error);
-  return strdup(tmp);			/* leaks memory */
+  return strdup(tmp);			/* leaks memory on unknown errors */
 }
 
 #endif /*BILLY_GETS_BETTER*/
@@ -1130,13 +1405,33 @@ nbio_error(int code, nbio_error_map mapid)
   return PL_raise_exception(except);
 }
 
+
+const char *
+nbio_last_error(nbio_sock_t socket)
+{
+#ifdef __WINDOWS__
+  plsocket *s;
+
+  if ( !(s=nbio_to_plsocket_raw(socket)) )
+    return NULL;
+
+  if ( s->error )
+    return WinSockError(s->error);
+#endif
+
+  return NULL;
+}
+
+
 		 /*******************************
 		 *	  INITIALISATION	*
 		 *******************************/
 
 int
 nbio_init(const char *module)
-{ LOCK();
+{ INITLOCK();				/* is this ok? */
+
+  LOCK();
   if ( initialised )
   { UNLOCK();
     return TRUE;
@@ -1194,7 +1489,7 @@ socket(-Socket)
     the format $socket(Id).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-int
+nbio_sock_t
 nbio_socket(int domain, int type, int protocol)
 { SOCKET sock;
   plsocket *s;
@@ -1206,7 +1501,7 @@ nbio_socket(int domain, int type, int protocol)
     return -1;
   }
 
-  if ( !(s=lookupSocket(sock)) )		/* register it */
+  if ( !(s=allocSocket(sock)) )		/* register it */
   { closesocket(sock);
     return -1;
   }
@@ -1216,15 +1511,18 @@ nbio_socket(int domain, int type, int protocol)
 		 FD_READ|FD_WRITE|FD_ACCEPT|FD_CONNECT|FD_CLOSE);
 #endif
 
-  return (int)sock;
+  return s->id;
 }
 
 
 int
-nbio_closesocket(int socket)
+nbio_closesocket(nbio_sock_t socket)
 { plsocket *s;
 
-  s = lookupSocket(socket);
+  if ( !(s = nbio_to_plsocket_raw(socket)) )
+  { DEBUG(1, Sdprintf("nbio_closesocket(%d): no plsocket\n", socket));
+    return -1;
+  }
 
   if ( true(s, SOCK_OUTSTREAM|SOCK_INSTREAM) )
   { int flags = s->flags;		/* may drop out! */
@@ -1241,10 +1539,12 @@ nbio_closesocket(int socket)
   {
 #ifdef __WINDOWS__
     if ( true(s, SOCK_CONNECT) )
-      shutdown(socket, SD_SEND);
+    { if ( s->socket >= 0 )
+	shutdown(s->socket, SD_SEND);
+    }
 #endif
 
-    freeSocket(socket);
+    freeSocket(s);
   }
 
   return 0;
@@ -1252,10 +1552,14 @@ nbio_closesocket(int socket)
 
 
 int
-nbio_setopt(int socket, nbio_option opt, ...)
-{ va_list args;
+nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
+{ plsocket *s;
+  va_list args;
   int rc;
 
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
   va_start(args, opt);
 
   switch(opt)
@@ -1265,7 +1569,7 @@ nbio_setopt(int socket, nbio_option opt, ...)
     case TCP_REUSEADDR:
     { int val = va_arg(args, int);
 
-      if( setsockopt(socket, SOL_SOCKET, SO_REUSEADDR,
+      if( setsockopt(s->socket, SOL_SOCKET, SO_REUSEADDR,
 		     (const char *)&val, sizeof(val)) == -1)
       { nbio_error(h_errno, TCP_HERRNO);
 	rc = -1;
@@ -1277,7 +1581,7 @@ nbio_setopt(int socket, nbio_option opt, ...)
     case UDP_BROADCAST:
     { int val = va_arg(args, int);
 
-      if( setsockopt(socket, SOL_SOCKET, SO_BROADCAST,
+      if( setsockopt(s->socket, SOL_SOCKET, SO_BROADCAST,
 		     (const char *)&val, sizeof(val)) == -1)
       { nbio_error(h_errno, TCP_HERRNO);
 	rc = -1;
@@ -1288,7 +1592,6 @@ nbio_setopt(int socket, nbio_option opt, ...)
     }
     case TCP_DISPATCH:
     { int val = va_arg(args, int);
-      plsocket *s = lookupSocket(socket);
       
       if ( val )
 	set(s, SOCK_DISPATCH);
@@ -1301,7 +1604,6 @@ nbio_setopt(int socket, nbio_option opt, ...)
     }
     case TCP_INSTREAM:
     { IOSTREAM *in = va_arg(args, IOSTREAM*);
-      plsocket *s = lookupSocket(socket);
 
       s->flags |= SOCK_INSTREAM;
       s->input = in;
@@ -1312,7 +1614,6 @@ nbio_setopt(int socket, nbio_option opt, ...)
     }
     case TCP_OUTSTREAM:
     { IOSTREAM *out = va_arg(args, IOSTREAM*);
-      plsocket *s = lookupSocket(socket);
 
       s->flags |= SOCK_OUTSTREAM;
       s->output = out;
@@ -1333,13 +1634,13 @@ nbio_setopt(int socket, nbio_option opt, ...)
 
 
 int
-nbio_get_flags(int socket)
-{ plsocket *s = lookupSocket(socket);
+nbio_get_flags(nbio_sock_t socket)
+{ plsocket *s;
   
-  if ( s )
-    return s->flags;
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
 
-  return -1;
+  return s->flags;
 } 
 
 
@@ -1465,33 +1766,38 @@ nbio_unify_ip4(term_t Ip, unsigned long hip)
 
 
 int
-nbio_bind(int socket, struct sockaddr *my_addr, size_t addrlen)
-{
+nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, size_t addrlen)
+{ plsocket *s;
+
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
 #ifdef __WINDOWS__
-  if ( bind(socket, my_addr, (int)addrlen) )
+  if ( bind(s->socket, my_addr, (int)addrlen) )
 #else
-  if ( bind(socket, my_addr, addrlen) )
+  if ( bind(s->socket, my_addr, addrlen) )
 #endif
   { nbio_error(errno, TCP_ERRNO);
     return -1;
   }
 
-  lookupSocket(socket)->flags |= SOCK_BIND;
+  s->flags |= SOCK_BIND;
 
   return 0;
 }
 
 
 int
-nbio_connect(int socket,
-	    const struct sockaddr *serv_addr,
-	    size_t addrlen)
+nbio_connect(nbio_sock_t socket,
+	     const struct sockaddr *serv_addr,
+	     size_t addrlen)
 { plsocket *s;
        
-  s = lookupSocket(socket);
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
   
 #ifdef __WINDOWS__
-  if ( connect(socket, serv_addr, (int)addrlen) )
+  if ( connect(s->socket, serv_addr, (int)addrlen) )
   { s->error = WSAGetLastError();
 
     if ( s->error == WSAEWOULDBLOCK )
@@ -1508,7 +1814,7 @@ nbio_connect(int socket,
   }
 #else /*!__WINDOWS__*/
   for(;;)
-  { if ( connect(socket, serv_addr, addrlen) )
+  { if ( connect(s->socket, serv_addr, addrlen) )
     { if ( need_retry(errno) )
       { if ( PL_handle_signals() < 0 )
 	  return -1;
@@ -1527,16 +1833,17 @@ nbio_connect(int socket,
 }
 
 
-int
-nbio_accept(int master, struct sockaddr *addr, socklen_t *addrlen)
+nbio_sock_t
+nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
 { SOCKET slave;
-  plsocket *m;
+  plsocket *m, *s;
 	
-  m = lookupSocket(master);
+  if ( !(m = nbio_to_plsocket(master)) )
+    return -1;
 
 #ifdef __WINDOWS__
-{ int alen;
-  slave = accept(master, addr, &alen);
+{ int alen = (int)*addrlen;
+  slave = accept(m->socket, addr, &alen);
   *addrlen = alen;
 }
 
@@ -1550,9 +1857,10 @@ nbio_accept(int master, struct sockaddr *addr, socklen_t *addrlen)
 	return FALSE;
       if ( m->error )
 	return nbio_error(m->error, TCP_ERRNO);
-      slave    = m->rdata.accept.slave;
       *addrlen = m->rdata.accept.addrlen;
       memcpy(addr, &m->rdata.accept.addr, m->rdata.accept.addrlen);
+
+      return m->rdata.accept.slave;		/* already registered */
     } else
     { nbio_error(m->error, TCP_ERRNO);
       return -1;
@@ -1562,10 +1870,10 @@ nbio_accept(int master, struct sockaddr *addr, socklen_t *addrlen)
 #else /*__WINDOWS__*/
 
   for(;;)
-  { if ( !wait_socket(m, master) )
+  { if ( !wait_socket(m) )
       return -1;
 
-    slave = accept(master, addr, addrlen);
+    slave = accept(m->socket, addr, addrlen);
 
     if ( slave == SOCKET_ERROR )
     { if ( need_retry(errno) )
@@ -1581,26 +1889,32 @@ nbio_accept(int master, struct sockaddr *addr, socklen_t *addrlen)
       break;
   }
 
-  { plsocket *s = lookupSocket(slave);
-    s->flags |= SOCK_ACCEPT;
-    if ( true(s, SOCK_NONBLOCK) )
-      nbio_setopt(slave, TCP_NONBLOCK);
-  }
-
 #endif /*__WINDOWS__*/
   
-  return (int)slave;
+  s = allocSocket(slave);
+  s->flags |= SOCK_ACCEPT;
+#ifndef __WINDOWS__
+  if ( true(s, SOCK_NONBLOCK) )
+    nbio_setopt(slave, TCP_NONBLOCK);
+#endif
+
+  return s->id;
 }
 
 
 int
-nbio_listen(int socket, int backlog)
-{ if( listen(socket, backlog) == -1 )
+nbio_listen(nbio_sock_t socket, int backlog)
+{ plsocket *s;
+
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
+  if( listen(s->socket, backlog) == -1 )
   { nbio_error(errno, TCP_ERRNO);
     return -1;
   }
 
-  lookupSocket(socket)->flags |= SOCK_LISTEN;
+  s->flags |= SOCK_LISTEN;
 
   return 0;
 }
@@ -1614,20 +1928,22 @@ nbio_listen(int socket, int backlog)
 
 ssize_t
 nbio_read(int socket, char *buf, size_t bufSize)
-{ plsocket *s = lookupSocket(socket);
+{ plsocket *s;
   int n;
 
-  if ( !s || s->magic != SOCK_MAGIC )
-  { errno = EINVAL;
-    DEBUG(1, Sdprintf("nbio_read(): Invalid socket: %d\n", socket));
+  if ( !(s = nbio_to_plsocket(socket)) )
     return -1;
-  }
 
 #ifdef __WINDOWS__
 
-  n = recv(socket, buf, (int)bufSize, 0);
+  DEBUG(3, Sdprintf("[%d] reading from socket %d\n",
+		    PL_thread_self(), socket));
+
+  n = recv(s->socket, buf, (int)bufSize, 0);
   if ( n < 0 )
-  { if ( WSAGetLastError() == WSAEWOULDBLOCK )
+  { int wsaerrno;
+
+    if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
     { s->rdata.read.buffer = buf;
       s->rdata.read.size   = bufSize;
       placeRequest(s, REQ_READ);
@@ -1639,18 +1955,22 @@ nbio_read(int socket, char *buf, size_t bufSize)
     }
 
     if ( n < 0 )
-      errno = EIO;			/* TBD: map s->error to POSIX errno */
+    { s->error = wsaerrno;
+      errno = EIO;
+    }
+  } else
+  { s->error = 0;
   }
 
 #else /*__WINDOWS__*/
 
   for(;;)
-  { if ( !wait_socket(s, socket) )
+  { if ( !wait_socket(s) )
     { errno = EPLEXCEPTION;
       return -1;
     }
 
-    n = recv(socket, buf, bufSize, 0);
+    n = recv(s->socket, buf, bufSize, 0);
 
     if ( n == -1 && need_retry(errno) )
     { if ( PL_handle_signals() < 0 )
@@ -1665,29 +1985,38 @@ nbio_read(int socket, char *buf, size_t bufSize)
 
 #endif /*__WINDOWS__*/
 
-  if ( n == 0 )
-    set(s, SOCK_EOF_SEEN);
-
   return n;
 }
 
 
 ssize_t
-nbio_write(int socket, char *buf, size_t bufSize)
-{ size_t len = bufSize;
+nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
+{ plsocket *s;
+  size_t len = bufSize;
   char *str = buf;
 
+  if ( !(s = nbio_to_plsocket(socket)) )
+    return -1;
+
 #ifdef __WINDOWS__
   while( len != 0 )
   { ssize_t n;
 
-    n = send(socket, str, (int)len, 0);
+    DEBUG(3, Sdprintf("[%d] sending %d bytes using socket %d\n",
+		      PL_thread_self(), (int)len, socket));
+
+    n = send(s->socket, str, (int)len, 0);
     if ( n < 0 )
     { int error = WSAGetLastError();
 
       if ( error == WSAEWOULDBLOCK )
 	break;
 
+      DEBUG(1, Sdprintf("[%d]: send(%d, %d bytes): %s\n",
+			PL_thread_self(), s->socket, (int)bufSize,
+			WinSockError(error)));
+
+      s->error = error;
       return -1;
     }
 
@@ -1696,18 +2025,16 @@ nbio_write(int socket, char *buf, size_t bufSize)
   }
 
   if ( len > 0 )			/* operation would block */
-  { plsocket *s = lookupSocket(socket);
-
-    s->rdata.write.buffer  = str;
+  { s->rdata.write.buffer  = str;
     s->rdata.write.size    = len;
     s->rdata.write.written = 0;
     placeRequest(s, REQ_WRITE);
     if ( !waitRequest(s) )
-    { errno = EPLEXCEPTION;
+    { errno = EPLEXCEPTION;		/* handled Prolog signal */
       return -1;
     }
     if ( s->rdata.write.bytes < 0 )
-      return -1;
+      return -1;			/* error in s->error */
   }
 
 #else /*__WINDOWS__*/
@@ -1715,7 +2042,7 @@ nbio_write(int socket, char *buf, size_t bufSize)
   while( len > 0 )
   { int n;
 
-    n = send(socket, str, len, 0);
+    n = send(s->socket, str, len, 0);
     if ( n < 0 )
     { if ( need_retry(errno) )
       { if ( PL_handle_signals() < 0 )
@@ -1738,41 +2065,63 @@ nbio_write(int socket, char *buf, size_t bufSize)
 
 
 int
-nbio_close_input(int socket)
+nbio_close_input(nbio_sock_t socket)
 { int rc = 0;
-  plsocket *s = lookupSocket(socket);
+  plsocket *s;
 
-  DEBUG(2, Sdprintf("[%d]: nbio_close_input(%d)\n", PL_thread_self(), socket));
+  if ( !(s = nbio_to_plsocket_raw(socket)) )
+    return -1;
+
+  DEBUG(2, Sdprintf("[%d]: nbio_close_input(%d, flags=0x%x)\n",
+		    PL_thread_self(), socket, s->flags));
   s->input = NULL;
   s->flags &= ~SOCK_INSTREAM;
 #ifdef __WINDOWS__
   if ( false(s, SOCK_LISTEN) )
-  { if ( shutdown(socket, SD_RECEIVE) == SOCKET_ERROR )
-      Sdprintf("shutdown(%d, SD_RECEIVE) failed: %s\n",
-	       socket,
-	       WinSockError(WSAGetLastError()));
+  { SOCKET sock;
+
+    if ( (sock=s->socket) < 0 )
+    { s->error = WSAECONNRESET;
+      rc = -1;
+    } else if ( shutdown(sock, SD_RECEIVE) == SOCKET_ERROR )
+    { DEBUG(1, Sdprintf("shutdown(%d=%d, SD_RECEIVE) failed: %s\n",
+			socket, s->socket,
+			WinSockError(WSAGetLastError())));
+    }
     rc = -1;
   }
 #endif
 
   if ( !(s->flags & (SOCK_INSTREAM|SOCK_OUTSTREAM)) )
-    return freeSocket(socket);
+    return freeSocket(s);
 
   return rc;
 }
 
 
 int
-nbio_close_output(int socket)
+nbio_close_output(nbio_sock_t socket)
 { int rc = 0;
-  plsocket *s = lookupSocket(socket);
+  plsocket *s;
 
-  DEBUG(2, Sdprintf("[%d]: nbio_close_output(%d)\n", PL_thread_self(), socket));
+  if ( !(s = nbio_to_plsocket_raw(socket)) )
+    return -1;
+
+  DEBUG(2, Sdprintf("[%d]: nbio_close_output(%d, flags=0x%x)\n",
+		    PL_thread_self(), socket, s->flags));
   if ( s->output )
-  { s->output = NULL;
+  {
+#if __WINDOWS__
+    SOCKET sock;
+#endif
+
+    s->output = NULL;
     s->flags &= ~SOCK_OUTSTREAM;
 #if __WINDOWS__
-    if ( shutdown(socket, SD_SEND) == SOCKET_ERROR )
+    if ( (sock=s->socket) < 0 )
+    { s->error = WSAECONNRESET;
+      rc = -1;
+    } else if ( shutdown(sock, SD_SEND) == SOCKET_ERROR )
     {
 #ifdef O_DEBUG
       if ( debugging )
@@ -1782,7 +2131,8 @@ nbio_close_output(int socket)
 #else
         msg = strerror(errno);
 #endif
-        Sdprintf("shutdown(%d, SD_SEND) failed: %s\n", socket, msg);
+        Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
+		 socket, s->socket, msg);
       }
 #endif
       rc = -1;
@@ -1792,7 +2142,7 @@ nbio_close_output(int socket)
 
   DEBUG(3, Sdprintf("%d->flags = 0x%x\n", socket, s->flags));
   if ( !(s->flags & (SOCK_INSTREAM|SOCK_OUTSTREAM)) )
-    return freeSocket(socket);
+    return freeSocket(s);
 
   return rc;
 }
