@@ -1,4 +1,4 @@
-/*  $Id: socket.c,v 1.65 2007/01/13 17:11:34 jan Exp $
+/*  $Id: socket.c,v 1.73 2007/04/23 09:56:45 jan Exp $
 
     Part of SWI-Prolog
 
@@ -28,6 +28,14 @@
 #include <config.h>
 #endif
 
+#ifdef __GNUC__
+# define alloca __builtin_alloca
+#else
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# endif
+#endif
+
 #ifdef __CYGWIN__
 #undef HAVE_H_ERRNO
 #endif
@@ -74,9 +82,10 @@ tcp_get_socket(term_t Socket, int *id)
       return TRUE;
   }
   
-  if ( PL_get_stream_handle(Socket, &s) &&
-       (socket = Sfileno(s)) >= 0 )
-  { *id = socket;
+  if ( PL_get_stream_handle(Socket, &s) )
+  { socket = (int)(intptr_t)s->handle;
+
+    *id = socket;
     return TRUE;
   }
 
@@ -100,7 +109,7 @@ pl_close_socket(term_t socket)
     return FALSE;
 
   if ( nbio_closesocket(sock) < 0 )
-    return FALSE;
+    return nbio_error(errno, TCP_ERRNO);;
 
   return TRUE;
 }
@@ -112,9 +121,6 @@ pl_host_to_address(term_t Host, term_t Ip)
   struct hostent *host;
   char *host_name;
 
-  if ( !nbio_init("socket") )
-    return FALSE;
-
   if ( PL_get_atom_chars(Host, &host_name) )
   { if ( (host = gethostbyname(host_name)) )
     { if ( sizeof(ip) == host->h_length )
@@ -179,11 +185,11 @@ pl_setopt(term_t Socket, term_t opt)
 		 *	  IO-STREAM STUFF	*
 		 *******************************/
 
-#define fdFromHandle(p) ((int)((long)(p)))
+#define fdFromHandle(p) ((nbio_sock_t)((long)(p)))
 
 static ssize_t
 tcp_read_handle(void *handle, char *buf, size_t bufSize)
-{ int sock = fdFromHandle(handle);
+{ nbio_sock_t sock = fdFromHandle(handle);
 
   return nbio_read(sock, buf, bufSize);
 }
@@ -191,7 +197,7 @@ tcp_read_handle(void *handle, char *buf, size_t bufSize)
 
 static ssize_t
 tcp_write_handle(void *handle, char *buf, size_t bufSize)
-{ int sock = fdFromHandle(handle);
+{ nbio_sock_t sock = fdFromHandle(handle);
 
   return nbio_write(sock, buf, bufSize);
 }
@@ -205,7 +211,7 @@ tcp_seek_null(void *handle, long offset, int whence)
 
 static int
 tcp_close_input_handle(void *handle)
-{ int socket = fdFromHandle(handle);
+{ nbio_sock_t socket = fdFromHandle(handle);
 
   return nbio_close_input(socket);
 }
@@ -213,17 +219,49 @@ tcp_close_input_handle(void *handle)
 
 static int
 tcp_close_output_handle(void *handle)
-{ int socket = fdFromHandle(handle);
+{ nbio_sock_t socket = fdFromHandle(handle);
 
   return nbio_close_output(socket);
 }
 
 
+static int
+tcp_control(void *handle, int action, void *arg)
+{ nbio_sock_t socket = fdFromHandle(handle);
+
+  switch(action)
+  { case SIO_GETFILENO:
+    { SOCKET fd = nbio_fd(socket);
+      SOCKET *fdp = arg;
+      *fdp = fd;
+      return 0;
+    }
+    case SIO_LASTERROR:
+    { const char *s;
+      
+      if ( (s=nbio_last_error(socket)) )
+      { const char **sp = arg;
+	*sp = s;
+	return 0;
+      }
+
+      return -1;
+    }
+    case SIO_SETENCODING:
+    case SIO_FLUSHOUTPUT:
+      return 0;
+    default:
+      return -1;
+  }
+}
+
+
 static IOFUNCTIONS readFunctions =
 { tcp_read_handle,
   tcp_write_handle,
   tcp_seek_null,
   tcp_close_input_handle,
+  tcp_control
 };
 
 
@@ -232,6 +270,7 @@ static IOFUNCTIONS writeFunctions =
   tcp_write_handle,
   tcp_seek_null,
   tcp_close_output_handle,
+  tcp_control
 };
 
 
@@ -245,14 +284,14 @@ pl_open_socket(term_t Socket, term_t Read, term_t Write)
     return FALSE;
   handle = (void *)(long)socket;
   
-  in  = Snew(handle, SIO_FILE|SIO_INPUT|SIO_RECORDPOS,  &readFunctions);
+  in  = Snew(handle, SIO_INPUT|SIO_RECORDPOS,  &readFunctions);
   in->encoding = ENC_OCTET;
   if ( !PL_open_stream(Read, in) )
     return FALSE;
   nbio_setopt(socket, TCP_INSTREAM, in);
 
   if ( !(nbio_get_flags(socket) & SOCK_LISTEN) )
-  { out = Snew(handle, SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS, &writeFunctions);
+  { out = Snew(handle, SIO_OUTPUT|SIO_RECORDPOS, &writeFunctions);
     out->encoding = ENC_OCTET;
     if ( !PL_open_stream(Write, out) )
       return FALSE;
@@ -299,6 +338,7 @@ udp_receive(term_t Socket, term_t Data, term_t From, term_t Options)
   socklen_t alen = sizeof(sockaddr);
 #endif
   int socket;
+  SOCKET fd;
   int flags = 0;
   char buf[UDP_MAXDATA];
   ssize_t n;
@@ -307,7 +347,10 @@ udp_receive(term_t Socket, term_t Data, term_t From, term_t Options)
        !nbio_get_sockaddr(From, &sockaddr) )
     return FALSE;
   
-  if ( (n=recvfrom(socket, buf, sizeof(buf), flags,
+  if ( (fd=nbio_fd(socket)) < 0 )
+    return nbio_error(errno, TCP_ERRNO);
+
+  if ( (n=recvfrom(fd, buf, sizeof(buf), flags,
 		   (struct sockaddr*)&sockaddr, &alen)) == -1 )
     return nbio_error(errno, TCP_ERRNO);
 
@@ -327,6 +370,7 @@ udp_send(term_t Socket, term_t Data, term_t To, term_t Options)
   int alen = sizeof(sockaddr);
 #endif
   int socket;
+  SOCKET fd;
   int flags = 0L;
   char *data;
   size_t dlen;
@@ -339,7 +383,10 @@ udp_send(term_t Socket, term_t Data, term_t To, term_t Options)
        !nbio_get_sockaddr(To, &sockaddr) )
     return FALSE;
 
-  if ( (n=sendto(socket, data,
+  if ( (fd=nbio_fd(socket)) < 0 )
+    return nbio_error(errno, TCP_ERRNO);
+
+  if ( (n=sendto(fd, data,
 #ifdef __WINDOWS__
 		 (int)dlen,
 #else
@@ -361,9 +408,6 @@ static foreign_t
 create_socket(term_t socket, int type)
 { int sock;
 
-  if ( !nbio_init("socket") )
-    return FALSE;
-
   sock = nbio_socket(AF_INET, type, 0);
   if ( sock < 0 )
     return FALSE;
@@ -413,14 +457,15 @@ pl_bind(term_t Socket, term_t Address)
     return FALSE;
 
   if ( PL_is_variable(Address) )
-  { struct sockaddr_in addr;
+  { SOCKET fd = nbio_fd(socket);
+    struct sockaddr_in addr;
 #ifdef __WINDOWS__
     int len = sizeof(addr);
 #else
     socklen_t len = sizeof(addr);
 #endif
 
-    if ( getsockname(socket, (struct sockaddr *) &addr, &len) )
+    if ( getsockname(fd, (struct sockaddr *) &addr, &len) )
       return nbio_error(errno, TCP_ERRNO);
     PL_unify_integer(Address, ntohs(addr.sin_port));
   }
@@ -472,9 +517,6 @@ static foreign_t
 pl_gethostname(term_t name)
 { char buf[256];
 
-  if ( !nbio_init("socket") )
-    return FALSE;
-
   if ( gethostname(buf, sizeof(buf)) == 0 )
   { struct hostent *he;
 
@@ -510,6 +552,16 @@ findmap(fdentry *map, int fd)
 }
 
 
+static int
+is_socket_stream(IOSTREAM *s)
+{ if ( s->functions == &readFunctions /* ||
+       s->functions == &writeFunctions */ )
+    return TRUE;
+
+  return FALSE;
+}
+
+
 static foreign_t
 tcp_select(term_t Streams, term_t Available, term_t timeout)
 { fd_set fds;
@@ -527,17 +579,27 @@ tcp_select(term_t Streams, term_t Available, term_t timeout)
   FD_ZERO(&fds);
   while( PL_get_list(streams, head, streams) )
   { IOSTREAM *s;
+#ifdef __WINDOWS__
+    nbio_sock_t fd;
+#else
     int fd;
+#endif
     fdentry *e;
 
     if ( !PL_get_stream_handle(head, &s) )
       return FALSE;
-    if ( (fd=Sfileno(s)) < 0 )
-    { PL_release_stream(s);
-      return pl_error("tcp_select", 3, NULL, ERR_DOMAIN,
-		      head, "file_stream");
-    }
+
+#ifdef __WINDOWS__
+    fd = fdFromHandle(s->handle);
+#else
+    fd = Sfileno(s);
+#endif
+
     PL_release_stream(s);
+    if ( fd < 0 || !is_socket_stream(s) )
+    { return pl_error("tcp_select", 3, NULL, ERR_DOMAIN,
+		      head, "socket_stream");
+    }
 					/* check for input in buffer */
     if ( s->bufp < s->limitp )
     { if ( !PL_unify_list(available, ahead, available) ||
@@ -595,12 +657,7 @@ tcp_select(term_t Streams, term_t Available, term_t timeout)
 
     FD_ZERO(&fds);			/* EINTR may leave fds undefined */
     for(e=map; e; e=e->next)		/* so we rebuild it to be safe */
-    {
-#ifdef __WINDOWS__
-      FD_SET((SOCKET)e->fd, &fds);
-#else
-      FD_SET(e->fd, &fds);
-#endif
+    { FD_SET((SOCKET)e->fd, &fds);
     }
   }
 
@@ -611,7 +668,7 @@ tcp_select(term_t Streams, term_t Available, term_t timeout)
     case 0: /* Timeout */
       break;
 
-    default: /* Something happend -> check fds */
+    default: /* Something happened -> check fds */
       for(n=min; n <= max; n++)
       { if ( FD_ISSET(n, &fds) )
 	{ if ( !PL_unify_list(available, ahead, available) ||
@@ -641,7 +698,9 @@ pl_debug(term_t val)
 
 install_t
 install_socket()
-{ ATOM_reuseaddr  = PL_new_atom("reuseaddr");
+{ nbio_init("socket");
+
+  ATOM_reuseaddr  = PL_new_atom("reuseaddr");
   ATOM_broadcast  = PL_new_atom("broadcast");
   ATOM_dispatch   = PL_new_atom("dispatch");
   ATOM_nonblock   = PL_new_atom("nonblock");
