@@ -1,11 +1,11 @@
-/*  $Id: html_write.pl,v 1.19 2006/08/21 10:25:07 jan Exp $
+/*  $Id: html_write.pl,v 1.33 2007/04/10 21:02:53 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -30,10 +30,15 @@
 */
 
 :- module(html_write,
-	  [ page/3,			% generate an HTML page
+	  [ reply_html_page/2,		% :Head, :Body
+
+	    page/3,			% generate an HTML page
 	    page/4,			% page from head and body
 	    html/3,
 
+	    html_set_options/1,		% +OptionList
+	    html_current_option/1,	% ?Option
+
 					% Useful primitives for expanding
 	    html_begin/3,		% +EnvName[(Attribute...)]
 	    html_end/3,			% +EnvName
@@ -45,10 +50,14 @@
 	    print_html/2,		% +Stream, +List
 	    html_print_length/2		% +List, -Length
 	  ]).
+:- use_module(library(error)).
+:- use_module(library(option)).
+:- use_module(library(sgml)).		% Quote output
 :- use_module(library(quintus)).	% for meta_predicate/1
 :- set_prolog_flag(generate_debug_info, false).
 
 :- meta_predicate
+	reply_html_page(:, :),
 	html(:, -, +),
 	page(:, -, +),
 	page(:, :, -, +),
@@ -68,34 +77,154 @@ directly, but this is generally not very satisfactory:
 
 This module tries to remedy these problems.   The idea is to translate a
 Prolog term into  an  HTML  document.  We   use  DCG  for  most  of  the
-generation. 
+generation.
+
+---++ International documents
+
+The library supports the generation of international documents, but this
+is currently limited to using UTF-8 encoded HTML or XHTML documents.  It
+is strongly recommended to use the following mime-type.
+
+==
+Content-type: text/html; charset=UTF-8
+==
+
+When generating XHTML documents, the output stream must be in UTF-8
+encoding.
 */
 
+
+		 /*******************************
+		 *	      SETTINGS		*
+		 *******************************/
+
+%%	html_set_options(+Options) is det.
+%
+%	Set options for the HTML output.   Options  are stored in prolog
+%	flags to ensure  with  proper   multi-threaded  behaviour  where
+%	setting an option is local to the   thread and new threads start
+%	with the options from the parent thread.  Defined options are:
+%	
+%		* dialect(Dialect)
+%		One of =html= (default) or =xhtml=.
+%		
+%		* doctype(+DocType)
+%		Set the =|<|DOCTYPE|= DocType =|>|= line for page//1 and
+%		page//2.
+%		
+%		* content_type(+ContentType)
+%		Set the =|Content-type|= for reply_html_page/2
+%		
+%	Note  that  the  doctype  is  covered    by  two  prolog  flags:
+%	=html_doctype= for the html dialect  and =xhtml_doctype= for the
+%	xhtml dialect. Dialect muct be switched before doctype.
+
+html_set_options([]).
+html_set_options([H|T]) :-
+	html_set_option(H),
+	html_set_options(T).
+
+html_set_option(dialect(Dialect)) :- !,
+	must_be(oneof([html,xhtml]), Dialect),
+	set_prolog_flag(html_dialect, Dialect).
+html_set_option(doctype(Atom)) :- !,
+	must_be(atom, Atom),
+	(   current_prolog_flag(html_dialect, html)
+	->  set_prolog_flag(html_doctype, Atom)
+	;   set_prolog_flag(xhtml_doctype, Atom)
+	).
+html_set_option(content_type(Atom)) :- !,
+	must_be(atom, Atom),
+	(   current_prolog_flag(html_dialect, html)
+	->  set_prolog_flag(html_content_type, Atom)
+	;   set_prolog_flag(xhtml_content_type, Atom)
+	).
+html_set_option(O) :-
+	domain_error(html_option, O).
+
+
+%%	html_current_option(?Option) is nondet.
+%
+%	True if Option is an active option for the HTML generator.
+
+html_current_option(dialect(Dialect)) :-
+	current_prolog_flag(html_dialect, Dialect).
+html_current_option(doctype(DocType)) :-
+	(   current_prolog_flag(html_dialect, html)
+	->  current_prolog_flag(html_doctype, DocType)
+	;   current_prolog_flag(xhtml_doctype, DocType)
+	).
+html_current_option(content_type(ContentType)) :-
+	(   current_prolog_flag(html_dialect, html)
+	->  current_prolog_flag(html_content_type, ContentType)
+	;   current_prolog_flag(xhtml_content_type, ContentType)
+	).
+
+
+option_default(html_dialect, html).
+option_default(html_doctype, 'HTML PUBLIC "-//IETF//DTD HTML 4.0//EN"').
+option_default(xhtml_doctype,
+	       'html PUBLIC "-//W3C//DTD XHTML 1.0 \
+	       Transitional//EN" \
+	       "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"').
+option_default(html_content_type, 'text/html').
+option_default(xhtml_content_type, 'application/xhtml+xml; charset=UTF-8').
+
+%%	init_options is det.
+%
+%	Initialise the HTML processing options.
+
+init_options :-
+	(   option_default(Name, Value),
+	    (	current_prolog_flag(Name, _)
+	    ->	true
+	    ;	set_prolog_flag(Name, Value)
+	    ),
+	    fail
+	;   true
+	).
+	
+:- initialization
+	init_options.
+
+%%	xml_header(-Header)
+%
+%	First line of XHTML document.  Added by print_html/1.
+
+xml_header('<?xml version=\'1.0\' encoding=\'UTF-8\'?>').
+
+%%	ns(?Which, ?Atom)
+%
+%	Namespace declarations
+
+ns(xhtml, 'http://www.w3.org/1999/xhtml').
+
+
+		 /*******************************
+		 *	       PAGE		*
+		 *******************************/
+
 %%	page(+Content:dom)// is det.
 %%	page(+Head:dom, +Body:dom)// is det.
 %
-%	Generate a page including the HTML =|<!DOCTYPE>|= header.
+%	Generate a page including the   HTML  =|<!DOCTYPE>|= header. The
+%	actual doctype is read from the   option =doctype= as defined by
+%	html_set_options/1.
 
 page(Content) -->
-	[ '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 4.0//EN">\n',
-	  '<html>',
-	  nl(1)
+	{ html_current_option(doctype(DocType)) },
+	[ '<!DOCTYPE ', DocType, '>'
 	],
-	html(Content),
-	[ nl(1),
-	  '</html>\n'
-	].
+	html(html(Content)).
 
 page(Head, Body) -->
-	[ '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 4.0//EN">\n',
-	  '<html>',
-	  nl(1)
+	{ html_current_option(doctype(DocType)) },
+	[ '<!DOCTYPE ', DocType, '>'
 	],
+	html_begin(html),
 	pagehead(Head),
 	pagebody(Body),
-	[ nl(1),
-	  '</html>\n'
-	].
+	html_end(html).
 
 pagehead(Head) -->
 	{ functor(Head, head, _)
@@ -156,7 +285,7 @@ html(X, M) -->
 do_expand(Token, _) -->			% call user hooks
 	expand(Token), !.
 do_expand(Fmt-Args, _) --> !,
-	{ sformat(String, Fmt, Args)
+	{ format(string(String), Fmt, Args)
 	},
 	html_quoted(String).
 do_expand(\List, _) -->
@@ -173,33 +302,56 @@ do_expand(script(Content), _) --> !,	% general CDATA declared content elements?
 	],
 	html_end(script).
 do_expand(&(Entity), _) --> !,
-	{ concat_atom([&, Entity, ;], HTML)
+	{ format(string(String), '&~w;', [Entity])
 	},
-	[ HTML ].
+	[ String ].
 do_expand(Token, _) -->
 	{ atomic(Token)
 	}, !,
 	html_quoted(Token).
 do_expand(element(Env, Attributes, Contents), M) --> !,
-	html_begin(Env, Attributes),
-	html(Contents, M),
-	html_end(Env).
+	(   { Contents == [],
+	      html_current_option(dialect(xhtml))
+	    }
+	->  xhtml_empty(Env, Attributes)
+	;   html_begin(Env, Attributes),
+	    html(Contents, M),
+	    html_end(Env)
+	).
 do_expand(Term, M) -->
 	{ Term =.. [Env, Contents]
 	}, !,
 	(   { layout(Env, _, empty)
 	    }
 	->  html_begin(Env, Contents)
-	;   html_begin(Env),
+	;   (   { Contents == [],
+		  html_current_option(dialect(xhtml))
+		}
+	    ->  xhtml_empty(Env, [])
+	    ;	html_begin(Env),
+		html(Contents, M),
+		html_end(Env)
+	    )
+	).
+do_expand(Term, M) -->
+	{ Term =.. [Env, Attributes, Contents],
+	  check_non_empty(Contents, Env, Term)
+	}, !,
+	(   { Contents == [],
+	      html_current_option(dialect(xhtml))
+	    }
+	->  xhtml_empty(Env, Attributes)
+	;   html_begin(Env, Attributes),
 	    html(Contents, M),
 	    html_end(Env)
 	).
-do_expand(Term, M) -->
-	{ Term =.. [Env, Attributes, Contents]
-	}, !,
-	html_begin(Env, Attributes),
-	html(Contents, M),
-	html_end(Env).
+
+check_non_empty([], _, _) :- !.
+check_non_empty(_, Tag, Term) :-
+	layout(Tag, _, empty), !,
+	print_message(warning, format('Using empty element with content: ~p', [Term])).
+check_non_empty(_, _, _).
+
 
 %%	html_begin(+Env)// is det.
 %%	html_end(+End)// is det
@@ -228,12 +380,18 @@ html_begin(Env, Attributes) -->
 	pre_open(Env),
 	[<],
 	[Env],
-	attributes(Attributes),
-	[>],
+	attributes(Env, Attributes),
+	(   { layout(Env, _, empty),
+	      html_current_option(dialect(xhtml))
+	    }
+	->  ['/>']
+	;   [>]
+	),
 	post_open(Env).
 
 html_end(Env)   -->			% empty element or omited close
-	{ layout(Env, _, -)
+	{ layout(Env, _, -),
+	  html_current_option(dialect(html))
 	; layout(Env, _, empty)
 	}, !,
 	[].
@@ -244,6 +402,33 @@ html_end(Env)   -->
 	['>'],
 	post_close(Env).
 
+%%	xhtml_empty(+Env, +Attributes)// is det.
+%
+%	Emit element in xhtml mode with empty content.
+
+xhtml_empty(Env, Attributes) -->
+	pre_open(Env),
+	[<],
+	[Env],
+	attributes(Attributes),
+	['/>'].
+
+
+%%	attributes(+Env, +Attributes)// is det.
+%
+%	Emit attributes for Env. Adds XHTML namespace declaration to the
+%	html tag if not provided by the caller.
+
+attributes(html, L) --> 
+	{ html_current_option(dialect(xhtml)) }, !,
+	(   { option(xmlns(_), L) }
+	->  attributes(L)
+	;   { ns(xhtml, NS) },
+	    attributes([xmlns(NS)|L])
+	).
+attributes(_, L) -->
+	attributes(L).
+
 attributes([]) --> !,
 	[].
 attributes([H|T]) --> !,
@@ -253,9 +438,13 @@ attributes(One) -->
 	attribute(One).
 
 attribute(Name=Value) --> !,
-	[' ', Name, '="' ],
-	html_quoted_attribute(Value),
+	[' '], name(Name), [ '="' ],
+	attribute_value(Value),
 	['"'].
+attribute(NS:Term) --> !,
+	{ Term =.. [Name, Value]
+	}, !,
+	attribute((NS:Name)=Value).
 attribute(Term) -->
 	{ Term =.. [Name, Value]
 	}, !,
@@ -265,6 +454,26 @@ attribute(Atom) -->			% Value-abbreviated attribute
 	},
 	[ ' ', Atom ].
 
+name(NS:Name) --> !,
+	[NS, :, Name].
+name(Name) -->
+	[ Name ].
+
+%%	attribute_value(+Value)
+%
+%	Print an attribute value.  Value  is   either  atomic  or a term
+%	=|A+B|=, concatenating A and B.
+
+attribute_value(Var) -->
+	{ var(Var), !,
+	  throw(error(instantiation_error, _))
+	}.
+attribute_value(A+B) --> !,
+	attribute_value(A),
+	attribute_value(B).
+attribute_value(Value) -->
+	html_quoted_attribute(Value).
+
 
 		 /*******************************
 		 *	   QUOTING RULES	*
@@ -280,29 +489,12 @@ attribute(Atom) -->			% Value-abbreviated attribute
 %	==
 %		html(b(Text))
 %	==
+%	
+%	@tbd	Assumes UTF-8 encoding of the output.
 
 html_quoted(Text) -->
-	{ sub_atom(Text, _, _, _, <)
-	; sub_atom(Text, _, _, _, >)
-	; sub_atom(Text, _, _, _, &)
-	}, !,
-	{ atom_chars(Text, Chars),
-	  quote_chars(Chars, QuotedChars),
-	  concat_atom(QuotedChars, Quoted)
-	},
+	{ xml_quote_cdata(Text, Quoted, utf8) },
 	[ Quoted ].
-html_quoted(Text) -->
-	[ Text ].
-
-quote_chars([], []).
-quote_chars([H0|T0], [H|T]) :-
-	quote_char(H0, H),
-	quote_chars(T0, T).
-
-quote_char(<, '&lt;') :- !.
-quote_char(>, '&gt;') :- !.
-quote_char(&, '&amp;') :- !.
-quote_char(X, X).
 
 %%	html_quoted_attribute(+Text)// is det.
 %
@@ -310,32 +502,12 @@ quote_char(X, X).
 %	included in double-quotes.  Note   that  -like  html_quoted//1-,
 %	attributed   values   printed   through   html//1   are   quoted
 %	atomatically.
+%	
+%	@tbd	Assumes UTF-8 encoding of the output.
 
 html_quoted_attribute(Text) -->
-	{ sub_atom(Text, _, _, _, <)
-	; sub_atom(Text, _, _, _, >)
-	; sub_atom(Text, _, _, _, '"')
-	; sub_atom(Text, _, _, _, '&')
-	}, !,
-	{ atom_chars(Text, Chars),
-	  quote_att_chars(Chars, QuotedChars),
-	  concat_atom(QuotedChars, Quoted)
-	},
+	{ xml_quote_attribute(Text, Quoted, utf8) },
 	[ Quoted ].
-html_quoted_attribute(Text) -->
-	[ Text ].
-	
-quote_att_chars([], []).
-quote_att_chars([H0|T0], [H|T]) :-
-	quote_att_char(H0, H),
-	quote_att_chars(T0, T).
-
-quote_att_char(<, '&lt;') :- !.
-quote_att_char(>, '&gt;') :- !.
-quote_att_char(&, '&amp;') :- !.
-quote_att_char('"', '&quot;') :- !.
-%quote_att_char('''', '&apos;') :- !.
-quote_att_char(X, X).
 
 
 		 /*******************************
@@ -381,6 +553,8 @@ post_close(_) -->
 %	@param Close	Either as Open, or the atom - (minus) to imit the
 %			close-tag or =empty= to indicate the element has
 %			no content model.
+%			
+% 	@tbd	Complete table
 
 :- multifile
 	layout/3.
@@ -401,6 +575,8 @@ layout(script,	   1-1,	1-1).
 layout(select,	   1-1,	1-1).
 layout(map,	   1-1,	1-1).
 layout(html,	   1-1,	1-1).
+layout(caption,	   1-1,	1-1).
+layout(applet,	   1-1,	1-1).
 
 layout(tr,	   1-0,	0-1).
 layout(option,	   1-0,	0-1).
@@ -425,12 +601,12 @@ layout(frame,	   1-1, empty).
 layout(col,	   0-0, empty).
 layout(area,	   1-0, empty).
 layout(input,	   1-0, empty).
-layout(option,	   1-0, empty).
 layout(param,	   1-0, empty).
 
 layout(p,	   2-1, -).		% omited close
 layout(td,	   0-0, 0-0).
 
+layout(div,	   1-0,	0-1).
 
 		 /*******************************
 		 *	     PRINTING		*
@@ -449,6 +625,16 @@ print_html(List) :-
 	current_output(Out),
 	write_html(List, Out).
 print_html(Out, List) :-
+	(   html_current_option(dialect(xhtml))
+	->  stream_property(Out, encoding(Enc)),
+	    (	Enc == utf8
+	    ->	true
+	    ;	print_message(warning, html(wrong_encoding(Out, Enc)))
+	    ),
+	    xml_header(Hdr),
+	    write(Out, Hdr), nl(Out)
+	;   true
+	),
 	write_html(List, Out).
 
 write_html([], _).
@@ -480,13 +666,19 @@ write_nl(N, Out) :-
 %	==
 %		phrase(html(DOM), Tokens),
 %		html_print_length(Tokens, Len),
-%		format('Content-type: text/html~n'),
+%		format('Content-type: text/html; charset=UTF-8~n'),
 %		format('Content-length: ~d~n~n', [Len]),
 %		print_html(Tokens)
 %	==
 
 html_print_length(List, Len) :-
-	html_print_length(List, 0, Len).
+	(   html_current_option(dialect(xhtml))
+	->  xml_header(Hdr),
+	    atom_length(Hdr, L0),
+	    L1 is L0+1			% one for newline
+	;   L1 = 0
+	),
+	html_print_length(List, L1, Len).
 
 html_print_length([], L, L).
 html_print_length([nl(N)|T], L0, L) :- !,
@@ -499,6 +691,19 @@ html_print_length([H|T], L0, L) :-
 	html_print_length(T, L1, L).
 
 
+%%	reply_html_page(:Head, :Body) is det.
+%
+%	Provide the complete reply as required  by http_wrapper.pl for a
+%	page constructed from Head and   Body. The HTTP =|Content-type|=
+%	is provided by html_current_option/1.
+
+reply_html_page(Head, Body) :-
+	html_current_option(content_type(Type)),
+	phrase(page(Head, Body), HTML),
+	format('Content-type: ~w~n~n', [Type]),
+	print_html(HTML).
+
+
 		 /*******************************
 		 *	PCE EMACS SUPPORT	*
 		 *******************************/
@@ -525,6 +730,10 @@ emacs_prolog_colours:goal_colours(pagehead(HTML,_,_),
 emacs_prolog_colours:goal_colours(pagebody(HTML,_,_),
 				  built_in-[Colours, classify, classify]) :-
 	html_colours(HTML, Colours).
+emacs_prolog_colours:goal_colours(reply_html_page(Head, Body),
+				  built_in-[HC, BC]) :-
+	html_colours(Head, HC),
+	html_colours(Body, BC).
 
 
 					% TBD: Check with do_expand!
@@ -571,6 +780,9 @@ attr_colours(Term, list-Elements) :-
 	Term = [_|_], !,
 	attr_list_colours(Term, Elements).
 attr_colours(Name=_, built_in-[html_attribute(Name), classify]) :- !.
+attr_colours(NS:Term, built_in-[html_xmlns(NS), html_attribute(Name)-[classify]]) :-
+	compound(Term),
+	Term =.. [Name,_], !.
 attr_colours(Term, html_attribute(Name)-[classify]) :-
 	compound(Term),
 	Term =.. [Name,_], !.
@@ -586,18 +798,22 @@ attr_list_colours([H0|T0], [H|T]) :-
 	attr_list_colours(T0, T).
 
 
+:- op(990, xfx, :=).			% allow compiling without XPCE
+:- op(200, fy, @).
+
 emacs_prolog_colours:style(html(_), style(bold := @on,
 					  colour := magenta4)).
 emacs_prolog_colours:style(entity(_), style(colour := magenta4)).
 emacs_prolog_colours:style(html_attribute(_), style(colour := magenta4)).
+emacs_prolog_colours:style(html_xmlns(_), style(colour := magenta4)).
 
 
 emacs_prolog_colours:identify(html(Element), Summary) :-
-	sformat(Summary, '~w: SGML element', [Element]).
+	format(string(Summary), '~w: SGML element', [Element]).
 emacs_prolog_colours:identify(entity(Entity), Summary) :-
-	sformat(Summary, '~w: SGML entity', [Entity]).
+	format(string(Summary), '~w: SGML entity', [Entity]).
 emacs_prolog_colours:identify(html_attribute(Attr), Summary) :-
-	sformat(Summary, '~w: SGML attribute', [Attr]).
+	format(string(Summary), '~w: SGML attribute', [Attr]).
 
 
 %	prolog:called_by(+Goal, -Called)
@@ -616,6 +832,8 @@ prolog:called_by(pagehead(HTML,_,_), Called) :-
 	phrase(called_by(HTML), Called).
 prolog:called_by(pagebody(HTML,_,_), Called) :-
 	phrase(called_by(HTML), Called).
+prolog:called_by(reply_html_page(Head,Body), Called) :-
+	phrase(called_by([Head,Body]), Called).
 
 called_by(Var) -->
 	{ var(Var) }, !,
@@ -648,3 +866,5 @@ called_by(_) -->
 
 prolog:message(html(expand_failed(What))) -->
 	[ 'Failed to translate to HTML: ~p'-[What] ].
+prolog:message(html(wrong_encoding(Stream, Enc))) -->
+	[ 'XHTML demands UTF-8 encoding; encoding of ~p is ~w'-[Stream, Enc] ].
