@@ -1,11 +1,11 @@
-/*  $Id: http_client.pl,v 1.14 2006/08/18 09:27:25 jan Exp $
+/*  $Id: http_client.pl,v 1.18 2007/06/21 07:03:49 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -39,6 +39,7 @@
 :- use_module(library(socket)).
 :- use_module(library(url)).
 :- use_module(http_header).
+:- use_module(http_chunked).
 :- use_module(library(debug)).
 :- use_module(library(memfile)).
 :- use_module(library(lists)).
@@ -183,14 +184,17 @@ http_read_reply(In, Data, Options) :-
 	between(0, 1, _),
 	    http_read_reply_header(In, Fields),
 	\+ memberchk(status(continue, _), Fields), !,
-	(   memberchk(status(302, _), Fields),
-	    memberchk(location(Location), Fields)
+	(   memberchk(location(Location), Fields),
+	    (   memberchk(status(moved, _), Fields)
+	    ;	memberchk(status(moved_temporary, _), Fields)
+	    ;	memberchk(status(see_other, _), Fields)
+	    )
 	->  Data = redirect(Location)
-	;   (   memberchk(reply_header(Fields), Options)
+	;   (   select(reply_header(Fields), Options, ReadOptions)
 	    ->  true
-	    ;   true
+	    ;   ReadOptions = Options
 	    ),
-	    http_read_data(In, Fields, Data, Options)
+	    http_read_data(In, Fields, Data, ReadOptions)
 	),
 	(   memberchk(connection(Connection), Fields),
 	    downcase_atom(Connection, 'keep-alive')
@@ -278,28 +282,9 @@ http_read_data(Fields, Data, Options) :-
 
 http_read_data(In, Fields, Data, Options) :-	% Transfer-encoding: chunked
 	select(transfer_encoding(chunked), Fields, RestFields), !,
-	(    memberchk(to(stream(Out)), Options)
-	->   copy_chunk_data(In, Out, _Foot)
-	;    new_memory_file(MemFile),
-	     open_memory_file(MemFile, write, Stream),
-	     copy_chunk_data(In, Stream, Foot),
-	     close(Stream),
-	     (	 memberchk(to(atom), Options)
-	     ->	 memory_file_to_atom(MemFile, Data0),
-		 free_memory_file(MemFile),
-		 Data = Data0
-	     ;	 memberchk(to(codes), Options)
-	     ->	 memory_file_to_codes(MemFile, Data0),
-		 free_memory_file(MemFile),
-		 Data = Data0
-	     ;   size_memory_file(MemFile, Length),
-		 append([content_length(Length)|RestFields], Foot, NewFields),
-		 open_memory_file(MemFile, read, NewIn),
-		 http_read_data(NewIn, NewFields, Data, Options),
-		 close(NewIn),
-		 free_memory_file(MemFile)
-	     )
-	).
+	http_chunked_open(In, DataStream, []),
+	call_cleanup(http_read_data(DataStream, RestFields, Data, Options),
+		     close(DataStream)).
 http_read_data(In, Fields, Data, Options) :-
 	memberchk(to(X), Options), !,
 	(   X = stream(Stream)
@@ -337,21 +322,6 @@ http_read_data(In, Fields, Data, Options) :-
 	http_read_data(In, Fields, Data, [to(atom)|Options]).
 
 
-%%	copy_chunk_data(+In, +Out, -Footer)
-%
-%	Copy chunk-encoded data (HTTP/1.1, RFC rfc2068) to a new stream
-%	and return the foot.
-
-copy_chunk_data(In, Out, Foot) :-
-	read_line_to_codes(In, Codes),
-	phrase(xinteger(Size), Codes, _), !, 	% _: skip extra header
-	(   Size == 0
-	->  http_read_header(In, Foot)
-	;   copy_stream_data(In, Out, Size),
-	    read_line_to_codes(In, ""),		% skip empty line
-	    copy_chunk_data(In, Out, Foot)
-	).
-
 		 /*******************************
 		 *	       POST		*
 		 *******************************/
