@@ -1,11 +1,11 @@
-/*  $Id: http_header.pl,v 1.22 2006/08/21 09:34:44 jan Exp $
+/*  $Id: http_header.pl,v 1.33 2007/04/10 20:40:48 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -66,8 +66,8 @@
 http_read_request(In, [input(In)|Request]) :-
 	read_line_to_codes(In, Codes),
 	(   Codes == end_of_file
-	->  debug(header, 'end-of-file', [])
-	;   debug(header, 'First line: ~s~n', [Codes]),
+	->  debug(http(header), 'end-of-file', [])
+	;   debug(http(header), 'First line: ~s~n', [Codes]),
 	    phrase(request(In, Request), Codes)
 	).
 
@@ -80,9 +80,9 @@ http_read_request(In, [input(In)|Request]) :-
 http_read_reply_header(In, [input(In)|Reply]) :-
 	read_line_to_codes(In, Codes),
 	(   Codes == end_of_file
-	->  debug(header, 'end-of-file', []),
+	->  debug(http(header), 'end-of-file', []),
 	    throw(error(syntax(http_reply_header, end_of_file), _))
-	;   debug(header, 'First line: ~s~n', [Codes]),
+	;   debug(http(header), 'First line: ~s~n', [Codes]),
 	    (   phrase(reply(In, Reply), Codes)
 	    ->  true
 	    ;   throw(error(syntax(http_reply_header, Codes), _))
@@ -162,11 +162,37 @@ http_reply(moved(To), Out, HrdExtra) :- !,
 	phrase(reply_header(moved(To, HTML), HrdExtra), Header),
 	format(Out, '~s', [Header]),
 	print_html(Out, HTML).
+http_reply(moved_temporary(To), Out, HrdExtra) :- !,
+	phrase(page([ title('302 Moved Temporary')
+		    ],
+		    [ h1('Moved Temporary'),
+		      p(['The document is currently ',
+			 a(href(To), ' Here')
+			]),
+		      address(httpd)
+		    ]),
+	       HTML),
+	phrase(reply_header(moved_temporary(To, HTML), HrdExtra), Header),
+	format(Out, '~s', [Header]),
+	print_html(Out, HTML).
+http_reply(see_other(To),Out,HdrExtra) :- !,
+       phrase(page([ title('303 See Other')
+                    ],
+                    [ h1('See Other'),
+                      p(['See other document ',
+                         a(href(To), ' Here')
+                        ]),
+                      address(httpd)
+                    ]),
+               HTML),
+        phrase(reply_header(see_other(To, HTML), HdrExtra), Header),
+        format(Out, '~s', [Header]),
+        print_html(Out, HTML).
 http_reply(not_found(URL), Out, HrdExtra) :- !,
 	phrase(page([ title('404 Not Found')
 		    ],
 		    [ h1('Not Found'),
-		      p(['The requested URL ', URL,
+		      p(['The requested URL ', tt(URL),
 			 ' was not found on this server'
 			]),
 		      address(httpd)
@@ -238,12 +264,12 @@ html_message_lines([nl|T]) --> !,
 html_message_lines([flush]) -->
 	[].
 html_message_lines([Fmt-Args|T]) --> !,
-	{ sformat(S, Fmt, Args)
+	{ format(string(S), Fmt, Args)
 	},
 	html([S]),
 	html_message_lines(T).
 html_message_lines([Fmt|T]) --> !,
-	{ sformat(S, Fmt, [])
+	{ format(string(S), Fmt, [])
 	},
 	html([S]),
 	html_message_lines(T).
@@ -276,7 +302,7 @@ http_update_encoding(Header0, utf8, [content_type(Type)|Header]) :-
 	select(content_type(Type0), Header0, Header),
 	sub_atom(Type0, 0, _, _, 'text/'), !,
 	(   sub_atom(Type0, S, _, _, ';')
-	->  sub_atom(Type0, 0, B, _, S)
+	->  sub_atom(Type0, 0, S, _, B)
 	;   B = Type0
 	),
 	atom_concat(B, '; charset=UTF-8', Type).
@@ -317,7 +343,45 @@ content_length_in_encoding(Enc, Stream, Bytes) :-
 
 %%	http_post_data(+Data, +Out:stream, +HdrExtra) is det.
 %
-%	@tbd	Complete docs
+%	Send data on behalf on an HTTP   POST request. This predicate is
+%	normally called by http_post/4 from   http_client.pl to send the
+%	POST data to the server.  Data is one of:
+%	
+%	  * html(+Tokens)
+%	  Result of html//1 from html_write.pl
+%	  
+%	  * file(+File)
+%	  Send contents of a file. Mime-type is determined by
+%	  file_mime_type/2.
+%	  
+%	  * file(+Type, +File)
+%	  Send file with content of indicated mime-type.
+%	  
+%	  * codes(+Codes)
+%	  As string(text/plain, Codes).
+%	  
+%	  * codes(+Type, +Codes)
+%	  Send Codes using the indicated MIME-type.
+%	  
+%	  * cgi_stream(+Stream, +Len)
+%	  Read the input from Stream which, like CGI data starts with a partial
+%	  HTTP header. The fields of this header are merged with the provided
+%	  HdrExtra fields. The first Len characters of Stream are used.
+%	  
+%	  * form(+ListOfParameter)
+%	  Send data of the MIME type application/x-www-form-urlencoded as
+% 	  produced by browsers issuing a POST request from an HTML form.
+%	  ListOfParameter is a list of Name=Value or Name(Value).
+%	  
+%	  * form_data(+ListOfData)
+%	  Send data of the MIME type multipart/form-data.  ListOfData is the same
+%	  as for the List alternative described below.
+%	  
+%	  * List
+%	  If the argument is a plain list, it is sent using the MIME type
+%	  multipart/mixed and packed using mime_pack/3. See mime_pack/3
+%	  for details on the argument format.
+
 
 http_post_data(html(HTML), Out, HdrExtra) :-
 	phrase(post_header(html(HTML), HdrExtra), Header),
@@ -335,6 +399,11 @@ http_post_data(file(Type, File), Out, HdrExtra) :- !,
 	open(File, read, In, [type(binary)]),
 	call_cleanup(copy_stream_data(In, Out),
 		     close(In)).
+http_post_data(codes(Codes), Out, HdrExtra) :- !,
+	http_post_data(codes(text/plain, Codes), Out, HdrExtra).
+http_post_data(codes(Type, Codes), Out, HdrExtra) :- !,
+	phrase(post_header(codes(Type, Codes), HdrExtra), Header),
+	format(Out, '~s~s', [Header, Codes]).
 http_post_data(cgi_stream(In, _Len), Out, HdrExtra) :- !,
 	debug(obsolete, 'Obsolete 2nd argument in cgi_stream(In,Len)', []),
 	http_post_data(cgi_stream(In), Out, HdrExtra).
@@ -363,7 +432,7 @@ http_post_data(form_data(Data), Out, HdrExtra) :- !,
 	mime_pack(Data, MimeOut, Boundary),
 	close(MimeOut),
 	size_memory_file(MemFile, Size),
-	sformat(ContentType, 'multipart/form-data; boundary=~w', [Boundary]),
+	format(string(ContentType), 'multipart/form-data; boundary=~w', [Boundary]),
 	http_join_headers(HdrExtra,
 			  [ mime_version('1.0'),
 			    content_type(ContentType)
@@ -381,7 +450,7 @@ http_post_data(List, Out, HdrExtra) :-		% multipart-mixed
 	mime_pack(List, MimeOut, Boundary),
 	close(MimeOut),
 	size_memory_file(MemFile, Size),
-	sformat(ContentType, 'multipart/mixed; boundary=~w', [Boundary]),
+	format(string(ContentType), 'multipart/mixed; boundary=~w', [Boundary]),
 	http_join_headers(HdrExtra,
 			  [ mime_version('1.0'),
 			    content_type(ContentType)
@@ -408,6 +477,11 @@ post_header(file(Type, File), HdrExtra) -->
 post_header(cgi_data(Size), HdrExtra) -->
 	header_fields(HdrExtra),
 	content_length(Size),
+	"\r\n". 
+post_header(codes(Type, Codes), HdrExtra) -->
+	header_fields(HdrExtra),
+	content_length(ascii_string(Codes)),
+	content_type(Type),
 	"\r\n". 
 
 
@@ -460,6 +534,22 @@ reply_header(moved(To, Tokens), HdrExtra) -->
 	content_length(html(Tokens)),
 	content_type(text/html),
 	"\r\n".
+reply_header(moved_temporary(To, Tokens), HdrExtra) -->
+	vstatus(moved_temporary),
+	date(now),
+	header_field('Location', To),
+	header_fields(HdrExtra),
+	content_length(html(Tokens)),
+	content_type(text/html),
+	"\r\n".
+reply_header(see_other(To,Tokens),HdrExtra) -->
+      vstatus(see_other),
+      date(now),
+      header_field('Location',To),
+      header_fields(HdrExtra),
+      content_length(html(Tokens)),
+      content_type(text/html),
+      "\r\n".
 reply_header(status(Status, Tokens), HdrExtra) -->
 	vstatus(Status),
 	date(now),
@@ -483,14 +573,16 @@ vstatus(Status) -->
 	status_comment(Status),
 	"\r\n".
 
-status_number(continue)	    --> "100".
-status_number(ok)	    --> "200".
-status_number(moved)	    --> "301".
-status_number(not_modified) --> "304". 
-status_number(not_found)    --> "404".
-status_number(forbidden)    --> "403".
-status_number(authorise)    --> "401".
-status_number(server_error) --> "500".
+status_number(continue)	       --> "100".
+status_number(ok)	       --> "200".
+status_number(moved)	       --> "301".
+status_number(moved_temporary) --> "302".
+status_number(see_other)       --> "303".
+status_number(not_modified)    --> "304". 
+status_number(not_found)       --> "404".
+status_number(forbidden)       --> "403".
+status_number(authorise)       --> "401".
+status_number(server_error)    --> "500".
 
 status_comment(continue) -->
 	"Continue".
@@ -498,6 +590,10 @@ status_comment(ok) -->
 	"OK".
 status_comment(moved) -->
 	"Moved Permanently".
+status_comment(moved_temporary) -->
+	"Moved Temporary".
+status_comment(see_other) -->
+      "See Other".
 status_comment(not_modified) --> 
 	"Not Modified".
 status_comment(not_found) -->
@@ -514,7 +610,7 @@ authenticate(Method, '') --> !,
 	atom(Method).
 authenticate(Method, Realm) -->
 	authenticate(Method, ''),
-	"Realm=\"", atom(Realm), "\"".
+	" Realm=\"", atom(Realm), "\"\r\n".
 
 date(Time) -->
 	"Date: ",
@@ -691,7 +787,9 @@ now -->
 
 rfc_date(Time, String, Tail) :-
 	stamp_date_time(Time, Date, 'UTC'),
-	format_time(codes(String, Tail), '%a, %d %b %Y %H:%M:%S GMT', Date).
+	format_time(codes(String, Tail),
+		    '%a, %d %b %Y %H:%M:%S GMT',
+		    Date, posix).
 	
 %%	http_timestamp(+Time, -Atom)
 %	
@@ -699,7 +797,9 @@ rfc_date(Time, String, Tail) :-
 
 http_timestamp(Time, Atom) :-
 	stamp_date_time(Time, Date, 'UTC'),
-	format_time(atom(Atom), '%a, %d %b %Y %H:%M:%S GMT', Date).
+	format_time(codes(Atom),
+		    '%a, %d %b %Y %H:%M:%S GMT',
+		    Date, posix).
 
 
 		 /*******************************
@@ -873,7 +973,7 @@ http_read_header(Fd, Header) :-
 read_header_data(Fd, Header) :-
 	read_line_to_codes(Fd, Header, Tail),
 	read_header_data(Header, Fd, Tail),
-	debug(header, 'Header = ~n~s~n', [Header]).
+	debug(http(header), 'Header = ~n~s~n', [Header]).
 
 read_header_data("\r\n", _, _) :- !.
 read_header_data("\n", _, _) :- !.
@@ -884,7 +984,7 @@ read_header_data(_, Fd, Tail) :-
 
 parse_header(Text, Header) :-
 	phrase(header(Header), Text),
-	debug(header, 'Fields: ~w~n', [Header]).
+	debug(http(header), 'Fields: ~w~n', [Header]).
 
 header(List) -->
 	header_field(Name, Value), !,
