@@ -1,4 +1,4 @@
-/*  $Id: doc_html.pl,v 1.13 2006/08/29 20:28:01 jan Exp $
+/*  $Id: doc_html.pl,v 1.22 2007/05/24 14:34:23 jan Exp $
 
     Part of SWI-Prolog
 
@@ -32,6 +32,7 @@
 :- module(pldoc_html,
 	  [ doc_for_file/3,		% +FileSpec, +Out, +Options
 	    doc_write_html/3,		% +Stream, +Title, +Term
+	    doc_for_wiki_file/3,	% +FileSpec, +Out, +Options
 	    				% Support doc_index
 	    doc_page_dom/3,		% +Title, +Body, -DOM
 	    print_html_head/1,		% +Stream
@@ -40,11 +41,13 @@
 	    module_info/3,		% +File, +Options0, -Options
 	    doc_hide_private/3,		% +Doc0, -Doc, +Options
 	    edit_button/4,		% +File, +Options, //
+	    source_button/4,		% +File, +Options, //
 	    pred_edit_button/4,		% +PredInd, +Options, //
 	    object_edit_button/4,	% +Obj, +Options, //
 
 	    private/2,			% +Obj, +Options
 	    file/3,			% +File, //
+	    image/3,			% +File, //
 	    tags/3,			% +Tags, //
 	    file_header/4,		% +File, +Options, //
 	    objects/4,			% +Objects, +Options, //
@@ -55,7 +58,10 @@
 :- use_module(library(lists)).
 :- use_module(library(option)).
 :- use_module(library(url)).
+:- use_module(library(readutil)).
 :- use_module(library('http/html_write')).
+:- use_module(library(doc_http)).
+:- use_module(library(debug)).
 :- use_module(doc_process).
 :- use_module(doc_modes).
 :- use_module(doc_wiki).
@@ -66,7 +72,7 @@
 /** <module> PlDoc HTML backend
 
 This  module  translates  the  Herbrand   term  from  the  documentation
-extracting module wiki.pl into HTML+CSS.
+extracting module doc_wiki.pl into HTML+CSS.
 
 @tbd	Split put generation from computation as computation is reusable
 	in other backends.
@@ -234,6 +240,7 @@ file_title(Title, File, Options) -->
 		[ span(style('float:right'),
 		       [ \reload_button(Base, Options),
 			 \zoom_button(Base, Options),
+			 \source_button(Base, Options),
 			 \edit_button(File, Options)
 		       ])
 		| Title
@@ -249,14 +256,15 @@ reload_button(Base, Options) -->
 	{ option(edit(true), Options), !,
 	  option(public_only(Public), Options, true),
 	  format(string(HREF), '~w?reload=true&public_only=~w', 
-		 [Base, Public])
+		 [Base, Public]),
+	  doc_server_root(Root)
 	},
 	html(a(href=HREF,
 	       img([ %class(icon),
 		     height=24,
 		     alt('Reload'),
 		     style('padding-top:4px; border:0;'),
-		     src('/reload.gif')
+		     src(Root+'reload.gif')
 		   ]))).
 reload_button(_, _) -->
 	[].
@@ -271,39 +279,67 @@ edit_button(File, Options) -->
 	{ option(edit(true), Options), !,
 	  option(button_height(H), Options, 24),
 	  www_form_encode(File, Enc),
-	  format(string(HREF), '/edit?file=~w', [Enc]),
+	  doc_server_root(Root),
+	  format(string(HREF), '~wedit?file=~w', [Root, Enc]),
 	  format(string(OnClick), 'HTTPrequest("~w")', [HREF])
 	},
 	html(a([ onClick(OnClick),
 		 onMouseOver('window.status=\'Edit file\'; return true;')
 	       ],
-	       img([ height=H,
+	       img([ height(H),
 		     alt(edit),
 		     style('border:0'),
-		     src='/edit.gif'
+		     src(Root+'edit.gif')
 		 ]))).
 edit_button(_, _) -->
 	[].
 
 
-%%	zoom_button(+Options)// is det.
+%%	zoom_button(BaseName, +Options)// is det.
 %
 %	Add zoom in/out button to show/hide the private documentation.
 
+zoom_button(_, Options) -->
+	{ option(files(_Map), Options) }, !.	% generating files
 zoom_button(Base, Options) -->
-	{   option(public_only(true), Options, true)
-	->  format(string(HREF), '~w?public_only=false', [Base]),
-	    Zoom = '/zoomin.gif',
-	    Alt = 'Show all'
-	;   format(string(HREF), '~w?public_only=true', [Base]),
-	    Zoom = '/zoomout.gif',
-	    Alt = 'Show public'
+	{   (   option(public_only(true), Options, true)
+	    ->  format(string(HREF), '~w?public_only=false', [Base]),
+		Zoom = 'zoomin.gif',
+		Alt = 'Show all'
+	    ;   format(string(HREF), '~w?public_only=true', [Base]),
+		Zoom = 'zoomout.gif',
+		Alt = 'Show public'
+	    ),
+	    doc_server_root(Root)
 	},
 	html(a(href=HREF,
 	       img([ height=24,
 		     alt(Alt),
 		     style('padding-top:4px; border:0;'),
-		     src(Zoom)
+		     src(Root+Zoom)
+		   ]))).
+	
+
+%%	source_button(+File, +Options)// is det.
+%
+%	Add show-source button.
+
+source_button(_File, Options) -->
+	{ option(files(_Map), Options) }, !.	% generating files
+source_button(File, Options) -->
+	{ (   is_absolute_file_name(File)
+	  ->  doc_file_href(File, HREF0),
+	      atom_concat(HREF0, '?source=true', HREF)
+	  ;   format(string(HREF), '~w?source=true', [File])
+	  ),
+	  option(button_height(H), Options, 24),
+	  doc_server_root(Root)
+	},
+	html(a(href=HREF,
+	       img([ height(H),
+		     alt('Show source'),
+		     style('padding-top:4px; border:0;'),
+		     src(Root+'source.gif')
 		   ]))).
 	
 
@@ -460,7 +496,8 @@ object_page(Obj, Options) -->
 	prolog:doc_object_page(Obj, Options).
 object_page(Obj, Options) -->
 	{ doc_comment(Obj, File:_Line, _Summary, _Comment),
-	  format(string(FileRef), '/doc~w', [File])
+	  doc_server_root(Root),
+	  format(string(FileRef), '~wdoc~w', [Root, File])
 	},
 	html([ div(class(navhdr),
 		   [ span(style('float:left'), a(href(FileRef), File)),
@@ -496,12 +533,13 @@ doc_write_html(Out, Title, Doc) :-
 %	links to the style-sheet and javaScript files.
 
 doc_page_dom(Title, Body, DOM) :-
+	doc_server_root(Root),
 	DOM = html([ head([ title(Title),
 			    link([ rel(stylesheet),
 				   type('text/css'),
 				   href('pldoc.css')
 				 ]),
-			    script([ src('/pldoc.js'),
+			    script([ src(Root+'pldoc.js'),
 				     type('text/javascript')
 				   ], [])
 			  ]),
@@ -532,7 +570,7 @@ tags(Tags) -->
 
 %%	tag(+Tag, +Value)// is det.
 %
-%	Called from \tag(Name,Value) terms produced by wiki.pl.
+%	Called from \tag(Name, Value) terms produced by doc_wiki.pl.
 
 tag(Tag, Value) -->
 	{   tag_title(Tag, Title)
@@ -551,12 +589,12 @@ tag_title(see,    'See also:').
 
 tag_class(tbd, 		warn).
 tag_class(bug, 		error).
-tag_class(depreciated,	warn).
+tag_class(deprecated,	warn).
 
 %%	params(+Params:list) is det.
 %
-%	Called from \params(List) created by wiki.pl.   Params is a list
-%	of param(Name, Descr).
+%	Called from \params(List) created by   doc_wiki.pl.  Params is a
+%	list of param(Name, Descr).
 
 params(Params) -->
 	html([ dt(class=tag, 'Parameters:'),
@@ -661,15 +699,16 @@ pred_edit_button(Name/Arity, Options) -->
 	     )
 	}, !.
 pred_edit_button(Name/Arity, Options) -->
-	{ www_form_encode(Name, QName),
+	{ doc_server_root(Root),
+	  www_form_encode(Name, QName),
 	  (   option(module(M), Options)
 	  ->  www_form_encode(M, QM),
 	      format(string(OnClick),
-		     'HTTPrequest("/edit?name=~w&arity=~w&module=~w")',
-		     [QName, Arity, QM])
+		     'HTTPrequest("~wedit?name=~w&arity=~w&module=~w")',
+		     [Root, QName, Arity, QM])
 	  ;   format(string(OnClick),
-		     'HTTPrequest("/edit?name=~w&arity=~w")',
-		     [QName, Arity])
+		     'HTTPrequest("~wedit?name=~w&arity=~w")',
+		     [Root, QName, Arity])
 	  )
 	},
 	html(span(style('float:right'),
@@ -677,7 +716,7 @@ pred_edit_button(Name/Arity, Options) -->
 		    ],
 		    img([ height=14,
 			  style('border:0;'),
-			  src='/edit.gif'
+			  src(Root+'edit.gif')
 			])))).
 pred_edit_button(_/_, _) --> !,
 	[].
@@ -743,10 +782,10 @@ pred_head(Head) -->			% Plain terms
 		   [ '(', \pred_args(Args, 1), ')' ])
 	     ]).
 
-op_type(xf,  prefix).
-op_type(yf,  prefix).
-op_type(fx,  postfix).
-op_type(fy,  postfix).
+op_type(fx,  prefix).
+op_type(fy,  prefix).
+op_type(xf,  postfix).
+op_type(yf,  postfix).
 op_type(xfx, infix).
 op_type(xfy, infix).
 op_type(yfx, infix).
@@ -802,7 +841,7 @@ pred_det(Det) -->
 
 %%	term(+Term, +Bindings)// is det.
 %
-%	Process the \term element as produced by wiki.pl.
+%	Process the \term element as produced by doc_wiki.pl.
 %	
 %	@tbd	Properly merge with pred_head//1
 
@@ -835,18 +874,17 @@ term(Term, Bindings) -->
 %	/doc/<file>#anchor.  Such links only work in the online browser.
 
 predref(Term) -->
-	predref(Term, []).
+	{ catch(nb_getval(pldoc_options, Options), _, Options = []) },
+	predref(Term, Options).
 
 predref(M:Term, Options) --> !,
 	predref(Term, M, Options).
 predref(Term, Options) -->
 	predref(Term, _, Options).
 
-predref(Name/Arity, _, _Options) -->		% Builtin; cannot be overruled
+predref(Name/Arity, _, Options) -->		% Builtin; cannot be overruled
 	{ prolog:doc_object_summary(Name/Arity, manual, _, _), !,
-	  format(string(FragmentId), '~w/~d', [Name, Arity]),
-	  www_form_encode(FragmentId, EncId),
-	  format(string(HREF), '/man?predicate=~w', [EncId])
+	  manref(Name/Arity, HREF, Options)
 	},
 	html(a([class=builtin, href=HREF], [Name, /, Arity])).
 predref(Obj, Module, Options) -->		% Local
@@ -861,11 +899,9 @@ predref(Name//Arity, Module, _Options) -->
 	  pred_href(Name/PredArity, Module, HREF)
 	}, !,
 	html(a(href=HREF, [Name, //, Arity])).
-predref(Name/Arity, _, _Options) -->		% From packages
+predref(Name/Arity, _, Options) -->		% From packages
 	{ prolog:doc_object_summary(Name/Arity, Category, _, _), !,
-	  format(string(FragmentId), '~w/~d', [Name, Arity]),
-	  www_form_encode(FragmentId, EncId),
-	  format(string(HREF), '/man?predicate=~w', [EncId])
+	  manref(Name/Arity, HREF, Options)
 	},
 	html(a([class=Category, href=HREF], [Name, /, Arity])).
 predref(Name/Arity, _, _Options) --> !,
@@ -879,6 +915,21 @@ predref(Callable, Module, Options) -->
 	predref(Name/Arity, Module, Options).
 
 
+%%	manref(+NameArity, -HREF, +Options) is det.
+%
+%	Create reference to a manual page.
+
+manref(Name/Arity, HREF, Options) :-
+	format(string(FragmentId), '~w/~d', [Name, Arity]),
+	www_form_encode(FragmentId, EncId),
+	(   option(files(_Map), Options)
+	->  option(man_server(Root), Options,
+		   'http://gollem.science.uva.nl/SWI-Prolog/pldoc/')
+	;   doc_server_root(Root)
+	),
+	format(string(HREF), '~wman?predicate=~w', [Root, EncId]).
+	
+
 %%	pred_href(+NameArity, +Module, -HREF) is semidet.
 %
 %	Create reference.  Prefer:
@@ -894,6 +945,7 @@ pred_href(Name/Arity, Module, HREF) :-
 	format(string(FragmentId), '~w/~d', [Name, Arity]),
 	www_form_encode(FragmentId, EncId),
 	functor(Head, Name, Arity),
+	doc_server_root(Root),
 	(   catch(relative_file(Module:Head, File), _, fail)
 	->  format(string(HREF), '~w#~w', [File, EncId])
 	;   in_file(Module:Head, File)
@@ -902,8 +954,8 @@ pred_href(Name/Arity, Module, HREF) :-
 		prolog:doc_object_summary(Name/Arity, packages, _, _)
 	    ->	format(string(FragmentId), '~w/~d', [Name, Arity]),
 		www_form_encode(FragmentId, EncId),
-		format(string(HREF), '/man?predicate=~w', [EncId])
-	    ;	format(string(HREF), '/doc~w#~w', [File, EncId])
+		format(string(HREF), '~wman?predicate=~w', [Root, EncId])
+	    ;	format(string(HREF), '~wdoc~w#~w', [Root, File, EncId])
 	    )
 	).
 
@@ -931,18 +983,42 @@ object_ref([H|T], Options) --> !,
 	;   []
 	).
 object_ref(Obj, Options) -->
-	{ object_href(Obj, HREF)
+	{ object_href(Obj, HREF, Options)
 	},
 	html(a(href(HREF), \object_link(Obj, Options))).
 	
-%%	object_href(+Object, -HREF)
+%%	object_href(+Object, -HREF) is det.
+%%	object_href(+Object, -HREF, +Options) is det.
 %
 %	HREF is the URL to access Object.
 
 object_href(Obj, HREF) :-
+	object_href(Obj, HREF, []).
+
+object_href(M:PI0, HREF, Options) :-
+	option(files(Map), Options),
+	current_module(M, File),
+	memberchk(file(File, DocFile), Map), !,
+	expand_pi(PI0, PI),
+	term_to_string(PI, PIS),
+	www_form_encode(PIS, PIEnc),
+	file_base_name(DocFile, LocalFile),	% TBD: proper directory index
+	format(string(HREF), '~w#~w', [LocalFile, PIEnc]).
+object_href(Obj, HREF, _Options) :-
 	term_to_string(Obj, String),
 	www_form_encode(String, Enc),
-	format(string(HREF), '/doc_for?object=~w', [Enc]).
+	doc_server_root(Root),
+	format(string(HREF), '~wdoc_for?object=~w', [Root, Enc]).
+
+expand_pi(Name//Arity0, Name/Arity) :- !,
+	Arity is Arity0+2.
+expand_pi(PI, PI).
+
+
+%%	term_to_string(+Term, -String) is det.
+%
+%	Convert Term, possibly  holding  variables,   into  a  canonical
+%	string using A, B, ... for variables and _ for singletons.
 
 term_to_string(Term, String) :-
 	State = state(-),
@@ -1036,18 +1112,66 @@ to_dot_dot([_|T0], ['..'|T], Tail) :-
 %      \file(File) terms in the DOM term generated by wiki.pl.
 
 file(File) -->
-	{ catch(b_getval(pldoc_file, CurrentFile), _, fail),
-	  CurrentFile \== [],
-	  absolute_file_name(File, _,
-			     [ relative_to(CurrentFile),
-			       access(read),
-			       file_errors(fail)
-			     ]), !
+	{ catch(nb_getval(pldoc_options, Options), _, Options = []) },
+	file(File, Options).
+	  
+file(File, Options) -->
+	{ existing_linked_file(File, Path) }, !,
+	{ (   option(files(Map), Options),
+	      memberchk(file(Path, DocFile), Map)
+	  ->  file_base_name(DocFile, HREF)	% TBD: proper location
+	  ;   HREF=File
+	  )
 	},
-	html(a([class(file), href(File)],File)).
-file(File) -->
+	html(a([class(file), href(HREF)], File)).
+file(File, _) -->
 	html(code(class(file), File)).
 
+existing_linked_file(File, Path) :-
+	catch(b_getval(pldoc_file, CurrentFile), _, fail),
+	CurrentFile \== [],
+	absolute_file_name(File, Path,
+			   [ relative_to(CurrentFile),
+			     access(read),
+			     file_errors(fail)
+			   ]).
+
+
+%%	image(+FileName)// is det.
+%
+%	Create an inline image from a link   to another file if the file
+%	exists. Called by \image(File) terms in   the DOM term generated
+%	by wiki.pl.
+
+image(File) -->
+	{ existing_linked_file(File, _) }, !,
+	html(img([src(File), alt(File)])).
+image(File) -->
+	html(code(class(file), ['[[',File,']]'])).
+
+
+		 /*******************************
+		 *	     WIKI FILES		*
+		 *******************************/
+
+
+%%	doc_for_wiki_file(+File, +Out:stream, +Options) is det.
+%
+%	Write HTML for the File containing wiki data.
+
+doc_for_wiki_file(FileSpec, Out, _Options) :-
+	absolute_file_name(FileSpec, File,
+			   [ access(read)
+			   ]),
+	read_file_to_codes(File, String, []),
+	b_setval(pldoc_file, File),
+	call_cleanup((wiki_string_to_dom(String, [], DOM),
+		      phrase(html(DOM), Tokens),
+		      print_html_head(Out),
+		      print_html(Out, Tokens)
+		     ),
+		     nb_delete(pldoc_file)).
+
 
 		 /*******************************
 		 *	      ANCHORS		*
