@@ -1,4 +1,4 @@
-/*  $Id: doc_http.pl,v 1.39 2006/11/13 13:25:14 jan Exp $
+/*  $Id: doc_http.pl,v 1.42 2007/06/28 11:42:26 jan Exp $
 
     Part of SWI-Prolog
 
@@ -41,6 +41,7 @@
 :- use_module(library('http/http_parameters')).
 :- use_module(library('http/html_write')).
 :- use_module(library('http/mimetype')).
+:- use_module(library('http/dcg_basics')).
 :- use_module(library(debug)).
 :- use_module(library(lists)).
 :- use_module(library(url)).
@@ -272,12 +273,41 @@ allow_edit(_) :-
 	can_edit(false), !, 
 	fail.
 allow_edit(Request) :-
-	memberchk(peer(Peer), Request),
+	(   memberchk(x_forwarded_for(Forwarded), Request),
+	    primary_forwarded_host(Forwarded, IPAtom),
+	    parse_ip(IPAtom, Peer)
+	->  true
+	;   memberchk(peer(Peer), Request)
+	),
 	match_peer(localhost, +, Peer).
 
+
+%%	primary_forwarded_host(+Spec, -Host) is det.
+%
+%	x_forwarded host contains multiple hosts seperated   by  ', ' if
+%	there are multiple proxy servers in   between.  The first one is
+%	the one the user's browser knows about.
+
+primary_forwarded_host(Spec, Host) :-
+	sub_atom(Spec, B, _, _, ','), !,
+	sub_atom(Spec, 0, B, _, Host).
+primary_forwarded_host(Host, Host).
+
+
 localhost(ip(127,0,0,1)).
 localhost(localhost).
 
+parse_ip(Atom, IP) :-
+	atom_codes(Atom, Codes),
+	phrase(ip(IP), Codes).
+
+%%	ip(?IP)// is semidet.
+%
+%	Parses A.B.C.D into ip(A,B,C,D)
+
+ip(ip(A,B,C,D)) -->
+	integer(A), ".", integer(B), ".", integer(C), ".", integer(D).
+	
 edit_option(Options0, Options) :-
 	select_option(edit(Bool), Options0, Options), !,
 	assert(can_edit(Bool)).
@@ -313,10 +343,10 @@ doc_server_root(Root) :-
 %	Make paths relative to / if it was moved.
 
 normalise_path(Path0, Path) :-
-	doc_server_root(Root),
 	(   doc_server_root(/)
 	->  Path = Path0
-	;   atom_concat(Root, Rest, Path0)
+	;   doc_server_root(Root),
+	    atom_concat(Root, Rest, Path0)
 	->  atom_concat(/, Rest, Path)
 	).
 	
@@ -340,7 +370,7 @@ reply(/, _) :-
 	ensure_slash_end(Dir0, Dir1),
 	doc_file_href(Dir1, Ref0),
 	atom_concat(Ref0, 'index.html', Index),
-	throw(http_reply(moved(Index))).
+	throw(http_reply(see_other(Index))).
 
 ensure_slash_end(Dir, Dir) :-
 	sub_atom(Dir, _, _, 0, /), !.
