@@ -1,4 +1,4 @@
-/*  $Id: doc_index.pl,v 1.24 2006/10/30 16:06:11 jan Exp $
+/*  $Id: doc_index.pl,v 1.27 2007/06/21 09:19:47 jan Exp $
 
     Part of SWI-Prolog
 
@@ -79,7 +79,8 @@ doc_for_dir(DirSpec, Out, Options) :-
 %	any directory contained in Dir.
 
 dir_index(Dir, Options) -->
-	{ dir_source_files(Dir, Files, Options),
+	{ dir_source_files(Dir, Files0, Options),
+	  sort(Files0, Files),
 	  atom_concat(Dir, '/index.html', File),
 	  b_setval(pldoc_file, File)	% for predref
 	},
@@ -180,18 +181,12 @@ file_index(File, Options) -->
 file_index_header(File, Options) -->
 	prolog:doc_file_index_header(File, Options), !.
 file_index_header(File, Options) -->
-	{ (   option(directory(Dir), Options),
-	      atom_concat(Dir, Local0, File),
-	      atom_concat(/, Local, Local0),
-	      HREF=Local
-	  ->  true
-	  ;   file_base_name(File, Local),
-	      doc_file_href(File, HREF)
-	  ),
+	{ file_base_name(File, Label),
+	  doc_file_href(File, HREF, Options),
 	  ButtonOptions = [button_height(16)|Options]
 	},
 	html(tr(th([colspan(3), class(file)],
-		   [ span(style('float:left'), a(href(HREF), Local)),
+		   [ span(style('float:left'), a(href(HREF), Label)),
 		     span(style('float:right'),
 			  [ \source_button(File, ButtonOptions),
 			    \edit_button(File, ButtonOptions)
@@ -199,6 +194,24 @@ file_index_header(File, Options) -->
 		   ]))).
 
 
+%%	doc_file_href(+File, -HREF, +Options) is det.
+%
+%	HREF is reference to documentation of File.
+
+doc_file_href(File, HREF, Options) :-
+	option(directory(Dir), Options),
+	atom_concat(Dir, Local0, File),
+	atom_concat(/, Local, Local0), !,
+	(   option(files(Map), Options),	% generating files
+	    memberchk(file(File, DocFile), Map)
+	->  file_base_name(DocFile, HREF)
+	;   HREF = Local
+	).
+doc_file_href(File, HREF, _) :-
+	doc_file_href(File, HREF).
+
+
+
 %%	doc_file_href(+Path, -HREF) is det.
 %
 %	Create a /doc HREF from Path.  There   are  some nasty things we
@@ -215,8 +228,8 @@ doc_file_href(File, HREF) :-
 
 %%	ensure_slash_start(+File0, -File) is det.
 %
-%	Ensure Fil starts with a /. This maps C:/foobar into /C:/foobar,
-%	so our paths start with /doc/ again ...
+%	Ensure  File  starts  with  a  /.    This  maps  C:/foobar  into
+%	/C:/foobar, so our paths start with /doc/ again ...
 
 ensure_slash_start(File, File) :-
 	sub_atom(File, 0, _, _, /), !.
@@ -288,18 +301,40 @@ object_summary(_, _, _) -->
 %
 %	Provide overview links and search facilities.
 
+doc_links(_Directory, Options) -->
+	{ option(files(_), Options), !
+	}.
 doc_links(Directory, Options) -->
 	{   Directory == ''
 	->  working_directory(Dir, Dir)
 	;   Dir = Directory
 	},
 	html(div(class(navhdr),
-		 [ span(style('float:left'),  \source_dir_menu(Dir)),
+		 [ span(style('float:left'),
+			div([ \source_dir_menu(Dir),
+			      \version
+			    ])),
 		   span(style('float:right'), \search_form(Options)),
 		   br(clear(both))
 		 ])).
 
 
+%%	version// is det.
+%
+%	Prolog version
+
+version -->
+	{ current_prolog_flag(version, Version),
+	  Major is Version // 10000,
+	  Minor is (Version // 100) mod 100,
+	  Patch is Version mod 100
+	},
+	html(a([ class(prolog_version),
+		 href('http://www.swi-prolog.org')
+	       ],
+	       [' SWI-Prolog ', Major, '.', Minor, '.', Patch])).
+
+
 %%	source_dir_menu(Current)// is det
 %
 %	Create a =select= menu with entries for all loaded directories
