@@ -1,4 +1,4 @@
-/*  $Id: doc_wiki.pl,v 1.2 2006/08/22 10:23:07 jan Exp $
+/*  $Id: doc_wiki.pl,v 1.8 2007/03/22 11:37:00 jan Exp $
 
     Part of SWI-Prolog
 
@@ -30,13 +30,14 @@
 */
 
 :- module(pldoc_wiki,
-	  [ wiki_lines_to_dom/3,	% +Lines, +Map, -DOM
-	    wiki_string_to_dom/3,	% +String, +Args, -DOM
+	  [ wiki_string_to_dom/3,	% +String, +Args, -DOM
+	    wiki_lines_to_dom/3,	% +Lines, +Map, -DOM
 	    section_comment_header/3,	% +Lines, -Header, -RestLines
 	    summary_from_lines/2,	% +Lines, -Summary
 	    indented_lines/3,		% +Text, +PrefixChars, -Lines
 	    strip_leading_par/2,	% +DOM0, -DOM
-	    normalise_white_space/3	% -Text, //
+	    normalise_white_space/3,	% -Text, //
+	    autolink_file/2		% +FileName, -Type
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(debug)).
@@ -310,7 +311,7 @@ tag_name(String, Tag, Order) :-
 	format(atom(Name), '~s', [String]),
 	(   renamed_tag(Name, Tag),
 	    tag_order(Tag, Order)
-	->  print_message(warning, pldoc(depreciated_tag(Name, Tag)))
+	->  print_message(warning, pldoc(deprecated_tag(Name, Tag)))
 	;   tag_order(Name, Order)
 	->  Tag = Name
 	;   print_message(warning, pldoc(unknown_tag(Name))),
@@ -326,9 +327,9 @@ rest_tag([L|Lines0], Indent, [L|VT], Lines) :-
 	rest_tag(Lines0, Indent, VT, Lines).
 
 
-%%	renamed_tag(+DepreciatedTag:atom, -Tag:atom) is semidet.
+%%	renamed_tag(+DeprecatedTag:atom, -Tag:atom) is semidet.
 %
-%	Declaration for depreciated tags.
+%	Declaration for deprecated tags.
 
 renamed_tag(exception, throws).
 
@@ -344,7 +345,7 @@ tag_order(throws,      2).
 tag_order(author,      3).
 tag_order(version,     4).
 tag_order(see,	       5).
-tag_order(depreciated, 6).
+tag_order(deprecated,  6).
 tag_order(compat,      7).		% PlDoc extension
 tag_order(copyright,   8).
 tag_order(license,     9).
@@ -484,12 +485,25 @@ wiki_face(\predref(Name//Arity), _) -->
 wiki_face(span(class=cvs, CVS), _) -->
 	[$, Word, :], {string(Word)}, wiki_faces(CVS0, []), [$], !,
 	{ strip_ws_tokens(CVS0, CVS) }.
+wiki_face(\image(Name), _) -->
+	['[','['], word_token(BaseS), ['.'], word_token(ExtS), [']',']'],
+	{  concat_atom([BaseS, '.', ExtS], Name),
+	   file_name_extension(_, Ext, Name),
+	   autolink_extension(Ext, image)
+	}.
 wiki_face(\file(Name), _) -->
 	word_token(BaseS), ['.'], word_token(ExtS),
-	{ string_to_atom(ExtS, Ext),
-	  autolink_extension(Ext), !,
-	  concat_atom([BaseS, '.', Ext], Name)
+	{ concat_atom([BaseS, '.', ExtS], Name),
+	  (   autolink_file(Name, _)
+	  ;   file_name_extension(_, Ext, Name),
+	      autolink_extension(Ext, _)
+	  ), !
 	}.
+wiki_face(\file(Name), _) -->
+	word_token(NameS),
+	{ autolink_file(Name, _),
+	  sub_atom(NameS, 0, _, 0, Name)
+	}, !.
 wiki_face(a(href=Ref, Ref), _) -->
 	word_token(ProtS), [:,/,/], { url_protocol(ProtS) },
 	string(Rest), peek_end_url, !,
@@ -520,13 +534,26 @@ peek_end_url -->
 peek_end_url -->
 	eos, !.
 
-%%	autolink_extension(?Ext) is nondet.
+%%	autolink_extension(?Ext, ?Type) is nondet.
 %
 %	True if Ext is a filename extensions that create automatic links
 %	in the documentation.
 
-autolink_extension(pl).
+autolink_extension(pl, prolog).
+autolink_extension(txt, wiki).
+autolink_extension(gif, image).
+autolink_extension(png, image).
+autolink_extension(jpg, image).
+autolink_extension(jpeg, image).
 
+%%	autolink_file(?File, -Type) is nondet.
+%
+%	Files to which we automatically create links, regardless of the
+%	extension.
+
+autolink_file('README', wiki).
+autolink_file('TODO', wiki).
+autolink_file('ChangeLog', wiki).
 
 		 /*******************************
 		 *	     SECTIONS		*
@@ -813,7 +840,7 @@ take_line([]) -->
 take_line(Line) -->
 	[H], { code_type(H, white) }, !,
 	take_white(White, WT),
-	(   peek_nl
+	(   nl
 	->  { Line = [] }
 	;   { Line = [H|White] },
 	    take_line(WT)
@@ -895,6 +922,15 @@ peek_nl -->
 	peek(H),
 	{ H == 0'\n ; H == 0'\r }, !.
 
+%%	nl//
+%
+%	Get end-of-line
+
+nl -->
+	"\r\n", !.
+nl -->
+	"\n".
+
 %%	peek(H)//
 %
 %	True if next token is H without eating it.
