@@ -1,4 +1,4 @@
-/*  $Id: plunit.pl,v 1.19 2006/11/23 14:22:09 jan Exp $
+/*  $Id: plunit.pl,v 1.21 2007/03/28 08:51:25 jan Exp $
 
     Part of SWI-Prolog
 
@@ -280,7 +280,7 @@ make_unit_module(Unit, Module) :-
 %
 %	@tbd	Verify options.
 
-expand_test(Name, Options, Body,
+expand_test(Name, Options0, Body,
 	    [ 'unit test'(Name, Line, Options, Module:'unit body'(Id, Vars)),
 	      ('unit body'(Id, Vars) :- !, Body)
 	    ]) :-
@@ -288,7 +288,11 @@ expand_test(Name, Options, Body,
 	prolog_load_context(module, Module),
 	concat_atom([Name, '@line ', Line], Id),
 	term_variables(Body, VarList),
-	Vars =.. [vars|VarList].
+	Vars =.. [vars|VarList],
+	(   is_list(Options0)		% allow for single option without list
+	->  Options = Options0
+	;   Options = [Options0]
+	).
 
 %%	expand(+Term, -Clauses) is semidet.
 
@@ -345,14 +349,17 @@ run_unit([H|T]) :- !,
 run_unit(Spec) :-
 	unit_from_spec(Spec, Unit, Tests, Module, UnitOptions),
 	(   option(blocked(Reason), UnitOptions)
-	->  print_message(informational, plunit(blocked(unit(Unit, Reason))))
+	->  info(plunit(blocked(unit(Unit, Reason))))
 	;   setup(Module, UnitOptions)
-	->  print_message(informational, plunit(begin(Spec))),
+	->  info(plunit(begin(Spec))),
 	    forall((Module:'unit test'(Name, Line, Options, Body),
 		    matching_test(Name, Tests)),
 		   run_test(Unit, Name, Line, Options, Body)),
-	    print_message(informational, plunit(end(Spec))),
-	    format(user_error, '~N', []),
+	    info(plunit(end(Spec))),
+	    (	message_level(silent)
+	    ->	true
+	    ;	format(user_error, '~N', [])
+	    ),
 	    cleanup(Module, UnitOptions)
 	;   true
 	).
@@ -671,7 +678,7 @@ report :-
 report_blocked :-
 	predicate_property(blocked(_,_,_,_), number_of_clauses(N)),
 	N > 0,
-	print_message(informational, plunit(blocked(N))),
+	info(plunit(blocked(N))),
 	(   blocked(Unit, Name, Line, Reason),
 	    unit_file(Unit, File),
 	    print_message(informational,
@@ -683,10 +690,10 @@ report_blocked.
 report_failed :-
 	predicate_property(failed(_,_,_,_), number_of_clauses(N)),
 	N > 0, !,
-	print_message(informational, plunit(failed(N))),
+	info(plunit(failed(N))),
 	fail.
 report_failed :-
-	print_message(informational, plunit(failed(0))).
+	info(plunit(failed(0))).
 
 report_failure(Unit, Name, Line, Error) :-
 	print_message(error, plunit(failed(Unit, Name, Line, Error))).
@@ -747,6 +754,24 @@ load_test_files(_Options) :-
 		 *	     MESSAGES		*
 		 *******************************/
 
+%%	info(+Term)
+%
+%	Runs print_message(Level, Term), where Level  is one of =silent=
+%	or =informational= (default).
+
+info(Term) :-
+	message_level(Level),
+	print_message(Level, Term).
+
+message_level(Level) :-
+	current_test_flag(test_options, Options),
+	option(silent(Silent), Options, false),
+	(   Silent == false
+	->  Level = informational
+	;   Level = silent
+	).
+
+
 message(error(context_error(plunit_close(Name, -)), _)) -->
 	[ 'PL-Unit: cannot close unit ~w: no open unit'-[Name] ].
 message(error(context_error(plunit_close(Name, Start)), _)) -->
