@@ -1,11 +1,11 @@
-/*  $Id: rdf_db.h,v 1.48 2007/01/13 21:23:51 jan Exp $
+/*  $Id: rdf_db.h,v 1.53 2007/07/09 13:56:38 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -30,7 +30,7 @@
 #endif
 #include "lock.h"
 
-#define RDF_VERSION 20500		/* 2.7.0 */
+#define RDF_VERSION 20700		/* 2.7.0 */
 
 #define URL_subPropertyOf \
 	"http://www.w3.org/2000/01/rdf-schema#subPropertyOf"
@@ -63,7 +63,7 @@
 #define INDEX_TABLES 		        7
 #define INITIAL_TABLE_SIZE   		8*1024
 #define INITIAL_PREDICATE_TABLE_SIZE	1024
-#define INITIAL_SOURCE_TABLE_SIZE	64
+#define INITIAL_GRAPH_TABLE_SIZE	64
 
 #define MAX_HASH_FACTOR 8		/* factor to trigger re-hash */
 #define MIN_HASH_FACTOR 4		/* factor after re-hash */
@@ -94,18 +94,19 @@ typedef struct bitmatrix
 
 typedef struct predicate
 { atom_t	    name;		/* name of the predicate */
-  list	            subPropertyOf;	/* the one I'm subPropertyOf */
-  list		    siblings;		/* my subProperties */
-  struct predicate *root;		/* Root of property tree */
   struct predicate *next;		/* next in hash-table */
-  struct predicate *oldroot;		/* from previous run */
-  bitmatrix	   *reachable;		/* Reachability matrix on root */
-  int		    label;		/* Numeric label */
-  int		    visited;		/* loop detection */
+					/* hierarchy */
+  list	            subPropertyOf;	/* the one I'm subPropertyOf */
+  list	            siblings;		/* reverse of subPropertyOf */
+  int		    label;		/* Numeric label in cloud */
+  struct predicate_cloud *cloud;	/* cloud I belong to */
+  unsigned long	    hash;		/* key used for hashing
+  					   (=hash if ->cloud is up-to-date) */
+					/* properties */
   struct predicate *inverse_of;		/* my inverse predicate */
   unsigned 	    transitive : 1;	/* P(a,b)&P(b,c) --> P(a,c) */
+					/* statistics */
   long		    triple_count;	/* # triples on this predicate */
-
   long		    distinct_updated[2];/* Is count still valid? */
   long		    distinct_count[2];  /* Triple count at last update */
   long		    distinct_subjects[2];/* # distinct subject values */
@@ -113,15 +114,26 @@ typedef struct predicate
 } predicate;
 
 
-typedef struct source
-{ struct source    *next;		/* next in table */
-  atom_t	    name;		/* name of the source */
+typedef struct predicate_cloud
+{ predicate   **members;		/* member predicates */
+  unsigned long hash;			/* hash-code */
+  int		size;			/* size of the cloud */
+  int		deleted;		/* See erase_predicates() */
+  bitmatrix    *reachable;		/* cloud reachability matrix */
+  unsigned	dirty : 1;		/* predicate hash not synchronised */
+} predicate_cloud;
+
+
+typedef struct graph
+{ struct graph    *next;		/* next in table */
+  atom_t	    name;		/* name of the graph */
+  atom_t	    source;		/* URL graph was loaded from */
   int		    triple_count;	/* # triples associated to it */
 #ifdef WITH_MD5
   unsigned	    md5 : 1;		/* do/don't record MD5 */
   md5_byte_t 	    digest[16];		/* MD5 digest */
 #endif
-} source;  
+} graph;  
 
 
 typedef struct literal
@@ -153,8 +165,8 @@ typedef struct triple
   { literal *	literal;
     atom_t	resource;
   } object;
-  atom_t	source;			/* where it comes from */
-  unsigned long line;			/* source-line number */
+  atom_t	graph;			/* where it comes from */
+  unsigned long line;			/* graph-line number */
 					/* indexing */
   struct triple*next[INDEX_TABLES];	/* hash-table next links */
 					/* flags */
@@ -174,6 +186,8 @@ typedef struct triple
 
 typedef enum
 { TR_MARK,				/* mark start for nesting */
+  TR_SUB_START,				/* start nested transaction */
+  TR_SUB_END,				/* end nested transaction */
   TR_ASSERT,				/* rdf_assert */
   TR_RETRACT,				/* rdf_retractall */
   TR_UPDATE,				/* rdf_update */
@@ -195,13 +209,20 @@ typedef struct transaction_record
       unsigned long		line;
     } src;
     struct
-    { source		       *source;
+    { graph		       *graph;
       md5_byte_t	       *digest;
     } md5;
+    record_t		       transaction_id;
   } update;
 } transaction_record;
 
 
+typedef struct active_transaction
+{ struct active_transaction *parent;
+  term_t id;
+} active_transaction;
+
+
 typedef struct rdf_db
 { triple       *by_none, *by_none_tail;
   triple      **table[INDEX_TABLES];
@@ -221,17 +242,21 @@ typedef struct rdf_db
   predicate   **pred_table;		/* Hash-table of predicates */
   int		pred_table_size;	/* #entries in the table */
   int		pred_count;		/* #predicates */
+  unsigned long next_hash;		/* cloud hash keys */
   int		active_queries;		/* Calls with choicepoints */
   int		need_update;		/* We need to update */
   long		agenda_created;		/* #visited nodes in agenda */
   long		duplicates;		/* #duplicate triples */
   long		generation;		/* generation-id of the database */
-  source      **source_table;		/* Hash table of sources */
-  int      	source_table_size;	/* Entries in table */
-  source	*last_source;		/* last accessed source */
+  graph      **graph_table;		/* Hash table of sources */
+  int      	graph_table_size;	/* Entries in table */
+
+  graph	*last_graph;		/* last accessed graph */
+  active_transaction *tr_active;	/* open transactions */
   transaction_record *tr_first;		/* first transaction record */
   transaction_record *tr_last;		/* last transaction record */
   int		tr_nesting;		/* nesting depth of transactions */
+  int		tr_reset;		/* transaction contains reset */
   int 		resetting;		/* We are in rdf_reset_db() */
 
   rwlock	lock;			/* threaded access */
