@@ -1,11 +1,11 @@
-/*  $Id: rdf_db.pl,v 1.84 2006/08/31 09:44:00 jan Exp $
+/*  $Id: rdf_db.pl,v 1.107 2007/07/09 13:56:38 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -50,9 +50,9 @@
 	    rdf_set_predicate/2,	% +Predicate, +Property
 	    rdf_predicate_property/2,	% +Predicate, ?Property
 	    rdf_current_predicate/1,	% -Predicate
-
 	    rdf_transaction/1,		% :Goal
 	    rdf_transaction/2,		% :Goal, +Id
+	    rdf_active_transaction/1,	% ?Id
 
 	    rdf_monitor/2,		% :Goal, +Options
 
@@ -74,7 +74,9 @@
 	    rdf_md5/2,			% +DB, -MD5
 	    rdf_atom_md5/3,		% +Text, +Times, -MD5
 
+	    rdf_graph/1,		% ?DB
 	    rdf_source/1,		% ?File
+	    rdf_source/2,		% ?DB, ?SourceURL
 	    rdf_make/0,			% Reload modified databases
 
 	    rdf_source_location/2,	% +Subject, -Source
@@ -122,21 +124,29 @@
 :- use_module(library(sgml_write)).
 :- use_module(library(option)).
 :- use_module(library(nb_set)).
+:- use_module(library(error)).
+:- use_module(library(url)).
+:- use_module(rdf_cache).
 
 :- initialization
    load_foreign_library(foreign(rdf_db)).
 
 :- multifile
 	ns/2,
-	rdf_meta_specification/2.	% Unbound Head, Head
+	rdf_meta_specification/2.	% UnboundHead, Head
 :- dynamic
-	ns/2,				% ID, URL
-	rdf_source/4.			% File, ModTimeAtLoad, Triples, MD5
+	ns/2,			% ID, URL
+	rdf_source/5.		% DB, SourceURL, ModTimeAtLoad, Triples, MD5
 :- volatile
-	rdf_source/4.
+	rdf_source/5.
 :- discontiguous
 	term_expansion/2.
 
+/** <module> Core RDF database
+
+@see Documentation for semweb package
+*/
+
 		 /*******************************
 		 *	     NAMESPACES		*
 		 *******************************/
@@ -152,12 +162,20 @@ ns(serql,'http://www.openrdf.org/schema/serql#').
 %%	rdf_register_ns(+Alias, +URI) is det.
 %%	rdf_register_ns(+Alias, +URI, +Options) is det.
 %
-%	Register a namespace. If force(true) is   provided as an option,
-%	the  existing  namespace  alias  is   replaced.  Otherwise  this
-%	predicate raises a permission error.  Please note that replacing
-%	a namespace is dangerous  as   namespaces  affect preprocessing.
-%	Make sure all code that depends on a namespace is compiled after
-%	changing the registration.
+%	Register Alias as an abbreviateion for URI. Options:
+%	
+%		* force(Boolean)
+%		If =true=, Replace existing namespace alias. Please note
+%		that replacing a namespace is dangerous as namespaces
+%		affect preprocessing. Make sure all code that depends on
+%		a namespace is compiled after changing the registration.
+%		
+%		* keep(Boolean)
+%		If =true= and Alias is already defined, keep the
+%		original message and succeed silently.
+%		
+%	Without options, an attempt  to  redefine   an  alias  raises  a
+%	permission error.
 
 rdf_register_ns(Alias, URI) :-
 	rdf_register_ns(Alias, URI, []).
@@ -180,7 +198,7 @@ rdf_register_ns(Alias, URI, _) :-
 	assert(ns(Alias, URI)).
 
 
-%%	rdf_global_id(?Id, ?GlobalId)
+%%	rdf_global_id(?Id, ?GlobalId) is det.
 %
 %	Convert between NS:Local and global atomic identifier.
 %	To be completed.
@@ -190,7 +208,7 @@ rdf_global_id(NS:Local, Global) :-
 rdf_global_id(Global, Global).
 
 
-%%	rdf_global_object(?Object, ?GlobalObject)
+%%	rdf_global_object(?Object, ?GlobalObject) is det.
 %	
 %	Same as rdf_global_id/2,  but  intended   for  dealing  with the
 %	object part of a  triple,  in   particular  the  type  for typed
@@ -216,7 +234,7 @@ global(NS, Local, Global) :-
 
 
 
-%%	rdf_global_term(+TermIn, -GlobalTerm)
+%%	rdf_global_term(+TermIn, -GlobalTerm) is det.
 %	
 %	Does rdf_global_id/2 on all terms NS:Local by recursively analysing
 %	the term.
@@ -352,13 +370,15 @@ must_be_atom(X) :-
 	rdf_retractall(r,r,o,?),
 	rdf_reachable(r,r,r),
 	rdf_update(r,r,o,t),
+	rdf_update(r,r,o,+,t),
 	rdf_equal(r,r),
 	rdf_source_location(r,-),
 	rdf_subject(r),
 	rdf_set_predicate(r, +),
 	rdf_predicate_property(r, -),
 	rdf_estimate_complexity(r,r,r,-),
-	rdf_transaction(:).
+	rdf_transaction(:),
+	rdf_transaction(:, +).
 
 %%	rdf_equal(?Resource1, ?Resource2)
 %	
@@ -406,16 +426,12 @@ rdf_member_property(P, N) :-
 
 %%	rdf_node(-Id)
 %
-%	Generate a unique identifier for a subject.  Obsolete.  New
-%	code should use rdf_bnode/1.
-
-rdf_node(Value) :-
-	repeat,
-	gensym('_:', Value),
-	\+ rdf_subject(Value),
-	\+ rdf(_, _, Value),
-	\+ rdf(_, Value, _), !.
+%	Generate a unique blank node identifier for a subject.
+%	
+%	@depricated	New code should use rdf_bnode/1.
 
+rdf_node(Resource) :-
+	rdf_bnode(Resource).
 
 %%	rdf_bnode(-Id)
 %
@@ -433,6 +449,8 @@ rdf_bnode(Value) :-
 %	
 %	Tests if a resource is a blank node (i.e. is an anonymous
 %	resource).
+%	
+%	@see rdf_bnode/1.
 
 rdf_is_bnode(Id) :-
 	atom(Id),
@@ -464,7 +482,7 @@ rdf_source_location(Subject, Source) :-
 %	@param KeyValue	Term of the form Key(Value).
 
 rdf_statistics(sources(Count)) :-
-	predicate_property(rdf_source(_,_,_,_), number_of_clauses(Count)).
+	predicate_property(rdf_source(_,_,_,_,_), number_of_clauses(Count)).
 rdf_statistics(subjects(Count)) :-
 	rdf_statistics_(subjects(Count)).
 rdf_statistics(properties(Count)) :-
@@ -548,15 +566,36 @@ rdf_predicate_property(P, Prop) :-
 		 *	    TRANSACTION		*
 		 *******************************/
 
-%%	rdf_transaction(:Goal)
+%%	rdf_transaction(:Goal) is semidet.
+%%	rdf_transaction(:Goal, +Id) is semidet.
 %	
 %	Backward compatibility
 
 rdf_transaction(Goal) :-
-	rdf_transaction(Goal, user).
+	rdf_transaction_(Goal, user).
+rdf_transaction(Goal, Id) :-
+	(   nonvar(Id),
+	    Id = log(_, DB)
+	->  must_be(atom, DB)
+	;   true
+	),
+	rdf_transaction_(Goal, Id).
+
+%%	rdf_active_transaction(?Id) is nondet.
+%
+%	True if Id is the identifier of a currently open transaction. If
+%	Id  is  not  instantiated,    backtracking   yields  transaction
+%	identifiers starting with  the   innermost  nested  transaction.
+%	Transaction identifier terms are not copied,  need not be ground
+%	and can be instantiated during the transaction.
+
+rdf_active_transaction(Id) :-
+	rdf_active_transactions_(List),
+	member(Id, List).
 
 %%	rdf_monitor(:Goal, +Options)
 %
+%	Call Goal if spefified actions occur on the database.
 
 rdf_monitor(Goal, Options) :-
 	monitor_mask(Options, 0xffff, Mask),
@@ -613,7 +652,8 @@ monitor_mask(all,	   0xffff).
 %	
 %	Save triples into File in a   quick-to-load binary format. If DB
 %	is supplied only triples flagged to originate from that database
-%	are added.
+%	are  added.  Files  created  this  way    can  be  loaded  using
+%	rdf_load_db/1.
 
 rdf_save_db(File) :-
 	open(File, write, Out, [type(binary)]),
@@ -629,15 +669,20 @@ rdf_load_db_no_admin(File, Id) :-
 	open(File, read, Out, [type(binary)]),
 	call_cleanup(rdf_load_db_(Out, Id), close(Out)).
 
+%%	rdf_load_db(+File) is det.
+%
+%	Load triples from a file created using rdf_save_db/2 and update
+%	the file administration.
 
 rdf_load_db(File) :-
-	rdf_load_db_no_admin(File, file(File)),
-	rdf_sources_(Sources),
-	(   member(Src, Sources),
-	    rdf_md5(Src, MD5),
-	    rdf_statistics_(triples(Src, Triples)),
-	    retractall(rdf_source(Src, _, _, _)),
-	    assert(rdf_source(Src, 0, Triples, MD5)),
+	file_name_to_url(File, URL),
+	rdf_load_db_no_admin(File, URL),
+	rdf_graphs_(DBList),
+	(   member(DB, DBList),
+	    rdf_md5(DB, MD5),
+	    rdf_statistics_(triples(DB, Triples)),
+	    retractall(rdf_source(DB, _, _, _, _)),
+	    assert(rdf_source(DB, -, 0, Triples, MD5)),
 	    fail
 	;   true
 	).
@@ -647,16 +692,44 @@ rdf_load_db(File) :-
 		 *	    LOADING RDF		*
 		 *******************************/
 
-%%	rdf_load(+FileOrList, +Options)
+:- multifile
+	rdf_open_hook/3,
+	rdf_load_stream/3,
+	rdf_input_info/3,
+	rdf_file_type/2,
+	url_protocol/1.
+
+%%	rdf_load(+FileOrList) is det.
+%%	rdf_load(+FileOrList, +Options) is det.
 %
 %	Load RDF file.  Options provides additional processing options.
 %	Currently defined options are:
 %	
-%	    # result(-Action, -Triples, -MD5)
-%	    	Return action taken (load, reload, none) and number
-%	    	of triples loaded from the file as well as the MD5
-%	    	digest.
-%	    	
+%	    * result(-Action, -Triples, -MD5)
+%	    Return action taken (load, reload, none) and number
+%	    of triples loaded from the file as well as the MD5
+%	    digest.
+%	    
+%	    * blank_nodes(+ShareMode)
+%	    How to handle equivalent blank nodes.  If =share= (default),
+%	    equivalent blank nodes are shared in the same resource.
+%	    
+%	    * base_uri(+URI)
+%	    URI that is used for rdf:about="" and other RDF constructs
+%	    that are relative to the base uri.  Default is the source
+%	    URL.
+%	    
+%	    * db(+DB)
+%	    Named graph in which to load the data.  It is *not* allowed
+%	    to load two sources into the same named graph.
+%	    
+%	    * if(Condition)
+%	    When to load the file. One of =true=, =changed= (default) or
+%	    =not_loaded=.
+%	    
+%	    * cache(Bool)
+%	    If =false=, do not use or create a cache file.
+%	    
 %	Other options are forwarded to process_rdf/3.
 
 rdf_load(Spec) :-
@@ -666,107 +739,222 @@ rdf_load([], _) :- !.
 rdf_load([H|T], Options) :- !,
 	rdf_load(H, Options),
 	rdf_load(T, Options).
-rdf_load(Spec, Options0) :-
-	statistics(cputime, CpuOld),
-	(   select(result(Action, Triples, MD5), Options0, Options1)
+rdf_load(Spec, Options) :-
+	statistics(cputime, T0),
+	(   select(result(Action, Triples, MD5), Options, Options1)
 	->  true
-	;   Options1 = Options0
+	;   Options1 = Options
 	),
-	(   memberchk(blank_nodes(_), Options1)
-	->  Options2 = Options1
-	;   Options2 = [ blank_nodes(share)|Options1 ]
+	rdf_input(Spec, Input, SourceURL),
+	(   rdf_input_info(Input, Modified, DefFormat)
+	->  true
+	;   Modified = 0,
+	    DefFormat = xml
 	),
-	(   (   is_stream(Spec)
-	    ->	In = Spec
-	    ;	Spec = stream(In)
-	    )
-	->  (   memberchk(base_uri(BaseURI), Options2)
-	    ->	Options = Options2
-	    ;	gensym('stream://', BaseURI),
-		Options = [base_uri(BaseURI)|Options2]
-	    ),
-	    process_rdf(In, assert_triples, Options),
-	    rdf_statistics_(triples(BaseURI, Triples)),
-	    rdf_md5(BaseURI, MD5),
-	    assert(rdf_source(BaseURI, Modified, Triples, MD5)),
-	    Source = BaseURI,
-	    Load = parsed(ParseTime),
-	    Action = load
-	;   Source = Spec,
-	    absolute_file_name(Spec,
-			       [ access(read),
-				 extensions([rdf,rdfs,owl,''])
-			       ], File),
-	    time_file(File, Modified),
-	    (	rdf_source(File, WhenLoaded, _, _)
-	    ->	(   Modified > WhenLoaded
-		->  rdf_retractall(_,_,_,File:_),
-		    Action = reload
-		;   Action = none
-		)
-	    ;	Action = load
-	    ),
-	    (	Action \== none
-	    ->  add_base_url(Options2, File, Options),
-		retractall(rdf_source(File, _, _, _)),
-		(   cache_file(File, Cache)
-		->  (   time_file(Cache, CacheTime),
-		        time_file(File, FileTime),
-			CacheTime >= FileTime,
-			catch(rdf_load_db_no_admin(Cache, cache(File)),
-			      _, fail),
-			ignore(memberchk(namespaces([]), Options)) % TBD
-		    ->  Load = cache(ParseTime)
-		    ;   parse_rdf_file(File, Options),
-			Load = parsed(ParseTime),
-			save_cache(File, Cache)
-		    )
-		;   parse_rdf_file(File, Options),
-		    Load = parsed(ParseTime)
-		),
-		rdf_statistics_(triples(File, Triples)),
-		rdf_md5(File, MD5),
-		assert(rdf_source(File, Modified, Triples, MD5))
-	    ;	rdf_source(File, _Modified, Triples, MD5)
-	    )
-	),
-	(   Action \== none
-	->  statistics(cputime, CpuLoaded),
-	    ParseTime is CpuLoaded - CpuOld,
-	    (	memberchk(silent(true), Options)
-	    ->	Level = silent
-	    ;	Level = informational
-	    ),
-	    print_message(Level,
-			  rdf(loaded(Source, Triples, Load)))
+	select_option(base_uri(BaseURI), Options1, Options2, SourceURL),
+	select_option(format(Format), Options2, Options3, DefFormat),
+	select_option(blank_nodes(ShareMode), Options3, Options4, share),
+	select_option(cache(Cache), Options4, Options5, true),
+	select_option(if(If), Options5, Options6, changed),
+	select_option(db(DB), Options6, RDFOptions, SourceURL),
+	(   var(BaseURI)
+	->  BaseURI = SourceURL
 	;   true
+	),
+	(   must_load(If, DB, Modified)
+	->  do_unload(DB),		% unload old
+	    (   Cache == true,
+		read_cache(SourceURL, Modified, CacheFile),
+	        catch(rdf_load_db_no_admin(CacheFile, cache(DB)), _, fail)
+	    ->	Action = load
+	    ;   rdf_input_open(Input, Stream, Format),
+		must_be(ground, Format),
+		call_cleanup(rdf_load_stream(Format, Stream,
+					     [ base_uri(BaseURI),
+					       blank_nodes(ShareMode),
+					       db(DB)
+					     | RDFOptions
+					     ]),
+			     close_input(Input, Stream)), !,
+		rdf_set_graph_source(DB, SourceURL),
+		(   Cache == true,
+		    rdf_cache_file(SourceURL, write, CacheFile)
+		->  catch(save_cache(DB, CacheFile), E,
+			  print_message(warning, E))
+		;   true
+		),
+		(   Format = triples
+		->  Action = load
+		;   Action = parsed
+		)
+	    ),
+	    rdf_statistics_(triples(DB, Triples)),
+	    rdf_md5(DB, MD5),
+	    assert(rdf_source(DB, SourceURL, Modified, Triples, MD5))
+	;   Action = none,
+	    rdf_source(DB, _, _, Triples, MD5)
+	),
+	report_loaded(Action, Input, DB, Triples, T0, Options).
+
+%%	close_input(+Input, +Stream) is det.
+%
+%	Close input if it was not specified as a stream.
+
+close_input(stream(_), _) :- !.
+close_input(_, Stream) :-
+	close(Stream).
+
+%%	rdf_input(+Term, -Input, -BaseURI)
+%
+%	Resolve input description Term.  Unify   Input  with a canonical
+%	description of the input, which is one of:
+%	
+%	  * stream(Stream)
+%	  * file(AbsolutePath)
+%	  * url(Protocol, URL)
+%	
+%	BaseURI is unified with a default base   URI. Note that this may
+%	be overruled from the options of rdf_load/2.
+
+rdf_input(stream(Stream), stream(Stream), BaseURI) :- !,
+	(   stream_property(Stream, file_name(File))
+	->  file_name_to_url(File, BaseURI)
+	;   gensym('stream://', BaseURI)
+	).
+rdf_input(Stream, stream(Stream), BaseURI) :-
+	is_stream(Stream), !,
+	rdf_input(stream(Stream), _, BaseURI).
+rdf_input(FileURL, file(File), FileURL) :-
+	atom(FileURL),
+	file_name_to_url(File, FileURL), !.
+rdf_input(URL, url(Protocol, URL), URL) :-
+	is_url(URL, Protocol), !.
+rdf_input(Spec, file(Path), BaseURI) :-
+	findall(Ext, (rdf_file_type(Ext, _);Ext=''), Exts),
+	absolute_file_name(Spec, Path,
+			   [ access(read),
+			     extensions(Exts)
+			   ]),
+	file_name_to_url(Path, BaseURI).
+	
+%%	is_url(+Term, -Protocol) is semidet.
+%
+%	True if Term is an atom denoting a URL of the given Protocol.
+%	We only support a limited set of protocols as defined by the
+%	extensible predicate url_protocol/1.
+
+is_url(URL, Protocol) :-
+	atom(URL),
+	sub_atom(URL, B, _, _, :), !,
+	sub_atom(URL, 0, B, _, RawProtocol),
+	downcase_atom(RawProtocol, Protocol),
+	url_protocol(Protocol).
+
+%%	url_protocol(+Protocol) is det.
+%
+%	True  if  Protocol  is  the  (lowercase)  name  of  a  supported
+%	protocol.  New  protocols  can  be    added  to  this  multifile
+%	predicate. In addition the plugin must define rdf_open_hook/3 to
+%	create a stream from the added protocol.
+
+%	(no default clauses)
+
+%%	rdf_input_info(+Input, -Modified, -Format) is semidet.
+%
+%	Return the last modification  time  of   Input  as  a POSIX time
+%	stamp as well as the format of the input.
+
+rdf_input_info(file(File), Modified, Format) :-
+	time_file(File, Modified),
+	file_name_extension(_, Ext, File),
+	(   rdf_file_type(Ext, Format)
+	->  true
+	;   Format = xml
 	).
 
+%%	rdf_input_open(+Input, -Stream, ?Format) is det.
+%
+%	Open given input as a stream.
 
-parse_rdf_file(File, Options) :-
-	source_descr(File, Id),
-	rdf_transaction(process_rdf(File, assert_triples, Options),
-			parse(Id)).
-
-source_descr(File, file(File)) :-
-	atom(File), !.
-source_descr(Stream, stream(Stream)) :-
-	is_stream(Stream), !.
-source_descr(Descr, Descr).
+rdf_input_open(Input, Stream, Format) :-
+	rdf_open_hook(Input, Stream, Format), !.
+rdf_input_open(stream(Stream), Stream, _) :- !.
+rdf_input_open(file(File), Stream, _) :-
+	open(File, read, Stream, [type(binary)]).
 
 
-%%	add_base_url(+Options0, +File, -Options)
+%%	rdf_file_type(+Extension, -Format) is semidet.
+%
+%	True if Format  is  the  format   belonging  to  the  given file
+%	extension.  This predicate is multifile and can thus be extended
+%	by plugins.
+
+rdf_file_type(rdf,   xml).
+rdf_file_type(rdfs,  xml).
+rdf_file_type(owl,   xml).
+rdf_file_type(htm,   xhtml).
+rdf_file_type(html,  xhtml).
+rdf_file_type(xhtml, xhtml).
+rdf_file_type(trp,   triples).
+
+
+%%	rdf_load_stream(+Format, +Stream, +Options)
+%
+%	Load RDF data from Stream.
 %	
-%	Add base_uri(file://File) to the options if  this is not already
-%	there. A base URI is needed  to   avoid  the  case where loading
-%	multiple  files  independently  cause    clashes   of  generated
-%	descriptions.
+%	@tbd	Handle mime-types?
+
+rdf_load_stream(xml, Stream, Options) :- !,
+	option(db(Id), Options),
+	rdf_transaction(process_rdf(Stream, assert_triples, Options),
+			parse(Id)).
+rdf_load_stream(xhtml, Stream, Options) :- !,
+	option(db(Id), Options),
+	rdf_transaction(process_rdf(Stream, assert_triples, [embedded(true)|Options]),
+			parse(Id)).
+rdf_load_stream(triples, Stream, Options) :- !,
+	option(db(Id), Options),
+	rdf_load_db_(Stream, Id).
+
+%%	read_cache(+BaseURI, +SourceModified, -CacheFile) is semidet.
+%
+%	True if CacheFile is a cache-file modified after SourceModified.
+
+read_cache(BaseURI, Modified, CacheFile) :-
+	rdf_cache_file(BaseURI, read, CacheFile),
+	time_file(CacheFile, CacheModified),
+	CacheModified >= Modified.
+
+
+%%	must_load(+Condition, +DB, +ModifiedSrc) is semidet.
+%
+%	True if source must be reloaded.
+
+must_load(true, _, _) :- !.
+must_load(changed, DB, ModifiedSrc) :- !,
+	(   rdf_source(DB, _SourceURL, ModifiedLoaded, _, _)
+	->  ModifiedSrc > ModifiedLoaded
+	;   true
+	).
+must_load(not_loaded, DB, _) :- !,
+	\+ rdf_source(DB, _SourceURL, _, _, _).
+must_load(Cond, _, _) :-
+	throw(eror(domain_error(condition, Cond), _)).
+
+
+%%	report_loaded(+Action, +Source, +DB, +Triples, +StartCPU, +Options)
+
+report_loaded(none, _, _, _, _, _) :- !.
+report_loaded(Action, Source, DB, Triples, T0, Options) :-
+	statistics(cputime, T1),
+	Time is T1 - T0,
+	(   option(silent(true), Options)
+	->  Level = silent
+	;   Level = informational
+	),
+	print_message(Level,
+		      rdf(loaded(Action, Source, DB, Triples, Time))).
 
-add_base_url(Options, _File, Options) :-
-	option(base_uri(Base), Options, Var),
-	Base \== Var, !.
-add_base_url(Options, File, [base_uri(Base)|Options]) :-
-	atom_concat('file://', File, Base).
 
 
 %%	rdf_unload(+Spec)
@@ -774,81 +962,80 @@ add_base_url(Options, File, [base_uri(Base)|Options]) :-
 %	Remove the triples loaded from the specified source and remove
 %	the source from the database.
 
+rdf_unload(DB) :-
+	atom(DB),
+	rdf_statistics_(triples(DB, _)),
+	(   rdf(_,_,_,DB)
+	;   rdf(_,_,_,DB:_)
+	), !,
+	do_unload(DB).
 rdf_unload(Spec) :-
-	(   (   is_stream(Spec)
-	    ;   Spec = stream(_)
-	    )
-	->  throw(error(permission_error(rdf_db, unload, Spec), _))
-	;   atom(Spec),
-	    rdf_statistics_(triples(Spec, _)),
-	    (	rdf(_,_,_,Spec)
-	    ;   rdf(_,_,_,Spec:_)
-	    )
-	->  do_unload(Spec)
-	;   absolute_file_name(Spec,
-			       [ access(read),
-				 extensions([rdf,rdfs,owl,'']),
-				 file_errors(fail)
-			       ], File)
-	->  do_unload(File)
-	;   true
-	).
-	
-do_unload(Spec) :-
-	retractall(rdf_source(Spec, _, _, _)),
-	rdf_transaction((rdf_retractall(_,_,_,Spec:_),
-			 rdf_retractall(_,_,_,Spec)),
-			unload(Spec)).
+	rdf_input(Spec, _, BaseURI),
+	do_unload(BaseURI).
 
+do_unload(DB) :-
+	retractall(rdf_source(DB, _, _, _, _)),
+	rdf_transaction((rdf_retractall(_,_,_,DB:_),
+			 rdf_retractall(_,_,_,DB)),
+			unload(DB)).
+
+
+%%	rdf_graph(+DB) is semidet.
+%%	rdf_graph(-DB) is nondet.
+%
+%	True if DB is a current named graph with at least one triple.
+
+rdf_graph(DB) :-
+	atom(DB), !,
+	rdf_statistics_(triples(DB, Triples)),
+	Triples > 0.
+rdf_graph(DB) :-
+	rdf_graphs_(Sources),
+	member(DB, Sources),
+	rdf_statistics_(triples(DB, Triples)),
+	Triples > 0.
+
+%%	rdf_source(?DB, ?Source) is nondet.
+%
+%	True if named graph DB is loaded from Source.
+
+rdf_source(DB, SourceURL) :-
+	rdf_graph(DB),
+	rdf_graph_source_(DB, SourceURL).
 
 %%	rdf_source(?Source)
 %	
-%	Query the loaded sources.  
+%	True if Source is a loaded source.
+%	
+%	@depricated	Use rdf_graph/1 or rdf_source/2.
 
-rdf_source(File) :-
-	rdf_source(File,_,_,_).		% loaded files
-rdf_source(File) :- 
-	rdf_sources_(Sources),		% other sources
-	member(File, Sources),
-	\+ rdf_source(File,_,_,_),
-	rdf_statistics_(triples(File, Triples)),
+rdf_source(DB) :-
+	rdf_source(DB,_,_,_,_).		% loaded files
+rdf_source(DB) :- 
+	rdf_graphs_(Sources),		% other sources
+	member(DB, Sources),
+	\+ rdf_source(DB,_,_,_,_),
+	rdf_statistics_(triples(DB, Triples)),
 	Triples > 0.
 
-%	rdf_make
+%%	rdf_make
 %	
 %	Reload all loaded files that have been modified since the last
 %	time they were loaded.
 
 rdf_make :-
-	forall((rdf_source(File, Time, _Triples, _),
+	forall((rdf_source(DB, SourceURL, Time, _Triples, _),
 		Time \== 0),
-	       catch(rdf_load(File), _, true)).
+	       catch(rdf_load(SourceURL, [db(DB)]), _, true)).
 
+%%	save_cache(+DB, +Cache) is det.
+%
+%	Save triples belonging to DB in the file Cache.
 
-%%	cache_file(+Base, -CacheFile)
-%	
-%	Deduce the name of the file used to cache the triples.
-
-cache_dir(CacheDir) :-
-	(   current_prolog_flag(windows, true)
-	->  CacheDir = '_cache'
-	;   CacheDir = '.cache'
-	).
-
-cache_file(Base, Cache) :-
-	file_directory_name(Base, BaseDir),
-	file_base_name(Base, File),
-	cache_dir(Dir),
-	concat_atom([BaseDir, /, Dir], CacheDir),
-	exists_directory(CacheDir),
-	concat_atom([CacheDir, /, File, '.trp'], Cache).
-
-%%	save_cache(+File, +Cache)
-
-save_cache(File, Cache) :-
+save_cache(DB, Cache) :-
 	catch(open(Cache, write, CacheStream, [type(binary)]), _, fail), !,
-	rdf_save_db_(CacheStream, File),
-	close(CacheStream).
+	call_cleanup(rdf_save_db_(CacheStream, DB),
+		     close(CacheStream)).
 
 %%	assert_triples(+Triples, +Source)
 %
@@ -874,7 +1061,7 @@ assert_triples([H|_], _) :-
 %	statistics.
 
 rdf_reset_db :-
-	retractall(rdf_source(_,_,_,_)),
+	retractall(rdf_source(_,_,_,_,_)),
 	rdf_transaction(rdf_reset_db_, reset).
 
 
@@ -888,26 +1075,44 @@ rdf_reset_db :-
 %	Save RDF data to file.  Options is a list of one or more of the
 %	following options:
 %	
-%		# db(+DB)
+%		* db(+DB)
 %		Save only triples associated to the given DB
 %		
-%		# anon(Bool)
+%		* anon(Bool)
 %		If false (default true) do not save blank nodes that do
 %		not appear (indirectly) as object of a named resource.
+%		
+%		* base_uri(URI)
+%		BaseURI used. If present, all URIs that can be
+%		represented relative to this base are written using
+%		their shorthand.  See also =write_xml_base= option
+%		
+%		* write_xml_base(Bool)
+%		If =false=, do _not_ include the =|xml:base|=
+%		declaration that is written normally when using the
+%		=base_uri= option.
 %
-%		# convert_typed_literal(:Convertor)
+%		* convert_typed_literal(:Convertor)
 %		Call Convertor(-Type, -Content, +RDFObject), providing
 %		the opposite for the convert_typed_literal option of
 %		the RDF parser.
 %		
-%		# encoding(Encoding)
+%		* encoding(Encoding)
 %		Encoding for the output.  Either utf8 or iso_latin_1
 %		
-%		# document_language(+Lang)
+%		* document_language(+Lang)
 %		Initial xml:lang saved with rdf:RDF element
+%		
+%		* sorted(+Boolean)
+%		If =true= (default =false=), emit subjects sorted on
+%		the full URI.  Useful to make file comparison easier.
+%	
+%	@param File	Location to save the data.  This can also be a 
+%			file-url (=|file://path|=).
 
 :- module_transparent
 	rdf_transaction/1,
+	rdf_transaction/2,
 	rdf_monitor/2,
 	rdf_save/2,
 	meta_options/2.
@@ -918,14 +1123,19 @@ rdf_reset_db :-
 rdf_save(File) :-
 	rdf_save2(File, []).
 
-rdf_save(File, Options0) :-
+rdf_save(Spec, Options0) :-
 	is_list(Options0), !,
 	meta_options(Options0, Options),
+	to_file(Spec, File),
 	rdf_save2(File, Options).
-rdf_save(File, DB) :-
+rdf_save(Spec, DB) :-
 	atom(DB), !,			% backward compatibility
+	to_file(Spec, File),
 	rdf_save2(File, [db(DB)]).
 
+to_file(URL, File) :-
+	file_name_to_url(File, URL), !.
+to_file(File, File).
 
 rdf_save2(File, Options) :-
 	option(encoding(Encoding), Options, utf8),
@@ -966,8 +1176,13 @@ cleanup_save(Reason,
 
 rdf_do_save(Out, Options) :-
 	rdf_save_header(Out, Options),
-	forall(rdf_subject(Subject, Options),
-	       rdf_save_non_anon_subject(Out, Subject, Options)),
+	(   option(sorted(true), Options, false)
+	->  setof(Subject, rdf_subject(Subject, Options), Subjects),
+	    forall(member(Subject, Subjects),
+		   rdf_save_non_anon_subject(Out, Subject, Options))
+	;   forall(rdf_subject(Subject, Options),
+		   rdf_save_non_anon_subject(Out, Subject, Options))
+	),
 	rdf_save_footer(Out), !.	% dubious cut; without the
 					% cleanup handlers isn't called!?
 
@@ -1014,13 +1229,16 @@ meta_options([H0|T0], [H|T]) :-
 
 rdf_save_header(Out, Options) :-
 	is_list(Options), !,
-	xml_encoding(Out, Encoding),
+	stream_property(Out, encoding(Enc)),
+	xml_encoding(Enc, Encoding),
 	format(Out, '<?xml version=\'1.0\' encoding=\'~w\'?>~n', [Encoding]),
 	format(Out, '<!DOCTYPE rdf:RDF [', []),
 	header_namespaces(Options, NSList),
 	(   member(Id, NSList),
 	    ns(Id, NS),
-	    format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NS]),
+	    rdf_quote_uri(NS, QNS),
+	    xml_quote_attribute(QNS, NSText, Enc),
+	    format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NSText]),
 	    fail
 	;   true
 	),
@@ -1031,6 +1249,13 @@ rdf_save_header(Out, Options) :-
 	    fail
 	;   true
 	),
+	(   option(base_uri(Base), Options),
+	    option(write_xml_base(true), Options, true)
+	->  rdf_quote_uri(Base, QBase),
+	    xml_quote_attribute(QBase, BaseText, Enc),
+	    format(Out, '~N    xml:base="~w"~n', [BaseText])
+	;   true
+	),
 	(   memberchk(document_language(Lang), Options)
 	->  format(Out, '~N    xml:lang="~w"', [Lang])
 	;   true
@@ -1040,8 +1265,7 @@ rdf_save_header(Out, FileRef) :-	% compatibility
 	atom(FileRef),
 	rdf_save_header(Out, [db(FileRef)]).
 	
-xml_encoding(Out, Encoding) :-
-	stream_property(Out, encoding(Enc)),
+xml_encoding(Enc, Encoding) :-
 	(   xml_encoding_name(Enc, Encoding)
 	->  true
 	;   throw(error(domain_error(rdf_encoding, Enc), _))
@@ -1157,9 +1381,13 @@ xml_codes([H|T]) :-
 
 xml_code(X) :-
 	code_type(X, csym), !.
-xml_code(0'-).
+xml_code(0'-).				% Match 0'-
 
 
+%%	rdf_save_footer(Out:stream) is det.
+%
+%	Finish XML generation and write the document footer.
+
 rdf_save_footer(Out) :-
 	retractall(named_anon(_, _)),
 	format(Out, '</rdf:RDF>~n', []).
@@ -1180,9 +1408,14 @@ rdf_save_non_anon_subject(Out, Subject, Options) :-
 	flag(rdf_db_saved_subjects, X, X+1).
 
 
+%%	rdf_save_subject(+Out, +Subject:resource, +Options) is det.
+%
+%	Save the triples associated to Subject to Out.
+
 rdf_save_subject(Out, Subject, Options) :-
 	is_list(Options), !,
-	(   rdf_save_subject(Out, Subject, -, 0, Options)
+	option(base_uri(BaseURI), Options, '-'),
+	(   rdf_save_subject(Out, Subject, BaseURI, 0, Options)
 	->  format(Out, '~n', [])
 	;   throw(error(rdf_save_failed(Subject), 'Internal error'))
 	).
@@ -1193,7 +1426,14 @@ rdf_save_subject(Out, Subject, DB) :-
 	).
 		  
 
-rdf_save_subject(Out, Subject, DefNS, Indent, Options) :-
+%%	rdf_save_subject(+Out:stream, +Subject:resource, +BaseURI,
+%%			 +Indent:int, +Options) is det.
+%
+%	Save properties of Subject.
+%	
+%	@param Indent	Current indentation
+
+rdf_save_subject(Out, Subject, BaseURI, Indent, Options) :-
 	db(Options, DB),
 	findall(Pred=Object, rdf_db(Subject, Pred, Object, DB), Atts0),
 	sort(Atts0, Atts),		% remove duplicates
@@ -1205,7 +1445,7 @@ rdf_save_subject(Out, Subject, DefNS, Indent, Options) :-
 			  rdf(save_removed_duplicates(Del, Subject)))
 	;   true
 	),
-	rdf_save_subject(Out, Subject, DefNS, Atts, Indent, Options),
+	rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options),
 	flag(rdf_db_saved_triples, X, X+L).
 
 rdf_db(Subject, Pred, Object, DB) :-
@@ -1214,49 +1454,59 @@ rdf_db(Subject, Pred, Object, DB) :-
 rdf_db(Subject, Pred, Object, DB) :-
 	rdf(Subject, Pred, Object, DB:_).
 
-rdf_save_subject(Out, Subject, DefNS, Atts, Indent, Options) :-
+%%	rdf_save_subject(+Out:stream, +Subject:resource, +BaseURI,
+%%			 +Atts:list(Pred=Obj), +Indent:int, +Options) is det.
+%
+%	Save triples defined by Atts on Subject.
+
+rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options) :-
 	rdf_equal(rdf:type, RdfType),
 	select(RdfType=Type, Atts, Atts1),
-	rdf_id(Type, DefNS, TypeId),
+	rdf_id(Type, BaseURI, TypeId),
 	xml_is_name(TypeId), !,
 	format(Out, '~*|<~w', [Indent, TypeId]),
-	save_about(Out, Subject),
-	save_attributes(Atts1, DefNS, Out, TypeId, Indent, Options).
-rdf_save_subject(Out, Subject, _DefNS, Atts, Indent, Options) :-
+	save_about(Out, BaseURI, Subject),
+	save_attributes(Atts1, BaseURI, Out, TypeId, Indent, Options).
+rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options) :-
 	format(Out, '~*|<rdf:Description', [Indent]),
-	save_about(Out, Subject),
-	save_attributes(Atts, rdf, Out, rdf:'Description', Indent, Options).
+	save_about(Out, BaseURI, Subject),
+	save_attributes(Atts, BaseURI, Out, rdf:'Description', Indent, Options).
 
 xml_is_name(_NS:Atom) :- !,
 	xml_name(Atom).
 xml_is_name(Atom) :-
 	xml_name(Atom).
 
-save_about(Out, Subject) :-
+%%	save_about(+Out, +BaseURI, +Subject) is det.
+%
+%	Save the rdf:about. If Subject is a  blank node, save the nodeID
+%	if any.
+
+save_about(Out, _, Subject) :-
 	rdf_is_bnode(Subject), !,
 	(   named_anon(Subject, NodeID)
 	->  format(Out, ' rdf:nodeID="~w"', [NodeID])
 	;   true
 	).
-save_about(Out, Subject) :-
+save_about(Out, BaseURI, Subject) :-
 	stream_property(Out, encoding(Encoding)),
-	rdf_value(Subject, QSubject, Encoding),
+	rdf_value(Subject, BaseURI, QSubject, Encoding),
 	format(Out, ' rdf:about="~w"', [QSubject]).
 
-%%	save_attributes(+List, +DefNS, +Stream, Element)
+%%	save_attributes(+List, +BaseURI, +Stream, Element)
 %
 %	Save the attributes.  Short literal attributes are saved in the
 %	tag.  Others as the content of the description element.  The
 %	begin tag has already been filled.
 
-save_attributes(Atts, DefNS, Out, Element, Indent, Options) :-
+save_attributes(Atts, BaseURI, Out, Element, Indent, Options) :-
 	split_attributes(Atts, InTag, InBody),
 	SubIndent is Indent + 2,
-	save_attributes2(InTag, DefNS, tag, Out, SubIndent, Options),
+	save_attributes2(InTag, BaseURI, tag, Out, SubIndent, Options),
 	(   InBody == []
 	->  format(Out, '/>~n', [])
 	;   format(Out, '>~n', []),
-	    save_attributes2(InBody, _, body, Out, SubIndent, Options),
+	    save_attributes2(InBody, BaseURI, body, Out, SubIndent, Options),
 	    format(Out, '~N~*|</~w>~n', [Indent, Element])
 	).
 
@@ -1312,23 +1562,23 @@ in_tag_attribute(_=literal(Text)) :-
 	atom_length(Text, Len),
 	Len < 60.
 
-%%	save_attributes(+List, +DefNS, +TagOrBody, +Stream)
+%%	save_attributes(+List, +BaseURI, +TagOrBody, +Stream)
 %
 %	Save a list of attributes.
 
 save_attributes2([], _, _, _, _, _).
-save_attributes2([H|T], DefNS, Where, Out, Indent, Options) :-
-	save_attribute(Where, H, DefNS, Out, Indent, Options),
-	save_attributes2(T, DefNS, Where, Out, Indent, Options).
+save_attributes2([H|T], BaseURI, Where, Out, Indent, Options) :-
+	save_attribute(Where, H, BaseURI, Out, Indent, Options),
+	save_attributes2(T, BaseURI, Where, Out, Indent, Options).
 
-save_attribute(tag, Name=literal(Value), DefNS, Out, Indent, _DB) :-
+save_attribute(tag, Name=literal(Value), BaseURI, Out, Indent, _DB) :-
 	AttIndent is Indent + 2,
-	rdf_att_id(Name, DefNS, NameText),
+	rdf_id(Name, BaseURI, NameText),
 	stream_property(Out, encoding(Encoding)),
 	xml_quote_attribute(Value, QVal, Encoding),
 	format(Out, '~N~*|~w="~w"', [AttIndent, NameText, QVal]).
-save_attribute(body, Name=literal(Literal0), DefNS, Out, Indent, Options) :- !,
-	rdf_id(Name, DefNS, NameText),
+save_attribute(body, Name=literal(Literal0), BaseURI, Out, Indent, Options) :- !,
+	rdf_id(Name, BaseURI, NameText),
 	(   memberchk(convert_typed_literal(Converter), Options),
 	    call(Converter, Type, Content, Literal0)
 	->  Literal = type(Type, Content)
@@ -1337,15 +1587,19 @@ save_attribute(body, Name=literal(Literal0), DefNS, Out, Indent, Options) :- !,
 	(   Literal = lang(Lang, Value)
 	->  (   memberchk(document_language(Lang), Options)
 	    ->  format(Out, '~N~*|<~w>', [Indent, NameText])
-	    ;   rdf_id(Lang, DefNS, LangText),
+	    ;   rdf_id(Lang, BaseURI, LangText),
 		format(Out, '~N~*|<~w xml:lang="~w">',
 		       [Indent, NameText, LangText])
 	    )
 	;   Literal = type(Type, Value)
-	->  stream_property(Out, encoding(Encoding)),
-	    rdf_value(Type, QVal, Encoding),
-	    format(Out, '~N~*|<~w rdf:datatype="~w">',
-		   [Indent, NameText, QVal])
+	->  (   rdf_equal(Type, rdf:'XMLLiteral')
+	    ->	format(Out, '~N~*|<~w rdf:parseType="Literal">',
+		       [Indent, NameText])
+	    ;	stream_property(Out, encoding(Encoding)),
+		rdf_value(Type, BaseURI, QVal, Encoding),
+		format(Out, '~N~*|<~w rdf:datatype="~w">',
+		       [Indent, NameText, QVal])
+	    )
 	;   atomic(Literal)
 	->  format(Out, '~N~*|<~w>', [Indent, NameText]),
 	    Value = Literal
@@ -1355,9 +1609,9 @@ save_attribute(body, Name=literal(Literal0), DefNS, Out, Indent, Options) :- !,
 	),
 	save_attribute_value(Value, Out, Indent),
 	format(Out, '</~w>', [NameText]).
-save_attribute(body, Name=Value, DefNS, Out, Indent, Options) :-
+save_attribute(body, Name=Value, BaseURI, Out, Indent, Options) :-
 	rdf_is_bnode(Value), !,
-	rdf_id(Name, DefNS, NameText),
+	rdf_id(Name, BaseURI, NameText),
 	(   named_anon(Value, NodeID)
 	->  format(Out, '~N~*|<~w rdf:nodeID="~w"/>',
 		   [Indent, NameText, NodeID])
@@ -1372,18 +1626,18 @@ save_attribute(body, Name=Value, DefNS, Out, Indent, Options) :-
 	    SubIndent is Indent + 2,
 	    (   rdf(Value, rdf:type, rdf:'List')
 	    ->  format(Out, '~N~*|<~w', [Indent, NameText]),
-		save_about(Out, Value),
+		save_about(Out, BaseURI, Value),
 		format(Out, ' rdf:parseType="Collection">~n', []),
-		rdf_save_list(Out, Value, DefNS, SubIndent, Options)
+		rdf_save_list(Out, Value, BaseURI, SubIndent, Options)
 	    ;   format(Out, '~N~*|<~w>~n', [Indent, NameText]),
-		rdf_save_subject(Out, Value, DefNS, SubIndent, Options)
+		rdf_save_subject(Out, Value, BaseURI, SubIndent, Options)
 	    ),
 	    format(Out, '~N~*|</~w>~n', [Indent, NameText])
 	).
-save_attribute(body, Name=Value, DefNS, Out, Indent, _DB) :-
+save_attribute(body, Name=Value, BaseURI, Out, Indent, _DB) :-
 	stream_property(Out, encoding(Encoding)),
-	rdf_value(Value, QVal, Encoding),
-	rdf_id(Name, DefNS, NameText),
+	rdf_value(Value, BaseURI, QVal, Encoding),
+	rdf_id(Name, BaseURI, NameText),
 	format(Out, '~N~*|<~w rdf:resource="~w"/>', [Indent, NameText, QVal]).
 
 save_attribute_value(Value, Out, _) :-	% strings
@@ -1406,32 +1660,36 @@ save_attribute_value(Value, _Out, _) :-
 
 rdf_save_list(_, List, _, _, _) :-
 	rdf_equal(List, rdf:nil), !.
-rdf_save_list(Out, List, DefNS, Indent, Options) :-
+rdf_save_list(Out, List, BaseURI, Indent, Options) :-
 	rdf_has(List, rdf:first, First),
 	(   rdf_is_bnode(First)
 	->  nl(Out),
-	    rdf_save_subject(Out, First, DefNS, Indent, Options)
+	    rdf_save_subject(Out, First, BaseURI, Indent, Options)
 	;   stream_property(Out, encoding(Encoding)),
-	    rdf_value(First, QVal, Encoding),
+	    rdf_value(First, BaseURI, QVal, Encoding),
 	    format(Out, '~N~*|<rdf:Description rdf:about="~w"/>',
 		   [Indent, QVal])
 	),
 	flag(rdf_db_saved_triples, X, X+3),
 	(   rdf_has(List, rdf:rest, List2),
 	    \+ rdf_equal(List2, rdf:nil)
-	->  rdf_save_list(Out, List2, DefNS, Indent, Options)
+	->  rdf_save_list(Out, List2, BaseURI, Indent, Options)
 	;   true
 	).
 
-%%	rdf_id(+Resource, +DefNS, -NSLocal)
-%	
-%	Generate a NS:Local name for Resource given the indicated
-%	default namespace.  This call is used for elements.
 
-rdf_id(Id, NS, NS:Local) :-
-	ns(NS, Full),
-	Full \== '',
-	atom_concat(Full, Local, Id), !.
+%%	rdf_id(+Resource, +BaseURI, -NSLocal)
+%	
+%	Generate a NS:Local  name  for   Resource  given  the  indicated
+%	default namespace. This call is used for elements.
+
+rdf_id(Id, BaseURI, Local) :-
+	assertion(atom(BaseURI)),
+	atom_concat(BaseURI, Local, Id),
+	sub_atom(Local, 0, 1, _, #), !.
+rdf_id(Id, _, NS:Local) :-
+	rdf_split_url(Full, Local, Id),
+	ns(NS, Full), !.
 rdf_id(Id, _, NS:Local) :-
 	ns(NS, Full),
 	Full \== '',
@@ -1439,27 +1697,26 @@ rdf_id(Id, _, NS:Local) :-
 rdf_id(Id, _, Id).
 
 
-rdf_att_id(Id, _, NS:Local) :-
-	ns(NS, Full),
-	Full \== '',
-	atom_concat(Full, Local, Id), !.
-rdf_att_id(Id, _, Id).
-
-
-%%	rdf_value(+Resource, -Text, +Encoding)
+%%	rdf_value(+Resource, +BaseURI, -Text, +Encoding)
 %	
 %	According  to  "6.4  RDF  URI  References"  of  the  RDF  Syntax
 %	specification, a URI reference is  UNICODE string not containing
 %	control sequences, represented as  UTF-8   and  then  as escaped
 %	US-ASCII.
 
-rdf_value(V, Text, Encoding) :-
+rdf_value(Base, Base, '', _) :- !.
+rdf_value(V, Base, Text, Encoding) :-
+	atom_concat(Base, Local, V),
+	sub_atom(Local, 0, _, _, #), !,
+	rdf_quote_uri(Local, Q0),
+	xml_quote_attribute(Q0, Text, Encoding).
+rdf_value(V, _, Text, Encoding) :-
 	ns(NS, Full),
 	atom_concat(Full, Local, V), !,
 	rdf_quote_uri(Local, QLocal0),
 	xml_quote_attribute(QLocal0, QLocal, Encoding),
 	concat_atom(['&', NS, (';'), QLocal], Text).
-rdf_value(V, Q, Encoding) :-
+rdf_value(V, _, Q, Encoding) :-
 	rdf_quote_uri(V, Q0),
 	xml_quote_attribute(Q0, Q, Encoding).
 
@@ -1471,18 +1728,11 @@ rdf_value(V, Q, Encoding) :-
 :- multifile
 	prolog:message/3.
 
-prolog:message(rdf(loaded(Spec, Triples, parsed(ParseTime)))) -->
-	[ 'Parsed' ],
-	source(Spec),
-	in_time(Triples, ParseTime).
-prolog:message(rdf(loaded(Spec, Triples, cache(ParseTime)))) -->
-	[ 'Loaded' ],
-	source(Spec),
-	in_time(Triples, ParseTime).
-prolog:message(rdf(loaded(Spec, Triples, snapshot(ParseTime)))) -->
-	[ 'Loaded snapshot' ],
-	source(Spec),
-	in_time(Triples, ParseTime).
+prolog:message(rdf(loaded(How, What, BaseURI, Triples, Time))) -->
+	how(How),
+	source(What),
+	into(What, BaseURI),
+	in_time(Triples, Time).
 prolog:message(rdf(save_removed_duplicates(N, Subject))) -->
 	[ 'Removed ~d duplicate triples about "~p"'-[N,Subject] ].
 prolog:message(rdf(saved(File, SavedSubjects, SavedTriples))) -->
@@ -1492,17 +1742,26 @@ prolog:message(rdf(saved(File, SavedSubjects, SavedTriples))) -->
 prolog:message(rdf(using_namespace(Id, NS))) -->
 	[ 'Using namespace id ~w for ~w'-[Id, NS] ].
 
-source(Spec) -->
-	{ atom(Spec),
-	  (   sub_atom(Spec, 0, _, _, 'stream://')
-	  ->  Base = Spec
-	  ;   file_base_name(Spec, Base)
-	  )
+how(load)   --> [ 'Loaded' ].
+how(parsed) --> [ 'Parsed' ].
+
+source(stream(Stream)) -->
+	{ stream_property(Stream, file_name(Path)), !,
+	  file_base_name(Path, Base)
 	},
 	[ ' "~w"'-[Base] ].
+source(file(Path)) -->
+	{ atom(Path), !,
+	  file_base_name(Path, Base)
+	},
+	[ ' "~w"'-[Base] ].
+source(url(_Protocol, URL)) -->
+	[ ' "~w"'-[URL] ].
 source(Spec) -->
 	[ ' "~p"'-[Spec] ].
 
+into(_, _) --> [].			% TBD
+
 in_time(Triples, ParseTime) -->
 	[ ' in ~2f sec; ~D triples'-[ParseTime, Triples]
 	].
