@@ -1,11 +1,11 @@
-/*  $Id: rdf_db_test.pl,v 1.30 2006/11/09 20:06:58 jan Exp $
+/*  $Id: rdf_db_test.pl,v 1.33 2007/07/09 13:56:38 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -36,6 +36,7 @@
 :- use_module(rdf_db).
 :- use_module(rdfs).
 :- use_module(library(xsdp_types)).
+:- use_module(library(lists)).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 RDF-DB test file.  A test is a clause of the form:
@@ -203,6 +204,7 @@ lang_data :-
 lang_data(S, A) :-
 	rdf_assert(S, A, literal(lang(nl, 'Jan'))),
 	rdf_assert(S, A, literal(lang(en, 'John'))),
+	rdf_assert(S, A, literal(lang(en, ''))),
 	rdf_assert(S, A, literal('Johannes')).
 
 same_set(S1, S2) :-
@@ -215,6 +217,7 @@ lang(1) :-
 	findall(X, rdf(x, a, literal(X)), Xs),
 	Xs == [ lang(nl, 'Jan'),
 		lang(en, 'John'),
+		lang(en, ''),
 		'Johannes'
 	      ]. 
 lang(2) :-
@@ -227,6 +230,7 @@ lang(3) :-
 	findall(X, rdf(x, a, literal(X)), Xs),
 	Xs =@= [ lang(nl, 'Jan'),
 		 lang(en, 'John'),
+		 lang(en, ''),
 		 lang(_,  'Johannes')
 	       ].
 lang(4) :-
@@ -238,7 +242,11 @@ lang(save_db) :-
 	save_reload_db,
 	X = lang(_,_),
 	findall(X, rdf(x, a, literal(X)), Xs),
-	(   Xs =@= [ lang(nl, 'Jan'),  lang(en, 'John'), lang(_, 'Johannes') ]
+	(   Xs =@= [ lang(nl, 'Jan'), 
+		     lang(en, 'John'),
+		     lang(en, ''),
+		     lang(_, 'Johannes') 
+		   ]
 	->  true
 	;   format(user_error, 'Xs = ~w~n', [Xs]),
 	    fail
@@ -248,7 +256,11 @@ lang(save) :-
 	save_reload,
 	findall(X, rdf(x, a, literal(X)), Xs),
 	(   same_set(Xs,
-		     [ lang(nl, 'Jan'),  lang(en, 'John'), 'Johannes' ])
+		     [ lang(nl, 'Jan'),
+		       lang(en, 'John'),
+		       lang(en, ''),
+		       'Johannes' 
+		     ])
 	->  true
 	;   format(user_error, 'Xs = ~q~n', [Xs]),
 	    fail
@@ -434,6 +446,19 @@ transaction(deadlock-2) :-
 	     ],
 	delete_file(F1),
 	delete_file(F2).
+transaction(active-1) :-
+	\+ rdf_active_transaction(_).
+transaction(active-2) :-
+	rdf_transaction(rdf_active_transaction(x), x).
+transaction(active-3) :-
+	rdf_transaction(findall(X, rdf_active_transaction(X), Xs), x),
+	Xs == [x].
+transaction(active-4) :-
+	rdf_transaction(rdf_active_transaction(Y), X),
+	X == Y.
+transaction(active-5) :-
+	rdf_transaction(rdf_active_transaction(x), X),
+	X == x.
 
 
 		 /*******************************
@@ -443,11 +468,11 @@ transaction(deadlock-2) :-
 label(1) :-
 	rdf_global_id(rdfs:label, Label),
 	lang_data(x, Label),
-	findall(L, rdfs_label(x, L), Ls), Ls = ['Jan', 'John', 'Johannes'].
+	findall(L, rdfs_label(x, L), Ls), Ls = ['Jan', 'John', '', 'Johannes'].
 label(2) :-
 	rdf_global_id(rdfs:label, Label),
 	lang_data(x, Label),
-	findall(L, rdfs_label(x, en, L), Ls), Ls = ['John'].
+	findall(L, rdfs_label(x, en, L), Ls), Ls = ['John', ''].
 
 
 		 /*******************************
@@ -586,10 +611,9 @@ subproperty(1) :-
 
 %%	ptree/1
 %
-%	Property hierarchy handling for  rdf_has/3.   The  routines must
-%	associate a unique root to all   networks of properties, dealing
-%	with   cycles   and   networks   with    multiple   roots.   See
-%	organise_predicates().
+%	Property hierarchy handling for rdf_has/3. The routines maintain
+%	clouds of connected properties and for   each  cloud a bitmatrix
+%	filled with the closure of the rdfs:subPropertyOf relation.
 
 ptree(1) :-
 	rdf_assert(a, rdfs:subPropertyOf, b),
@@ -632,6 +656,11 @@ ptree(6) :-				% two root cycles
 	rdf_has(x, d, y),
 	rdf_has(x, dc, y),
 	rdf_has(x, bc, y).
+ptree(7) :-				% create and break the cycles
+	rdf_assert(x, a, y),
+	rdf_assert(a, rdfs:subPropertyOf, b),
+	rdf_retractall(a, rdfs:subPropertyOf, b),
+	\+ rdf_has(x, b, y).
 
 
 
@@ -674,6 +703,18 @@ duplicates(1) :-
 	rdf_retractall(a, b, literal(lang(en, l))).
 
 
+		 /*******************************
+		 *	      SOURCE		*
+		 *******************************/
+
+source(1) :-
+	rdf_assert(a,b,c,test),
+	rdf_db:rdf_set_graph_source(test, 'test.rdf'),
+	rdf_source(test, X),
+	X == 'test.rdf'.
+
+
+
 		 /*******************************
 		 *	      SCRIPTS		*
 		 *******************************/
@@ -764,6 +805,7 @@ testset(subproperty).
 testset(ptree).
 testset(reachable).
 testset(duplicates).
+testset(source).
 
 %	testdir(Dir)
 %	
