@@ -1,11 +1,11 @@
-/*  $Id: rdf_persistency.pl,v 1.5 2006/08/31 09:50:20 jan Exp $
+/*  $Id: rdf_persistency.pl,v 1.26 2007/07/09 13:56:38 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -33,6 +33,7 @@
 	  [ rdf_attach_db/2,		% +Directory, +Options
 	    rdf_detach_db/0,		% +Detach current DB
 	    rdf_current_db/1,		% -Directory
+	    rdf_persistency/2,		% +DB, +Bool
 	    rdf_flush_journals/1,	% +Options
 	    rdf_journal_file/2,		% ?DB, ?JournalFile
 	    rdf_db_to_file/2		% ?DB, ?FileBase
@@ -41,8 +42,11 @@
 :- use_module(library(lists)).
 :- use_module(library(url)).
 :- use_module(library(debug)).
+:- use_module(library(error)).
+:- use_module(library(thread)).
+
+/** <module> RDF persistency plugin
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This  module  provides  persistency   for    rdf_db.pl   based   on  the
 rdf_monitor/2 predicate to  track  changes   to  the  repository.  Where
 previous  versions  used  autosafe  of  the  whole  database  using  the
@@ -61,9 +65,11 @@ is restored by loading  the  snapshot   and  replaying  the journal. The
 predicate rdf_flush_journals/1 can be used to create fresh snapshots and
 delete the journals.
 
-TBD: if there is a complete `.new'   snapshot  and no journal, we should
+@tbd if there is a complete `.new'   snapshot  and no journal, we should
 move the .new to the plain snapshot name as a means of recovery.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+@see	rdf_edit.pl
+*/
 
 :- volatile
 	rdf_directory/1,
@@ -79,32 +85,85 @@ move the .new to the plain snapshot name as a means of recovery.
 %%	rdf_attach_db(+Directory, +Options)
 %	
 %	Start persistent operations using Directory   as  place to store
-%	files.   There are several cases.
+%	files.   There are several cases:
 %	
 %		* Empty DB, existing directory
 %		Load the DB from the existing directory
 %		
 %		* Full DB, empty directory
 %		Create snapshots for all sources in directory
+%		
+%	Options:
+%	
+%		* concurrency(+Jobs)
+%		Number of threads to use for loading the initial
+%		database.  If not provided it is the number of CPUs
+%		as optained from the flag =cpu_count=.
+%
+%		* max_open_journals(+Count)
+%		Maximum number of journals kept open.  If not provided,
+%		the default is 10.  See limit_fd_pool/0.
+%		
+%		* silent(+Boolean)
+%		If =true= (default =false=), do not print informational
+%		messages.
+%		
+%		* log_nested_transactions(+Boolean)
+%		If =true=, nested _log_ transactions are added to the
+%		journal information.  By default (=false=), no log-term
+%		is added for nested transactions.
 
 rdf_attach_db(DirSpec, Options) :-
 	absolute_file_name(DirSpec,
 			   Directory,
 			   [ access(write),
-			     file_type(directory)
-			   ]),
+			     file_type(directory),
+			     file_errors(fail)
+			   ]), !,
 	(   rdf_directory(Directory)
 	->  true			% update settings?
 	;   rdf_detach_db,
 	    mkdir(Directory),
-	    assert(rdf_directory(Directory)),
-	    forall(member(O, Options), assert(rdf_option(O))),
 	    lock_db(Directory),
+	    assert(rdf_directory(Directory)),
+	    assert_options(Options),
 	    stop_monitor,		% make sure not to register load
 	    load_db,
 	    at_halt(rdf_detach_db),
 	    start_monitor
 	).
+rdf_attach_db(DirSpec, Options) :-
+	absolute_file_name(DirSpec,
+			   Directory,
+			   [ solutions(all)
+			   ]),
+	(   exists_directory(Directory)
+	->  access_file(Directory, write)
+	;   catch(make_directory(Directory), _, fail)
+	), !,
+	rdf_attach_db(Directory, Options).
+
+
+assert_options([]).
+assert_options([H|T]) :-
+	(   option_type(H, Check)
+	->  Check,
+	    assert(rdf_option(H))
+	;   domain_error(rdf_option, H)
+	),
+	assert_options(T).
+	
+option_type(concurrency(X),		must_be(positive_integer, X)).
+option_type(max_open_journals(X),	must_be(positive_integer, X)).
+option_type(silent(X),			must_be(boolean, X)).
+option_type(log_nested_transactions(X),	must_be(boolean, X)).
+
+
+%%	rdf_detach_db is det.
+%
+%	Detach from the  current  database.   Succeeds  silently  if  no
+%	database is attached. Normally called at  the end of the program
+%	through at_halt/1. 
 
 rdf_detach_db :-
 	debug(halt, 'Detaching database', []),
@@ -133,7 +192,7 @@ rdf_current_db(Directory) :-
 %	
 %	Flush dirty journals.  Options:
 %	
-%		# min_size(+KB)
+%		* min_size(+KB)
 %		Only flush if journal is over KB in size.
 %		TBD: sensible default size
 
@@ -157,14 +216,42 @@ rdf_flush_journal(DB, Options) :-
 		 *	       LOAD		*
 		 *******************************/
 
-%%	load_db(-DBs)
+%%	load_db is det.
+%
+%	Reload database from the directory specified by rdf_directory/1.
+%	First we find all names graphs using find_dbs/1 and then we load
+%	them.
 
 load_db :-
 	rdf_directory(Dir),
 	working_directory(Old, Dir),
 	call_cleanup(find_dbs(DBs), working_directory(_, Old)),
-	forall(member(DB, DBs),
-	       load_source(DB)).
+	make_goals(DBs, Goals),
+	concurrency(Jobs),
+	concurrent(Jobs, Goals, []).
+
+make_goals([], []).
+make_goals([DB|T0], [load_source(DB)|T]) :-
+	make_goals(T0, T).
+
+%%	concurrency(-Jobs)
+%
+%	Number of jobs to run concurrently.
+
+concurrency(Jobs) :-
+	rdf_option(concurrency(Jobs)), !.
+concurrency(Jobs) :-
+	current_prolog_flag(cpu_count, Jobs),
+	Jobs > 0, !.
+concurrency(1).
+
+
+%%	find_dbs(-DBs:list(atom)) is det.
+%
+%	DBs is a  list  of  database   (named  graph)  names,  sorted in
+%	increasing file-size. Small files  are   loaded  first  as these
+%	typically contain the schemas and we   want  to avoid re-hashing
+%	large databases due to added rdfs:subPropertyOf triples.
 
 find_dbs(DBs) :-
 	expand_file_name(*, Files),
@@ -204,24 +291,28 @@ unkey([_-H|T0], [H|T]) :-
 	unkey(T0, T).
 
 
-%%	load_source(+SnapshotFile)
+%%	load_source(+DB) is det.
 %	
 %	Load triples and reload  journal   from  the  indicated snapshot
 %	file.
 
 load_source(DB) :-
+	(   rdf_option(silent(true))
+	->  Level = silent
+	;   Level = informational
+	),
 	db_files(DB, SnapshotFile, JournalFile),
 	rdf_retractall(_,_,_,DB),
 	statistics(cputime, T0),
-	print_message(informational, rdf(restore(source(DB)))),
+	print_message(Level, rdf(restore(source(DB)))),
 	(   exists_db(SnapshotFile)
-	->  print_message(informational, rdf(restore(snapshot(SnapshotFile)))),
+	->  print_message(Level, rdf(restore(snapshot(SnapshotFile)))),
 	    db_file(SnapshotFile, AbsFile),
 	    rdf_load_db(AbsFile)
 	;   true
 	),
 	(   exists_db(JournalFile)
-	->  print_message(informational, rdf(restore(journal(JournalFile)))),
+	->  print_message(Level, rdf(restore(journal(JournalFile)))),
 	    load_journal(JournalFile, DB)
 	;   true
 	),
@@ -231,17 +322,20 @@ load_source(DB) :-
 	->  true
 	;   Count = 0
 	),
-	print_message(informational, rdf(restore(done(DB, T, Count)))).
+	print_message(Level, rdf(restore(done(DB, T, Count)))).
 	
 	
 		 /*******************************
 		 *	   LOAD JOURNAL		*
 		 *******************************/
 	
+%%	load_journal(+File:atom, +DB:atom) is det.
+%
+%	Process transactions from the RDF journal File, adding the given
+%	named graph.
+
 load_journal(File, DB) :-
-	open_db(File, read, In,
-		[ encoding(utf8)
-		]),
+	open_db(File, read, In, []),
 	call_cleanup((  read(In, T0),
 			process_journal(T0, In, DB)
 		     ),
@@ -269,10 +363,12 @@ process_journal_term(update(S,P,O,Action), DB) :-
 	->  true
 	;   print_message(warning, rdf(update_failed(S,P,O,Action)))
 	).
-process_journal_term(start(_), _).
+process_journal_term(start(_), _).	% journal open/close
 process_journal_term(end(_), _).
-process_journal_term(begin(_), _).
+process_journal_term(begin(_), _).	% logged transaction (compatibility)
 process_journal_term(end, _).
+process_journal_term(begin(_,_,_,_), _).% logged transaction (current)
+process_journal_term(end(_,_,_), _).
 
 
 		 /*******************************
@@ -280,9 +376,37 @@ process_journal_term(end, _).
 		 *******************************/
 
 :- dynamic
-	blocked_db/1,			% DB
-	transaction_message/2,		% Id, Message
-	transaction_db/2.		% Id, DB
+	blocked_db/2,			% DB, Reason
+	transaction_message/3,		% Nesting, Time, Message
+	transaction_db/3.		% Nesting, DB, Id
+
+%%	rdf_persistency(+DB, Bool)
+%
+%	Specify whether a database is persistent.  Switching to =false=
+%       kills the persistent state.  Switching to =true= creates it.
+
+rdf_persistency(DB, Bool) :-
+	must_be(atom, DB),
+	must_be(boolean, Bool),
+	fail.
+rdf_persistency(DB, false) :- !,
+	(   blocked_db(DB, persistency)
+	->  true
+	;   assert(blocked_db(DB, persistency)),
+	    delete_db(DB)
+	).
+rdf_persistency(DB, true) :-
+	(   retract(blocked_db(DB, persistency))
+	->  create_db(DB)
+	;   true
+	).
+
+
+%%	start_monitor is det.
+%%	stop_monitor is det.
+%
+%	Start/stop monitoring the RDF database   for  changes and update
+%	the journal.
 
 start_monitor :-
 	rdf_monitor(monitor,
@@ -293,35 +417,42 @@ stop_monitor :-
 		    [ -all
 		    ]).
 
+%%	monitor(+Term) is semidet.
+%
+%	Handle an rdf_monitor/2 callback to  deal with persistency. Note
+%	that the monitor calls that come   from rdf_db.pl that deal with
+%	database changes are synchronised. They   do come from different
+%	threads though.
+
 monitor(Msg) :-
 	debug(monitor, 'Monitor: ~p~n', [Msg]),
 	fail.
 monitor(assert(S,P,O,DB:Line)) :- !,
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	journal_fd(DB, Fd),
 	open_transaction(DB, Fd),
 	format(Fd, '~q.~n', [assert(S,P,O,Line)]),
 	sync_journal(DB, Fd).
 monitor(assert(S,P,O,DB)) :-
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	journal_fd(DB, Fd),
 	open_transaction(DB, Fd),
 	format(Fd, '~q.~n', [assert(S,P,O)]),
 	sync_journal(DB, Fd).
 monitor(retract(S,P,O,DB:Line)) :- !,
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	journal_fd(DB, Fd),
 	open_transaction(DB, Fd),
 	format(Fd, '~q.~n', [retract(S,P,O,Line)]),
 	sync_journal(DB, Fd).
 monitor(retract(S,P,O,DB)) :-
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	journal_fd(DB, Fd),
 	open_transaction(DB, Fd),
 	format(Fd, '~q.~n', [retract(S,P,O)]),
 	sync_journal(DB, Fd).
 monitor(update(S,P,O,DB:Line,Action)) :- !,
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	(   Action = source(NewDB)
 	->  monitor(assert(S,P,O,NewDB)),
 	    monitor(retract(S,P,O,DB:Line))
@@ -330,7 +461,7 @@ monitor(update(S,P,O,DB:Line,Action)) :- !,
 	    sync_journal(DB, Fd)
 	).
 monitor(update(S,P,O,DB,Action)) :-
-	\+ blocked_db(DB),
+	\+ blocked_db(DB, _),
 	(   Action = source(NewDB)
 	->  monitor(assert(S,P,O,NewDB)),
 	    monitor(retract(S,P,O,DB))
@@ -347,47 +478,161 @@ monitor(load(BE, Id)) :-
 monitor(transaction(BE, Id)) :-
 	monitor_transaction(Id, BE).
 
-monitor_transaction(load_journal(DB), begin) :- !,
-	assert(blocked_db(DB)).
-monitor_transaction(load_journal(DB), end) :- !,
-	retractall(blocked_db(DB)).
+monitor_transaction(load_journal(DB), begin(_)) :- !,
+	assert(blocked_db(DB, journal)).
+monitor_transaction(load_journal(DB), end(_)) :- !,
+	retractall(blocked_db(DB, journal)).
 
-monitor_transaction(parse(file(DB)), begin) :- !,
-	assert(blocked_db(DB)).
-monitor_transaction(parse(file(DB)), end) :- !,
-	retractall(blocked_db(DB)),
-	create_db(DB).
-monitor_transaction(unload(DB), begin) :- !,
-	assert(blocked_db(DB)).
-monitor_transaction(unload(DB), end) :- !,
-	retractall(blocked_db(DB)),
-	delete_db(DB).
-monitor_transaction(log(Msg), begin) :- !,
-	(   predicate_property(transaction_message(_,_),
-			       number_of_clauses(N))
+monitor_transaction(parse(URI), begin(_)) :- !,
+	(   blocked_db(URI, persistency)
 	->  true
-	;   N = 0
-	),
-	asserta(transaction_message(N, Msg)).
-monitor_transaction(log(_), end) :- !,
-	retract(transaction_message(N, _)),
-	(   retract(transaction_db(N, DB)),
-	    journal_fd(DB, Fd),
-	    format(Fd, 'end.~n', []),
-	    sync_journal(DB, Fd),
-	    fail
+	;   assert(blocked_db(URI, parse))
+	).
+monitor_transaction(parse(URI), end(_)) :- !,
+	(   retract(blocked_db(URI, parse))
+	->  create_db(URI)
 	;   true
 	).
+monitor_transaction(unload(DB), begin(_)) :- !,
+	(   blocked_db(DB, persistency)
+	->  true
+	;   assert(blocked_db(DB, unload))
+	).
+monitor_transaction(unload(DB), end(_)) :- !,
+	(   retract(blocked_db(DB, unload))
+	->  delete_db(DB)
+	;   true
+	).
+monitor_transaction(log(Msg), begin(N)) :- !,
+	check_nested(N),
+	get_time(Time),
+	asserta(transaction_message(N, Time, Msg)).
+monitor_transaction(log(_), end(N)) :-
+	check_nested(N),
+	retract(transaction_message(N, _, _)), !,
+	findall(DB:Id, retract(transaction_db(N, DB, Id)), DBs),
+	end_transactions(DBs, N).    
+monitor_transaction(log(Msg, DB), begin(N)) :- !,
+	check_nested(N),
+	get_time(Time),
+	asserta(transaction_message(N, Time, Msg)),
+	journal_fd(DB, Fd),
+	open_transaction(DB, Fd).
+monitor_transaction(log(Msg, _DB), end(N)) :-
+	monitor_transaction(log(Msg), end(N)).
+monitor_transaction(reset, begin(L)) :-
+	forall(rdf_graph(DB),
+	       monitor_transaction(unload(DB), begin(L))).
+monitor_transaction(reset, end(L)) :-
+	forall(blocked_db(DB, unload),
+	       monitor_transaction(unload(DB), end(L))).
+	       
+
+%%	check_nested(+Level) is semidet.
+%
+%	True if we must log this transaction.   This  is always the case
+%	for toplevel transactions. Nested transactions   are only logged
+%	if log_nested_transactions(true) is defined.
+
+check_nested(0) :- !.
+check_nested(_) :-
+	rdf_option(log_nested_transactions(true)).
+
+
+%%	open_transaction(+DB, +Fd) is det.
+%
+%	Add a begin(Id, Level, Time,  Message)   term  if  a transaction
+%	involves DB. Id is an incremental   integer, where each database
+%	has its own counter. Level is the nesting level, Time a floating
+%	point timestamp and Message te message   provided as argument to
+%	the log message.
 
 open_transaction(DB, Fd) :-
-	transaction_message(N, Msg), !,
-	(   transaction_db(N, DB)
+	transaction_message(N, Time, Msg), !,
+	(   transaction_db(N, DB, _)
 	->  true
-	;   assert(transaction_db(N, DB)),
-	    format(Fd, 'begin(~q).~n', [Msg])
+	;   next_transaction_id(DB, Id),
+	    assert(transaction_db(N, DB, Id)),
+	    format(Fd, 'begin(~q, ~q, ~2f, ~q).~n', [Id, N, Time, Msg])
 	).
 open_transaction(_,_).
 
+
+%%	next_transaction_id(+DB, -Id) is det.
+%
+%	Id is the number to user for  the next logged transaction on DB.
+%	Transactions in each  named  graph   are  numbered  in sequence.
+%	Searching the Id of the last transaction is performed by the 2nd
+%	clause starting 1Kb from the end   and doubling this offset each
+%	failure.
+
+:- dynamic
+	current_transaction_id/2.
+
+next_transaction_id(DB, Id) :-
+	retract(current_transaction_id(DB, Last)), !,
+	Id is Last + 1,
+	assert(current_transaction_id(DB, Id)).
+next_transaction_id(DB, Id) :-
+	db_files(DB, _, Journal),
+	catch(open_db(Journal, read, In, []),
+	      error(existence_error(_,_),_),
+	      fail), !,
+	call_cleanup(iterative_expand(In, Last), close(In)),
+	Id is Last + 1,
+	assert(current_transaction_id(DB, Id)).
+next_transaction_id(DB, 1) :-
+	assert(current_transaction_id(DB, 1)).
+
+iterative_expand(In, Last) :-
+	between(10, 62, Step),		% 62 is big enough and safer than infinite
+	    Offset is -(1<<Step),
+	    catch(seek(In, Offset, eof, _), E, true),
+	    (	var(E)
+	    ->	skip(In, 10)		% records are line-based
+	    ;	seek(In, 0, bof, _), !,
+		WholeFile = true
+	    ),
+	    read(In, T0),
+	    last_transaction_id(T0, In, 0, Last),
+	    (	WholeFile == true	% we scanned the whole file
+	    ;   Last > 0		% we found a transaction
+	    ), !.
+iterative_expand(_, 0).
+
+last_transaction_id(end_of_file, _, Last, Last) :- !.
+last_transaction_id(end(Id, _, _), In, _, Last) :-
+	read(In, T1),
+	last_transaction_id(T1, In, Id, Last).
+last_transaction_id(_, In, Id, Last) :-
+	read(In, T1),
+	last_transaction_id(T1, In, Id, Last).
+
+
+%%	end_transactions(+DBs:list(atom:id)) is det.
+%
+%	End a transaction that affected the  given list of databases. We
+%	write the list of other affected databases as an argument to the
+%	end-term to facilitate fast finding of the related transactions.
+%	
+%	In each database, the transaction is   ended with a term end(Id,
+%	Nesting, Others), where  Id  and   Nesting  are  the transaction
+%	identifier and nesting (see open_transaction/2)  and Others is a
+%	list of DB:Id,  indicating  other   databases  affected  by  the
+%	transaction.
+
+end_transactions(DBs, N) :-
+	end_transactions(DBs, DBs, N).
+
+end_transactions([], _, _).
+end_transactions([DB:Id|T], DBs, N) :-
+	journal_fd(DB, Fd),
+	once(select(DB:Id, DBs, Others)),
+	format(Fd, 'end(~q, ~q, ~q).~n', [Id, N, Others]),
+	sync_journal(DB, Fd),
+	end_transactions(T, DBs, N).
+
+
 %	State  handling.  We  use   this    for   trapping   changes  by
 %	rdf_load_db/1. In theory, loading such files  can add triples to
 %	multiple sources. In practice this rarely   happens. We save the
@@ -404,7 +649,7 @@ push_state(Id) :-
 	asserta(pre_load_state(Id, State)).
 
 get_state(State) :-
-	findall(DB-MD5, (rdf_source(DB), rdf_md5(DB, MD5)), State0),
+	findall(DB-MD5, (rdf_graph(DB), rdf_md5(DB, MD5)), State0),
 	keysort(State0, State).
 
 sync_state(Id) :-
@@ -426,9 +671,13 @@ sync_state([DB-MD5|TA], Pre) :-
 		 *	   JOURNAL FILES	*
 		 *******************************/
 
-%%	journal_db(+DB, -Stream)
+%%	journal_fd(+DB, -Stream) is det.
 %	
-%	Open existing journal or create new journal for database DB
+%	Get an open stream to a journal. If the journal is not open, old
+%	journals are closed to satisfy   the =max_open_journals= option.
+%	Then the journal is opened in   =append= mode. Journal files are
+%	always encoded as UTF-8 for  portability   as  well as to ensure
+%	full coverage of Unicode.
 
 journal_fd(DB, Fd) :-
 	source_journal_fd(DB, Fd), !.
@@ -440,14 +689,39 @@ journal_fd(DB, Fd) :-
 journal_fd_(DB, Fd) :-
 	source_journal_fd(DB, Fd), !.
 journal_fd_(DB, Fd) :-
+	limit_fd_pool,
 	db_files(DB, _Snapshot, Journal),
 	open_db(Journal, append, Fd,
-		[ encoding(utf8),
-		  close_on_abort(false)
+		[ close_on_abort(false)
 		]),
 	time_stamp(Now),
 	format(Fd, '~q.~n', [start([time(Now)])]),
-	assert(source_journal_fd(DB, Fd)).
+	assert(source_journal_fd(DB, Fd)).		% new one at the end
+
+%%	limit_fd_pool is det.
+%
+%	Limit the number of  open   journals  to max_open_journals (10).
+%	Note that calls  from  rdf_monitor/2   are  issued  in different
+%	threads, but as they are part of write operations they are fully
+%	synchronised.
+
+limit_fd_pool :-
+	predicate_property(source_journal_fd(_, _), number_of_clauses(N)), !,
+	(   rdf_option(max_open_journals(Max))
+	->  true
+	;   Max = 10
+	),
+	Close is N - Max,
+	forall(between(1, Close, _),
+	       close_oldest_journal).
+limit_fd_pool.
+
+close_oldest_journal :-
+	source_journal_fd(DB, _Fd), !,
+	debug(rdf_persistency, 'Closing old journal for ~q', [DB]),
+	close_journal(DB).
+close_oldest_journal.
+		       
 
 %%	sync_journal(+DB, +Fd)
 %	
@@ -456,13 +730,13 @@ journal_fd_(DB, Fd) :-
 %	end of the transaction.
 
 sync_journal(DB, _) :-
-	transaction_db(_, DB), !.
+	transaction_db(_, DB, _), !.
 sync_journal(_, Fd) :-
 	flush_output(Fd).
 
-%%	close_journal(+DB)
+%%	close_journal(+DB) is det.
 %	
-%	Close the journal associated with DB if it is open
+%	Close the journal associated with DB if it is open.
 
 close_journal(DB) :-
 	with_mutex(rdf_journal_file,
@@ -534,8 +808,12 @@ delete_db(DB) :-
 lock_db(Dir) :-
 	lockfile(Dir, File),
 	exists_file(File), !,
-	throw(error(permission_error(lock, Dir),
-		    context(_, 'Database is in use'))).
+	(   catch(read_file_to_terms(File, Terms, []), _, fail),
+	    Terms = [locked(Args)]
+	->  Context = rdf_locked(Args)
+	;   Context = context(_, 'Database is in use')
+	),
+	throw(error(permission_error(lock, rdf_db, Dir), Context)).
 lock_db(Dir) :-
 	lockfile(Dir, File),
 	open(File, write, Out),
@@ -572,7 +850,7 @@ db_file(Base, File) :-
 
 open_db(Base, Mode, Stream, Options) :-
 	db_file(Base, File),
-	open(File, Mode, Stream, Options).
+	open(File, Mode, Stream, [encoding(utf8)|Options]).
 
 exists_db(Base) :-
 	db_file(Base, File),
@@ -581,6 +859,10 @@ exists_db(Base) :-
 %%	db_files(+DB, -Snapshot, -Journal).
 %%	db_files(-DB, +Snapshot, -Journal).
 %%	db_files(-DB, -Snapshot, +Journal).
+%
+%	True if named graph DB is represented  by the files Snapshot and
+%	Journal. The filenames are local   to the directory representing
+%	the store.
 
 db_files(DB, Snapshot, Journal) :-
 	nonvar(DB), !,
@@ -603,13 +885,21 @@ db_abs_files(DB, Snapshot, Journal) :-
 	db_file(Snapshot0, Snapshot),
 	db_file(Journal0, Journal).
 
-%%	rdf_journal_file(?DB, -File)
+
+%%	rdf_journal_file(+DB, -File) is semidet.
+%%	rdf_journal_file(-DB, -File) is nondet.
 %	
-%	Return the journal files of  the   current  server. Intended for
-%	external modules for merging journals.
+%	True if File is the absolute  file   name  of  an existing named
+%	graph DB. 
+%	
+%	@tbd	Avoid using private rdf_db:rdf_graphs_/1.
 
 rdf_journal_file(DB, Journal) :-
-	rdf_source(DB),
+	(   var(DB)
+	->  rdf_db:rdf_graphs_(All),	% also pick the empty graphs
+	    member(DB, All)
+	;   true
+	),
 	db_abs_files(DB, _Snapshot, Journal),
 	exists_file(Journal).
 
@@ -681,7 +971,7 @@ alphanum(C) -->
 	  code_type(C, alnum)
 	}.
 
-no_enc_extra(0'_) --> "_".
+no_enc_extra(0'_) --> "_".		%'
 
 
 		 /*******************************
@@ -711,7 +1001,8 @@ time_stamp(Int) :-
 		 *******************************/
 
 :- multifile
-	prolog:message/3.
+	prolog:message/3,
+	prolog:message_context/3.
 
 prolog:message(rdf(restore(source(DB)))) -->
 	{ file_base_name(DB, Base) },
@@ -724,3 +1015,12 @@ prolog:message(rdf(restore(done(_, Time, Count)))) -->
 	[ at_same_line, '~D triples in ~2f sec.'-[Count, Time] ].
 prolog:message(rdf(update_failed(S,P,O,Action))) -->
 	[ 'Failed to update <~p ~p ~p> with ~p'-[S,P,O,Action] ].
+
+prolog:message_context(rdf_locked(Args)) -->
+	{ memberchk(time(Time), Args),
+	  memberchk(pid(Pid), Args),
+	  format_time(string(S), '%+', Time)
+	},
+	[ nl,
+	  'locked at ~s by process id ~w'-[S,Pid] 
+	].
