@@ -1,4 +1,4 @@
-/*  $Id: rdf_parser.pl,v 1.41 2006/08/31 09:44:01 jan Exp $
+/*  $Id: rdf_parser.pl,v 1.43 2007/07/09 13:44:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -250,11 +250,14 @@ propertyElt(Id, Name, unparsed(Value), Options) ::=
 		       ]),
 		Value).
 
-emptyPropertyElt(Id, literal(''), Options) ::=
+emptyPropertyElt(Id, Literal, Options) ::=
 	\attrs([ \?idAttr(Id, Options),
 		 \?parseLiteral
 	       | \noMoreAttrs
-	       ]), !.
+	       ]),
+	{ !,
+	  mkliteral('', Literal, Options)
+	}.
 emptyPropertyElt(Id,
 		 description(description, About, BagID, Properties),
 		 Options) ::=
@@ -272,20 +275,19 @@ aboutResourceEmptyElt(node(URI), _Options) ::=
 
 %%	literal_value(+In, -Value, +Options)
 %	
-%	Translate a literal into its  value.   Notably  if  the value is
-%	plain  CDATA,  remove  the   list.    This   predicate   handles
-%	parseType="Literal" attributes.
-
-literal_value([Value], Literal, Options) :-
-	atomic(Value), !,
-	mkliteral(Value, Literal, Options).
-literal_value(Value, literal(Value), _).
+%	Create the literal value for rdf:parseType="Literal" attributes.
+%	The content is the Prolog XML DOM tree for the literal.
+%	
+%	@tbd	Note that the specs demand a canonical textual representation
+%		of the XML data as a Unicode string.  For now the user can
+%		achieve this using the convert_typed_literal hook.
 
+literal_value(Value, literal(type(rdf:'XMLLiteral', Value)), _).
 
 %%	mkliteral(+Atom, -Object, +Options)
 %	
 %	Translate attribute value Atom into an RDF object using the
-%%	lang(Lang) option from Options.
+%	lang(Lang) option from Options.
 
 mkliteral(Text, literal(Val), Options) :-
 	atom(Text),
