@@ -1,4 +1,4 @@
-/*  $Id: ssllib.h,v 1.5 2007/01/21 13:29:23 jan Exp $
+/*  $Id: ssllib.h,v 1.6 2007/02/26 13:50:02 jan Exp $
 
     Part of SWI-Prolog
 
@@ -24,6 +24,7 @@
 
 #ifndef SSLLIBH__
 #define SSLLIBH__
+#include "../clib/nonblockio.h"
 
 #ifdef __WINDOWS__
 typedef size_t socklen_t;
@@ -104,7 +105,7 @@ typedef struct pl_ssl {
 typedef struct ssl_instance {
     PL_SSL              *config;
     SSL                 *ssl;
-    int                  sock;
+    nbio_sock_t          sock;
     int                  close_needed;
 } PL_SSL_INSTANCE;
 
