@@ -1,4 +1,4 @@
-/*  $Id: prolog_colour.pl,v 1.78 2004/09/21 14:22:25 jan Exp $
+/*  $Id: prolog_colour.pl,v 1.100 2007/07/09 14:51:45 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -43,6 +43,8 @@ User extension hooks.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 :- multifile
+	style/2,
+	identify/2,
 	term_colours/2,
 	goal_colours/2,
 	goal_classification/2.
@@ -80,9 +82,12 @@ reload_styles(M) :->
 	"Force reloading the styles"::
 	retractall(style_name(_,_)),
 	(   style(Class, Name, Style),
-	    assert(style_name(Class, Name)),
-	    Style \== @default,
-	    send(M, style, Name, Style),
+	    (	style_name(_, Name)
+	    ->  true			% redefined
+	    ;   assert(style_name(Class, Name)),
+		Style \== @default,
+		send(M, style, Name, Style)
+	    ),
 	    fail
 	;   true
 	).
@@ -177,8 +182,10 @@ colourise_buffer(Fd, M) :-
 	    (	colourise_term(Term, TB, TermPos),
 		send(M, mark_singletons, Term, Singletons, TermPos)
 	    ->	true
-	    ;	print_message(warning,
-			      format('Failed to colourise ~p~n', [Term]))
+	    ;	arg(1, TermPos, From),
+	        print_message(warning,
+			      format('Failed to colourise ~p at index ~d~n',
+				     [Term, From]))
 	    ),
 	    Term == end_of_file, !.
 
@@ -232,6 +239,8 @@ fix_operators(_, _).
 
 process_directive(style_check(X), _) :- !,
 	style_check(X).
+process_directive(system_mode(on), _) :- !,
+	style_check(+dollar).
 process_directive($(Name), _) :-
 	atom(Name),
 	style_check(+dollar),
@@ -260,10 +269,11 @@ process_use_module([H|T], Src) :-
 	process_use_module(H, Src),
 	process_use_module(T, Src).
 process_use_module(File, Src) :-
-	xref_public_list(File, _Path, Public, Src),
-	forall(member(op(P,T,N), Public),
-	       emacs_push_op(P,T,N)).
-
+	(   xref_public_list(File, _Path, Public, Src)
+	->  forall(member(op(P,T,N), Public),
+		   emacs_push_op(P,T,N))
+	;   true
+	).
 
 %	colourise(+TB, +Stream)
 %
@@ -498,6 +508,13 @@ colourise_dcg_goal(Goal, _, TB, Pos) :-
 %	
 %	Colourise access to a single goal.
 
+					% Deal with list as goal (consult)
+colourise_goal(Goal, _, TB, list_position(F,T,Elms,_)) :- !,
+	FT is F + 1,
+	AT is T - 1,
+	colour_item(goal(built_in, Goal), TB, F-FT),
+	colour_item(goal(built_in, Goal), TB, AT-T),
+	colourise_file_list(Goal, TB, Elms).
 colourise_goal(Goal, Origin, TB, Pos) :-
 	nonvar(Goal),
 	goal_colours(Goal, ClassSpec-ArgSpecs), !, % specified
@@ -620,7 +637,8 @@ colourise_files(Var, TB, P) :-
 	colour_item(var, TB, P).
 colourise_files(Spec0, TB, Pos) :-
 	strip_module(Spec0, _, Spec),
-	(   catch(xref_source_file(Spec, Path, TB), _, fail)
+	(   TB = @SourceId,		% HACK
+	    catch(xref_source_file(Spec, Path, SourceId), _, fail)
 	->  colour_item(file(Path), TB, Pos)
 	;   colour_item(nofile, TB, Pos)
 	).
@@ -630,6 +648,21 @@ colourise_file_list([H|T], TB, [PH|PT]) :-
 	colourise_files(H, TB, PH),
 	colourise_file_list(T, TB, PT).
 
+
+%%	colourise_directory(+Arg, +TB, +Pos)
+%
+%	Colourise argument that should be an existing directory.
+
+colourise_directory(Spec, TB, Pos) :-
+	(   TB = @SourceId,		% HACK
+	    catch(xref_source_file(Spec, Path, SourceId,
+				   [file_type(directory)]),
+		  _, fail)
+	->  colour_item(directory(Path), TB, Pos)
+	;   colour_item(nofile, TB, Pos)
+	).
+
+
 %	colourise_class(ClassName, TB, Pos)
 %
 %	Colourise an XPCE class.  
@@ -700,6 +733,33 @@ colourise_exports2([G0|GT], TB, [P0|PT]) :- !,
 	colourise_exports2(GT, TB, PT).
 colourise_exports2(_, _, _).
 
+
+%	colourise_imports(+List, +File, +TB, +Pos)
+%	
+%	Colourise import list from use_module/2, importing from File.
+
+colourise_imports(List, File, TB, Pos) :-
+	(   xref_public_list(File, _, Public, TB)
+	->  true
+	;   Public = []
+	),
+	colourise_imports(List, File, Public, TB, Pos).
+
+colourise_imports([], _, _, _, _).
+colourise_imports(List, File, Public, TB, list_position(_,_,ElmPos,Tail)) :- !,
+	(   Tail == none
+	->  true
+	;   colour_item(type_error(list), TB, Tail)
+	),
+	colourise_imports2(List, File, Public, TB, ElmPos).
+colourise_imports(_, _, _, TB, Pos) :-
+	colour_item(type_error(list), TB, Pos).	
+
+colourise_imports2([G0|GT], File, Public, TB, [P0|PT]) :- !,
+	colourise_declaration(G0, TB, P0),
+	colourise_imports2(GT, File, Public, TB, PT).
+colourise_imports2(_, _, _, _, _).
+
 %	colourise_declarations(+Term, +TB, +Pos)
 %	
 %	Colourise the Name/Arity lists of dynamic, multifile, etc
@@ -713,7 +773,8 @@ colourise_declarations(Last, TB, Pos) :-
 	colourise_declaration(Last, TB, Pos).
 
 colourise_declaration($(Name)/Arity, TB, Pos) :-
-	style_check(?dollar), !,	% deal with system boot-files
+	atom(Name),
+	style_check(?(dollar)), !,	% deal with system boot-files
 	atom_concat($, Name, TheName),
 	colourise_declaration(TheName/Arity, TB, Pos).
 colourise_declaration(Name/Arity, TB, Pos) :-
@@ -834,8 +895,13 @@ goal_classification(SS, expanded) :-	% XPCE (TBD)
 
 classify_head(TB, Goal, exported) :-
 	xref_exported(TB, Goal), !.
+classify_head(_TB, Goal, hook) :-
+	xref_hook(Goal), !.
+classify_head(TB, Goal, hook) :-
+	xref_module(TB, M),
+	xref_hook(M:Goal), !.
 classify_head(TB, Goal, unreferenced) :-
-	\+ xref_called(TB, Goal), !.
+	\+ (xref_called(TB, Goal, By), By \= Goal), !.
 classify_head(TB, Goal, How) :-
 	xref_defined(TB, Goal, How), !.
 classify_head(_TB, Goal, built_in) :-
@@ -852,16 +918,14 @@ built_in_predicate(module(_, _)).
 
 goal_colours(module(_,_),	     built_in-[identifier,exports]).
 goal_colours(use_module(_),	     built_in-[file]).
-goal_colours(use_module(_,_),	     built_in-[file,classify]).
+goal_colours(use_module(File,_),     built_in-[file,imports(File)]).
 goal_colours(dynamic(_),	     built_in-[predicates]).
 goal_colours(thread_local(_),	     built_in-[predicates]).
 goal_colours(multifile(_),	     built_in-[predicates]).
 goal_colours(volatile(_),	     built_in-[predicates]).
 goal_colours(consult(_),	     built_in-[file]).
-goal_colours([_|_],		     built_in-file).
 goal_colours(include(_),	     built_in-[file]).
 goal_colours(ensure_loaded(_),	     built_in-[file]).
-goal_colours(load_files(_),	     built_in-[file]).
 goal_colours(load_files(_,_),	     built_in-[file,classify]).
 % Database access
 goal_colours(assert(_),		     built_in-[db]).
@@ -876,13 +940,14 @@ goal_colours(clause(_,_),	     built_in-[db,classify]).
 goal_colours(clause(_,_,_),	     built_in-[db,classify,classify]).
 % XPCE stuff
 goal_colours(pce_autoload(_,_),	     classify-[classify,file]).
-goal_colours(pce_image_directory(_), classify-[file]).
+goal_colours(pce_image_directory(_), classify-[directory]).
 goal_colours(new(_, _),		     built_in-[classify,pce_new]).
 goal_colours(send_list(_,_,_),	     built_in-pce_arg_list).
 goal_colours(send(_,_),		     built_in-[pce_arg,pce_selector]).
 goal_colours(get(_,_,_),	     built_in-[pce_arg,pce_selector,pce_arg]).
 goal_colours(send_super(_,_),	     built_in-[pce_arg,pce_selector]).
 goal_colours(get_super(_,_),	     built_in-[pce_arg,pce_selector,pce_arg]).
+goal_colours(get_chain(_,_,_),	     built_in-[pce_arg,pce_selector,pce_arg]).
 goal_colours(Pce,		     built_in-pce_arg) :-
 	compound(Pce),
 	functor(Pce, Functor, _),
@@ -900,13 +965,12 @@ pce_functor(get_super).
 
 head_colours(file_search_path(_,_), hook-[identifier,classify]).
 head_colours(library_directory(_),  hook-[file]).
-head_colours(message_hook(_,_,_),   hook-[classify,classify,classify]).
-head_colours(portray(_),	    hook-[classify]).
 head_colours(resource(_,_,_),	    hook-[identifier,classify,file]).
-head_colours(term_expansion(_,_),   hook-[classify,classify]).
-head_colours(goal_expansion(_,_),   hook-[classify,classify]).
-head_colours(prolog_trace_interception(_,_,_,_),
-				    hook-classify).
+
+head_colours(M:H, Colours) :-
+	atom(M), callable(H),
+	xref_hook(M:H), !,
+	Colours = hook - [ hook, hook-classify ].
 head_colours(M:H, Colours) :-
 	M == user,
 	head_colours(H, HC),
@@ -919,79 +983,85 @@ head_colours(M:_,		    meta-[module(M),extern(M)]).
 		 *	       STYLES		*
 		 *******************************/
 
-%	style(+Pattern, -Style)
+%	def_style(+Pattern, -Style)
 %
-%	Define the style used for the given pattern.
+%	Define the style used for the   given  pattern. Definitions here
+%	can     be     overruled     by       defining     rules     for
+%	emacs_prolog_colours:style/2
 
-style(goal(built_in,_),	  style(colour	   := blue)).
-style(goal(imported(_),_),style(colour	   := blue)).
-style(goal(autoload,_),	  style(colour	   := navy_blue)).
-style(goal(global,_),	  style(colour	   := navy_blue)).
-style(goal(undefined,_),  style(colour	   := red)).
-style(goal(thread_local,_), style(colour   := magenta,
-				  underline:= @on)).
-style(goal(dynamic,_),	  style(colour	   := magenta)).
-style(goal(multifile,_),  style(colour	   := navy_blue)).
-style(goal(expanded,_),	  style(colour	   := blue,
-				underline  := @on)).
-style(goal(extern(_),_),  style(colour	   := blue,
-				underline  := @on)).
-style(goal(recursion,_),  style(underline  := @on)).
-style(goal(meta,_),	  style(colour	   := red4)). % same as var
-style(goal(local(_),_),	  @default).
-style(goal(constraint(_),_), style(colour := darkcyan)).
+def_style(goal(built_in,_),	style(colour := blue)).
+def_style(goal(imported(_),_),	style(colour := blue)).
+def_style(goal(autoload,_), 	style(colour := navy_blue)).
+def_style(goal(global,_),	style(colour := navy_blue)).
+def_style(goal(undefined,_),	style(colour := red)).
+def_style(goal(thread_local(_),_), style(colour := magenta,
+				      underline:= @on)).
+def_style(goal(dynamic(_),_), 	style(colour := magenta)).
+def_style(goal(multifile(_),_),	style(colour := navy_blue)).
+def_style(goal(expanded,_),	style(colour := blue,
+				      underline := @on)).
+def_style(goal(extern(_),_),	style(colour := blue,
+				      underline := @on)).
+def_style(goal(recursion,_),	style(underline := @on)).
+def_style(goal(meta,_),		style(colour := red4)). % same as var
+def_style(goal(foreign(_),_),	style(colour := darkturquoise)).
+def_style(goal(local(_),_),	@default).
+def_style(goal(constraint(_),_), style(colour := darkcyan)).
 
-style(head(exported),	  style(bold	   := @on, colour := blue)).
-style(head(extern(_)),	  style(bold	   := @on, colour := blue)).
-style(head(dynamic),	  style(bold	   := @on, colour := magenta)).
-style(head(multifile),	  style(bold	   := @on, colour := navy_blue)).
-style(head(unreferenced), style(bold	   := @on, colour := red)).
-style(head(hook),	  style(underline  := @on, colour := blue)).
-style(head(meta),	  @default).
-style(head(constraint(_)),style(bold	   := @on, colour := darkcyan)).
-style(head(_),	  	  style(bold	   := @on)).
+def_style(head(exported),	style(bold := @on, colour := blue)).
+def_style(head(extern(_)),	style(bold := @on, colour := blue)).
+def_style(head(dynamic),	style(bold := @on, colour := magenta)).
+def_style(head(multifile),	style(bold := @on, colour := navy_blue)).
+def_style(head(unreferenced),	style(bold := @on, colour := red)).
+def_style(head(hook),	  	style(underline  := @on, colour := blue)).
+def_style(head(meta),	  	@default).
+def_style(head(constraint(_)),	style(bold := @on, colour := darkcyan)).
+def_style(head(_),	  	style(bold := @on)).
 
-style(comment,		  style(colour	   := dark_green)).
+def_style(comment,		style(colour := dark_green)).
 
-style(directive,	  style(background := grey90)).
-style(method(_),	  style(bold       := @on)).
+def_style(directive,	  	style(background := grey90)).
+def_style(method(_),	  	style(bold := @on)).
 
-style(var,		  style(colour	   := red4)).
-style(unbound,		  style(bold	   := @on, colour := red)).
-style(quoted_atom,        style(colour	   := navy_blue)).
-style(string,		  style(colour	   := navy_blue)).
-style(nofile,		  style(colour	   := red)).
-style(file(_),		  style(colour	   := blue,
-				underline  := @on)).
-style(class(built_in,_),  style(colour	   := blue,
-				underline  := @on)).
-style(class(library(_),_),style(colour	   := navy_blue,
-				underline  := @on)).
-style(class(local(_,_,_),_),	  style(underline  := @on)).
-style(class(user(_),_),	  style(underline  := @on)).
-style(class(user,_),	  style(underline  := @on)).
-style(class(undefined,_), style(colour	   := red,
-				underline  := @on)).
-style(prolog_data,	  style(colour	   := blue,
-				underline  := @on)).
+def_style(var,		  	style(colour := red4)).
+def_style(unbound,		style(bold := @on, colour := red)).
+def_style(quoted_atom,        	style(colour := navy_blue)).
+def_style(string,		style(colour := navy_blue)).
+def_style(nofile,		style(colour := red)).
+def_style(file(_),		style(colour := blue,
+				      underline  := @on)).
+def_style(directory(_),		style(colour := blue)).
+def_style(class(built_in,_),	style(colour := blue,
+				      underline := @on)).
+def_style(class(library(_),_),	style(colour := navy_blue,
+				      underline := @on)).
+def_style(class(local(_,_,_),_), style(underline := @on)).
+def_style(class(user(_),_),	style(underline := @on)).
+def_style(class(user,_), 	style(underline := @on)).
+def_style(class(undefined,_),	style(colour := red,
+				      underline  := @on)).
+def_style(prolog_data,		style(colour := blue,
+				      underline  := @on)).
 
-style(identifier,	  style(bold       := @on)).
-style(delimiter,	  style(bold       := @on)).
-style(expanded,		  style(colour	   := blue,
-				underline  := @on)).
+def_style(identifier, 		style(bold := @on)).
+def_style(delimiter,		style(bold := @on)).
+def_style(expanded,		style(colour := blue,
+				      underline  := @on)).
 
-style(hook,		  style(colour	   := blue,
-				underline  := @on)).
+def_style(hook,			style(colour := blue,
+				      underline := @on)).
 
-style(error,		  style(background := orange)).
-style(type_error(_),	  style(background := orange)).
-style(syntax_error,	  style(background := red)).
+def_style(error,		style(background := orange)).
+def_style(type_error(_),	style(background := orange)).
+def_style(syntax_error,	  	style(background := red)).
 
 :- dynamic
 	style_name/2.
 
 style(Class, Name, Style) :-
-	style(Class, Style),
+	(   style(Class, Style)		% user hook
+	;   def_style(Class, Style)	% system default
+	),
 	copy_term(Class, Copy),
 	numbervars(Copy, 0, _),
 	term_to_atom(Copy, Name).
@@ -1017,6 +1087,13 @@ term_colours((prolog:error_message(_) --> _),
 				     ],
 			  classify
 			]).
+term_colours((prolog:message_context(_) --> _),
+	     expanded - [ expanded - [ expanded,
+				       expanded - [ identifier
+						  ]
+				     ],
+			  classify
+			]).
 
 %	XPCE rules
 
@@ -1056,6 +1133,10 @@ term_colours(class_variable(_,_,_),
 term_colours(delegate_to(_),
 	     expanded - [ classify
 			]).
+term_colours((:- encoding(_)),
+	     expanded - [ expanded - [ classify
+				     ]
+			]).
 term_colours((:- pce_begin_class(_, _, _)),
 	     expanded - [ expanded - [ identifier,
 				       pce_new,
@@ -1134,15 +1215,24 @@ specified_item(extern(M), Term, TB, Pos) :- !,
 					% classify as body
 specified_item(body, Term, TB, Pos) :- !,
 	colourise_body(Term, TB, Pos).
+					% DCG goal in body
+specified_item(dcg, Term, TB, Pos) :- !,
+	colourise_dcg(Term, [], TB, Pos).
 					% assert/retract arguments
 specified_item(db, Term, TB, Pos) :- !,
 	colourise_db(Term, TB, Pos).
 					% files
 specified_item(file, Term, TB, Pos) :- !,
 	colourise_files(Term, TB, Pos).
+					% directory
+specified_item(directory, Term, TB, Pos) :- !,
+	colourise_directory(Term, TB, Pos).
 					% [Name/Arity, ...]
 specified_item(exports, Term, TB, Pos) :- !,
 	colourise_exports(Term, TB, Pos).
+					% [Name/Arity, ...]
+specified_item(imports(File), Term, TB, Pos) :- !,
+	colourise_imports(Term, File, TB, Pos).
 					% Name/Arity, ...
 specified_item(predicates, Term, TB, Pos) :- !,
 	colourise_declarations(Term, TB, Pos).
@@ -1199,6 +1289,12 @@ specified_item(FuncSpec-[ArgSpec], {Term}, TB,
 	specified_item(FuncSpec, {Term}, TB, F-T),
 	specified_item(ArgSpec, Term, TB, ArgPos).
 					% Specified
+specified_item(FuncSpec-ElmSpec, List, TB, list_position(F,T,ElmPos,TailPos)) :- !,
+	FT is F + 1,
+	AT is T - 1,
+	colour_item(FuncSpec, TB, F-FT),
+	colour_item(FuncSpec, TB, AT-T),
+	specified_list(ElmSpec, List, TB, ElmPos, TailPos).
 specified_item(Class, _, TB, Pos) :-
 	colour_item(Class, TB, Pos).
 
@@ -1212,6 +1308,7 @@ specified_items(Spec, Term, TB, PosList) :-
 
 
 specified_arglist([], _, _, _, _).
+specified_arglist(_, _, _, _, []) :- !.			% Excess specification args
 specified_arglist([S0|ST], N, T, TB, [P0|PT]) :-
 	arg(N, T, Term),
 	specified_item(S0, Term, TB, P0),
@@ -1225,6 +1322,20 @@ specified_argspec([P0|PT], Spec, N, T, TB) :-
 	NN is N + 1,
 	specified_argspec(PT, Spec, NN, T, TB).
 
+
+%	specified_list(+Spec, +List, +TB, +PosList, TailPos)
+
+specified_list([], [], _, [], _).
+specified_list([HS|TS], [H|T], TB, [HP|TP], TailPos) :- !,
+	specified_item(HS, H, TB, HP),
+	specified_list(TS, T, TB, TP, TailPos).
+specified_list(Spec, [H|T], TB, [HP|TP], TailPos) :-
+	specified_item(Spec, H, TB, HP),
+	specified_list(Spec, T, TB, TP, TailPos).
+specified_list(_, _, _, [], none) :- !.
+specified_list(Spec, Tail, TB, [], TailPos) :-
+	specified_item(Spec, Tail, TB, TailPos).
+
 :- emacs_end_mode.
 
 
@@ -1318,15 +1429,11 @@ has_source(F) :->
 	"Test if there is source available"::
 	get(F, text_buffer, TB),
 	get(F, head, Head),
-	(   xref_defined(TB, Head, local(_Line))
-	->  true
-	;   xref_defined(TB, Head, constraint(_Line))
-	->  true
+	(   xref_defined(TB, Head, How),
+	    xref_definition_line(How, _)
 	;   xref_defined(TB, Head, imported(_From))
-	->  true
 	;   get(prolog_predicate(Head), source, _)
-	).
-
+	), !.
 
 %	->edit
 %	
@@ -1395,6 +1502,9 @@ identify(F) :->
 %	
 %	Generate an identifying description for the predicate.
 
+identify_pred(Term, _, Summary) :-
+	identify(Term, String),
+	new(Summary, string(String)).
 identify_pred(built_in, F, Summary) :-	% SWI-Prolog documented built-in
 	get(F, predicate, Pred),
 	get(Pred, summary, Summary0), !,
@@ -1420,12 +1530,14 @@ identify_pred(autoload, F, Summary) :-	% Autoloaded predicates
 	new(Summary, string('%N: autoload from %s', F, Source)).
 identify_pred(local(Line), F, Summary) :-	% Local predicates
 	new(Summary, string('%N: locally defined at line %d', F, Line)).
+identify_pred(foreign(Line), F, Summary) :-	% Foreign predicates
+	new(Summary, string('%N: foreign (C/C++) loaded at line %d', F, Line)).
 identify_pred(constraint(Line), F, Summary) :-	% Local constraint
 	new(Summary, string('%N: constraint defined at line %d', F, Line)).
 identify_pred(imported(From), F, Summary) :-
 	new(Summary, string('%N: imported from %s', F, From)).
 identify_pred(recursion, _, 'Recursive reference') :- !.
-identify_pred(dynamic, F, Summary) :-
+identify_pred(dynamic(_Line), F, Summary) :-
 	get(F, loaded_specifier, Spec),
 	(   predicate_property(Spec, number_of_clauses(N))
 	->  new(Summary, string('%N: dynamic predicate with %d clauses',
@@ -1433,6 +1545,7 @@ identify_pred(dynamic, F, Summary) :-
 				N))
 	;   new(Summary, string('%N: dynamic predicate', F))
 	).
+identify_pred(global, _, 'Predicate defined in global module user') :- !.
 identify_pred(Class, _, ClassName) :-
 	term_to_atom(Class, ClassName).
 
@@ -1649,9 +1762,13 @@ identify(F) :->
 	identify_fragment(Term, F, Summary), !,
 	send(F, report, status, Summary).
 
+identify_fragment(Term,  _, Summary) :-
+	identify(Term, Summary), !.
 identify_fragment(var,  _, 'Variable').
 identify_fragment(file(Path), _, Summary) :-
 	new(Summary, string('File %s', Path)).
+identify_fragment(directory(Path), _, Summary) :-
+	new(Summary, string('Directory %s', Path)).
 identify_fragment(type_error(Type), _, Summary) :-
 	new(Summary, string('Type error: argument must be a %s', Type)).
 identify_fragment(module(Module), _, Summary) :-
