@@ -1,4 +1,4 @@
-/*  $Id: prolog_mode.pl,v 1.70 2004/09/21 14:22:25 jan Exp $
+/*  $Id: prolog_mode.pl,v 1.92 2007/03/25 15:29:34 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -31,6 +31,7 @@
 
 :- module(emacs_prolog_mode, []).
 :- use_module(library(pce)).
+:- use_module(library(operators)).
 :- use_module(library(emacs_extend)).
 :- use_module(library(prolog_predicate)).
 :- use_module(library(pce_prolog_xref)).
@@ -42,6 +43,7 @@
 	   , default/3
 	   , forall/2
 	   , ignore/1
+	   , last/2
 	   , list_to_set/2
 	   , member/2
 	   , memberchk/2
@@ -58,6 +60,10 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 		    "Mode for editing XPCE/Prolog sources",
 					% BINDINGS
 	[ insert_if_then_else	       = key('(') + key(';') + key('>'),
+	  insert_quote		       = key('"'),
+	  insert_percent	       = key('%'),
+
+	  newline_and_indent	       = key('RET'),
 
 					% delete some things
 	  manual_entry		       = -button(help),
@@ -70,7 +76,11 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 	  trace			       = button(prolog),
 	  break_at		       = key('\\C-cb') + button(prolog),
 	  delete_breakpoint	       = button(prolog),
+	  -			       = button(prolog),
 	  edit_breakpoints	       = button(prolog),
+	  edit_exceptions	       = button(prolog),
+	  view_threads		       = button(prolog),
+	  view_debug_messages	       = button(prolog),
 	  -			       = button(prolog),
 	  check_clause		       = key('\\C-c\\C-s') + button(prolog),
 	  insert_full_stop	       = key(.),
@@ -110,12 +120,12 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 	  '/'  + comment_start('*'),
 	  '*'  + comment_end('/'),
 
-	  paragraph_end([ '\\s *$',		% empty line
-			  '^/\\*',		% comment start
-			  '.*\\*/\\s *$', 	% comment end
-			  '^%',			% line comment
-			  '^	',		% indented line
-			  '.*:<?->?\\s *$'	% clause head
+	  paragraph_end([ '\\s*$',		% empty line
+			  '/\\*',		% comment start
+			  '[^\n]*\\*/\\s*$', 	% comment end
+			  '%',			% line comment
+			  '\t',			% indented line
+			  '[^\n]*:<?->?\\s*$'	% clause head
 			])
 	]).
 		 
@@ -123,8 +133,11 @@ class_variable(varmark_style, style*,
 	       style(background := honeydew,
 		     underline := @on)).
 
-variable(varmark_style, style*,       get, "How to mark variables").
-variable(has_var_marks, bool := @off, get, "Optimise a bit").
+variable(varmark_style,    style*,       get, "How to mark variables").
+variable(has_var_marks,    bool := @off, get, "Optimise a bit").
+variable(var_marked_caret, int*,	 get, "Last caret at ->mark_variable").
+variable(var_marked_gen,   int*,	 get, "Last generation").
+
 
 icon(_, I:image) :<-
 	"Return icon for mode"::
@@ -147,7 +160,7 @@ setup_mode(M) :->
 
 
 :- send(@class, attribute, outline_regex_list,
-	chain(regex('\\(^\\w+.*:<?->?\\)\\([^.]+\\.\\(\\s *\n\\)*\\)\\s '))).
+	chain(regex('(^\\w+.*:<?->?)([^.]+\\.(\\s*\n)*)\\s'))).
 
 source_file_chain(Ch) :-
 	new(Ch, chain),
@@ -179,9 +192,9 @@ expand_path(Term, D) :-
 :- pce_group(indent).
 
 :- pce_global(@prolog_neck_regex,
-	      new(regex(':-\\|:->\\|:<-\\|-->'))).
+	      new(regex(':-|:->|:<-|-->'))).
 :- pce_global(@prolog_full_stop,
-	      new(regex('[^-#$&*+./:<=>?@\\^`~]\\.\\($\\|\\s \\)'))).
+	      new(regex('[^-#$&*+./:<=>?@\\\\^`~]\\.($|\\s)'))).
 
 indent_line(E) :->
 	"Indent current line (Prolog indentation)"::
@@ -212,11 +225,12 @@ beginning_of_clause(E, Start:int, BOP:int) :<-
 		),
 		get(TB, skip_comment, BOP0, BOP)
 	    ;   get(@prolog_full_stop, register_start, SReg),
-		get(@prolog_full_stop, register_end, P0),
+	        get(@prolog_full_stop, register_end, EReg),
 		send(Here, value, SReg),
-		get(TB, skip_comment, P0, BOP),
+	        get(TB, scan_syntax, 0, SReg, tuple(code,_)),
+		get(TB, skip_comment, EReg, BOP),
 		BOP =< Start,
-		get(TB, scan_syntax, 0, BOP, tuple(code,_))
+		get(TB, scan_syntax, SReg, BOP, tuple(code,_))
 	    ).
 beginning_of_clause(E) :->
 	"Goto start of clause"::
@@ -244,9 +258,9 @@ indent_if_then_else(E) :->
 	get(E, caret, Caret),
 	get(E, text_buffer, TB),
 	get(E, skip_comment, Caret-1, OpenPos, EndOfPreviousTerm),
-	(   send(regex('\\,'), match, TB, EndOfPreviousTerm)
+	(   send(regex(','), match, TB, EndOfPreviousTerm)
 	->  get(TB, scan, Caret, line, -1, start, StartOfPrevLine),
-	    get(regex('\\s *\\(->\\|;\\)\\s *'), match, TB, StartOfPrevLine, L),
+	    get(regex('\\s*(->|;)\\s*'), match, TB, StartOfPrevLine, L),
 	    get(E, column, L+StartOfPrevLine, PrevExprCol),
 	    send(E, align_line, PrevExprCol)
 	;   get(E, column, OpenPos, OpenCol),
@@ -259,9 +273,9 @@ indent_clause_line(E) :->
 	get(E, caret, Caret),
 	get(E, text_buffer, TB),
 	get(E, skip_comment, Caret-1, 0, Glue),
-	(   send(regex('\\.'), match, TB, Glue)		% new clause
+	(   send(regex(\.), match, TB, Glue)		% new clause
 	->  send(E, align_line, 0)
-	;   send(regex('\\,'), match, TB, Glue)	  	% Next subclause
+	;   send(regex(','), match, TB, Glue)	  	% Next subclause
 	->  get(E, alignment_of_previous_line, N),
 	    (	N == 0					% head :- !,
 	    ->	send(E, align_line, 8)
@@ -279,19 +293,11 @@ insert_if_then_else(E, Times:[int], Char:char) :->
 	get(E, caret, Caret),
 	get(E, text_buffer, TB),
 	get(TB, scan, Caret, line, 0, start, SOL),
-	(   get(regex('\\s *\\((\\|->\\|;\\)'), match, TB, SOL, L),
+	(   get(regex('\\s*(\\(|->|;)'), match, TB, SOL, L),
 	    Caret =:= SOL + L,
 	    get(E, beginning_of_if_then_else, OpenPos)
-	->  get(E, text_buffer, TB),
-	    get(TB, scan, Caret, line, 0, start, SOL),
-	    (   (   send(regex('\\s *\\((\\|->\\|;\\)$'), match,
-			 TB, SOL, Caret)
-		;   Caret =:= 1 + OpenPos
-		)
-	    ->  get(E, column, OpenPos, Col),
-		send(E, align, Col+4)
-	    ;   true
-	    )
+	->  get(E, column, OpenPos, Col),
+	    send(E, align, Col+4)
 	;   true
 	).
 
@@ -304,12 +310,15 @@ indent_clause(E) :->
 	between(0, 1000, _),		% avoid loops on errors
 	    send(E, indent_line),
 	    get(E, caret, Caret),
-	    (	get(regex('.*\\S.\\.'), match, TB, Caret, Size),
+	    (	get(regex('[^\n]*[^[:punct:]]\\.[[:space:]]'), match,
+		    TB, Caret, Size0),
+		Size is Size0 - 1,	% subtract matched space
 		End is Caret + Size,
 		get(TB, scan_syntax, Start, End, tuple(code,_))
 	    ->	!
 	    ;	get(TB, size, Caret)
-	    ->	!
+	    ->	!,
+		Size = 0
 	    ;   send(E, next_line),
 		fail
 	    ),
@@ -328,6 +337,39 @@ fill_paragraph(M, Justify:[int]) :->
 	).
 
 
+insert_quote(E, Times:[int], Char:char) :->
+	"Complete quote"::
+	send(E, insert_self, Times, Char),
+	get(E, caret, Here),
+	(   send(E, looking_at, ':->\n\\s*"', Here, 0)
+	->  send(E, insert, '"::'),
+	    send(E, caret, Here)
+	;   true
+	).
+
+
+insert_percent(E, Times:[int], Char:char) :->
+	"Deal with %% comments"::
+	send(E, insert_self, Times, Char),
+	get(E, caret, Here),
+	(   send(E, looking_at, '\n%%', Here, 0),
+	    send(E, looking_at, '\\s*$', Here)
+	->  send(E, insert, '\t')
+	;   true
+	).
+
+
+indent_comment_line(E) :->
+	"Deal with %% comments"::
+	send_super(E, indent_comment_line),
+	get(E, caret, Here),
+	(   send(E, looking_at, '\n%%[^\n]*\n%\\s*\n%', Here, 0)
+	->  send(E, insert, '\t')
+	;   true
+	).
+
+
+
 		 /*******************************
 		 *          COMPILATION		*
 		 *******************************/
@@ -373,6 +415,7 @@ find_definition(M, For:prolog_predicate, NewWindow:[bool]) :->
 	get(For, head, @off, Head),
 	(   (   xref_defined(TB, Head, local(Line))		% local
 	    ;	xref_defined(TB, Head, constraint(Line))
+	    ;   xref_defined(TB, Head, foreign(Line))
 	    )
 	->  (   NewWindow == @on
 	    ->	get(M, text_buffer, TB),
@@ -414,11 +457,11 @@ find_local_definition(M, For:prolog_predicate) :->
 		 *	    PCE CLASSES		*
 		 *******************************/
 
-class_regex(':-\\s *pce_begin_class(\\(\\w+\\)',
-	    ':-\\s *pce_end_class\\s *.',
+class_regex(':-\\s*pce_begin_class\\((\\w+)',
+	    ':-\\s*pce_end_class\\s*.',
 	    A-[A]).
-class_regex(':-\\s *emacs_begin_mode(\\(\\w+\\)',
-	    ':-\\s *emacs_end_mode\\s *.',
+class_regex(':-\\s*emacs_begin_mode\\((\\w+)',
+	    ':-\\s*emacs_end_mode\\s*.',
 	    A-[emacs_, A, '_mode']).
 
 what_class(E, ClassName:name) :<-
@@ -508,6 +551,21 @@ edit_breakpoints(_M) :->
 	prolog_ide(open_debug_status).
 
 
+edit_exceptions(_M) :->
+	"Open Prolog Exception editor"::
+	prolog_ide(open_exceptions(@on)).
+
+
+view_threads(_M) :->
+	"View running threads"::
+	prolog_ide(thread_monitor).
+
+
+view_debug_messages(_M) :->
+	"View debug/3 messages"::
+	prolog_ide(debug_monitor).
+
+
 		 /*******************************
 		 *	   COMPILATION		*
 		 *******************************/
@@ -588,7 +646,7 @@ pce_check_require(M, File:file) :->
 	    (   send(Message, sub, 'up-to-date')
 	    ->  true
 	    ;   new(B, emacs_buffer(File)),
-		(   get(regex('^:-\\s *require('), search, B, Index)
+		(   get(regex('^:-\\s*require\\('), search, B, Index)
 		->  true
 		;   Index = 0
 		),
@@ -756,7 +814,7 @@ check_clause(M, From:from=[int], Repair:repair=[bool], End:int) :<-
 	),
 	get(M, text_buffer, TB),
 	pce_open(TB, read, Fd),
-	read_term_from_stream(Fd, Start, T, Error, S, P),
+	read_term_from_stream(TB, Fd, Start, T, Error, S, P),
 	close(Fd),
 	(   Error == none
 	->  (	send(M, has_send_method, colourise_term)
@@ -776,7 +834,9 @@ check_clause(M, From:from=[int], Repair:repair=[bool], End:int) :<-
 	    ),
 	    arg(2, P, E0),
 	    get(TB, find, E0, '.', 1, end, End)
-	;   Error = EPos:Msg,
+	;   get(M, forward_clause, Start, End),
+	    send(M, remove_syntax_fragments, Start, End),
+	    Error = EPos:Msg,
 	    (	Repair \== @off
 	    ->  send(M, caret, EPos)
 	    ;	true
@@ -785,26 +845,30 @@ check_clause(M, From:from=[int], Repair:repair=[bool], End:int) :<-
 	    fail
 	).
 
-read_term_from_stream(Fd, Start, T, Error, S, P) :-
+read_term_from_stream(TB, Fd, Start, T, Error, S, P) :-
 	retractall(syntax_error(_)),
 	alternate_syntax(_Name, Setup, Restore),
+	findall(Op, xref_op(TB, Op), Ops),
+	push_operators(Ops),
 	Setup,
 	seek(Fd, Start, bof, _),
 	read_with_errors(Fd, Start, T, Error, S, P),
 	Restore,
+	pop_operators,
 	(   Error == none
 	->  true
 	;   assert(syntax_error(Error)),
 	    fail
 	), !.
-read_term_from_stream(_, _, _, Error, _, _) :-
+read_term_from_stream(_, _, _, _, Error, _, _) :-
 	setof(E, retract(syntax_error(E)), Es),
 	last(Es, Error).
 
 pce_ifhostproperty(prolog(swi),
 (read_with_errors(Fd, _Start, T, Error, Singletons, TermPos) :-
 	catch(read_term(Fd, T, [ singletons(Singletons),
-				 subterm_positions(TermPos)
+				 subterm_positions(TermPos),
+				 module(emacs_prolog_mode)
 			       ]),
 	      Error0,
 	      true),
@@ -860,9 +924,9 @@ pce_ifhostproperty(prolog(quintus),
 	EP is EP0 + Start)).
 
 
-check_clause(M, From:[int]) :->
+check_clause(M, From:from=[int], Repair:repair=[bool]) :->
 	"Check syntax of clause"::
-	get(M, check_clause, From, _).
+	get(M, check_clause, From, Repair, _).
 
 
 unmark_singletons(M, P) :-
@@ -990,7 +1054,7 @@ prolog_term(M, From:[int], Silent:[bool], TermPos:[prolog], Clause:prolog) :<-
 	  ),
 	  get(M, text_buffer, TB),
 	  pce_open(TB, read, Fd),
-	  read_term_from_stream(Fd, Start, Clause, Error, _S, P),
+	  read_term_from_stream(TB, Fd, Start, Clause, Error, _S, P),
 	  close(Fd),
 	  ignore(P = TermPos),
 	  (   Error == none
@@ -1012,10 +1076,9 @@ typed(M, Id:'event|event_id', Editor:editor) :->
 	"Extend variable marks"::
 	send_super(M, typed, Id, Editor),
 	(   object(M)			% Control-x k destroys the mode
-	->  get(M, caret, Caret),
-	    get(M, text_buffer, TB),
-	    (   send(regex('[_A-Z][a-zA-Z_0-9]*.?'), match, TB, Caret, 0)
-	    ->  send(M, mark_variable)
+	->  (   get(M, varmark_style, Style),
+		Style \== @nil
+	    ->  send(M, mark_variable, @on)
 	    ;   true
 	    )
 	;   true
@@ -1025,8 +1088,9 @@ new_caret_position(M, NewCaret:int) :->
 	"Mark variables around caret"::
 	send_super(M, new_caret_position, NewCaret),
 	(   get(M, varmark_style, Style),
-	    Style \== @nil
-	->  send(M, mark_variable)
+	    Style \== @nil,
+	    send(M, mark_variable, @on),
+	    true
 	;   true
 	).
 
@@ -1035,21 +1099,53 @@ varmark_style(M, Style:style*) :->
 	send(M, slot, varmark_style, Style),
 	send(M, style, varmark, Style).
 
-mark_variable(M) :->
+%	(*)  Creating  fragments  finally  calls  ChangedRegionEditor(),
+%	which clears the kill-location. We  preserve   it  over the mark
+%	variables as this only concerns  colouring.   We  should fix the
+%	editor/fragment/textbuffer  interfaces  to  have  two  types  of
+%	changes  and  skip  resetting  the   kill-location  on  fragment
+%	changes. This however involves a lot of  changes while it is not
+%	likely to be a frequent  problem.   Therefore  this hack for the
+%	moment.
+
+mark_variable(M, Check:[bool]) :->
 	"Mark variable around caret"::
-	send(M, unmark_variables),
 	get(M, caret, Caret),
-	(   get(M, prolog_term, @default, @on, Pos, Clause),
-	    find_variable(Pos, Clause, Caret, Var)
-	->  get(M, text_buffer, TB),
-	    send(M, slot, has_var_marks, @on),
-	    (   subterm_position(Var, Clause, Pos, F-T),
-		Len is T - F,
-		new(_, emacs_colour_fragment(TB, F, Len, varmark)),
-		fail
-	    ;   true
-	    )
-	;   true
+	get(M, text_buffer, TB),
+	get(M, generation, Gen),
+	(   get(M, var_marked_caret, Caret),
+	    get(M, var_marked_gen, Gen)
+	->  true
+	;   send(M, slot, var_marked_caret, Caret),
+	    send(M, slot, var_marked_gen, Gen),
+
+	    get(M, editor, E),
+	    get(E, slot, kill_location, KillLocation), 		%  (*)
+	    send(M, unmark_variables),
+	    get(M, beginning_of_clause, Caret, Start),
+	    (   get(M, prolog_term, Start, @on, Pos, Clause)
+	    ->  (   Check == @on
+		->  send(M, check_clause, Start, @off)
+		;   true
+		),
+		(   find_variable(Pos, Clause, Caret, Var)
+		->  get(M, text_buffer, TB),
+		    send(M, slot, has_var_marks, @on),
+		    (   subterm_position(Var, Clause, Pos, F-T),
+			Len is T - F,
+			new(_, emacs_colour_fragment(TB, F, Len, varmark)),
+			fail
+		    ;   true
+		    )
+		;   true
+		)
+	    ;   (   get(M, forward_clause, Start, End)
+		->  true
+		;   End = Caret		% or end of buffer?
+		),
+		send(M, remove_syntax_fragments, Start, End)
+	    ),
+	    send(E, slot, kill_location, KillLocation)
 	).
 
 unmark_variables(M) :->
@@ -1075,6 +1171,8 @@ find_variable(term_position(_,_,_,_,ArgPos), Compound, Caret, Var) :-
 find_variable(list_position(_,_,EP,TP), List, Caret, Var) :-
 	list_pos(EP, TP, List, P, E),
 	find_variable(P, E, Caret, Var).
+find_variable(brace_term_position(_,_,TP), {Term}, Caret, Var) :-
+	find_variable(TP, Term, Caret, Var).
 
 list_pos([], P, T, P, T) :-
 	P \== none, !.
@@ -1095,6 +1193,8 @@ subterm_position(Search, Term, term_position(_,_,_,_,ArgPos), Pos) :- !,
 subterm_position(Search, List, list_position(_,_,EP,TP), Pos) :-
 	list_pos(EP, TP, List, P, E),
 	subterm_position(Search, E, P, Pos).
+subterm_position(Search, {Term}, brace_term_position(_,_,TP), Pos) :-
+	subterm_position(Search, Term, TP, Pos).
 
 
 		 /*******************************
@@ -1165,16 +1265,17 @@ at_start_of_clause(M, Pos:[int]) :->
 	;   get(M, text_buffer, TB),
 	    get(TB, scan, C, term, 1, end, TE),
 	    get(TB, skip_comment, TE, Neck),
-	    send(M, looking_at, ':-\\|-->\\|:<-\\|\\.', Neck)
+	    send(M, looking_at, ':-|-->|:<-|\\.', Neck)
 	).
 
 backward_clause(M, Start:int, BOC:int) :<-
 	"Find start of clause or previous clause"::
-	(   send(M, at_start_of_clause, Start)
-	->  From is Start - 1
-	;   From is Start
-	),
-	get(M, beginning_of_clause, From, BOC).
+	get(M, beginning_of_clause, Start, BOC0),
+	(   BOC0 == Start
+	->  Start2 is max(0, BOC0-1),
+	    get(M, beginning_of_clause, Start2, BOC)
+	;   BOC = BOC0
+	).
 
 forward_clause(M, Arg:[int]) :->
 	"Go forwards by <arg> clauses"::
