@@ -1,4 +1,4 @@
-/*  $Id: gui.pl,v 1.40 2006/12/13 13:01:27 jan Exp $
+/*  $Id: gui.pl,v 1.42 2007/06/22 14:29:06 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -414,6 +414,7 @@ fill_menu_bar(F) :->
 	get(MBD, member, menu_bar, MB),
 	send(MB, append, new(Tool, popup(tool))),
 	send(MB, append, new(Edit, popup(edit))),
+	send(MB, append, new(View, popup(view))),
 	send(MB, append, new(Comp, popup(compile))),
 	send(MB, append, new(Help, popup(help)), right),
 	send_list(Tool, append,
@@ -428,11 +429,17 @@ fill_menu_bar(F) :->
 		  ]),
 	send_list(Edit, append,
 		  [ menu_item(breakpoints,
-			      message(F, breakpoints),
+			      message(F, breakpoints)),
+		    menu_item(exceptions,
+			      message(F, exceptions),
 			      end_group := @on),
 		    menu_item(toggle_edit_mode,
 			      message(F, edit))
 		  ]),
+	send_list(View, append,
+		  [ menu_item(threads,
+			      message(F, show_threads))
+		  ]),
 	send_list(Comp, append,
 		  [ menu_item(make,
 			      message(F, make),
@@ -477,6 +484,10 @@ show_stack(GUI, CallFrames:prolog, ChoiceFrames:prolog) :->
 	send(StackWindow, clear),
 	display_stack(StackWindow, CallFrames, ChoiceFrames).
 
+show_threads(_GUI) :->
+	"Open Thread monitor"::
+	prolog_ide(thread_monitor).
+
 
 		 /*******************************
 		 *	      EVENT		*
@@ -594,6 +605,10 @@ breakpoints(_F) :->
 	"Edit spy/break/trace-points"::
 	prolog_ide(open_debug_status).
 
+exceptions(_F) :->
+	"Edit exceptions"::
+	prolog_ide(open_exceptions).
+
 make(_) :->
 	"Run Prolog make"::
 	(   object(@emacs)
