@@ -1,4 +1,4 @@
-/*  $Id: postscript.c,v 1.39 2007/01/14 20:36:28 jan Exp $
+/*  $Id: postscript.c,v 1.41 2007/06/25 08:20:30 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -506,8 +506,8 @@ header(Any gr, Area area, Bool ls)
   { scale = 1.0;
   } else
   { scale = min( (float)w / (float)wgr, (float)h/(float)hgr );
-    paperW = (int)((float)paperW * scale);
-    paperH = (int)((float)paperH * scale);
+    paperW = (int)((float)paperW * scale + 0.999);
+    paperH = (int)((float)paperH * scale + 0.999);
   }
 
   if ( ls == ON )
@@ -1645,6 +1645,9 @@ drawPostScriptImage(Image image, Name hb)
 }
 
 
+#define Wrapped(t)      ((t)->wrap == NAME_wrap || \
+			 (t)->wrap == NAME_wrapFixedWidth)
+
 status
 drawPostScriptText(TextObj t, Name hb)
 { String s = &t->string->data;
@@ -1701,7 +1704,7 @@ drawPostScriptText(TextObj t, Name hb)
       if ( t->underline == ON )
 	flags |= TXT_UNDERLINED;
 
-      if ( t->wrap == NAME_wrap )
+      if ( Wrapped(t) )
       { LocalString(buf, s->iswide, s->size + MAX_WRAP_LINES);
       
 	str_format(buf, s, valInt(t->margin), t->font);
