@@ -1,11 +1,11 @@
-/*  $Id: kernel.h,v 1.66 2004/09/21 14:22:25 jan Exp $
+/*  $Id: kernel.h,v 1.92 2007/01/23 19:22:17 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -36,6 +36,12 @@
 #include <time.h>			/* do that before we do it here */
 #endif
 #include <h/stream.h>			/* IOSTREAM interface */
+#include <limits.h>
+
+#ifndef INT_MAX
+#define INT_MAX	    ((int)(((unsigned int)1<<(sizeof(int)*8-1))-1))
+#define INT_MIN     (-(INT_MIN)-1)
+#endif
 
 #ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>			/* Use www.dmalloc.com debugger */
@@ -57,7 +63,7 @@
 #define PUBLIC_GLOBAL extern		/* exported global variables */
 #endif
 
-#if defined(WIN32) || defined(__CYGWIN32__)
+#if defined(WIN32) || defined(__CYGWIN__)
 #define __pce_export __declspec(dllexport)
 #else
 #define __pce_export extern
@@ -93,10 +99,6 @@ generate warnings on accidental use of them.
 
 #ifdef SYSLIB_H
 #include SYSLIB_H
-#endif
-
-#ifndef HAVE_GETDTABLESIZE		/* avoid clash with SWI-Prolog */
-#define getdtablesize PCE_getdtablesize
 #endif
 
 		/********************************
@@ -107,7 +109,7 @@ generate warnings on accidental use of them.
 #define METHOD_MAX_ARGS		16	/* maximum # args for C-method */
 #define FWD_PCE_MAX_ARGS	10	/* @arg1 ... @arg10 */
 #define SCAN_MAX_ARGS 		32	/* scanstr maximum arguments */
-#define PCE_MAX_INT		((long)((1L<<(sizeof(Any)*8 - TAG_BITS-1))-1))
+#define PCE_MAX_INT		((intptr_t)(((intptr_t)1<<(sizeof(Any)*8 - TAG_BITS-1))-1))
 #define PCE_MIN_INT		(-(PCE_MAX_INT-1))
 #ifndef INT_MAX
 #define INT_MAX			((int)(((unsigned int)1<<(sizeof(int)*8-1))-1))
@@ -154,7 +156,7 @@ different types when passed to a vararg function.
 		********************************/
 
 #define SAVEMAGIC   		"PCE version 4"
-#define SAVEVERSION 		17
+#define SAVEVERSION 		18	/* last increment for 5.6.14 */
 
 		/********************************
 		*             ASSERTS		*
@@ -173,7 +175,7 @@ different types when passed to a vararg function.
 		********************************/
 
 #if __STRICT_ANSI__
-#undef TAGGED_LVALUE 1
+#undef TAGGED_LVALUE
 #endif
 
 #ifdef __GNUC__
@@ -188,7 +190,7 @@ different types when passed to a vararg function.
 # define Promote(type) type
 #endif
 
-#ifdef HAVE_HIDDEN_ATTRIBUTE
+#ifdef HAVE_VISIBILITY_ATTRIBUTE
 #define SO_LOCAL __attribute__((visibility("hidden")))
 #else
 #define SO_LOCAL
@@ -198,10 +200,10 @@ different types when passed to a vararg function.
 #define inline
 #endif
 
-#if O_CONST_FUNCTION
-#define constf const
+#ifdef __GNUC__
+#define PURE_FUNCTION __attribute__((pure))
 #else
-#define constf
+#define PURE_FUNCTION
 #endif
 
 #define forwards	static		/* Local functions */
@@ -274,6 +276,7 @@ cpdata(to, from, type, n)
 #define	CtoInt(i)	toInt(i)		/* int --> Int */
 #define CtoName(s)	(Name)cToPceName((s))	/* const char * --> Name */
 #define CtoType(s)	nameToType(CtoName(s))  /* char * --> type object */
+#define WCtoType(s)	nameToType(WCToName(s, -1)) /* wchar_t * --> type object */
 #define	pp(x)		pcePP((Any)(x))		/* interface name */
 #define get             getPCE          	/* avoid common name-conflict */
 #define send            sendPCE         	/* same */
@@ -309,7 +312,7 @@ typedef struct dCell 	      **DelegateList;   /* See msg-passing.c */
 
 #ifdef VARIABLE_POINTER_OFFSET
 #undef POINTER_OFFSET
-GLOBAL unsigned long pce_data_pointer_offset;
+GLOBAL uintptr_t pce_data_pointer_offset;
 #define POINTER_OFFSET pce_data_pointer_offset
 #else
 #ifndef POINTER_OFFSET
@@ -317,7 +320,7 @@ GLOBAL unsigned long pce_data_pointer_offset;
 #endif
 #endif
 
-#define PointerToCInt(p) (((unsigned long)(p) - POINTER_OFFSET)/sizeof(int))
+#define PointerToCInt(p) (((uintptr_t)(p) - POINTER_OFFSET)/sizeof(int))
 #define PointerToInt(p)	 toInt(PointerToCInt(p))
 #define longToPointer(i) ((Any) (i * sizeof(int) + POINTER_OFFSET))
 #define IntToPointer(i)  longToPointer(valInt(i))
@@ -331,8 +334,8 @@ GLOBAL unsigned long pce_data_pointer_offset;
 #define MASK_MASK	0x00000001	/* 11 Mask Mask */
 #define TAG_BITS	1		/* number of mask bits for INT */
 
-#define MaskOf(obj)		((unsigned long)(obj) & MASK_MASK)
-#define UnMask(obj)		((unsigned long)(obj) & ~MASK_MASK)
+#define MaskOf(obj)		((uintptr_t)(obj) & MASK_MASK)
+#define UnMask(obj)		((uintptr_t)(obj) & ~MASK_MASK)
 
 
 		/********************************
@@ -369,14 +372,22 @@ integer is declared as of type Int (for casting purposes). The following
 test, conversion and computation macro's are provided.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#ifndef INTPTR_FORMAT
+#if SIZEOF_INT == SIZEOF_VOIDP
+#define INTPTR_FORMAT "%d"		/* printf format for intptr_t */
+#else
+#define INTPTR_FORMAT "%ld"		/* printf format for intptr_t */
+#endif
+#endif
+
 #undef max
 #undef min
 #define max(a, b)	((a) > (b) ? (a) : (b))
 #define min(a, b)	((a) < (b) ? (a) : (b))
 
-#define isInteger(i)	((unsigned long)(i) & INT_MASK)
-#define toInt(i)	((Int)(((long)(i)<<TAG_BITS)|INT_MASK))
-#define valInt(i)	(((long)(i))>>TAG_BITS)
+#define isInteger(i)	((uintptr_t)(i) & INT_MASK)
+#define toInt(i)	((Int)(((intptr_t)(i)<<TAG_BITS)|INT_MASK))
+#define valInt(i)	(((intptr_t)(i))>>TAG_BITS)
 #define incrInt(i)	((i) = toInt(valInt(i)+1))
 #define decrInt(i)	((i) = toInt(valInt(i)-1))
 #define addInt(i, j)	((i) = toInt(valInt(i) + valInt(j)))
@@ -410,8 +421,6 @@ test, conversion and computation macro's are provided.
 #define makeDFlag(n)		(1L << ((n) - 1 + TAG_BITS))
 #define DFlags(obj)		(((ProgramObject)(obj))->dflags)
 #ifndef TAGGED_LVALUE
-void	setDFlagProgramObject(Any, unsigned long);
-void	clearDFlagProgramObject(Any, unsigned long);
 #define setDFlag(obj, mask)     setDFlagProgramObject((obj), (mask))
 #define clearDFlag(obj, mask)	clearDFlagProgramObject((obj), (mask))
 #else
@@ -466,30 +475,17 @@ void	clearDFlagProgramObject(Any, unsigned long);
 
 
 		/********************************
-		*    CHAR_ARRAY, STRING, NAME	*
+		*       CHAR_ARRAY, STRING	*
 		********************************/
 
 #include "str.h"			/* string type and friends */
 #include "../txt/proto.h"		/* prototypes */
 
-#define LocalString(name, proto, size) \
+#define LocalString(name, iswide, size) \
   string _s_ ## name ## _hdr; \
-  void  *_s_ ## name ## _buf = (void *)alloca((proto)->b16 ? (size) * sizeof(char16) \
-						   : (size) * sizeof(char8)); \
-  String name = str_init(&_s_ ## name ## _hdr, (proto), _s_ ## name ## _buf)
-
-#ifndef NO_BUILT_IN_DECL
-extern struct name builtin_names[];	/* object-array of built-in's */
-#endif
-#include "names.ih"			/* #defines for code used names */
-
-#define isName(name)	(isObject(name) && onFlag((name), F_ISNAME))
-#define notName(name)	(!isName(name))
-#define equalName(a, b) ((a) == (b))
-#define strName(s)	((char *)((Name)(s))->data.s_text8)
-
-#define getAppendName(n, s) \
-	((Name) getAppendCharArray((CharArray)(n), (CharArray)(s)))
+  void  *_s_ ## name ## _buf = (void *)alloca(iswide ? (size) * sizeof(charW) \
+						   : (size) * sizeof(charA)); \
+  String name = fstr_inithdr(&_s_ ## name ## _hdr, iswide, _s_ ## name ## _buf, size)
 
 		/********************************
 		*         OBJECT HEADER		*
@@ -538,8 +534,8 @@ extern struct name builtin_names[];	/* object-array of built-in's */
 #define F_ISHOSTDATA		makeFlag(23) /* instanceOf(x, ClassHostData) */
 #define F_NOTANY		makeFlag(24) /* Not acceptable to any/object */
 
-#define OBJ_MAGIC		((unsigned long)(0x94L << 25))
-#define OBJ_MAGIC_MASK		((unsigned long)(0xfeL << 25))
+#define OBJ_MAGIC		((uintptr_t)(0x94L << 25))
+#define OBJ_MAGIC_MASK		((uintptr_t)(0xfeL << 25))
 
 #define hasObjectMagic(obj)	((((Instance)(obj))->flags&OBJ_MAGIC_MASK) == \
 					OBJ_MAGIC)
@@ -620,10 +616,10 @@ extern struct name builtin_names[];	/* object-array of built-in's */
 		*         CAREFUL CHECKERS	*
 		********************************/
 
-#define validAddress(a)	((unsigned long)(a) >= allocBase && \
-			 (unsigned long)(a) < allocTop)
+#define validAddress(a)	((uintptr_t)(a) >= allocBase && \
+			 (uintptr_t)(a) < allocTop)
 #define isAddress(a)	(validAddress(a) && \
-			 !((unsigned long)(a) & (sizeof(Any)-1)))
+			 !((uintptr_t)(a) & (sizeof(Any)-1)))
 #define validPceDatum(x) (isInteger(x) || isProperObject(x))
 
 
@@ -649,14 +645,14 @@ extern struct name builtin_names[];	/* object-array of built-in's */
 		********************************/
 
 #define OBJECT_HEADER \
-  unsigned long flags;			/* general flag field */ \
-  unsigned long references;		/* reference count */ \
+  uintptr_t     flags;			/* general flag field */ \
+  uintptr_t     references;		/* reference count */ \
   Class		class;			/* Associated class */
 
 #define ABSTRACT_OBJECT
 
 #define ABSTRACT_PROGRAM_OBJECT \
-  unsigned long dflags;			/* Debugging flags */
+  uintptr_t	dflags;			/* Debugging flags */
 
 #define ABSTRACT_RECOGNISER \
   Bool		active;			/* Does accept events? */
@@ -910,6 +906,7 @@ typedef struct _classdecl
 #define SMEntries(l)		(sizeof(l) / sizeof(senddecl))
 #define GMEntries(l)		(sizeof(l) / sizeof(getdecl))
 #define RCEntries(l)		(sizeof(l) / sizeof(classvardecl))
+#define TNEntries(l)		(sizeof(l) / sizeof(Name))
 
 #ifndef UXWIN
 #ifdef WIN32_GRAPHICS
@@ -1124,7 +1121,8 @@ NewClass(minus)
   ABSTRACT_BINARY_EXPRESSION
 End;
 
-#define ABSTRACT_SOURCE_SINK
+#define ABSTRACT_SOURCE_SINK \
+  Name		encoding;		/* used encoding */
 
 #define ABSTRACT_CHAR_ARRAY \
   string	data;			/* the represented data */
@@ -1146,7 +1144,7 @@ NewClass(source_sink)
 End;
 
 NewClass(number)
-  long		value;			/* value of the number */
+  intptr_t	value;			/* value of the number */
 End;
 
 NewClass(pce)
@@ -1198,8 +1196,13 @@ NewClass(host_data)
 End;
 
 NewClass(real)
-  unsigned long value1;			/* 1-st part of double */
-  unsigned long value2;			/* 2nd-part of double */
+#if SIZEOF_VOIDP == SIZEOF_DOUBLE
+#define REAL_IN_ONE 1
+  double	value;			/* can store in one slot */
+#else
+  uintptr_t value1;			/* 1-st part of double */
+  uintptr_t value2;			/* 2nd-part of double */
+#endif
 End;
 
 NewClass(recogniser)
@@ -1406,6 +1409,23 @@ struct class_definition
   char *	summary;		/* Summary description */
 };
 
+		 /*******************************
+		 *	       NAMES		*
+		 *******************************/
+
+#ifndef NO_BUILT_IN_DECL
+extern struct name builtin_names[];	/* object-array of built-in's */
+#endif
+#include "names.ih"			/* #defines for code used names */
+
+#define isName(name)	(isObject(name) && onFlag((name), F_ISNAME))
+#define notName(name)	(!isName(name))
+#define equalName(a, b) ((a) == (b))
+#define strName(s)	((char *)((Name)(s))->data.s_textA)
+
+#define getAppendName(n, s) \
+	((Name) getAppendCharArray((CharArray)(n), (CharArray)(s)))
+
 
 		/********************************
 		*         FORWARDING		*
@@ -1511,7 +1531,7 @@ struct to_cell
 GLOBAL ToCell	AnswerStack;		/* Stack of `answer objects' */
 GLOBAL int	deferredUnalloced;	/* # deferred unallocs in ->free */
 
-typedef long	AnswerMark;
+typedef intptr_t AnswerMark;
 
 #define markAnswerStack(mark)	{(mark) = AnswerStack->index;}
 #define rewindAnswerStack(mark, obj) \
@@ -1526,6 +1546,9 @@ typedef long	AnswerMark;
 #include "../ker/proto.h"
 #include "../msg/proto.h"
 #include "../adt/proto.h"
+#include "../rel/proto.h"
+
+#define getSubName(n, f, t) (Name)getSubCharArray((CharArray)(n), f, t)
 
 					/* Interface callback stubs */
 __pce_export void	Cprintf(const char *fmt, ...);
@@ -1534,34 +1557,40 @@ __pce_export int	Cputchar(int chr);
 __pce_export char *	Cgetline(char *line, int size);
 
 					/* interface prototypes */
-CPointer	CtoCPointer(void *);
 __pce_export Any	cToPceName(const char *text);
-status		makeClassC(Class class);
-status		makeClassRC(Class class);
-status		makeClassCPointer(Class class);
-status		initialiseHost(Host h, Name which);
-status		makeClassHost(Class class);
-status		makeClassHostData(Class class);
-status		makeClassSourceSink(Class class);
-Host		HostObject(void);
-int		hostGetc(void);
-void		pceWriteErrorGoal(void);
-
-status		initialiseSourceSink(SourceSink ss);
-status		checkErrorSourceSink(SourceSink ss, IOSTREAM *fd);
-IOSTREAM *	Sopen_object(Any obj, const char *mode);
-IOSTREAM *	Sopen_FILE(FILE *fd, int flags);
+COMMON(CPointer) CtoCPointer(void *);
+COMMON(status)	makeClassC(Class class);
+COMMON(status)	makeClassRC(Class class);
+COMMON(status)	makeClassCPointer(Class class);
+COMMON(status)	initialiseHost(Host h, Name which);
+COMMON(status)	makeClassHost(Class class);
+COMMON(status)	makeClassHostData(Class class);
+COMMON(status)	makeClassSourceSink(Class class);
+COMMON(Host)	HostObject(void);
+COMMON(int)	hostGetc(void);
+COMMON(void)	pceWriteErrorGoal(void);
+COMMON(int)	initPublicInterface(void);
 
+COMMON(status)	initialiseSourceSink(SourceSink ss);
+COMMON(status)	checkErrorSourceSink(SourceSink ss, IOSTREAM *fd);
+COMMON(status)	initialiseSourceSink(SourceSink ss);
+COMMON(IOSTREAM *)	Sopen_object(Any obj, const char *mode);
+COMMON(status) 	setStreamEncodingSourceSink(SourceSink ss, IOSTREAM *fd);
+COMMON(Name)	encoding_to_name(IOENC encoding);
 
 #if O_CPLUSPLUS
-status 	callCPlusPlusProc(void *f, int ac, const Any av[]);
-Any	callCPlusPlusFunc(void *f, int ac, const Any av[]);
-status 	callCPlusPlusPceMethodProc(Any o, void *f, int ac, const Any av[]);
-Any 	callCPlusPlusPceMethodFunc(Any o, void *f, int ac, const Any av[]);
-status 	callCPlusPlusMethodProc(Any o, void *f, int ac, const Any av[]);
-Any 	callCPlusPlusMethodFunc(Any o, void *f, int ac, const Any av[]);
+COMMON(status )	callCPlusPlusProc(void *f, int ac, const Any av[]);
+COMMON(Any)	callCPlusPlusFunc(void *f, int ac, const Any av[]);
+COMMON(status )	callCPlusPlusPceMethodProc(Any o, void *f,
+					   int ac, const Any av[]);
+COMMON(Any )	callCPlusPlusPceMethodFunc(Any o, void *f,
+					   int ac, const Any av[]);
+COMMON(status )	callCPlusPlusMethodProc(Any o, void *f,
+					int ac, const Any av[]);
+COMMON(Any )	callCPlusPlusMethodFunc(Any o, void *f,
+					int ac, const Any av[]);
 #endif
-void	initCGlobals(void);
+COMMON(void)	initCGlobals(void);
 
 		/********************************
 		*       GLOBAL VARIABLES	*
@@ -1608,8 +1637,8 @@ GLOBAL int	restoreVersion;		/* Version of save file */
 GLOBAL SourceSink LoadFile;		/* Current file for <-object */
 GLOBAL char    *SaveMagic;		/* Magic string for saved objects */
 GLOBAL int	inBoot;			/* is the system in the boot cycle? */
-GLOBAL unsigned long allocBase;		/* lowest allocated memory */
-GLOBAL unsigned long allocTop;		/* highest allocated memory */
+GLOBAL uintptr_t allocBase;		/* lowest allocated memory */
+GLOBAL uintptr_t allocTop;		/* highest allocated memory */
 #ifndef O_RUNTIME
 GLOBAL int	PCEdebugging;		/* PCE->debugging == ON */
 GLOBAL int	PCEdebugBoot;		/* Debug booting phase? */
@@ -1783,6 +1812,7 @@ extern char *T_report[];		/* ->report: kind, format, args... */
 		********************************/
 
 #include <h/syntax.h>
+#include <h/utf8.h>
 
 		 /*******************************
 		 *	 SPEEDUP MACROS		*
@@ -1816,8 +1846,8 @@ GLOBAL int hash_shifts;			/* Shifts in append */
 #define COUNT(g)
 #endif
 
-#define unboundedKey(name) (isInteger(name) ? (unsigned long)(name)>>1 \
-					    : (unsigned long)(name)>>2)
+#define unboundedKey(name) (isInteger(name) ? (uintptr_t)(name)>>1 \
+					    : (uintptr_t)(name)>>2)
 
 #if USE_PRIMES
 #define hashKey(name, buckets) (unboundedKey(name) % (buckets))
