@@ -1,4 +1,4 @@
-/*  $Id: classvar.c,v 1.12 2006/08/16 20:09:15 jan Exp $
+/*  $Id: classvar.c,v 1.14 2007/01/18 12:13:22 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -455,7 +455,7 @@ getManIdClassVariable(ClassVariable cv)
 { wchar_t buf[LINESIZE];
   wchar_t *nm, *o;
   Name ctx = ((Class)cv->context)->name;
-  unsigned len;
+  size_t len;
   Name rc;
 
   len = 4 + ctx->data.size + cv->name->data.size;
@@ -520,7 +520,7 @@ getPrintNameClassVariable(ClassVariable cv)
 { wchar_t buf[LINESIZE];
   wchar_t *nm, *o;
   Name ctx = ((Class)cv->context)->name;
-  unsigned len;
+  size_t len;
   Name rc;
   
   len = 2 + ctx->data.size + cv->name->data.size;
@@ -791,7 +791,7 @@ loadDefaultClassVariables(SourceSink f)
 	  for(;;)
 	  { for(s++; isblank(*s); s++)
 	      ;
-	    l = strlen(s);
+	    l = (int)strlen(s);
 					/* delete [\r\n]*$ */
 	    while( l > 0 && (s[l-1] == '\n' || s[l-1] == '\r') )
 	      s[--l] = EOS;
