@@ -1,4 +1,4 @@
-/*  $Id: menu.c,v 1.40 2004/09/21 14:22:25 jan Exp $
+/*  $Id: menu.c,v 1.44 2005/08/23 18:13:03 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -1087,28 +1087,8 @@ openComboBoxMenu(Menu m)
 
 
 static status
-executeMenu(Menu m, EventObj ev)
-{ MenuItem mi;
-
-  if ( m->feedback == NAME_showSelectionOnly )
-  { Any img = getClassVariableValueObject(m, NAME_cycleIndicator); /* TBD */
-
-    if ( img == NAME_comboBox )
-      return openComboBoxMenu(m);
-    else
-    { nextMenu(m);
-      flushGraphical(m);
-      if ( !send(m->device, NAME_modifiedItem, m, ON, EAV) )
-	forwardMenu(m, m->message, ev);
-      succeed;
-    }
-  }
-
-  if ( isDefault(ev) )
-    ev = EVENT->value;			/* @event */
-  TRY((mi = getItemFromEventMenu(m, ev)) && mi->active == ON);
-    
-  if ( m->multiple_selection == ON )
+executeMenuItem(Menu m, MenuItem mi, EventObj ev)
+{ if ( m->multiple_selection == ON )
   { toggleMenu(m, mi);
     flushGraphical(m);
     send(m->device, NAME_modifiedItem, m, ON, EAV);
@@ -1140,6 +1120,32 @@ executeMenu(Menu m, EventObj ev)
 }
 
 
+static status
+executeMenu(Menu m, EventObj ev)
+{ MenuItem mi;
+
+  if ( m->feedback == NAME_showSelectionOnly )
+  { Any img = getClassVariableValueObject(m, NAME_cycleIndicator); /* TBD */
+
+    if ( img == NAME_comboBox )
+      return openComboBoxMenu(m);
+    else
+    { nextMenu(m);
+      flushGraphical(m);
+      if ( !send(m->device, NAME_modifiedItem, m, ON, EAV) )
+	forwardMenu(m, m->message, ev);
+      succeed;
+    }
+  }
+
+  if ( isDefault(ev) )
+    ev = EVENT->value;			/* @event */
+  TRY((mi = getItemFromEventMenu(m, ev)) && mi->active == ON);
+    
+  return executeMenuItem(m, mi, ev);
+}
+
+
 		/********************************
 		*           SELECTION		*
 		********************************/
@@ -1225,8 +1231,18 @@ index_item_menu(Menu m, Any spec)
     }
   } else
   { Cell cell;
-    int n = 1;
+    int n;
     
+    n = 1;				/* exact match */
+    for_cell(cell, m->members)
+    { MenuItem mi = cell->value;
+
+      if ( mi->value == spec )
+        return n;
+      n++;
+    }
+
+    n = 1;				/* equal string representation */
     for_cell(cell, m->members)
     { if ( hasValueMenuItem(cell->value, spec) )
         return n;
@@ -1305,12 +1321,28 @@ selectionMenu(Menu m, Any selection)
     if ( is_set[n] & NEEDS_SET )
     { if ( !(is_set[n] & IS_SET) )
       { assign(mi, selected, ON);
-	ChangedItemMenu(m, mi);
       }
     } else
     { if ( is_set[n] & IS_SET )
-      { assign(((MenuItem) cell->value), selected, OFF);
-	ChangedItemMenu(m, mi);
+      { assign(mi, selected, OFF);
+      }
+    }
+    n++;
+  }
+
+					/* 2nd pass to avoid a call-back */
+					/* through ensureSingleSelectionMenu */
+  n = 1;
+  for_cell(cell, m->members)
+  { MenuItem mi = cell->value;
+
+    if ( is_set[n] & NEEDS_SET )
+    { if ( !(is_set[n] & IS_SET) )
+      { ChangedItemMenu(m, mi);
+      }
+    } else
+    { if ( is_set[n] & IS_SET )
+      { ChangedItemMenu(m, mi);
       }
     }
     n++;
@@ -1609,8 +1641,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_box);
       assign(m, pen, ONE);
       assign(m, border, TWO);
-      assign(m, multiple_selection, kind == NAME_toggle ? ON : OFF);
-
+      multipleSelectionMenu(m, kind == NAME_toggle ? ON : OFF);
       assign(m, kind, kind);
       return requestComputeGraphical(m, DEFAULT);
     } else if ( kind == NAME_marked )
@@ -1619,7 +1650,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_image);
       assign(m, pen, ZERO);
       assign(m, border, TWO);
-      assign(m, multiple_selection, OFF);
+      multipleSelectionMenu(m, OFF);
       assign(m, kind, kind);
 
       return requestComputeGraphical(m, DEFAULT);
@@ -1633,7 +1664,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_image);
       assign(m, pen, ZERO);
       assign(m, border, toInt(3));
-      assign(m, multiple_selection, kind == NAME_toggle ? ON : OFF);
+      multipleSelectionMenu(m, kind == NAME_toggle ? ON : OFF);
 
       assign(m, kind, kind);
       return requestComputeGraphical(m, DEFAULT);
@@ -1643,7 +1674,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_box);
       assign(m, pen, ONE);
       assign(m, border, TWO);
-      assign(m, multiple_selection, OFF);
+      multipleSelectionMenu(m, OFF);
 
       assign(m, kind, kind);
       return requestComputeGraphical(m, DEFAULT);
@@ -1662,7 +1693,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_image);
       assign(m, pen, ZERO);
       assign(m, border, toInt(3));
-      assign(m, multiple_selection, kind == NAME_toggle ? ON : OFF);
+      multipleSelectionMenu(m, kind == NAME_toggle ? ON : OFF);
 
       assign(m, kind, kind);
       return requestComputeGraphical(m, DEFAULT);
@@ -1672,7 +1703,7 @@ kindMenu(Menu m, Name kind)
       assign(m, feedback, NAME_box);
       assign(m, pen, ONE);
       assign(m, border, TWO);
-      assign(m, multiple_selection, OFF);
+      multipleSelectionMenu(m, OFF);
 
       assign(m, kind, kind);
       return requestComputeGraphical(m, DEFAULT);
@@ -1764,7 +1795,9 @@ ensureSingleSelectionMenu(Menu m)
 
   if ( sel == 0 )
   { if ( notNil(first) )
+    { DEBUG(NAME_popup, Cprintf("%s: selecting first\n", pp(m)));
       return selectionMenu(m, first);
+    }
 
     fail;
   }
@@ -1775,7 +1808,11 @@ ensureSingleSelectionMenu(Menu m)
 
 static status
 multipleSelectionMenu(Menu m, Bool val)
-{ return assignGraphical(m, NAME_multipleSelection, val);
+{ assignGraphical(m, NAME_multipleSelection, val);
+  
+  get(m, NAME_selection, EAV);		/* update <-selection */
+
+  succeed;
 }
 
 
@@ -2011,7 +2048,23 @@ applyMenu(Menu m, Bool always)
 
 static status
 assignAcceletatorsMenu(Menu m)
-{ return assignAccelerators(m->members, NAME_, NAME_label);
+{ return assignAccelerators(m->members, CtoName("\\e"), NAME_label);
+}
+
+
+static status
+keyMenu(Menu m, Name key)
+{ Cell cell;
+
+  for_cell(cell, m->members)
+  { MenuItem mi = cell->value;
+
+    if ( mi->accelerator == key )
+    { return executeMenuItem(m, mi, EVENT->value);
+    }
+  }
+
+  fail;
 }
 
 
@@ -2104,6 +2157,8 @@ static senddecl send_menu[] =
      DEFAULT, "Unlink from menu-items"),
   SM(NAME_assignAccelerators, 0, NULL, assignAcceletatorsMenu,
      NAME_accelerator, "Assign accelerators for the items"),
+  SM(NAME_key, 1, "key=name", keyMenu,
+     NAME_accelerator, "Handle accelerator key `name'"),
   SM(NAME_activeAllItems, 1, "active=bool", activeAllItemsMenu,
      NAME_active, "(De)activate all items in the menu"),
   SM(NAME_activeItem, 2, T_activeItem, activeItemMenu,
