@@ -1,4 +1,4 @@
-/*  $Id: msdraw.c,v 1.55 2004/10/12 10:30:07 jan Exp $
+/*  $Id: msdraw.c,v 1.73 2007/01/14 16:36:15 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -88,6 +88,7 @@ typedef struct
   DisplayObj	display;		/* The XPCE display */
   int		depth;			/* # bits/pixel */
   int		transformed;		/* A scaling-mapping is active */
+  int		compatible;		/* HDC is compatible to screen */
 
   Elevation	elevation;		/* current elevation context */
   HPEN		relief_pen;		/* standing-edge pen */
@@ -101,11 +102,11 @@ typedef struct
 
 } wdraw_context, *WDrawContext;
 
-#ifdef __WIN32__
+#ifdef __WINDOWS__
 #define MoveTo(hdc, x, y) MoveToEx((hdc), (x), (y), NULL);
 #define SetWindowOrg(hdc, x, y) SetWindowOrgEx((hdc), (x), (y), NULL);
 #define SetViewportOrg(hdc, x, y) SetViewportOrgEx((hdc), (x), (y), NULL);
-#endif /*__WIN32__*/
+#endif /*__WINDOWS__*/
 
 static int		cache = 1;	/* Do or don't */
 static int		quick;		/* Prefer speed */
@@ -167,6 +168,7 @@ reset_context()
   context.r_offset_x	     = 0;
   context.r_offset_y	     = 0;
   context.fixed_colours	     = 0;
+  context.compatible	     = TRUE;
 }
 
 
@@ -551,7 +553,7 @@ d_image(Image i, int x, int y, int w, int h)
 
 
 void
-d_hdc(HDC hdc, Colour fg, Colour bg)
+d_hdc(HDC hdc, Colour fg, Colour bg, int compatible)
 { push_context();
 
   d_display(DEFAULT);
@@ -563,6 +565,7 @@ d_hdc(HDC hdc, Colour fg, Colour bg)
   context.open++;
   context.hdc = hdc;
   context.device = NIL;			/* anonymous device */
+  context.compatible = compatible;
 
   if ( notNil(bg) )
     r_default_background(bg);
@@ -575,16 +578,15 @@ void
 d_screen(DisplayObj d)
 { HDC hdc = GetDC(NULL);
 
-  d_hdc(hdc, DEFAULT, DEFAULT);
+  d_hdc(hdc, DEFAULT, DEFAULT, TRUE);
 }
 
 
 status
-d_winmf(const char *fn, int x, int y, int w, int h, const char *descr)
+d_winmf(const wchar_t *fn, int x, int y, int w, int h, const wchar_t *descr)
 { HDC hdc;
   RECT bb;
   HDC refdc = GetDC(NULL);
-#if 1
   int hmm   = GetDeviceCaps(refdc, HORZSIZE);
   int hpxl  = GetDeviceCaps(refdc, HORZRES);
   int vmm   = GetDeviceCaps(refdc, VERTSIZE);
@@ -602,27 +604,18 @@ d_winmf(const char *fn, int x, int y, int w, int h, const char *descr)
 		x, y, x+w, y+h,
 		hmm, hpxl, vmm, vpxl,
 		bb.left, bb.top, bb.right, bb.bottom));
-#else
-  int rx    = GetDeviceCaps(refdc, LOGPIXELSX);
-  int ry    = GetDeviceCaps(refdc, LOGPIXELSY);
-
-  bb.left   = ((x*254)+rx-1)/rx;
-  bb.right  = (((x+w)*254)+rx-1)/rx;
-  bb.top    = ((y*254)+ry-1)/ry;
-  bb.bottom = (((y+h)*254)+ry-1)/ry;
-#endif
 
   DEBUG(NAME_winMetafile,
 	Cprintf("BB in .01 mm = %d,%d,%d,%d\n",
 		bb.left, bb.top, bb.right, bb.bottom));
 
-  if ( (hdc = CreateEnhMetaFile(refdc,		/* HDC reference */
-				fn,		/* name of the metafile */
-				&bb,		/* bounding box */
-//				NULL,		/* or let GDI compute BB */
-				descr)) )
+  if ( (hdc = CreateEnhMetaFileW(refdc,		/* HDC reference */
+				 fn,		/* name of the metafile */
+				 &bb,		/* bounding box */
+//				 NULL,		/* or let GDI compute BB */
+				 descr)) )
   { ReleaseDC(NULL, refdc);
-    d_hdc(hdc, DEFAULT, DEFAULT);
+    d_hdc(hdc, DEFAULT, DEFAULT, FALSE);
     SetBkMode(hdc, TRANSPARENT);
 
     succeed;
@@ -2301,6 +2294,91 @@ r_op_image(Image image, int sx, int sy, int x, int y, int w, int h, Name op)
 #include "imglib.h"
 #endif
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void *getDIBImage(Image, BITMAPINFO *dib)
+    This function creates a DIB from Image. The DIB is needed as an
+    intermediate if Image needs to be painted on a HDC of different
+    kind, such as a printer or metafile.  If Image has a mask, all
+    masked pixels are set to white (255,255,255), to (generally)
+    improve drawing on devices that do not deal with transparency,
+    such as PostScript printers (at least, their Windows drivers :-)
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void *
+getDIBImage(Image image, BITMAPINFO *dib)
+{ HDC hdc;
+  HBITMAP bm = (HBITMAP) getXrefObject(image, context.display);
+  int delete = FALSE;
+  void *data = NULL;
+  int w = valInt(image->size->w);
+  int h = valInt(image->size->h);
+  HPALETTE ohpal = NULL;
+
+  hdc = CreateCompatibleDC(NULL);
+  if ( context.hpal )
+    ohpal = SelectPalette(hdc, context.hpal, FALSE);
+
+  if ( notNil(image->mask) )
+  { HBITMAP mask = (HBITMAP) getXrefObject(image->mask, context.display);
+    HBITMAP copy = CopyImage(bm, IMAGE_BITMAP, 0, 0, LR_COPYRETURNORG);
+    HDC mskhdc = CreateCompatibleDC(hdc);
+    HBITMAP obm = ZSelectObject(hdc, copy);
+    HBITMAP mobm;
+    COLORREF oldbg, oldfg;
+    HPALETTE ohpal2 = NULL;
+
+    if ( context.hpal )
+      ohpal2 = SelectPalette(mskhdc, context.hpal, FALSE);
+    mobm = ZSelectObject(mskhdc, mask);
+
+    oldbg = SetBkColor(hdc, RGB(255,255,255));
+    oldfg = SetTextColor(hdc, RGB(0,0,0));
+    if ( !BitBlt(hdc, 0, 0, w, h, mskhdc, 0, 0, SRCPAINT) )
+      Cprintf("BitBlt() failed\n");
+    SetTextColor(hdc, oldfg);
+    SetBkColor(hdc, oldbg);
+
+    ZSelectObject(mskhdc, mobm);
+    if ( ohpal2 )
+      SelectPalette(mskhdc, ohpal2, FALSE);
+    DeleteDC(mskhdc);
+    ZSelectObject(hdc, obm);
+
+    bm = copy;
+    delete = TRUE;
+  }
+
+  memset(dib, 0, sizeof(*dib));
+  dib->bmiHeader.biSize = sizeof(dib->bmiHeader);
+  dib->bmiHeader.biWidth = w;
+  dib->bmiHeader.biHeight = -h;		/* work top-down */
+  dib->bmiHeader.biPlanes = 1;
+  dib->bmiHeader.biBitCount = 24;
+  dib->bmiHeader.biCompression = BI_RGB; /* from WINGDI, this is 0 */
+
+  if ( GetDIBits(hdc, bm, 0, h, NULL, dib, DIB_RGB_COLORS) )
+  { data = pceMalloc(dib->bmiHeader.biSizeImage);
+
+    if ( !GetDIBits(hdc, bm, 0, h, data, dib, DIB_RGB_COLORS) )
+    { pceFree(data);
+      data = NULL;
+    }
+  }
+
+  if ( delete )
+    DeleteObject(bm);
+  if ( ohpal )
+    SelectPalette(hdc, ohpal, FALSE);
+  DeleteDC(hdc);
+
+  if ( !data )
+    DEBUG(NAME_image, Cprintf("Failed to get DIB\n"));
+
+  return data;
+}
+
+
 void
 r_image(Image image,
 	int sx, int sy,
@@ -2308,46 +2386,17 @@ r_image(Image image,
 	Bool transparent)
 { DEBUG(NAME_image, Cprintf("r_image(%s, %d, %d, %d+%d, %dx%d, %s)\n",
 			    pp(image), sx, sy, x, y, w, h, pp(transparent)));
+
   if ( w > 0 && h > 0 &&
        image->size->w != ZERO && image->size->h != ZERO )
-  {
-#if 0					/* very slow */
-    if ( image->ws_ref && ((WsImage)image->ws_ref)->msw_info &&
-	 isNil(image->mask) && 0 )
-    { WsImage wsi = image->ws_ref;
-
-      if ( transparent == ON )
-      { HBRUSH hbrush = ZCreateSolidBrush(context.rgb);
-	HBRUSH oldbrush = ZSelectObject(context.hdc, hbrush);
-	COLORREF oldbk = SetBkColor(context.hdc, RGB(255,255,255));
-	COLORREF oldtx = SetTextColor(context.hdc, RGB(0,0,0));
-
-	StretchDIBits(context.hdc, x, y, w, h, sx, sy, w, h,
-		      wsi->data, wsi->msw_info, DIB_RGB_COLORS, 0xB8074AL);
-
-	SetTextColor(context.hdc, oldtx);
-	SetBkColor(context.hdc, oldbk);
-  
-	ZSelectObject(context.hdc, oldbrush);
-	ZDeleteObject(hbrush);
-      } else
-      { DEBUG(NAME_image, Cprintf("Using StretchDIBits()\n"));
-	StretchDIBits(context.hdc,
-		      x, y, w, h,
-		      sx, /*valInt(image->size->h)-*/sy, w, h,
-		      wsi->data, wsi->msw_info, DIB_RGB_COLORS, SRCCOPY);
-      }
-    } else
-#endif
+  { if ( (transparent == ON && image->kind == NAME_bitmap) ||
+	 context.compatible )
     { HBITMAP obm, bm = (HBITMAP) getXrefObject(image, context.display);
       HDC mhdc = CreateCompatibleDC(context.hdc);
       HPALETTE ohpal = NULL;
-  
-      DEBUG(NAME_image, Cprintf("from HBITMAP\n"));
 
       if ( context.hpal )
 	ohpal = SelectPalette(mhdc, context.hpal, FALSE);
-
       obm = ZSelectObject(mhdc, bm);
 
       if ( transparent == ON && image->kind == NAME_bitmap )
@@ -2355,60 +2404,74 @@ r_image(Image image,
 	HBRUSH oldbrush = ZSelectObject(context.hdc, hbrush);
 	COLORREF oldbk = SetBkColor(context.hdc, RGB(255,255,255));
 	COLORREF oldtx = SetTextColor(context.hdc, RGB(0,0,0));
-	
+      
 	BitBlt(context.hdc, x, y, w, h, mhdc, sx, sy, 0xB8074AL);
-	/* ROP from "Programming Windows3.1" */
-	/* 3-rd edition, page 633 */
+      /* ROP from "Programming Windows3.1" */
+      /* 3-rd edition, page 633 */
 	SetTextColor(context.hdc, oldtx);
 	SetBkColor(context.hdc, oldbk);
-  
+
 	ZSelectObject(context.hdc, oldbrush);
 	ZDeleteObject(hbrush);
-      } else if ( notNil(image->mask) )
-      { HBITMAP msk = (HBITMAP) getXrefObject(image->mask, context.display);
-	HDC chdc = CreateCompatibleDC(context.hdc);
-	HBITMAP obm = ZSelectObject(chdc, msk);
-	COLORREF oldbg, oldfg;
-
-	DEBUG(NAME_image, Cprintf("Using BitBlt\n"));
-
-	oldbg = SetBkColor(context.hdc, RGB(255,255,255));
-	oldfg = SetTextColor(context.hdc, RGB(0,0,0));
-	BitBlt(context.hdc, x, y, w, h, chdc, sx, sy, SRCAND);
-	SetBkColor(context.hdc, oldbg);
-	SetTextColor(context.hdc, oldfg);
-
-	if ( image->kind == NAME_bitmap )
-	{ HDC hdc = CreateCompatibleDC(context.hdc);
-	  HBITMAP otbm, tbm = ZCreateCompatibleBitmap(context.hdc, w, h);
-	  otbm = ZSelectObject(hdc, tbm);
-
-	  oldbg = SetBkColor(hdc, context.background_rgb);
-	  oldfg = SetTextColor(hdc, context.rgb);
-	  BitBlt(hdc, 0, 0, w, h, mhdc, sx, sy, SRCCOPY);
-	  SetBkColor(hdc, RGB(0,0,0));
-	  SetTextColor(hdc, RGB(255,255,255));
-	  BitBlt(hdc, 0, 0, w, h, chdc, sx, sy, SRCAND);
-	  SetBkColor(hdc, oldbg);
-	  SetTextColor(hdc, oldfg);
-	  BitBlt(context.hdc, x, y, w, h, hdc, 0, 0, SRCPAINT);
-	  ZSelectObject(hdc, otbm);
-	  ZDeleteObject(tbm);
-	  DeleteDC(hdc);
+      } else if ( context.compatible )
+      { if ( notNil(image->mask) )
+	{ HBITMAP msk = (HBITMAP) getXrefObject(image->mask, context.display);
+	  HDC chdc = CreateCompatibleDC(context.hdc);
+	  HBITMAP obm2 = ZSelectObject(chdc, msk);
+	  COLORREF oldbg, oldfg;
+  
+	  DEBUG(NAME_image, Cprintf("Using BitBlt\n"));
+  
+					/* Make non-mask part black */
+	  oldbg = SetBkColor(context.hdc, RGB(255,255,255));
+	  oldfg = SetTextColor(context.hdc, RGB(0,0,0));
+	  BitBlt(context.hdc, x, y, w, h, chdc, sx, sy, SRCAND);
+	  SetBkColor(context.hdc, oldbg);
+	  SetTextColor(context.hdc, oldfg);
+  
+	  if ( image->kind == NAME_bitmap )
+	  { HDC hdc = CreateCompatibleDC(context.hdc);
+	    HBITMAP otbm, tbm = ZCreateCompatibleBitmap(context.hdc, w, h);
+	    otbm = ZSelectObject(hdc, tbm);
+    
+	    oldbg = SetBkColor(hdc, context.background_rgb);
+	    oldfg = SetTextColor(hdc, context.rgb);
+	    BitBlt(hdc, 0, 0, w, h, mhdc, sx, sy, SRCCOPY);
+	    SetBkColor(hdc, RGB(0,0,0));
+	    SetTextColor(hdc, RGB(255,255,255));
+	    BitBlt(hdc, 0, 0, w, h, chdc, sx, sy, SRCAND);
+	    SetBkColor(hdc, oldbg);
+	    SetTextColor(hdc, oldfg);
+	    BitBlt(context.hdc, x, y, w, h, hdc, 0, 0, SRCPAINT);
+	    ZSelectObject(hdc, otbm);
+	    ZDeleteObject(tbm);
+	    DeleteDC(hdc);
+	  } else 
+	  { BitBlt(context.hdc, x, y, w, h, mhdc, sx, sy, SRCPAINT);
+	  }
+	  ZSelectObject(chdc, obm2);
+	  DeleteDC(chdc);
 	} else
-	{ BitBlt(context.hdc, x, y, w, h, mhdc, sx, sy, SRCPAINT);
+	{ BitBlt(context.hdc, x, y, w, h, mhdc, sx, sy, SRCCOPY);
 	}
-
-	ZSelectObject(chdc, obm);
-	DeleteDC(chdc);
-      } else
-      { BitBlt(context.hdc, x, y, w, h, mhdc, sx, sy, SRCCOPY);
       }
-	
+
+      ZSelectObject(mhdc, obm);
       if ( ohpal )
 	SelectPalette(mhdc, ohpal, FALSE);
-      ZSelectObject(mhdc, obm);
       DeleteDC(mhdc);
+    } else				/* incompatible HDC, make DIB */
+    { void *data;
+      BITMAPINFO dib;
+
+      DEBUG(NAME_mask, Cprintf("Using DIB\n"));
+      if ( (data = getDIBImage(image, &dib)) )
+      { StretchDIBits(context.hdc, x, y, w, h,
+		      sx, sy, w, h,
+		      data, &dib, DIB_RGB_COLORS, SRCCOPY);
+		      
+        pceFree(data);
+      }
     }
   }
 }
@@ -2613,9 +2676,9 @@ r_pixel(int x, int y, Any val)
 
   if ( isBoolean(val) )
   { if ( val == ON )
-      c = RGB(255, 255, 255);
-    else
       c = RGB(0, 0, 0);
+    else
+      c = RGB(255, 255, 255);
   } else
     c = cref_colour(val);
 
@@ -2724,39 +2787,16 @@ s_height(FontObj f)
 
 
 int
-c_width(unsigned int c, FontObj font)
-{ s_font(font);
-  
-  return context.wsf->widths[c];
-}
+c_width(wint_t c, FontObj font)
+{ INT w[1];
 
-				/* should move out of window module! */
-
-
-String
-str_bits_as_font(String s, FontObj f, int *shift)
-{ static string s2;
-
-  Bool b16 = getB16Font(f);
-  if ( b16 == ON && isstr8(s) )
-  { s2 = *s;
-    s2.b16 = TRUE;
-    s2.size /= 2;
-    if ( shift )
-      *shift = -1;
-    return &s2;
-  } else if ( b16 != ON && !isstr8(s) )
-  { s2 = *s;
-    s2.b16 = FALSE;
-    s2.size *= 2;
-    if ( shift )
-      *shift = 1;
-    return &s2;
-  }
-  
-  if ( shift )
-    *shift = 0;
-  return s;
+  s_font(font);
+  if ( GetCharWidth32W(context.hdc, c, c, w) )
+    return w[0];
+  else if ( GetCharWidth32W(context.hdc, 'x', 'x', w) )
+    return w[0];
+  else
+    return 10;				/* avoid crashes */
 }
 
 
@@ -2765,56 +2805,29 @@ s_width_(String s, int from, int to)
 { if ( !context.wsf )
   { return 0;			/* TBD */
   } else
-  { cwidth *widths = context.wsf->widths;
-    int width;
-    int n = to-from;
+  { SIZE size;
 
-    if ( isstr8(s) )
-    { char8 *q = &s->s_text8[from];
-
-      for(width = 0; n-- > 0; q++)
-	width += widths[*q];
+    if ( isstrA(s) )
+    { GetTextExtentPoint32A(context.hdc, s->s_textA+from, to-from, &size);
     } else
-    { char16 *q = &s->s_text16[from];
-
-      for(width = 0; n-- > 0; q++)
-	width += widths[*q];
+    { GetTextExtentPoint32W(context.hdc, s->s_textW+from, to-from, &size);
     }
 
-    return width;
+    return size.cx;
   }
 }
 
 
 int
 str_width(String s, int from, int to, FontObj f)
-{ string s2;
-
-  s_font(f);
-
-  if ( f->b16 == ON && isstr8(s) )
-  { s2 = *s;
-    s2.b16 = TRUE;
-    s2.size /= 2;
-    from /= 2;
-    to /= 2;
-    s = &s2;
-  } else if ( f->b16 != ON && !isstr8(s) )
-  { s2 = *s;
-    s2.b16 = FALSE;
-    s2.size *= 2;
-    from *= 2;
-    to *= 2;
-    s = &s2;
-  }
-
-  if ( from < 0 )
+{ if ( from < 0 )
     from = 0;
   if ( from >= s->size || to <= from )
     return 0;
   if ( to > s->size )
     to = s->size;
 
+  s_font(f);
   return s_width_(s, from, to);
 }
 
@@ -2829,7 +2842,7 @@ str_advance(String s, int from, int to, FontObj f)
 
 
 void
-s_print8(char8 *s, int l, int x, int y, FontObj f)
+s_printA(charA *s, int l, int x, int y, FontObj f)
 { if ( l > 0 )
   { s_font(f);
     y -= context.wsf->ascent;
@@ -2839,17 +2852,21 @@ s_print8(char8 *s, int l, int x, int y, FontObj f)
 
 
 void
-s_print16(char16 *s, int l, int x, int y, FontObj f)
-{ Cprintf("16-bits characters are not supported on XPCE for Windows\n");
+s_printW(charW *s, int l, int x, int y, FontObj f)
+{ if ( l > 0 )
+  { s_font(f);
+    y -= context.wsf->ascent;
+    TextOutW(context.hdc, x, y, s, l);
+  }
 }
 
 
 void
 s_print(String s, int x, int y, FontObj f)
-{ if ( isstr8(s) )
-    s_print8(s->s_text8, s->size, x, y, f);
+{ if ( isstrA(s) )
+    s_printA(s->s_textA, s->size, x, y, f);
   else
-    s_print16(s->s_text16, s->size, x, y, f);
+    s_printW(s->s_textW, s->size, x, y, f);
 }
 
 
@@ -2866,9 +2883,10 @@ the baseline.
 
 static void
 str_text(String s, int x, int y)
-{ assert(isstr8(s));
-
-  TextOut(context.hdc, x, y, s->s_text8, s->size);
+{ if ( isstrA(s) )
+    TextOutA(context.hdc, x, y, s->s_textA, s->size);
+  else
+    TextOutW(context.hdc, x, y, s->s_textW, s->size);
 }    
 
 
@@ -2885,11 +2903,15 @@ str_size(String s, FontObj font, int *width, int *height)
     rect.bottom = 0;
 
     s_font(font);
-    rval = DrawText(context.hdc, s->s_text8, s->size, &rect, flags);
+    if ( isstrA(s) )
+      rval = DrawTextA(context.hdc, s->s_textA, s->size, &rect, flags);
+    else
+      rval = DrawTextW(context.hdc, s->s_textW, s->size, &rect, flags);
+
     DEBUG(NAME_font,
 	  { char buf[32];
 	    int n = min(s->size, 25);
-	    strncpy(buf, s->s_text8, n);
+	    strncpy(buf, s->s_textA, n);
 	    buf[n] = EOS;
 	    if ( s->size > 25 )
 	      strcat(buf, " ...");
@@ -2923,9 +2945,10 @@ typedef struct
 
 
 static void
-str_break_into_lines(String s, strTextLine *line, int *nlines)
+str_break_into_lines(String s, strTextLine *line, int *nlines, int maxlines)
 { int here = 0;
   int size = s->size;
+  int nls = 0;
 
   *nlines = 0;
 
@@ -2937,7 +2960,7 @@ str_break_into_lines(String s, strTextLine *line, int *nlines)
     return;
   }
 
-  for( ; here < size; line++, (*nlines)++ )
+  for( ; here < size && nls < maxlines; line++, nls++ )
   { int el;
 
     str_cphdr(&line->text, s);
@@ -2947,7 +2970,7 @@ str_break_into_lines(String s, strTextLine *line, int *nlines)
     { line->text.size = el - here;
       here = el + 1;
       if ( here == size )		/* last char is newline: add a line */
-      { line++, (*nlines)++;
+      { line++, nls++;
 	str_cphdr(&line->text, s);
 	line->text.s_text = str_textp(s, here);
 	line->text.size = 0;
@@ -2957,6 +2980,8 @@ str_break_into_lines(String s, strTextLine *line, int *nlines)
       here = size;
     }
   }
+
+  *nlines = nls;
 }
 
 
@@ -2969,11 +2994,11 @@ str_compute_lines(strTextLine *lines, int nlines, FontObj font,
   strTextLine *line;
   int n;
 
-  if ( equalName(vadjust, NAME_top) )
+  if ( vadjust == NAME_top )
     cy = y;
-  else if ( equalName(vadjust, NAME_center) )
+  else if ( vadjust == NAME_center )
     cy = y + (h - nlines*th)/2;
-  else /*if ( equalName(vadjust, NAME_bottom) )*/
+  else /*if ( vadjust == NAME_bottom )*/
     cy = y + h - nlines*th;
 
   for( n = 0, line = lines; n++ < nlines; line++, cy += th )
@@ -2981,11 +3006,11 @@ str_compute_lines(strTextLine *lines, int nlines, FontObj font,
     line->height = th;
     line->width  = str_width(&line->text, 0, line->text.size, font);
 
-    if ( equalName(hadjust, NAME_left) )
+    if ( hadjust == NAME_left )
       line->x = x;
-    else if ( equalName(hadjust, NAME_center) )
+    else if ( hadjust == NAME_center )
       line->x = x + (w - line->width)/2;
-    else /*if ( equalName(hadjust, NAME_right) )*/
+    else /*if ( hadjust == NAME_right )*/
       line->x = x + w - line->width;
   }
 }
@@ -3005,7 +3030,7 @@ ps_string(String s, FontObj font, int x, int y, int w, Name format, int flags)
   ps_font(font);
 
   baseline = s_ascent(font);
-  str_break_into_lines(s, lines, &nlines);
+  str_break_into_lines(s, lines, &nlines, MAX_TEXT_LINES);
   str_compute_lines(lines, nlines, font, x, y, w, 0, format, NAME_top);
 
   for(n=0, line = lines; n++ < nlines; line++)
@@ -3038,7 +3063,7 @@ str_string(String s, FontObj font,
 
   s_font(font);
   baseline = s_ascent(font);
-  str_break_into_lines(s, lines, &nlines);
+  str_break_into_lines(s, lines, &nlines, MAX_TEXT_LINES);
   str_compute_lines(lines, nlines, font, x, y, w, h, hadjust, vadjust);
 
   oalign = SetTextAlign(context.hdc, TA_BASELINE|TA_LEFT|TA_NOUPDATECP);
@@ -3056,40 +3081,6 @@ str_string(String s, FontObj font,
   SetTextAlign(context.hdc, oalign);
 }
 
-#if 0
-{ RECT rect;
-  UINT flags = DT_EXTERNALLEADING|DT_NOCLIP|DT_NOPREFIX;
-  
-  if ( hadjust == NAME_left )
-    flags |= DT_LEFT;
-  else if ( hadjust == NAME_center )
-    flags |= DT_CENTER;
-  else
-    flags |= DT_RIGHT;
-
-  if ( vadjust != NAME_top )
-  { int th = valInt(getHeightFont(font)) *
-             (str_count_chr(s, 0, s->size, '\n')+1);
-
-    if ( vadjust == NAME_center )
-      y += (h-th)/2;
-    else
-      y += h-th;
-  } else
-    h = 10000;				/* hack: 0x0 does not draw??? */
-  if ( hadjust == NAME_left )
-    w = 10000;
-
-  rect.left   = x;
-  rect.top    = y;
-  rect.right  = rect.left + w;
-  rect.bottom = rect.top + h;
-
-  s_font(font);
-  DrawText(context.hdc, s->s_text8, s->size, &rect, flags);
-}
-#endif
-
 
 static void
 str_draw_text_lines(int acc, FontObj font,
@@ -3107,7 +3098,7 @@ str_draw_text_lines(int acc, FontObj font,
       int cn;
 
       for(cn=0; cn<line->text.size; cn++)
-      { int c  = str_fetch(&line->text, cn);
+      { wint_t c  = str_fetch(&line->text, cn);
 	int cw = c_width(c, font);
 
 	if ( tolower(c) == acc )
@@ -3154,7 +3145,7 @@ str_label(String s, int acc, FontObj font, int x, int y, int w, int h,
     return;
 
   s_font(font);
-  str_break_into_lines(s, lines, &nlines);
+  str_break_into_lines(s, lines, &nlines, MAX_TEXT_LINES);
   str_compute_lines(lines, nlines, font, x, y, w, h, hadjust, vadjust);
 
   oalign = SetTextAlign(context.hdc, TA_BASELINE|TA_LEFT|TA_NOUPDATECP);
@@ -3195,10 +3186,10 @@ str_stext(String s, int f, int len, Style style)
 	ofg = r_colour(style->colour);
     }
 
-    if ( isstr8(s) )
-    { TextOut(context.hdc, 0, 0, s->s_text8+f, len);
+    if ( isstrA(s) )
+    { TextOut(context.hdc, 0, 0, s->s_textA+f, len);
     } else
-    { Cprintf("16-bits characters are not supported on XPCE for Windows\n");
+    { TextOutW(context.hdc, 0, 0, s->s_textW+f, len);
     }
 
     if ( notNil(style) && (style->attributes & TXT_UNDERLINED) )
@@ -3227,7 +3218,7 @@ str_selected_string(String s, FontObj font,
 
   s_font(font);
   baseline = context.wsf->ascent;
-  str_break_into_lines(s, lines, &nlines);
+  str_break_into_lines(s, lines, &nlines, MAX_TEXT_LINES);
   str_compute_lines(lines, nlines, font, x, y, w, h, hadjust, vadjust);
 
   oalign = SetTextAlign(context.hdc, TA_BASELINE|TA_LEFT|TA_UPDATECP);
