@@ -1,4 +1,4 @@
-/*  $Id: mswindow.c,v 1.39 2004/09/21 14:22:25 jan Exp $
+/*  $Id: mswindow.c,v 1.44 2007/01/15 20:43:26 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -24,7 +24,7 @@
 
 #include "include.h"
 
-static int WINAPI window_wnd_proc(HWND win, UINT msg, UINT wP, LONG lP);
+static LRESULT WINAPI window_wnd_proc(HWND w, UINT m, WPARAM wP, LPARAM lP);
 
 static int clearing_update;		/* from ws_redraw_window() */
 static int invert_window = FALSE;	/* invert the window */
@@ -115,7 +115,7 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 	   (nfiles = DragQueryFile(hdrop, (UINT)-1, NULL, 0)) >= 0 )
       { Chain files;
       	Point pos;
-	char buf[MAXPATHLEN];
+	wchar_t buf[MAXPATHLEN];
 	AnswerMark mark;
 	int i;
 
@@ -128,9 +128,8 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 		      for(i=0; i<nfiles; i++)
 		      { int namlen;
 
-			namlen = DragQueryFile(hdrop, i, buf, sizeof(buf)-1);
-			buf[namlen] = EOS;
-			appendChain(files, CtoName(buf));
+			namlen = DragQueryFileW(hdrop, i, buf, MAXPATHLEN);
+			appendChain(files, WCToName(buf, namlen));
 		      }
 			  
 		      DragFinish(hdrop);		/* reclaims memory */
@@ -206,7 +205,7 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
     { HDC hdc = (HDC) wParam;
       RECT rect;
 
-      d_hdc(hdc, NIL, sw->background);
+      d_hdc(hdc, NIL, sw->background, TRUE);
       GetClipBox(hdc, &rect);
       r_clear(rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top);
       d_done();
@@ -267,7 +266,12 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 	
 	if ( oproc )			/* refining alien window */
 					/* see winHandleWindow() below */
-	{ SetWindowLong(hwnd, GWL_WNDPROC, (LONG) oproc);
+	{
+#if (_MSC_VER < 1400)
+	  SetWindowLong(hwnd, GWL_WNDPROC, (LONG) oproc);
+#else
+	  SetWindowLongPtr(hwnd, GWLP_WNDPROC, (LONG_PTR) oproc);
+#endif
 	  return CallWindowProc(oproc, hwnd, message, wParam, lParam);
 	}
 
@@ -375,8 +379,8 @@ cascade:
 }
 
 
-static int WINAPI
-window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
+static LRESULT WINAPI
+window_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 { int rval;
 
   if ( InSendMessage() )
@@ -845,8 +849,13 @@ winHandleWindow(PceWindow sw, Int handle)
   setHwndWindow(sw, hwnd);
   assocObjectToHWND(hwnd, sw);
   w = sw->ws_ref;
+#if (_MSC_VER < 1400)
   w->saved_window_procedure = (WNDPROC)GetWindowLong(hwnd, GWL_WNDPROC);
   SetWindowLong(hwnd, GWL_WNDPROC, (LONG) window_wnd_proc);
+#else
+  w->saved_window_procedure = (WNDPROC)GetWindowLongPtr(hwnd, GWLP_WNDPROC);
+  SetWindowLongPtr(hwnd, GWLP_WNDPROC, (LONG_PTR) window_wnd_proc);
+#endif
   GetWindowRect(hwnd, &rect);
   ServiceMode(is_service_window(sw),
 	      { Area a = sw->area;
