@@ -1,4 +1,4 @@
-/*  $Id: textbuffer.c,v 1.61 2006/06/13 08:13:17 jan Exp $
+/*  $Id: textbuffer.c,v 1.64 2007/01/18 12:13:23 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -119,7 +119,7 @@ getConvertTextBuffer(Any ctx, Editor e)
 static status
 storeTextBuffer(TextBuffer tb, FileObj file)
 { IOENC oenc = file->fd->encoding;
-  size_t i;
+  int i;
 
   TRY(storeSlotsObject(tb, file));
   storeIntFile(file, toInt(tb->size));
@@ -984,9 +984,9 @@ scan_syntax_textbuffer(TextBuffer tb,
 
 					/* Prolog 0'char syntax */
       if ( c == '\'' && syntax->name == NAME_prolog && here > 0 )
-      { int c0 = fetch(here-1);
+      { wint_t c0 = fetch(here-1);
 
-	if ( isdigit(c0) )		/* or <digit><number> */
+	if ( iswdigit(c0) )		/* or <digit><number> */
 	{ if ( c0 == '0' )
 	  { here++;			/* ignore this char */
 	    if ( here >= to )
@@ -1108,9 +1108,9 @@ inStringTextBuffer(TextBuffer tb, Int pos, Int from)
 
 					/* Prolog 0'char syntax */
       if ( c == '\'' && syntax->name == NAME_prolog && here > 0 )
-      { int c0 = fetch(here-1);
+      { wint_t c0 = fetch(here-1);
 
-	if ( isdigit(c0) )
+	if ( iswdigit(c0) )
 	{ if ( c0 == '0' && idx == here+1 )
 	    succeed;
 	  continue;
@@ -1491,7 +1491,7 @@ textbuffer.  Characters after to are always left untouched.
 
 static void
 distribute_spaces(TextBuffer tb, int spaces, int nbreaks, long int *breaks)
-{ int s = spaces / (nbreaks-1);
+{ int s = (nbreaks > 1 ? (spaces / (nbreaks-1)) : 1);
   int n, m;
   int *extra = (int *)alloca(nbreaks * sizeof(int));
   String space = str_spc(&tb->buffer);
