@@ -1,4 +1,4 @@
-/*  $Id: file.c,v 1.72 2005/11/15 13:22:07 jan Exp $
+/*  $Id: file.c,v 1.76 2007/01/23 19:22:18 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -102,7 +102,7 @@ initialiseFile(FileObj f, Name name, Name encoding)
     { return errorPce(f, NAME_noTempFile, getOsErrorPce(PCE));
     }
 #else
-#ifdef HAVE_TEMPNAM			/* Prefer this on WIN32 */
+#ifdef HAVE_TEMPNAM			/* Prefer this on __WINDOWS__ */
     char *s = tempnam("c:\\tmp", "xpce");
 
     if ( s )
@@ -339,7 +339,7 @@ isAbsoluteFile(FileObj f)
 #define O_BINARY 0
 #endif
 
-#ifdef __WIN32__
+#ifdef __WINDOWS__
 #include <fcntl.h>
 #endif
 
@@ -840,7 +840,7 @@ only slow ...
 static int
 statFile(FileObj f, STAT_TYPE *buf)
 { 
-#ifndef WIN32
+#ifndef __WINDOWS__
   int fno;
 
   if ( f->fd != NULL && (fno = Sfileno(f->fd)) >= 0)
@@ -937,8 +937,7 @@ getReadLineFile(FileObj f)
 
 static StringObj
 getReadFile(FileObj f, Int n)
-{ int size;
-  int m;
+{ size_t size;
   StringObj s;
 
   TRY( check_file(f, NAME_read) );
@@ -958,19 +957,16 @@ getReadFile(FileObj f, Int n)
   }
 
   if ( f->encoding == NAME_octet )
-  { s = answerObject(ClassString, EAV);
+  { size_t m;
+
+    s = answerObject(ClassString, EAV);
     str_unalloc(&s->data);
     str_inithdr(&s->data, FALSE);
-    s->data.size = size;
+    s->data.size = (int)size;
     str_alloc(&s->data);
 
     if ( (m = Sfread(s->data.s_textA, 1, size, f->fd)) != size )
-    { if ( m >= 0 )
-	deleteString(s, toInt(m), DEFAULT);
-      else
-      { errorPce(f, NAME_ioError, getOsErrorPce(PCE), 0);
-	fail;
-      }
+    { deleteString(s, toInt(m), DEFAULT); 			/* TBD: error? */
     }
   } else
   { tmp_string tmp;
@@ -1069,6 +1065,25 @@ storeWordFile(FileObj f, Any w)
   return checkErrorFile(f);
 }
 
+#ifdef WORDS_BIGENDIAN
+static const int double_byte_order[] = { 7,6,5,4,3,2,1,0 };
+#else
+static const int double_byte_order[] = { 0,1,2,3,4,5,6,7 };
+#endif
+
+#define BYTES_PER_DOUBLE (sizeof(double_byte_order)/sizeof(int))
+
+status
+storeDoubleFile(FileObj file, double f)
+{ unsigned char *cl = (unsigned char *)&f;
+  unsigned int i;
+
+  for(i=0; i<BYTES_PER_DOUBLE; i++)
+    Sputc(cl[double_byte_order[i]], file->fd);
+
+  return checkErrorFile(file);
+}
+
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 storeStringFile() stores a string to a   file. For compatibility reasons
@@ -1197,7 +1212,7 @@ findFile(FileObj f, CharArray path, Name mode)
     const wchar_t *end = pathstr;
     size_t l;
 
-#ifdef __WIN32__
+#ifdef __WINDOWS__
     if ( end[0] < 0x80 && isalpha(end[0]) && end[1] == ':' )
       end += 2;
 #endif
@@ -1368,7 +1383,7 @@ makeClassFile(Class class)
 { declareClass(class, &file_decls);
   setLoadStoreFunctionClass(class, loadFile, storeFile);
 
-#if defined(__WIN32__)
+#if defined(__WINDOWS__)
   featureClass(class, NAME_caseSensitive,  OFF);
   featureClass(class, NAME_casePreserving, ON);
   featureClass(class, NAME_8plus3names,    OFF);
