@@ -1,4 +1,4 @@
-/*  $Id: socket.c,v 1.41 2007/01/18 12:13:23 jan Exp $
+/*  $Id: socket.c,v 1.42 2007/03/08 18:08:21 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -494,7 +494,7 @@ bindSocket(Socket s, Bool reuse)
 	{ if ( s->address == ZERO )
 	    assign(s, address, toInt(ntohs(addr.sin_port)));
 	  else
-	    assign((Tuple)s->address, second, toInt(ntohs(address.sin_port)));
+	    assign((Tuple)s->address, second, toInt(ntohs(addr.sin_port)));
 	}
       }
     }
