@@ -1,4 +1,4 @@
-/*  $Id: frame.c,v 1.67 2006/08/11 08:01:15 jan Exp $
+/*  $Id: frame.c,v 1.69 2007/01/14 20:16:01 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -244,7 +244,7 @@ openFrame(FrameObj fr, Point pos, Bool grab, Bool normalise)
   { x = pos->x;
     y = pos->y;
 
-#ifdef WIN32
+#ifdef __WINDOWS__
   setpos:
 #endif
     if ( normalise == ON )
@@ -278,7 +278,7 @@ openFrame(FrameObj fr, Point pos, Bool grab, Bool normalise)
   } else if ( notNil(fr->geometry) )
   { ws_x_geometry_frame(fr, fr->geometry, DEFAULT);
   }					
-#ifdef WIN32				/* But in Windows `do-it-yourself' */
+#ifdef __WINDOWS__				/* But in Windows `do-it-yourself' */
   else if ( notNil(fr->transient_for) )
   { Area pa = fr->transient_for->area;
     int xb, yb, ycap;
@@ -791,7 +791,7 @@ getGeometryFrame(FrameObj fr)
     { Int n = getIndexChain(fr->display->monitors, mon);
 
       if ( n )
-	sprintf(buf+strlen(buf), "@%ld", valInt(n)-1);
+	sprintf(buf+strlen(buf), "@" INTPTR_FORMAT, valInt(n)-1);
     }
 
     answer(CtoName(buf));
