@@ -1,4 +1,4 @@
-/*  $Id: zlib4pl.c,v 1.23 2007/01/23 19:22:18 jan Exp $
+/*  $Id: zlib4pl.c,v 1.27 2007/05/02 19:39:18 jan Exp $
 
     Part of SWI-Prolog
 
@@ -86,6 +86,18 @@ domain_error(term_t actual, const char *domain)
 }
 
 
+static int
+instantiation_error()
+{ term_t ex = PL_new_term_ref();
+
+  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
+		      PL_CHARS, "inistantiation_error",
+		      PL_VARIABLE);
+
+  return PL_raise_exception(ex);
+}
+
+
 static int
 get_atom_ex(term_t t, atom_t *a)
 { if ( PL_get_atom(t, a) )
@@ -128,6 +140,7 @@ typedef struct z_context
 { IOSTREAM	   *stream;		/* Original stream */
   IOSTREAM	   *zstream;		/* Compressed stream (I'm handle of) */
   int		    close_parent;	/* close parent on close */
+  int		    initialized;	/* did inflateInit()? */
   zformat	    format;		/* current format */
   uLong		    crc;		/* CRC check */
   z_stream	    zstate;		/* Zlib state */
@@ -148,7 +161,11 @@ alloc_zcontext(IOSTREAM *s)
 
 static void
 free_zcontext(z_context *ctx)
-{ PL_release_stream(ctx->stream);
+{ if ( ctx->stream->upstream )
+    Sset_filter(ctx->stream, NULL);
+  else
+    PL_release_stream(ctx->stream);
+
   PL_free(ctx);
 }
 
@@ -358,13 +375,17 @@ zread(void *handle, char *buf, size_t size)
   ctx->zstate.next_out  = (Bytef*)buf;
   ctx->zstate.avail_out = (long)size;
   
-  if ( ctx->format == F_UNKNOWN )
+  if ( ctx->initialized == FALSE )
   { Bytef *p;
 
     DEBUG(1, Sdprintf("Trying gzip header\n"));
-    while( (p = gz_skip_header(ctx, ctx->zstate.next_in,
-			       ctx->zstate.avail_in)) == HDR_SHORT )
-    { 					/* TBD: read more */
+    if ( ctx->format == F_DEFLATE )
+    { p = NULL;
+    } else
+    { while( (p = gz_skip_header(ctx, ctx->zstate.next_in,
+				 ctx->zstate.avail_in)) == HDR_SHORT )
+      { 					/* TBD: read more */
+      }
     }
 
     if ( p )
@@ -378,6 +399,7 @@ zread(void *handle, char *buf, size_t size)
 					/* init without header */
       switch(inflateInit2(&ctx->zstate, -MAX_WBITS))
       { case Z_OK:
+	  ctx->initialized = TRUE;
 	  ctx->crc = crc32(0L, Z_NULL, 0);
 	  DEBUG(1, Sdprintf("inflateInit2(): Z_OK\n"));
 	  break;
@@ -393,6 +415,7 @@ zread(void *handle, char *buf, size_t size)
     { switch(inflateInit(&ctx->zstate))
       { case Z_OK:
 	  ctx->format = F_DEFLATE;
+	  ctx->initialized = TRUE;
 	  DEBUG(1, Sdprintf("inflateInit(): Z_OK\n"));
 	  break;
 	case Z_MEM_ERROR:		/* no memory */
@@ -441,21 +464,32 @@ zread(void *handle, char *buf, size_t size)
 	if ( ctx->format == F_GZIP )
 	  ctx->format = F_GZIP_CRC;
       } else
-      { DEBUG(1, Sdprintf("Z_OK: %d bytes\n", n));
+      { DEBUG(1, Sdprintf("inflate(): Z_OK: %d bytes\n", n));
       }
 
       return n;
     }
     case Z_NEED_DICT:
+      DEBUG(1, Sdprintf("Z_NEED_DICT\n"));
+      break;
     case Z_DATA_ERROR:
+      DEBUG(1, Sdprintf("Z_DATA_ERROR\n"));
+      break;
     case Z_STREAM_ERROR:
+      DEBUG(1, Sdprintf("Z_STREAM_ERROR\n"));
+      break;
     case Z_MEM_ERROR:
+      DEBUG(1, Sdprintf("Z_MEM_ERROR\n"));
+      break;
     case Z_BUF_ERROR:
+      DEBUG(1, Sdprintf("Z_BUF_ERROR\n"));
+      break;
     default:
-      if ( ctx->zstate.msg )
-	Sdprintf("ERROR: zread(): %s\n", ctx->zstate.msg);
-      return -1;
+      DEBUG(1, Sdprintf("Inflate error: %d\n", rc));
   }
+  if ( ctx->zstate.msg )
+    Sdprintf("ERROR: zread(): %s\n", ctx->zstate.msg);
+  return -1;
 }
 
 
@@ -464,6 +498,7 @@ zwrite4(void *handle, char *buf, size_t size, int flush)
 { z_context *ctx = handle;
   Bytef buffer[SIO_BUFSIZE];
   int rc;
+  int loops = 0;
 
   ctx->zstate.next_in = (Bytef*)buf;
   ctx->zstate.avail_in = (long)size;
@@ -471,34 +506,38 @@ zwrite4(void *handle, char *buf, size_t size, int flush)
     ctx->crc = crc32(ctx->crc, ctx->zstate.next_in, ctx->zstate.avail_in);
 
   DEBUG(1, Sdprintf("Compressing %d bytes\n", ctx->zstate.avail_in));
-  ctx->zstate.next_out  = buffer;
-  ctx->zstate.avail_out = sizeof(buffer);
-  switch( (rc = deflate(&ctx->zstate, flush)) )
-  { case Z_OK:
-    case Z_STREAM_END:
-    { size_t n = sizeof(buffer) - ctx->zstate.avail_out;
+  do
+  { loops++;
+    ctx->zstate.next_out  = buffer;
+    ctx->zstate.avail_out = sizeof(buffer);
 
-      DEBUG(1, Sdprintf("Compressed (%s) to %d bytes; left %d\n",
-			rc == Z_OK ? "Z_OK" : "Z_STREAM_END",
-			n, ctx->zstate.avail_in));
-
-      if ( Sfwrite(buffer, 1, n, ctx->stream) != n )
+    switch( (rc = deflate(&ctx->zstate, flush)) )
+    { case Z_OK:
+      case Z_STREAM_END:
+      { size_t n = sizeof(buffer) - ctx->zstate.avail_out;
+  
+	DEBUG(1, Sdprintf("Compressed (%s) to %d bytes; left %d\n",
+			  rc == Z_OK ? "Z_OK" : "Z_STREAM_END",
+			  n, ctx->zstate.avail_in));
+  
+	if ( Sfwrite(buffer, 1, n, ctx->stream) != n )
+	  return -1;
+  
+	break;
+      }
+      case Z_BUF_ERROR:
+	DEBUG(1, Sdprintf("zwrite4(): Z_BUF_ERROR\n"));
+        break;
+      case Z_STREAM_ERROR:
+      default:
+	Sdprintf("ERROR: zwrite(): %s\n", ctx->zstate.msg);
 	return -1;
-      if ( size == 0 && Sflush(ctx->stream) < 0 )
-	return -1;
-
-      break;
     }
-    case Z_BUF_ERROR:
-      if ( ctx->zstate.avail_in == 0 && ctx->zstate.avail_out > 0 )
-	return size;			/* nothing to progress: not an error */
-					/* i.e. flush twice.  Better check? */
-      /*FALLTHROUGH*/
-    case Z_STREAM_ERROR:
-    default:
-      Sdprintf("ERROR: zwrite(): %s\n", ctx->zstate.msg);
-      return -1;
-  }
+  } while ( ctx->zstate.avail_in > 0 ||
+	    (flush != Z_NO_FLUSH && rc == Z_OK) );
+
+  if ( flush != Z_NO_FLUSH && Sflush(ctx->stream) < 0 )
+    return -1;
 
   return size;
 }
@@ -516,6 +555,7 @@ zcontrol(void *handle, int op, void *data)
 
   switch(op)
   { case SIO_FLUSHOUTPUT:
+      DEBUG(1, Sdprintf("Flushing output\n"));
       return (int)zwrite4(handle, NULL, 0, Z_SYNC_FLUSH);
     case SIO_SETENCODING:
       return 0;				/* allow switching encoding */
@@ -548,6 +588,9 @@ zclose(void *handle)
 
   switch(rc)
   { case Z_OK:
+      DEBUG(1, Sdprintf("%s(): Z_OK\n", 
+		        (ctx->stream->flags & SIO_INPUT) ? "inflateEnd"
+							 : "deflateEnd"));
       if ( ctx->close_parent )
       { IOSTREAM *parent = ctx->stream;
 	free_zcontext(ctx);
@@ -638,10 +681,10 @@ pl_zopen(term_t org, term_t new, term_t options)
     return FALSE;			/* Error */
   ctx = alloc_zcontext(s);
   ctx->close_parent = close_parent;
+  ctx->format = fmt;
   if ( (s->flags & SIO_OUTPUT) )
   { int rc;
 
-    ctx->format = fmt;
     if ( fmt == F_GZIP )
     { if ( write_gzip_header(ctx) < 0 )
       { free_zcontext(ctx);
@@ -669,7 +712,14 @@ pl_zopen(term_t org, term_t new, term_t options)
   }
 
   ctx->zstream = s2;
-  return PL_unify_stream(new, s2);
+  if ( PL_unify_stream(new, s2) )
+  { Sset_filter(s, s2);
+    PL_release_stream(s);
+
+    return TRUE;
+  } else
+  { return instantiation_error();
+  }    
 }
 
 
@@ -696,7 +746,7 @@ install_zlib4pl()
   ATOM_level        = PL_new_atom("level");
   ATOM_close_parent = PL_new_atom("close_parent");
   ATOM_gzip	    = PL_new_atom("gzip");
-  ATOM_deflate	    = PL_new_atom("define");
+  ATOM_deflate	    = PL_new_atom("deflate");
 
   PL_register_foreign("zopen",  3, pl_zopen,  0);
 #ifdef O_DEBUG
