@@ -1,4 +1,4 @@
-/*  $Id: pl-alloc.c,v 1.121 2007/01/23 20:39:40 jan Exp $
+/*  $Id: pl-alloc.c,v 1.129 2007/06/19 09:05:30 jan Exp $
 
     Part of SWI-Prolog
 
@@ -526,7 +526,7 @@ allocBigHeap(size_t size)
   if ( !hTop )
   { hBase = (char *)h;
     hTop = (char *)h + size;
-    heap_base = (ulong)h & ~0x007fffffL; /* 8MB */
+    heap_base = (uintptr_t)h & ~(uintptr_t)0x007fffff; /* 8MB */
   } else
   { SetHBase(h);
     SetHTop((char *)h + size);
@@ -582,7 +582,7 @@ allocHeap__LD(size_t n ARG_LD)
     if ( !hTop )
     { hBase = mem;
       hTop = (char *)mem + n;
-      heap_base = (ulong)mem & ~0x007fffffL; /* 8MB */
+      heap_base = (uintptr_t)mem & ~(uintptr_t)0x007fffff; /* 8MB */
     } else
     { SetHBase(mem);
       SetHTop((char *)mem + n);
@@ -625,14 +625,16 @@ mergeAllocPool(AllocPool to, AllocPool from)
 		*********************************/
 
 word
-outOfStack(Stack s, stack_overflow_action how)
+outOfStack(void *stack, stack_overflow_action how)
 { GET_LD
+  Stack s = stack;
+
   LD->trim_stack_requested = TRUE;
 
   switch(how)
   { case STACK_OVERFLOW_FATAL:
       LD->outofstack = s;
-      warning("Out of %s stack", s->name);
+      Sdprintf("ERROR: Out of %s stack (ungraceful overflow)", s->name);
 
       pl_abort(ABORT_FATAL);
       assert(0);
@@ -686,9 +688,8 @@ Actually, for normal operation, consPtr() is a macro from pl-data.h
 #undef consPtr
 
 static inline word
-__consPtr(void *p, int ts)
-{
-  GET_LD
+__consPtr(void *p, word ts)
+{ GET_LD
   uintptr_t v = (uintptr_t) p;
 
   v -= base_addresses[ts&STG_MASK];
@@ -697,30 +698,13 @@ __consPtr(void *p, int ts)
 }
 
 word
-consPtr(void *p, int ts)
+consPtr(void *p, word ts)
 { return __consPtr(p, ts);
 }
 
 #define consPtr(p, s) __consPtr(p, s)
 #endif
 
-static inline word
-__makeRef(Word p ARG_LD)
-{ if ( p >= (Word) lBase )
-    return makeRefL(p);
-  else
-    return makeRefG(p);
-}
-
-
-word
-makeRef(Word p)
-{ GET_LD
-  return __makeRef(p PASS_LD);		/* public version */
-}
-
-#define makeRef(p)  __makeRef(p PASS_LD)
-
 
 		/********************************
 		*        GLOBAL STACK           *
@@ -738,10 +722,10 @@ allocGlobal__LD(size_t n ARG_LD)
 { Word result;
 
   if ( roomStack(global) < (intptr_t) (n * sizeof(word)) )
-  { growStacks(NULL, NULL, NULL, FALSE, TRUE, FALSE);
+  { growStacks(NULL, NULL, NULL, 0, n * sizeof(word), 0);
 
     if ( roomStack(global) < (intptr_t) (n * sizeof(word)) )
-      outOfStack((Stack) &LD->stacks.global, STACK_OVERFLOW_FATAL);
+      outOfStack((Stack) &LD->stacks.global, STACK_OVERFLOW_THROW);
   }
 
   result = gTop;
@@ -755,9 +739,7 @@ allocGlobalNoShift__LD(size_t n ARG_LD)
 { Word result;
 
   if ( roomStack(global) < (intptr_t) (n * sizeof(word)) )
-  { assert(0);				/* TBD: see copy_term/2 */
     return NULL;
-  }
 
   result = gTop;
   gTop += n;
