@@ -1,11 +1,11 @@
-/*  $Id: pl-arith.c,v 1.118 2007/01/23 14:21:59 jan Exp $
+/*  $Id: pl-arith.c,v 1.130 2007/07/02 07:40:23 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -41,9 +41,12 @@ numbers.  This should be changed to return arbitrary Prolog  terms  some
 day.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#include <math.h>			/* avoid abs() problem with MSVC++ */
-#include <limits.h>
 #include "pl-incl.h"
+#undef abs			/* avoid abs() problem with MSVC++ */
+#include <math.h>
+#undef abs
+#define abs(a) ((a) < 0 ? -(a) : (a))
+#include <limits.h>
 #ifdef HAVE_FLOAT_H
 #include <float.h>
 #ifdef __WINDOWS__
@@ -128,7 +131,7 @@ static int		registerFunction(ArithFunction f, int index);
 static int		getCharExpression(term_t t, Number r ARG_LD);
 static int		ar_sign_i(Number n1);
 static int		ar_add(Number n1, Number n2, Number r);
-static int		ar_add_ui(Number n, intptr_t val);
+static int		ar_add_ui(Number n, long val);
 static int		ar_minus(Number n1, Number n2, Number r);
 
 
@@ -465,6 +468,72 @@ PRED_IMPL("=:=", 2, eq, PL_FA_TRANSPARENT)
   return compareNumbers(A1, A2, EQ PASS_LD);
 }
 
+
+		 /*******************************
+		 *	 ARITHMETIC STACK	*
+		 *******************************/
+
+Number
+allocArithStack(ARG1_LD)
+{ Number n;
+
+  if ( LD->arith.stack.top == LD->arith.stack.max )
+  { size_t size;
+
+    if ( LD->arith.stack.base )
+    { size = (size_t)(LD->arith.stack.max - LD->arith.stack.base);
+      LD->arith.stack.base = PL_realloc(LD->arith.stack.base, size*sizeof(number)*2);
+      LD->arith.stack.top  = LD->arith.stack.base+size;
+      size *= 2;
+    } else
+    { size = 16;
+      LD->arith.stack.base = PL_malloc(size*sizeof(number));
+      LD->arith.stack.top  = LD->arith.stack.base;
+    }
+
+    LD->arith.stack.max = LD->arith.stack.base+size;
+  }
+
+  n = LD->arith.stack.top;
+  LD->arith.stack.top++;
+
+  return n;
+}
+
+
+void
+pushArithStack(Number n ARG_LD)
+{ Number np = allocArithStack(PASS_LD1);
+
+  *np = *n;				/* structure copy */
+}
+
+
+void
+resetArithStack(ARG1_LD)
+{ LD->arith.stack.top = LD->arith.stack.base;
+}
+
+
+Number
+argvArithStack(int n ARG_LD)
+{ assert(LD->arith.stack.top - n >= LD->arith.stack.base);
+  
+  return LD->arith.stack.top - n;
+}
+
+
+void
+popArgvArithStack(int n ARG_LD)
+{ assert(LD->arith.stack.top - n >= LD->arith.stack.base);
+  
+  for(; n>0; n--)
+  { LD->arith.stack.top--;
+    clearNumber(LD->arith.stack.top);
+  }
+}
+
+
 		/********************************
 		*           FUNCTIONS           *
 		*********************************/
@@ -576,6 +645,13 @@ check_float(double f)
     case FP_PINF:
       return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
       break;
+    case FP_NDENORM:			/* pos/neg denormalized non-zero */
+    case FP_PDENORM:
+    case FP_NNORM:			/* pos/neg normalized non-zero */
+    case FP_PNORM:
+    case FP_NZERO:			/* pos/neg zero */
+    case FP_PZERO:
+      break;
   }
 #else
 #ifdef HAVE__FPCLASS
@@ -720,6 +796,32 @@ valueExpression(term_t t, Number r ARG_LD)
 	PL_reset_term_refs(a);
 	break;
       }
+      case 3:
+      { term_t a = PL_new_term_ref();
+	number n1, n2, n3;
+
+	_PL_get_arg(1, t, a);
+	if ( valueExpression(a, &n1 PASS_LD) )
+	{ _PL_get_arg(2, t, a);
+	  if ( valueExpression(a, &n2 PASS_LD) )
+	  { _PL_get_arg(3, t, a);
+	    if ( valueExpression(a, &n3 PASS_LD) )
+	    { rval = (*f->function)(&n1, &n2, &n3, r);
+	      clearNumber(&n3);
+	    } else
+	    { rval = FALSE;
+	    }
+	    clearNumber(&n2);
+	  } else
+	  { rval = FALSE;
+	  }
+	  clearNumber(&n1);
+	} else
+	  rval = FALSE;
+
+	PL_reset_term_refs(a);
+	break;
+      }
       default:
 	sysError("Illegal arity for arithmic function");
         rval = FALSE;
@@ -808,15 +910,40 @@ getCharExpression(term_t t, Number r ARG_LD)
 		 *	     CONVERSION		*
 		 *******************************/
 
+static int
+double_in_int64_range(double x)
+{ int k;
+  double y = frexp(x, &k);
+
+  if ( k < 8*(int)sizeof(int64_t) ||
+       (y == -0.5 && k == 8*(int)sizeof(int64_t)) )
+    return TRUE;
+
+  return FALSE;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+toIntegerNumber(Number n, int flags)
+
+Convert a number to an integer. Default,   only rationals that happen to
+be integer are converted. If   TOINT_CONVERT_FLOAT  is present, floating
+point  numbers  are  converted  if  they  represent  integers.  If  also
+TOINT_TRUNCATE is provided non-integer floats are truncated to integers.
+
+Note that if a double is  out  of   range  for  int64_t,  it never has a
+fractional part.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
-toIntegerNumber(Number n)
+toIntegerNumber(Number n, int flags)
 { switch(n->type)
   { case V_INTEGER:
       succeed;
 #ifdef O_GMP
     case V_MPZ:
       succeed;
-    case V_MPQ:
+    case V_MPQ:				/* never from stacks iff integer */
       if ( mpz_cmp_ui(mpq_denref(n->value.mpq), 1L) == 0 )
       { mpz_clear(mpq_denref(n->value.mpq));
 	n->value.mpz[0] = mpq_numref(n->value.mpq)[0];
@@ -826,22 +953,28 @@ toIntegerNumber(Number n)
       fail;
 #endif
     case V_REAL:
-    { int64_t l;
-
-#ifdef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
-      if ( !((n->value.f >= PLMININT) && (n->value.f <= PLMAXINT)) )
-	fail;
+      if ( (flags & TOINT_CONVERT_FLOAT) )
+      { if ( double_in_int64_range(n->value.f) )
+	{ int64_t l = (int64_t)n->value.f;
+	  
+	  if ( (flags & TOINT_TRUNCATE) ||
+	       (double)l == n->value.f )
+	  { n->value.i = l;
+	    n->type = V_INTEGER;
+	    
+	    return TRUE;
+	  }
+	  return FALSE;
+#ifdef O_GMP
+	} else
+	{ mpz_init_set_d(n->value.mpz, n->value.f);
+	  n->type = V_MPZ;
+	  
+	  return TRUE;
 #endif
-
-      l = (int64_t)n->value.f;
-      if ( n->value.f == (real) l )
-      { n->value.i = l;
-	n->type = V_INTEGER;
-	succeed;
+	}
       }
-  
-      fail;
-    }
+      return FALSE;
   }
 
   assert(0);
@@ -878,7 +1011,7 @@ promoteIntNumber(Number n)
 		*********************************/
 
 static int
-ar_add_ui(Number n, intptr_t add)
+ar_add_ui(Number n, long add)
 { switch(n->type)
   { case V_INTEGER:
     { int64_t r = n->value.i + add;
@@ -895,9 +1028,9 @@ ar_add_ui(Number n, intptr_t add)
 #ifdef O_GMP
     case V_MPZ:
     { if ( add > 0 )
-	mpz_add_ui(n->value.mpz, n->value.mpz, add);
+	mpz_add_ui(n->value.mpz, n->value.mpz, (unsigned long)add);
       else
-	mpz_sub_ui(n->value.mpz, n->value.mpz, -add);
+	mpz_sub_ui(n->value.mpz, n->value.mpz, (unsigned long)-add);
 
       succeed;
     }
@@ -1001,9 +1134,9 @@ ar_minus(Number n1, Number n2, Number r)
 
 static int
 ar_mod(Number n1, Number n2, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("mod", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
-  if ( !toIntegerNumber(n2) )
+  if ( !toIntegerNumber(n2, 0) )
     return PL_error("mod", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
 
   same_type_numbers(n1, n2);
@@ -1054,12 +1187,12 @@ msb64(int64_t i)
 static int
 ar_shift(Number n1, Number n2, Number r, int dir) 
 { GET_LD
-  intptr_t shift;
+  long shift;
   const char *plop = (dir < 0 ? "<<" : ">>");
 
-  if ( !toIntegerNumber(n1) ) 
+  if ( !toIntegerNumber(n1, 0) ) 
     return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n1); 
-  if ( !toIntegerNumber(n2) ) 
+  if ( !toIntegerNumber(n2, 0) ) 
     return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n2); 
 
   if ( ar_sign_i(n1) == 0 )		/* shift of 0 is always 0 */
@@ -1073,7 +1206,7 @@ ar_shift(Number n1, Number n2, Number r, int dir)
 	   n2->value.i > LONG_MAX )
 	return (int)outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_RAISE);
       else
-	shift = (intptr_t)n2->value.i;
+	shift = (long)n2->value.i;
       break;
 #ifdef O_GMP
     case V_MPZ:
@@ -1163,9 +1296,9 @@ ar_shift_right(Number n1, Number n2, Number r)
 #define BINAIRY_INT_FUNCTION(name, plop, op, mpop) \
   static int \
   name(Number n1, Number n2, Number r) \
-  { if ( !toIntegerNumber(n1) ) \
+  { if ( !toIntegerNumber(n1, 0) ) \
       return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n1); \
-    if ( !toIntegerNumber(n2) ) \
+    if ( !toIntegerNumber(n2, 0) ) \
       return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n2); \
     same_type_numbers(n1, n2); \
     switch(n1->type) \
@@ -1189,9 +1322,9 @@ ar_shift_right(Number n1, Number n2, Number r)
 #define BINAIRY_INT_FUNCTION(name, plop, op, mpop) \
   static int \
   name(Number n1, Number n2, Number r) \
-  { if ( !toIntegerNumber(n1) ) \
+  { if ( !toIntegerNumber(n1, 0) ) \
       return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n1); \
-    if ( !toIntegerNumber(n2) ) \
+    if ( !toIntegerNumber(n2, 0) ) \
       return PL_error(plop, 2, NULL, ERR_AR_TYPE, ATOM_integer, n2); \
     same_type_numbers(n1, n2); \
     switch(n1->type) \
@@ -1233,13 +1366,13 @@ ar_pow(Number n1, Number n2, Number r)
 { 
 #ifdef O_GMP
   if ( intNumber(n1) && intNumber(n2) )
-  { uintptr_t exp;
+  { unsigned long exp;
 
     switch(n2->type)
     { case V_INTEGER:
 	if ( n2->value.i < 0 || n2->value.i > LONG_MAX )
 	  goto doreal;
-	exp = (uintptr_t)n2->value.i;
+	exp = (long)n2->value.i;
 	break;
       case V_MPZ:
 	if ( mpz_sgn(n2->value.mpz) < 0 ||
@@ -1258,7 +1391,7 @@ ar_pow(Number n1, Number n2, Number r)
     switch(n1->type)
     { case V_INTEGER:
 	if ( n1->value.i >= 0L && n1->value.i <= LONG_MAX )
-	{ mpz_ui_pow_ui(r->value.mpz, (intptr_t)n1->value.i, exp);
+	{ mpz_ui_pow_ui(r->value.mpz, (unsigned long)n1->value.i, exp);
 	  succeed;
 	} else
 	{ promoteToMPZNumber(n1);
@@ -1284,6 +1417,33 @@ doreal:
 }
 
 
+static int
+ar_powm(Number base, Number exp, Number mod, Number r)
+{ 
+  if ( !intNumber(base) )
+    PL_error("powm", 3, NULL, ERR_AR_TYPE, ATOM_integer, base);
+  if ( !intNumber(exp) )
+    PL_error("powm", 3, NULL, ERR_AR_TYPE, ATOM_integer, exp);
+  if ( !intNumber(exp) )
+    PL_error("powm", 3, NULL, ERR_AR_TYPE, ATOM_integer, exp);
+
+#ifdef O_GMP
+  promoteToMPZNumber(base);
+  promoteToMPZNumber(exp);
+  promoteToMPZNumber(mod);
+
+  r->type = V_MPZ;
+  mpz_init(r->value.mpz);
+  mpz_powm(r->value.mpz, base->value.mpz, exp->value.mpz, mod->value.mpz);
+
+  succeed;
+#else
+  return PL_error("powm", 3, "requires unbounded arithmetic (GMP) support",
+		  ERR_NOT_IMPLEMENTED_FEATURE, "powm/3");
+#endif
+}
+
+
 static int
 ar_sqrt(Number n1, Number r)
 { promoteToRealNumber(n1);
@@ -1341,9 +1501,9 @@ ar_log10(Number n1, Number r)
 
 static int
 ar_div(Number n1, Number n2, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("//", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
-  if ( !toIntegerNumber(n2) )
+  if ( !toIntegerNumber(n2, 0) )
     return PL_error("//", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
 
 #ifdef O_GMP
@@ -1405,9 +1565,9 @@ static int
 ar_rem(Number n1, Number n2, Number r)
 { real f;
 
-  if ( !toIntegerNumber(n1) )
+  if ( !toIntegerNumber(n1, 0) )
     return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
-  if ( !toIntegerNumber(n2) )
+  if ( !toIntegerNumber(n2, 0) )
     return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
 
   same_type_numbers(n1, n2);
@@ -1511,8 +1671,8 @@ ar_rationalize(Number n1, Number r)
 
 static int
 ar_rdiv(Number n1, Number n2, Number r)
-{ if ( toIntegerNumber(n1) &&
-       toIntegerNumber(n2) )
+{ if ( toIntegerNumber(n1, 0) &&
+       toIntegerNumber(n2, 0) )
   { promoteToMPZNumber(n1);
     promoteToMPZNumber(n2);
 
@@ -1769,7 +1929,7 @@ ar_min(Number n1, Number n2, Number r)
 
 static int
 ar_negation(Number n1, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("\\", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   switch(n1->type)
@@ -1816,7 +1976,7 @@ mustBePositive(const char *f, int a, Number n)
 
 static int
 ar_msb(Number n1, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("msb", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   switch(n1->type)
@@ -1866,7 +2026,7 @@ lsb64(int64_t i)
 
 static int
 ar_lsb(Number n1, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("lsb", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   switch(n1->type)
@@ -1909,7 +2069,7 @@ popcount64(int64_t i)
 
 static int
 ar_popcount(Number n1, Number r)
-{ if ( !toIntegerNumber(n1) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("popcount", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   switch(n1->type)
@@ -1976,8 +2136,13 @@ ar_u_minus(Number n1, Number r)
 }
 
 
-#undef abs
-#define abs(a) ((a) < 0 ? -(a) : (a))
+static int
+ar_u_plus(Number n1, Number r)
+{ cpNumber(r, n1);
+
+  succeed;
+}
+
 
 static int
 ar_abs(Number n1, Number r)
@@ -2126,14 +2291,8 @@ ar_floor(Number n1, Number r)
 #ifdef HAVE_FLOOR
       r->type = V_REAL;
       r->value.f = floor(n1->value.f);
-      if ( !toIntegerNumber(r) )
-      {
-#ifdef O_GMP
-	mpz_init_set_d(r->value.mpz, n1->value.f);
-	r->type = V_MPZ;
-#else
-	return PL_error("floor", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
-#endif
+      if ( !toIntegerNumber(r, TOINT_CONVERT_FLOAT|TOINT_TRUNCATE) )
+      { return PL_error("floor", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
       }
 #else /*HAVE_FLOOR*/
       if ( n1->value.f > (double)PLMININT && n1->value.f < (double)PLMAXINT )
@@ -2143,7 +2302,7 @@ ar_floor(Number n1, Number r)
 	r->type = V_INTEGER;
       } else
       { 
-#ifdef O_GMP:
+#ifdef O_GMP
 	r->type = V_MPZ;
 	mpz_init_set_d(r->value.mpz, n1->value.f);
 	if ( n1->value.f < 0 && 
@@ -2185,14 +2344,8 @@ ar_ceil(Number n1, Number r)
 #ifdef HAVE_CEIL
        r->type = V_REAL;
        r->value.f = ceil(n1->value.f);
-       if ( !toIntegerNumber(r) )
-       {
-#ifdef O_GMP
-         mpz_init_set_d(r->value.mpz, n1->value.f);
-	 r->type = V_MPZ;
-#else
-         return PL_error("ceil", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
-#endif
+       if ( !toIntegerNumber(r, TOINT_CONVERT_FLOAT|TOINT_TRUNCATE) )
+       { return PL_error("ceil", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
        }
 #else /*HAVE_CEIL*/
        if ( n1->value.f > (double)PLMININT && n1->value.f < (double)PLMAXINT )
@@ -2202,7 +2355,7 @@ ar_ceil(Number n1, Number r)
 	 r->type = V_INTEGER;
        } else
        {
-#ifdef O_GMP:
+#ifdef O_GMP
          r->type = V_MPZ;
 	 mpz_init_set_d(r->value.mpz, n1->value.f);
 	 if ( mpz_get_d(r->value.mpz) < n1->value.f )
@@ -2318,7 +2471,7 @@ static int
 ar_random(Number n1, Number r)
 { uint64_t bound;
 
-  if ( !toIntegerNumber(n1) )
+  if ( !toIntegerNumber(n1, TOINT_CONVERT_FLOAT) )
     return PL_error("random", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
   switch(n1->type)
@@ -2521,6 +2674,7 @@ PRED_IMPL("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
 { PRED_LD
   int i, mx;
   term_t tmp;
+  fid_t fid;
 
   switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
@@ -2537,11 +2691,10 @@ PRED_IMPL("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
   tmp = PL_new_term_ref();
   mx = (int)entriesBuffer(function_array, ArithFunction);
 
+  fid = PL_open_foreign_frame();
   for( ; i<mx; i++ )
   { ArithFunction f = FunctionFromIndex(i);
-    mark m;
 
-    Mark(m);
     PL_put_functor(tmp, f->functor);
     if ( f->proc &&
 	 PL_unify_term(A1,
@@ -2553,7 +2706,8 @@ PRED_IMPL("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
 	succeed;
       ForeignRedoInt(i);
     }
-    Undo(m);
+
+    PL_rewind_foreign_frame(fid);
   }
 
   fail;
@@ -2579,6 +2733,7 @@ static const ar_funcdef ar_funcdefs[] = {
   ADD(FUNCTOR_rdiv2,		ar_rdiv),
 #endif
   ADD(FUNCTOR_minus1,		ar_u_minus),
+  ADD(FUNCTOR_plus1,		ar_u_plus),
   ADD(FUNCTOR_abs1,		ar_abs),
   ADD(FUNCTOR_max2,		ar_max),
   ADD(FUNCTOR_min2,		ar_min),
@@ -2626,7 +2781,8 @@ static const ar_funcdef ar_funcdefs[] = {
   ADD(FUNCTOR_cputime0,		ar_cputime),
   ADD(FUNCTOR_msb1,		ar_msb),
   ADD(FUNCTOR_lsb1,		ar_lsb),
-  ADD(FUNCTOR_popcount1,	ar_popcount)
+  ADD(FUNCTOR_popcount1,	ar_popcount),
+  ADD(FUNCTOR_powm3,		ar_powm)
 };
 
 #undef ADD
@@ -2734,65 +2890,60 @@ functorArithFunction(int n)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-ar_func_n(code, argc, Number *stack) is executed  by the A_FUNC* virtual
-machine instructions. It invalidates all numbers  it pops from the stack
-using clearNumber()
+ar_func_n(code,  argc)  is  executed  by  the  A_FUNC*  virtual  machine
+instructions. It invalidates all numbers it   pops  from the stack using
+clearNumber()
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 bool
-ar_func_n(code n, int argc, Number *stack)
+ar_func_n(int findex, int argc ARG_LD)
 { number result;
   int rval;
-  ArithFunction f = FunctionFromIndex((int)n);
-  Number sp = *stack;
+  ArithFunction f = FunctionFromIndex(findex);
+  Number argv = argvArithStack(argc PASS_LD);
 
-  sp -= argc;				/* step to start of argv */
+  if ( f->proc )			/* Prolog implementation */
+  { intptr_t lsafe = (char*)lTop - (char*)lBase;
 
-  if ( f->proc )
-  { GET_LD
-    LocalFrame lSave = lTop;		/* TBD (check with stack!) */
-    term_t h0;
-    int n;
-    fid_t fid;
-
-    lTop = (LocalFrame) (*stack);
-    fid  = PL_open_foreign_frame();
-    h0   = PL_new_term_refs(argc+1);
+    lTop = (LocalFrame)argFrameP(lTop, 1); /* needed for is/2.  See A_IS */
+    { fid_t fid = PL_open_foreign_frame();
+      term_t h0 = PL_new_term_refs(argc+1);
+      int n;
     
-    for(n=0; n<argc; n++)
-    { _PL_put_number(h0+n, &sp[n]);
-      clearNumber(&sp[n]);
-    }
+      for(n=0; n<argc; n++)
+	_PL_put_number(h0+n, argv+n);
 
-    rval = prologFunction(f, h0, &result PASS_LD);
-    PL_close_foreign_frame(fid);
-    lTop = lSave;
+      rval = prologFunction(f, h0, &result PASS_LD);
+      PL_close_foreign_frame(fid);
+    }
+    lTop = addPointer(lBase, lsafe);
   } else
   { switch(argc)
     { case 0:
 	rval = (*f->function)(&result);
         break;
       case 1:
-	rval = (*f->function)(sp, &result);
-        clearNumber(sp);
+	rval = (*f->function)(argv, &result);
         break;
       case 2:
-	rval = (*f->function)(sp, &sp[1], &result);
-        clearNumber(sp);
-	clearNumber(&sp[1]);
+	rval = (*f->function)(argv, argv+1, &result);
+	break;
+      case 3:
+	rval = (*f->function)(argv, argv+1, argv+2, &result);
         break;
       default:
 	rval = FALSE;
         sysError("Too many arguments to arithmetic function");
     }
   }
+    
+  popArgvArithStack(argc PASS_LD);
 
   if ( rval )
   { if ( result.type == V_REAL && !check_float(result.value.f) )
       return FALSE;
 
-    *sp++ = result;
-    *stack = sp;
+    pushArithStack(&result PASS_LD);
   }
 
   return rval;
