@@ -1,11 +1,11 @@
-/*  $Id: pl-attvar.c,v 1.7 2005/10/31 13:34:49 jan Exp $
+/*  $Id: pl-attvar.c,v 1.10 2007/06/01 14:23:15 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -93,6 +93,8 @@ assignAttVar(Word av, Word value ARG_LD)
 
   a = valPAttVar(*av);
   wake    = allocGlobalNoShift(4);	/* may NOT shift the stacks!!! */
+  if ( !wake )
+    return outOfStack(&LD->stacks.global, STACK_OVERFLOW_THROW);
   wake[0] = FUNCTOR_wakeup3;
   wake[1] = needsRef(*a) ? makeRef(a) : *a;
   wake[2] = needsRef(*value) ? makeRef(value) : *value;
@@ -338,7 +340,7 @@ saveWakeup(ARG1_LD)
   { fid_t fid = PL_open_foreign_frame();
     term_t s = PL_new_term_refs(2);
     
-    pl_write(LD->attvar.head), pl_nl();
+    DEBUG(1, pl_write(LD->attvar.head); pl_nl());
 
     *valTermRef(s+0) = *h;
     setVar(*h);
@@ -365,7 +367,7 @@ restoreWakeup(fid_t fid ARG_LD)
     *valTermRef(LD->attvar.head) = p[0];
     *valTermRef(LD->attvar.tail) = p[1];
 
-    pl_write(LD->attvar.head), pl_nl();
+    DEBUG(1, pl_write(LD->attvar.head); pl_nl());
 
     PL_discard_foreign_frame(fid);
   }
