@@ -1,11 +1,11 @@
-/*  $Id: pl-data.h,v 1.27 2007/01/23 20:39:40 jan Exp $
+/*  $Id: pl-data.h,v 1.32 2007/06/06 10:05:40 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -23,8 +23,6 @@
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Proposal for new Prolog data-structures.
-
 Aim
 ===
 
@@ -58,22 +56,21 @@ Sorted to standard order of terms:
 Storage places:
 
 	S	Static (global variable)
-	H	Heap
 	L	Local
 	G	Global
 	T	Trail
-	-	`In the variable'
+	-	Inline
 
-	      INDEX  STORAGE  L  G  H  T  S  -  I
-----------------------------------------------------------------
-Var		0      -                    00
-Integer		1      HG-      10 01       00
-Float		2      HG       10 01
-Atom		3      HS          01    00
-String		4      HG       10 01
-List		5      HG       10 01
-Term		6      HG       10 01
-Reference	7      HLG   11 10 01
+	      INDEX  STORAGE  L  G  T  S  -  I
+-------------------------------------------------------------
+Var		0      -                 00
+Integer		1      G-       01       00
+Float		2      G        01 
+Atom		3      S              00
+String		4      G        01
+List		5      G        01
+Term		6      G        01
+Reference	7      LG    10 01
 ----------------------------------------------------------------
 
 Adding 2 bits for the garbage collector, this adds up to 7-bits tag info,
@@ -132,15 +129,15 @@ be kept consistent.
 #define TAG_TRAILMASK	0x00000001L	/* mask for tag */
 #define TAG_TRAILADDR	0x00000000L	/* Trail-only: address */
 #define TAG_TRAILVAL	0x00000001L	/* Trail-only: value */
-#define tagTrailPtr(p)	((Word)((ulong)(p)|TAG_TRAILVAL))
-#define isTrailVal(p)	((ulong)(p)&TAG_TRAILVAL)
-#define trailValP(p)	((Word)((ulong)(p)&~TAG_TRAILMASK))
+#define tagTrailPtr(p)	((Word)((uintptr_t)(p)|TAG_TRAILVAL))
+#define isTrailVal(p)	((uintptr_t)(p)&TAG_TRAILVAL)
+#define trailValP(p)	((Word)((uintptr_t)(p)&~TAG_TRAILMASK))
 #define trailVal(p)	(*trailValP(p))
 
 #define STG_MASK	(0x3<<3)
 #define STG_STATIC	(0x0<<3)	/* storage masks */
-#define STG_GLOBAL	(0x1<<3)
-#define STG_LOCAL	(0x2<<3)
+#define STG_GLOBAL	(0x1<<3)	/* global stack */
+#define STG_LOCAL	(0x2<<3)	/* local stack */
 #define STG_RESERVED	(0x3<<3)
 
 #define STG_INLINE	STG_STATIC
@@ -207,6 +204,7 @@ and while loading .wic files.  It comes at no price.
 #define deRef2(p, d)	{ (d) = (p); deRef(d); }
 #define makeRefL(p)	consPtr(p, TAG_REFERENCE|STG_LOCAL)
 #define makeRefG(p)	consPtr(p, TAG_REFERENCE|STG_GLOBAL)
+#define makeRef(p)	((p) >= (Word)lBase ? makeRefL(p) : makeRefG(p))
 #ifdef O_ATTVAR
 #define needsRef(w)	(tag(w) <= TAG_ATTVAR)
 #else
@@ -306,7 +304,7 @@ and while loading .wic files.  It comes at no price.
 
 #define F_ARITY_BITS	5		/* upto 32 inlined arity */
 #define F_ARITY_MASK	((1<<F_ARITY_BITS)-1)
-#define MK_FUNCTOR(n, a) (((((n)<<F_ARITY_BITS)|(a))<<LMASK_BITS) | \
+#define MK_FUNCTOR(n, a) (functor_t)(((((n)<<F_ARITY_BITS)|(a))<<LMASK_BITS) | \
 			  TAG_ATOM|STG_GLOBAL)
 #define functorHashValue(f, n)	((f)>>(LMASK_BITS) & ((n)-1))
 #define indexFunctor(w)	((w)>>(LMASK_BITS+F_ARITY_BITS))
