@@ -1,4 +1,4 @@
-/*  $Id: pl-error.c,v 1.35 2006/07/27 09:22:42 jan Exp $
+/*  $Id: pl-error.c,v 1.38 2007/01/12 19:52:55 jan Exp $
 
     Part of SWI-Prolog
 
@@ -543,7 +543,7 @@ printMessage(atom_t severity, ...)
 		 *******************************/
 
 int
-PL_get_nchars_ex(term_t t, unsigned int *len, char **s, unsigned int flags)
+PL_get_nchars_ex(term_t t, size_t *len, char **s, unsigned int flags)
 { return PL_get_nchars(t, len, s, flags|CVT_EXCEPTION);
 }
 
@@ -588,6 +588,26 @@ PL_get_long_ex(term_t t, long *i)
 }
 
 
+int
+PL_get_int64_ex(term_t t, int64_t *i)
+{ if ( PL_get_int64(t, i) )
+    succeed;
+
+  return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, t);
+}
+
+
+int
+PL_get_intptr_ex(term_t t, intptr_t *i)
+{
+#if SIZEOF_LONG != SIZEOF_VOIDP && SIZEOF_VOIDP == 8
+   return PL_get_int64_ex(t, i);
+#else
+   return PL_get_long_ex(t, (long*)i);
+#endif
+}
+
+
 int
 PL_get_bool_ex(term_t t, int *i)
 { if ( PL_get_bool(t, i) )
