@@ -1,11 +1,11 @@
-/*  $Id: pl-ext.c,v 1.154 2007/01/13 12:49:07 jan Exp $
+/*  $Id: pl-ext.c,v 1.157 2007/06/01 14:24:31 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -237,7 +237,6 @@ static const PL_extension foreigns[] = {
   FRG("$clause_from_source",	3, pl_clause_from_source,	0),
 
   FRG("unify_with_occurs_check",2, pl_unify_with_occurs_check,  0),
-  FRG("\\=",			2, pl_notunify,			0),
 
   FRG("repeat",			0, pl_repeat,		     NDET),
   FRG("fail",			0, pl_fail,			0),
@@ -560,6 +559,7 @@ DECL_PLIST(file);
 DECL_PLIST(btree);
 DECL_PLIST(ctype);
 DECL_PLIST(tai);
+DECL_PLIST(gc);
 
 void
 initBuildIns(void)
@@ -584,6 +584,7 @@ initBuildIns(void)
   REG_PLIST(btree);
   REG_PLIST(ctype);
   REG_PLIST(tai);
+  REG_PLIST(gc);
 #ifdef O_ATTVAR
   REG_PLIST(attvar);
 #endif
