@@ -1,11 +1,11 @@
-/*  $Id: pl-feature.c,v 1.55 2007/01/13 13:17:45 jan Exp $
+/*  $Id: pl-feature.c,v 1.59 2007/06/26 08:55:07 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -166,9 +166,17 @@ defFeature(const char *name, int flags, ...)
       break;
     }
     case FT_ATOM:
-    { const char *text = va_arg(args, const char *);
+    { PL_chars_t text;
+
+      text.text.t    = va_arg(args, char *);
+      text.encoding  = ENC_UTF8;
+      text.storage   = PL_CHARS_HEAP;
+      text.length    = strlen(text.text.t);
+      text.canonical = FALSE;
+
+      f->value.a = textToAtom(&text);	/* registered: ok */
+      PL_free_text(&text);
 
-      f->value.a = PL_new_atom(text);	/* registered: ok */
       break;
     }
     case FT_TERM:
@@ -593,7 +601,7 @@ pl_feature5(term_t key, term_t value,
 	    control_t h)
 { feature_enum *e;
   Symbol s;
-  mark m;
+  fid_t fid;
   Module module;
 
   switch( ForeignControl(h) )
@@ -658,7 +666,7 @@ pl_feature5(term_t key, term_t value,
       succeed;
   }
 
-  Mark(m);
+  fid = PL_open_foreign_frame();
   LOCK();
   for(;;)
   { while( (s=advanceTableEnum(e->table_enum)) )
@@ -682,7 +690,7 @@ pl_feature5(term_t key, term_t value,
       { exception_term = 0;
 	setVar(*valTermRef(exception_bin));
       }
-      Undo(m);
+      PL_rewind_foreign_frame(fid);
     }
 
     if ( e->scope == ATOM_local )
@@ -780,6 +788,9 @@ initFeatures()
 #ifdef O_PLMT
   defFeature("threads",	FT_BOOL, TRUE, 0);	/* FF_READONLY? */
   defFeature("system_thread_id", FT_INTEGER|FF_READONLY, 0, 0);
+#ifdef MAX_THREADS
+  defFeature("max_threads", FT_INTEGER|FF_READONLY, MAX_THREADS);
+#endif
 #else
   defFeature("threads",	FT_BOOL|FF_READONLY, FALSE, 0);
 #endif
@@ -877,6 +888,7 @@ initFeatures()
 
   setArgvFeature();
   setTZFeature();
+  setOSFeatures();
 }
 
 
