@@ -1,4 +1,4 @@
-/*  $Id: pl-file.c,v 1.202 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-file.c,v 1.216 2007/07/02 20:33:53 jan Exp $
 
     Part of SWI-Prolog
 
@@ -757,7 +757,7 @@ closeStream(IOSTREAM *s)
       Sclose(s);
       return FALSE;
     }
-    Sclose(s);
+    Sclose(s);				/* will unlock as well */
   }
 
   succeed;
@@ -833,6 +833,7 @@ protocol(const char *str, size_t n)
   if ( LD && (s = getStream(Sprotocol)) )
   { while( n-- > 0 )
       Sputcode(*str++&0xff, s);
+    Sflush(s);
     releaseStream(s);			/* we don not check errors */
   }
 }
@@ -2124,24 +2125,51 @@ word
 pl_tab2(term_t out, term_t spaces)
 { GET_LD
   number n;
+  int rval = FALSE;
+  IOSTREAM *s;
 
-  if ( valueExpression(spaces, &n PASS_LD) &&
-       toIntegerNumber(&n) )
-  { int64_t m = n.value.i;
-    IOSTREAM *s;
+  if ( !getOutputStream(out, &s) )
+    fail;
 
-    if ( !getOutputStream(out, &s) )
-      fail;
+  if ( valueExpression(spaces, &n PASS_LD) )
+  { if ( toIntegerNumber(&n, 0) )
+    { int64_t m;
 
-    while(m-- > 0)
-    { if ( Sputcode(' ', s) < 0 )
-	break;
+      switch(n.type)
+      { case V_INTEGER:
+	  m = n.value.i;
+	  break;
+#ifdef O_GMP
+	case V_MPZ:
+	{ if ( !mpz_to_int64(n.value.mpz, &m) )
+	  { PL_error(NULL, 0, NULL, ERR_EVALUATION, ATOM_int_overflow);
+	    goto error;
+	  }
+	}
+#endif
+	default:
+	  assert(0);
+      }
+  
+      while(m-- > 0)
+      { if ( Sputcode(' ', s) < 0 )
+	  break;
+      }
+  
+      rval = TRUE;
     }
 
-    return streamStatus(s);
+    clearNumber(&n);
+  } else
+  { rval = PL_error("tab", 1, NULL, ERR_TYPE, ATOM_integer, spaces);
   }
 
-  return PL_error("tab", 1, NULL, ERR_TYPE, ATOM_integer, spaces);
+  if ( rval )
+    return streamStatus(s);
+
+error:
+  (void)streamStatus(s);
+  fail;
 }
 
 
@@ -2498,7 +2526,7 @@ pl_seeing(term_t f)
 word
 pl_seen()
 { GET_LD
-  IOSTREAM *s = Scurin;
+  IOSTREAM *s = getStream(Scurin);
 
   pl_pop_input_context();
 
@@ -2508,6 +2536,9 @@ pl_seen()
   return closeStream(s);
 }
 
+/* MT: Does not create a lock on the stream
+*/
+
 static word
 do_tell(term_t f, atom_t m)
 { GET_LD
@@ -2569,7 +2600,7 @@ pl_telling(term_t f)
 word
 pl_told()
 { GET_LD
-  IOSTREAM *s = Scurout;
+  IOSTREAM *s = getStream(Scurout);
 
   popOutputContext();
 
@@ -2814,6 +2845,10 @@ stream_eof_action_prop(IOSTREAM *s, term_t prop ARG_LD)
 #include <sys/stat.h>
 #endif
 
+#if !defined(S_ISREG) && defined(S_IFREG)
+#define S_ISREG(m) ((m&S_IFMT) == S_IFREG)
+#endif
+
 static int
 stream_reposition_prop(IOSTREAM *s, term_t prop ARG_LD)
 { atom_t val;
@@ -2824,7 +2859,7 @@ stream_reposition_prop(IOSTREAM *s, term_t prop ARG_LD)
     int fd = Sfileno(s);
     struct stat buf;
 
-    if ( fstat(fd, &buf) == 0 && (buf.st_mode & S_IFMT) == S_IFREG )
+    if ( fstat(fd, &buf) == 0 && S_ISREG(buf.st_mode) )
       val = ATOM_true;
     else
       val = ATOM_false;
@@ -2955,7 +2990,7 @@ pl_stream_property(term_t stream, term_t property, control_t h)
 { GET_LD
   IOSTREAM *s;
   prop_enum *pe;
-  mark m;
+  fid_t fid;
   term_t a1;
 
   switch( ForeignControl(h) )
@@ -3055,17 +3090,18 @@ pl_stream_property(term_t stream, term_t property, control_t h)
   }
 
 
-  Mark(m);
+  fid = PL_open_foreign_frame();
+
   for(;;)
   { if ( pe->s )				/* given stream */
-    { mark m2;
+    { fid_t fid2;
   
       if ( PL_is_variable(stream) )
       { if ( !PL_unify_stream(stream, pe->s) )
 	  goto enum_e;
       }
 
-      Mark(m2);
+      fid2 = PL_open_foreign_frame();
       for( ; pe->p->functor ; pe->p++ )
       { if ( PL_unify_functor(property, pe->p->functor) )
 	{ int rval;
@@ -3089,7 +3125,7 @@ pl_stream_property(term_t stream, term_t property, control_t h)
 	    ForeignRedoPtr(pe);
 	  }
 	}
-	Undo(m2);
+	PL_rewind_foreign_frame(fid2);
       }
       pe->s = NULL;
     }
@@ -3099,7 +3135,7 @@ pl_stream_property(term_t stream, term_t property, control_t h)
     { Symbol symb;
 
       while ( (symb=advanceTableEnum(pe->e)) )
-      { Undo(m);
+      { PL_rewind_foreign_frame(fid);
 	if ( PL_unify_stream(stream, symb->name) )
 	{ pe->s = symb->name;
 	  pe->p = sprop_list;
@@ -3224,8 +3260,7 @@ pl_set_stream_position(term_t stream, term_t pos)
 		    ERR_DOMAIN, ATOM_stream_position, pos);
   }
 
-					/* TBD: use byteno */
-  if ( Sseek64(s, charno, SIO_SEEK_SET) != charno )
+  if ( Sseek64(s, byteno, SIO_SEEK_SET) != 0 )
     return PL_error(NULL, 0, MSG_ERRNO, ERR_FILE_OPERATION,
 		    ATOM_reposition, ATOM_stream, stream);
 
@@ -3248,8 +3283,8 @@ pl_seek(term_t stream, term_t offset, term_t method, term_t newloc)
   int64_t off, new;
   IOSTREAM *s;
 
-  if ( !(PL_get_atom(method, &m)) )
-    goto badmethod;
+  if ( !(PL_get_atom_ex(method, &m)) )
+    return FALSE;
 
   if ( m == ATOM_bof )
     whence = SIO_SEEK_SET;
@@ -3258,23 +3293,30 @@ pl_seek(term_t stream, term_t offset, term_t method, term_t newloc)
   else if ( m == ATOM_eof )
     whence = SIO_SEEK_END;
   else
-  { badmethod:
     return PL_error("seek", 4, NULL, ERR_DOMAIN, ATOM_seek_method, method);
-  }
   
   if ( !PL_get_int64(offset, &off) )
     return PL_error("seek", 4, NULL, ERR_DOMAIN, ATOM_integer, offset);
 
   if ( PL_get_stream_handle(stream, &s) )
-  { if ( Sseek64(s, off, whence) < 0 )
-    { PL_error("seek", 4, OsError(), ERR_PERMISSION,
-	       ATOM_reposition, ATOM_stream, stream);
+  { int unit = Sunit_size(s);
+
+    off *= unit;
+    if ( Sseek64(s, off, whence) < 0 )
+    { if ( errno == EINVAL )
+	PL_error("seek", 4, "offset out of range", ERR_DOMAIN,
+		 ATOM_position, offset);
+      else
+	PL_error("seek", 4, OsError(), ERR_PERMISSION,
+		 ATOM_reposition, ATOM_stream, stream);
       releaseStream(s);
       fail;
     }
 
     new = Stell64(s);
     releaseStream(s);
+    new /= unit;
+
     return PL_unify_int64(newloc, new);
   }
 
@@ -3414,9 +3456,10 @@ pl_at_end_of_stream1(term_t stream)
       rval = FALSE;
     }
     
-    releaseStream(s);
     if ( rval && Sferror(s) )		/* due to error */
       return streamStatus(s);
+    else
+      releaseStream(s);
 
     return rval;
   }
@@ -3582,7 +3625,7 @@ PRED_IMPL("set_prolog_IO", 3, set_prolog_IO, 0)
     goto out;
 
   if ( PL_compare(A2, A3) == 0 )	/* == */
-  { error = Snew(out->handle, out->flags, out->functions);
+  { error = getStream(Snew(out->handle, out->flags, out->functions));
     error->flags &= ~SIO_ABUF;		/* disable buffering */
     error->flags |= SIO_NBUF;
   } else
