@@ -1,11 +1,11 @@
-/*  $Id: pl-fli.c,v 1.176 2007/01/23 09:58:02 jan Exp $
+/*  $Id: pl-fli.c,v 1.185 2007/06/19 09:05:31 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -74,9 +74,6 @@ Prolog int) is used by the garbage collector to update the stack frames.
 #endif
 #define valHandleP(h)		valTermRef(h)
 
-#undef ulong
-#define ulong uintptr_t
-
 static inline word
 valHandle__LD(term_t r ARG_LD)
 { Word p = valTermRef(r);
@@ -184,6 +181,9 @@ term_t
 PL_new_term_refs(int n)
 { GET_LD
 
+  if ( (void*)fli_context <= (void*)environment_frame )
+    fatalError("PL_new_term_refs(): No foreign environment");
+
   return PL_new_term_refs__LD(n PASS_LD);
 }
 
@@ -192,6 +192,9 @@ term_t
 PL_new_term_ref()
 { GET_LD
 
+  if ( (void*)fli_context <= (void*)environment_frame )
+    fatalError("PL_new_term_ref(): No foreign environment");
+
   return PL_new_term_ref__LD(PASS_LD1);
 }
 
@@ -773,7 +776,7 @@ it is divided by 4 and the low 2   bits  are placed at the top (they are
 normally 0). longToPointer() does the inverse operation.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static inline ulong
+static inline uintptr_t
 pointerToLong(void *ptr)
 { uintptr_t p   = (uintptr_t) ptr;
   uintptr_t low = p & 0x3L;
@@ -1356,12 +1359,24 @@ PL_get_float(term_t t, double *f)
 }
 
 
+#ifdef __WINDOWS__
+#define ULL(x) x ## ui64
+#else
+#define ULL(x) x ## ULL
+#endif
+
 int
 PL_get_pointer__LD(term_t t, void **ptr ARG_LD)
 { int64_t p;
 
   if ( PL_get_int64(t, &p) )
-  { *ptr = longToPointer((uintptr_t)p);
+  {
+#if SIZEOF_VOIDP == 4
+    if ( p & ULL(0xffffffff00000000) )
+      fail;
+#endif    
+
+    *ptr = longToPointer((uintptr_t)p);
 
     succeed;
   }
@@ -1748,6 +1763,15 @@ PL_is_compound(term_t t)
 }
 
 
+int
+PL_is_callable(term_t t)
+{ GET_LD
+  word w = valHandle(t);
+
+  return (isTerm(w) || isTextAtom(w)) ? TRUE : FALSE;
+}
+
+
 int
 PL_is_functor__LD(term_t t, functor_t f ARG_LD)
 { word w = valHandle(t);
@@ -1963,14 +1987,20 @@ PL_put_list_chars(term_t t, const char *chars)
 
 
 void
-PL_put_integer__LD(term_t t, intptr_t i ARG_LD)
+PL_put_integer__LD(term_t t, long i ARG_LD)
+{ setHandle(t, makeNum(i));
+}
+
+
+void
+PL_put_intptr__LD(term_t t, intptr_t i ARG_LD)
 { setHandle(t, makeNum(i));
 }
 
 
 #undef PL_put_integer
 void
-PL_put_integer(term_t t, intptr_t i)
+PL_put_integer(term_t t, long i)
 { GET_LD
   setHandle(t, makeNum(i));
 }
@@ -1993,7 +2023,7 @@ _PL_put_number__LD(term_t t, Number n ARG_LD)
 void
 PL_put_pointer(term_t t, void *ptr)
 { GET_LD
-  PL_put_integer(t, pointerToLong(ptr));
+  PL_put_intptr(t, pointerToLong(ptr));
 }
 
 
@@ -2113,15 +2143,17 @@ PL_unify_functor(term_t t, functor_t f)
 
       bindConst(p, name);
     } else
-    { 
+    { intptr_t needed = (1+arity) * sizeof(word);
+
 #ifdef O_SHIFT_STACKS
-      if ( roomStack(global) < (1+arity) * (intptr_t)sizeof(word) )
-      { growStacks(environment_frame, NULL, NULL, FALSE, TRUE, FALSE);
+      if ( roomStack(global) < needed )
+      { if ( !growStacks(NULL, NULL, NULL, 0, needed, 0) )
+	  fail;				/* error */
 	p = valHandleP(t);
 	deRef(p);
       }
 #else 
-      requireStack(global, sizeof(word)*(1+arity));
+      requireStack(global, needed);
 #endif
 
       { Word a = gTop;
@@ -2520,6 +2552,11 @@ cont:
     case PL_INT64:
       rval = PL_unify_int64(t, va_arg(args, int64_t));
       break;
+    case PL_INTPTR:
+    { int64_t i = va_arg(args, intptr_t);
+      rval = PL_unify_int64(t, i);
+      break;
+    }
     case PL_POINTER:
       rval = PL_unify_pointer(t, va_arg(args, void *));
       break;
@@ -2888,15 +2925,8 @@ int
 PL_unify__LD(term_t t1, term_t t2 ARG_LD)
 { Word p1 = valHandleP(t1);
   Word p2 = valHandleP(t2);
-  mark m;
 
-  Mark(m);
-  if ( !unify(p1, p2 PASS_LD) )
-  { Undo(m);
-    fail;
-  }
-
-  succeed;
+  return unify_ptrs(p1, p2 PASS_LD);
 }
 
 
@@ -3052,8 +3082,8 @@ _PL_retry(intptr_t v)
 
 foreign_t
 _PL_retry_address(void *v)
-{ if ( (ulong)v & FRG_REDO_MASK )
-    PL_fatal_error("PL_retry_address(0x%lx): bad alignment", (ulong)v);
+{ if ( (uintptr_t)v & FRG_REDO_MASK )
+    PL_fatal_error("PL_retry_address(%p): bad alignment", v);
 
   ForeignRedoPtr(v);
 }
@@ -3106,18 +3136,30 @@ PL_raise_exception(term_t exception)
 int
 PL_throw(term_t exception)
 { GET_LD
-  QueryFrame QF = find_query(environment_frame);
 
   PL_put_term(exception_bin, exception);
   exception_term = exception_bin;
   assert(exception_term);
 
-  if ( QF )
-    longjmp(QF->exception_jmp_env, 1);
+  if ( LD->exception.throw_environment )
+    longjmp(LD->exception.throw_environment->exception_jmp_env, 1);
 
   fail;
 }
 
+
+int
+PL_rethrow(void)
+{ GET_LD
+
+  if ( LD->exception.throw_environment )
+    longjmp(LD->exception.throw_environment->exception_jmp_env, 1);
+
+  fail;
+}
+
+
+
 		/********************************
 		*      REGISTERING FOREIGNS     *
 		*********************************/
