@@ -1,11 +1,11 @@
-/*  $Id: pl-fmt.c,v 1.72 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-fmt.c,v 1.76 2007/06/19 09:05:31 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2004, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -230,7 +230,7 @@ word
 pl_current_format_predicate(term_t chr, term_t descr, control_t h)
 { Symbol s = NULL;
   TableEnum e;
-  mark m;
+  fid_t fid;
 
   switch( ForeignControl(h) )
   { case FRG_FIRST_CALL:
@@ -248,14 +248,14 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
       succeed;
   }
 
+  fid = PL_open_foreign_frame();
   while( (s=advanceTableEnum(e)) )
-  { Mark(m);
-
-    if ( PL_unify_integer(chr, (intptr_t)s->name) &&
+  { if ( PL_unify_integer(chr, (intptr_t)s->name) &&
 	 unify_definition(descr, ((Procedure)s->value)->definition, 0, 0) )
     { ForeignRedoPtr(e);
     }
-    Undo(m);
+    
+    PL_rewind_foreign_frame(fid);
   }
 
   freeTableEnum(e);
@@ -309,18 +309,18 @@ pl_format3(term_t out, term_t format, term_t args)
 { redir_context ctx;
   word rc;
 
-  startCritical;
-  if ( !setupOutputRedirect(out, &ctx, FALSE) )
-  { endCritical;
-    fail;
-  }
-
-  if ( (rc = format_impl(ctx.stream, format, args)) )
-    rc = closeOutputRedirect(&ctx);
-  else
-    discardOutputRedirect(&ctx);
-       
-  endCritical;
+  EXCEPTION_GUARDED(/*code*/
+		    if ( setupOutputRedirect(out, &ctx, FALSE) )
+		    { if ( (rc = format_impl(ctx.stream, format, args)) )
+			rc = closeOutputRedirect(&ctx);
+		      else
+			discardOutputRedirect(&ctx);
+		    } else
+		      rc = FALSE;
+		    /*cleanup*/,
+		    DEBUG(1, Sdprintf("Cleanup after throw()\n"));
+		    discardOutputRedirect(&ctx);
+		    PL_rethrow(););
 
   return rc;
 }
@@ -502,7 +502,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 
 		  NEED_ARG;
 		  if ( !valueExpression(argv, &i PASS_LD) ||
-		       !toIntegerNumber(&i) )
+		       !toIntegerNumber(&i, 0) )
 		  { char f[2];
 		    
 		    f[0] = c;
@@ -876,11 +876,11 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
     }
 #ifdef O_GMP
     case V_MPZ:
-    { int len = mpz_sizeinbase(i->value.mpz, radix);
+    { size_t len = mpz_sizeinbase(i->value.mpz, radix);
       char tmp[256];
       char *buf;
 
-      if ( len+2 > (int)sizeof(tmp) )
+      if ( len+2 > sizeof(tmp) )
 	buf = PL_malloc(len+2);
       else
 	buf = tmp;
@@ -893,7 +893,7 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 	  *s = toupper(*s);
       }
       if ( split || div > 0 )
-      { int before = len-div;
+      { int before = (int)(len-div);
 	int leading;
 	char *s = buf;
 
@@ -906,7 +906,7 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 	  if ( leading == 0 )
 	    leading = 3;
 	} else
-	{ leading = len;
+	{ leading = (int)len;
 	}
 	for(s=buf; *s; s++)
 	{ if ( before-- == 0 && div > 0 )
@@ -963,7 +963,7 @@ formatFloat(int how, int arg, Number f, Buffer out)
 
 	  mpz_init(iv);
 	  mpz_set_q(iv, f->value.mpq);
-	  fbits = mpz_sizeinbase(iv, 2) + 4*arg;
+	  fbits = (int)mpz_sizeinbase(iv, 2) + 4*arg;
 	  mpz_clear(iv);
 	  break;
 	}
