@@ -1,11 +1,11 @@
-/*  $Id: pl-funcs.h,v 1.275 2007/01/23 20:39:40 jan Exp $
+/*  $Id: pl-funcs.h,v 1.289 2007/06/26 08:55:07 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -31,12 +31,9 @@ libraries and avoids Prolog picking wrong   symbols. It also reduces ELF
 symbol lookup and relocations.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define COMMON(type) SO_LOCAL type
-
-
 /* pl-alloc.c */
 COMMON(void) 		freeHeap__LD(void *mem, size_t n ARG_LD);
-COMMON(word) 		outOfStack(Stack s, stack_overflow_action how);
+COMMON(word) 		outOfStack(void *stack, stack_overflow_action how);
 COMMON(void)		outOfCore(void) NORETURN;
 COMMON(Word) 		allocGlobal__LD(size_t words ARG_LD);
 COMMON(Word) 		allocGlobalNoShift__LD(size_t words ARG_LD);
@@ -60,9 +57,8 @@ COMMON(int64_t)		valBignum__LD(word w ARG_LD);
 COMMON(word) 		globalIndirect(word in);
 COMMON(int) 		equalIndirect(word r1, word r2);
 COMMON(word) 		globalIndirectFromCode(Code *PC);
-COMMON(word) 		makeRef(Word p);
 #ifndef consPtr
-COMMON(word) 		consPtr(void *p, int ts);
+COMMON(word) 		consPtr(void *p, word ts);
 #endif
 COMMON(char *) 		store_string(const char *s);
 COMMON(void) 		remove_string(char *s);
@@ -85,14 +81,13 @@ COMMON(void)		freezeGlobal(ARG1_LD);
 
 /* pl-wam.c */
 COMMON(word) 		pl_count(void);
-COMMON(bool) 		unify_ptrs(Word t1, Word t2 ARG_LD);
-COMMON(bool) 		can_unify(Word t1, Word t2);
 COMMON(void) 		TrailAssignment(Word p);
 COMMON(void) 		do_undo(mark *m);
 COMMON(Definition) 	getProcDefinition(Procedure proc);
 COMMON(void) 		fix_term_ref_count(void);
 COMMON(word) 		pl_unify_with_occurs_check(term_t t1, term_t t2);
 COMMON(fid_t) 		PL_open_signal_foreign_frame(void);
+COMMON(bool)		foreignWakeup(ARG1_LD);
 
 /* pl-atom.c */
 COMMON(word) 		lookupAtom(const char *s, size_t len);
@@ -126,10 +121,15 @@ COMMON(void) 		initArith(void);
 COMMON(void) 		cleanupArith(void);
 COMMON(int) 		indexArithFunction(functor_t fdef, Module m);
 COMMON(functor_t) 	functorArithFunction(int n);
-COMMON(bool) 		ar_func_n(code n, int argc, Number *stack);
+COMMON(bool) 		ar_func_n(int findex, int argc ARG_LD);
 COMMON(int) 		valueExpression(term_t p, Number n ARG_LD);
-COMMON(int) 		toIntegerNumber(Number n);
+COMMON(int) 		toIntegerNumber(Number n, int flags);
 COMMON(int) 		arithChar(Word p ARG_LD);
+COMMON(Number)		allocArithStack(ARG1_LD);
+COMMON(void)		pushArithStack(Number n ARG_LD);
+COMMON(void)		resetArithStack(ARG1_LD);
+COMMON(Number)		argvArithStack(int n ARG_LD);
+COMMON(void)		popArgvArithStack(int n ARG_LD);
 
 /* pl-bag.c */
 COMMON(void) 		resetBags(void);
@@ -186,10 +186,9 @@ COMMON(word) 		parseSaveProgramOptions(term_t args,
 			bool *tty, bool *standalone);
 
 /* pl-index.c */
-COMMON(int) 		cardinalityPattern(uintptr_t pattern);
-COMMON(void) 		getIndex(Word argv, uintptr_t pattern, int card,
-			 struct index *
-			 ARG_LD);
+COMMON(int) 		cardinalityPattern(unsigned long pattern);
+COMMON(void) 		getIndex(Word argv, unsigned long pattern, int card,
+				 struct index * ARG_LD);
 COMMON(word) 		getIndexOfTerm(term_t t);
 COMMON(ClauseRef) 	firstClause(Word argv, LocalFrame fr, Definition def,
 			    ClauseRef *next ARG_LD);
@@ -376,7 +375,8 @@ COMMON(int) 		PL_unify__LD(term_t t1, term_t t2 ARG_LD);
 COMMON(int) 		PL_unify_integer__LD(term_t t1, intptr_t i ARG_LD);
 COMMON(int) 		PL_get_atom__LD(term_t t1, atom_t *a ARG_LD);
 COMMON(void) 		PL_put_atom__LD(term_t t1, atom_t a ARG_LD);
-COMMON(void) 		PL_put_integer__LD(term_t t1, intptr_t i ARG_LD);
+COMMON(void) 		PL_put_integer__LD(term_t t1, long i ARG_LD);
+COMMON(void) 		PL_put_intptr__LD(term_t t1, intptr_t i ARG_LD);
 COMMON(int) 		PL_is_atomic__LD(term_t t ARG_LD);
 COMMON(int) 		PL_is_functor__LD(term_t t, functor_t f ARG_LD);
 COMMON(int) 		PL_is_variable__LD(term_t t ARG_LD);
@@ -406,6 +406,7 @@ COMMON(void) 		registerForeignLicenses(void);
 COMMON(void)            bindExtensions(const char *module,
 				       const PL_extension *ext);
 COMMON(void)		initForeign(void);
+COMMON(int)		PL_rethrow(void);
 
 
 /* pl-fmt.c */
@@ -431,6 +432,7 @@ COMMON(word) 		pl_garbage_collect(term_t d);
 COMMON(void) 		blockGC(ARG1_LD);
 COMMON(void) 		unblockGC(ARG1_LD);
 COMMON(Word) 		findGRef(int n);
+COMMON(size_t)		nextStackSizeAbove(size_t n);
 COMMON(int) 		growStacks(LocalFrame fr, Choice ch, Code PC,
 			   intptr_t lminfree, intptr_t gminfree, intptr_t tminfree);
 COMMON(void) 		clearUninitialisedVarsFrame(LocalFrame, Code);
@@ -502,7 +504,8 @@ COMMON(uint64_t)	_PL_Random(void);
 COMMON(char *) 		canonisePath(char *path);
 COMMON(char *) 		OsPath(const char *plpath, char *ospath);
 COMMON(char *) 		PrologPath(const char *ospath, char *plpath, size_t len);
-COMMON(intptr_t) 		LastModifiedFile(char *f);
+COMMON(void)		setOSFeatures(void);
+COMMON(intptr_t) 	LastModifiedFile(char *f);
 COMMON(bool) 		ExistsFile(const char *path);
 COMMON(bool) 		AccessFile(const char *path, int mode);
 COMMON(bool) 		ExistsDirectory(const char *path);
@@ -536,7 +539,9 @@ COMMON(char *) 		findExecutable(const char *module, char *buf);
 COMMON(int) 		Pause(real time);
 
 /* pl-prims.c */
-COMMON(word) 		pl_notunify(term_t t1, term_t t2);
+COMMON(bool) 		raw_unify_ptrs(Word t1, Word t2 ARG_LD);
+COMMON(bool) 		unify_ptrs(Word t1, Word t2 ARG_LD);
+COMMON(bool) 		can_unify(Word t1, Word t2);
 COMMON(int) 		compareStandard(Word t1, Word t2, int eq ARG_LD);
 COMMON(int) 		lengthList(term_t list, int errors);
 COMMON(word) 		pl_univ(term_t t, term_t l);
@@ -624,7 +629,7 @@ COMMON(void) 		gcClausesDefinitionAndUnlock(Definition def);
 COMMON(void) 		destroyDefinition(Definition def);
 COMMON(void) 		resetReferences(void);
 COMMON(Procedure) 	resolveProcedure(functor_t f, Module module);
-COMMON(Definition) 	trapUndefined(LocalFrame fr, Code PC, Procedure proc ARG_LD);
+COMMON(Definition) 	trapUndefined(LocalFrame *frp, Code PC, Procedure proc ARG_LD);
 COMMON(word) 		pl_retract(term_t term, control_t h);
 COMMON(word) 		pl_retractall(term_t head);
 COMMON(word) 		pl_abolish(term_t atom, term_t arity);
@@ -703,6 +708,7 @@ COMMON(void) 		install_rl(void);
 
 /* pl-setup.c */
 COMMON(void) 		setupProlog(void);
+COMMON(void)		endCritical__LD(ARG1_LD);
 COMMON(foreign_t) 	pl_on_signal(term_t sig, term_t name, term_t old, term_t new);
 COMMON(handler_t) 	set_sighandler(int sig, handler_t func);
 COMMON(void) 		blockSignals(sigset_t *mask);
