@@ -1,4 +1,4 @@
-/*  $Id: pl-gc.c,v 1.109 2007/01/12 21:12:29 jan Exp $
+/*  $Id: pl-gc.c,v 1.128 2007/06/01 20:30:02 jan Exp $
 
     Part of SWI-Prolog
 
@@ -31,8 +31,6 @@
 #ifndef HAVE_MEMMOVE			/* Note order!!!! */
 #define memmove(dest, src, n) bcopy(src, dest, n)
 #endif
-#undef ulong
-#define ulong uintptr_t
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This module is based on
@@ -91,7 +89,7 @@ surrounded by a word at the start and end, indicating their length.
 			      DEBUGGING
 
 Debugging a garbage collector is a difficult job.  Bugs --like  bugs  in
-memory  allocation--  usually  cause  crashes  intptr_t  after  the  garbage
+memory  allocation--  usually  cause  crashes  long  after  the  garbage
 collection has finished.   To  simplify  debugging  a  large  number  of
 actions  are  counted  during garbage collection.  At regular points the
 consistency between these counts  is  verified.   This  causes  a  small
@@ -112,7 +110,7 @@ Marking, testing marks and extracting values from GC masked words.
 #if O_SECURE
 char tmp[256];				/* for calling print_val(), etc. */
 #define check_relocation(p) do_check_relocation(p, __FILE__, __LINE__ PASS_LD)
-#define recordMark(p)   { if ( (p) < gLimit ) \
+#define recordMark(p)   { if ( (char*)(p) < (char*)lBase ) \
 			  { assert(onStack(global, p)); \
 			    *mark_top++ = (p); \
 			  } \
@@ -139,7 +137,7 @@ char tmp[256];				/* for calling print_val(), etc. */
 
 #define get_value(p)	(*(p) & VALUE_MASK)
 #define set_value(p, w)	{ *(p) &= GC_MASK; *(p) |= w; }
-#define val_ptr2(w, s)	((Word)((ulong)valPtr2((w), (s)) & ~0x3L))
+#define val_ptr2(w, s)	((Word)((uintptr_t)valPtr2((w), (s)) & ~(uintptr_t)0x3))
 #define val_ptr(w)	val_ptr2((w), storage(w))
 
 #define inShiftedArea(area, shift, ptr) \
@@ -263,7 +261,7 @@ print_val(word val, char *buf)
   Ssprintf(buf, "%s at %s(%ld)",
 	   tag_name[tag(val)],
 	   stg_name[storage(val) >> 3],
-	   val >> LMASK_BITS);
+	   (val >> LMASK_BITS)/sizeof(word));
   if ( val & MARK_MASK )
     strcat(buf, "M");
   if ( val & FIRST_MASK )
@@ -836,32 +834,33 @@ trail-ranged described by the choicepoint.
 As far as I can  see  the  only   first-marks  in  use  at this time are
 references to the trail-stack and we use   the first marks on the global
 stack.
+
+Older versions used the argument stack. We   now use the segmented cycle
+stack to avoid allocation issues on the argument stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #if O_DESTRUCTIVE_ASSIGNMENT
-static void
+static inline void
 push_marked(Word p ARG_LD)
-{ requireStack(argument, sizeof(Word));
-  *aTop++ = p;
+{ pushSegStack(&LD->cycle.stack, &p);
 }
 
+
 static void
-popall_marked(Word *base ARG_LD)
-{ Word *p = aTop;
+popall_marked(ARG1_LD)
+{ Word p;
 
-  while(p>base)
-  { p--;
-    unmark_first(*p);
+  while( popSegStack(&LD->cycle.stack, &p) )
+  { unmark_first(p);
   }
-
-  aTop = base;
 }
 
+
 static void
 mergeTrailedAssignments(GCTrailEntry top, GCTrailEntry mark,
 			int assignments ARG_LD)
 { GCTrailEntry te;
-  Word *m = aTop;
+  LD->cycle.stack.unit_size = sizeof(Word);
 
   DEBUG(2, Sdprintf("Scanning %d trailed assignments\n", assignments));
 
@@ -891,7 +890,7 @@ mergeTrailedAssignments(GCTrailEntry top, GCTrailEntry mark,
     }
   }
 
-  popall_marked(m PASS_LD);
+  popall_marked(PASS_LD1);
   assert(assignments == 0);
 }
 #endif
@@ -1799,7 +1798,7 @@ scan_global(int marked)
   intptr_t cells = 0;
 
   for( current = gBase; current < gTop; current += (offset_cell(current)+1) )
-  { int offset;
+  { size_t offset;
 
     cells++;
 
@@ -1825,8 +1824,8 @@ scan_global(int marked)
     { if ( offset_cell(next-1) != offset )
       { errors++;
 	Sdprintf("ERROR: Illegal indirect cell on global stack at %p-%p\n"
-		 "       tag=%d, offset=%d\n",
-		 current, next, tag(*current), offset);
+		 "       tag=%d, offset=%ld\n",
+		 current, next, tag(*current), (long)offset);
 	trap_gdb();
       }
     }
@@ -2021,6 +2020,21 @@ checkStacks(LocalFrame frame, Choice choice)
   return key;
 }
 
+
+static
+PRED_IMPL("$check_stacks", 1, check_stacks, 0)
+{ char *s = NULL;
+
+  if ( PL_get_atom_chars(A1, &s) )
+    Sdprintf("Checking stacks [%s] ...", s);
+
+  checkStacks(NULL, NULL);
+  if ( s )
+     Sdprintf(" (done)\n");
+
+  succeed;
+}
+
 #endif /*O_SECURE || O_DEBUG*/
 
 #ifdef O_PLMT
@@ -2161,13 +2175,13 @@ garbageCollect(LocalFrame fr, Choice ch)
     printMessage(ATOM_informational,
 		 PL_FUNCTOR_CHARS, "gc", 1,
 		   PL_FUNCTOR_CHARS, "done", 7,
-		     PL_LONG, ggar,
-		     PL_LONG, tgar,
+		     PL_INTPTR, ggar,
+		     PL_INTPTR, tgar,
 		     PL_DOUBLE, (double)t,
-		     PL_LONG, usedStack(global),
-		     PL_LONG, usedStack(trail),
-		     PL_LONG, roomStack(global),
-		     PL_LONG, roomStack(trail));
+		     PL_INTPTR, usedStack(global),
+		     PL_INTPTR, usedStack(trail),
+		     PL_INTPTR, roomStack(global),
+		     PL_INTPTR, roomStack(trail));
 
 #ifdef O_PROFILE
   if ( fr && LD->profile.active )
@@ -2207,7 +2221,7 @@ void
 blockGC(ARG1_LD)
 { gc_status.blocked++;
 #if O_SHIFT_STACKS
-  shift_status.blocked++;
+  LD->shift_status.blocked++;
 #endif
 }
 
@@ -2216,7 +2230,7 @@ void
 unblockGC(ARG1_LD)
 { gc_status.blocked--;
 #if O_SHIFT_STACKS
-  shift_status.blocked--;
+  LD->shift_status.blocked--;
 #endif
 }
 
@@ -2394,10 +2408,10 @@ update_argument(intptr_t ls, intptr_t gs)
 		 *******************************/
 
 static void
-update_trail(intptr_t ls, intptr_t gs)
+update_trail(TrailEntry tb, intptr_t ls, intptr_t gs)
 { GET_LD
-  TrailEntry p = tBase;
-  TrailEntry t = tTop;
+  TrailEntry p = tb;			/* new base */
+  TrailEntry t = tb+(tTop-tBase);	/* new top */
 
   for( ; p < t; p++ )
   { if ( onGlobal(trailValP(p->address)) )
@@ -2426,6 +2440,22 @@ update_foreign(intptr_t ts, intptr_t ls, intptr_t gs)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Update global variables. As our pointers   areoffsets  to the stacks, we
+don't actually need to update the variables   themselves.  We do need to
+update the frozen bar however.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+update_gvars(intptr_t gs)
+{ GET_LD
+  
+  if ( LD->frozen_bar )
+  { update_pointer(&LD->frozen_bar, gs);
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Entry-point.   Update the  stacks to  reflect  their current  positions.
 This function should be called *after*  the  stacks have been relocated.
@@ -2487,9 +2517,11 @@ update_stacks(LocalFrame frame, Choice choice, Code PC,
 
     if ( gs || ls )
     { update_argument(ls, gs);
-      update_trail(ls, gs);
+      update_trail(tb, ls, gs);
     }
     update_foreign(ts, ls, gs);
+    if ( gs )
+      update_gvars(gs);
 
     updateStackHeader(local,  ls);
     updateStackHeader(global, gs);
@@ -2501,9 +2533,9 @@ update_stacks(LocalFrame frame, Choice choice, Code PC,
   }
 
   if ( ls )
-  { update_pointer(&environment_frame, ls);
-    update_pointer(&fli_context, ls);
-    update_pointer(&LD->choicepoints, ls);
+  { update_pointer(&LD->environment,         ls);
+    update_pointer(&LD->foreign_environment, ls);
+    update_pointer(&LD->choicepoints,        ls);
   }
   if ( gs )
   { update_pointer(&LD->mark_bar, gs);
@@ -2511,120 +2543,210 @@ update_stacks(LocalFrame frame, Choice choice, Code PC,
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+nextStackSize() computes the size to use for s, given it should at least
+have minfree space after the stack  expansion.   We  want stacks to grow
+along a fixed set of sizes to maximize reuse of abandoned stacks.
+
+Note that we allocate local and  global   stacks  in one chunk, so their
+combined size should come from a fixed maximum.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#undef K
+#undef MB
+#define K * 1024
+#define MB * (1024L * 1024L)
+
+size_t
+nextStackSizeAbove(size_t n)
+{ size_t size;
+
+  if ( n < 4 MB )
+  { size = 8192;
+    while ( size <= n )
+      size *= 2;
+  } else
+  { size = 4 MB;
+
+    while ( size <= n )
+    { if ( (size + size/2) > n )
+	return size + size/2;
+
+      size *= 2;
+    }
+  }
+					/* enforce real limit */
+  if ( size > (size_t)(MAXTAGGEDPTR+1) )
+    size = (size_t)(MAXTAGGEDPTR+1);
+  if ( size < n )
+    return 0;				/* still too small */
+
+  return size;
+}
+
+
 static intptr_t
 nextStackSize(Stack s, intptr_t minfree)
-{ intptr_t size  = diffPointers(s->max, s->base);
-  intptr_t limit = diffPointers(s->limit, s->base);
-  intptr_t grow;
+{ intptr_t size;
+  intptr_t limit = limitStackP(s);
 
-  grow = ((minfree+8192) > size/2 ? minfree+8192 : size/2);
-  grow = ROUND(grow, 8192);
+  if ( minfree > 0 )
+    size = nextStackSizeAbove(sizeStackP(s) + minfree);
+  else
+    size = nextStackSizeAbove(usedStackP(s) + s->minfree);
 
-  if ( size + grow > limit )
-  { if ( size + grow > (limit*3)/2 )
-      outOfStack(s, STACK_OVERFLOW_THROW);
-
-    outOfStack(s, STACK_OVERFLOW_SIGNAL);
-    grow = limit - size;
-    if ( grow < 32 * 1024 )
-      grow = 32 * 1024;
-    s->limit = addPointer(s->base, size+grow);
+  if ( size == 0 )
+  { outOfStack(s, STACK_OVERFLOW_THROW);
+    return 0;
   }
 
-  size += grow;
+  if ( size > limit )
+  { if ( size > limit+limit/2 )
+    { outOfStack(s, STACK_OVERFLOW_THROW); /* _RAISE? */
+      return 0;
+    } else
+      outOfStack(s, STACK_OVERFLOW_SIGNAL);
+  }
 
   return size;
 }
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Entry point from interpret()
+Stack shifter entry point. The arguments l, g and t request expansion of
+the local, global and trail-stacks. Non-0 versions   ask the stack to be
+modified. Positive values enlarge the stack to the next size that has at
+least the specified value free space (i.e. min-free).
+
+Negative values cause the stack to shrink to the value nearest above the
+current usage and the minimum free stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define GL_SEPARATION sizeof(word)
-
 int
-growStacks(LocalFrame fr, Choice ch, Code PC, intptr_t l, intptr_t g, intptr_t t)
+growStacks(LocalFrame fr, Choice ch, Code PC, 
+	   intptr_t l, intptr_t g, intptr_t t)
 { GET_LD
-  
+  sigset_t mask;
+  intptr_t lsize, gsize, tsize;
+  void *fatal = NULL;	/* stack we couldn't expand due to lack of memory */
+#if O_SECURE
+  word key;
+#endif
+
+  if ( LD->shift_status.blocked ||
+       PC != NULL )			/* for now, only at the call-port */
+    return FALSE;
+
+  if ( t )
+  { if ( !(tsize = nextStackSize((Stack) &LD->stacks.trail, t)) )
+      fail;
+    if ( tsize == sizeStack(trail) )
+      t = 0;
+  } else
+  { tsize = sizeStack(trail);
+  }
+
+  if ( l )
+  { if ( !(lsize = nextStackSize((Stack) &LD->stacks.local, l)) )
+      fail;
+    if ( lsize == sizeStack(local) )
+      l = 0;
+  } else
+  { lsize = sizeStack(local);
+  }
+
+  if ( g )
+  { if ( !(gsize = nextStackSize((Stack) &LD->stacks.global, g)) )
+      fail;
+    if ( gsize == sizeStack(global) )
+      g = 0;
+  } else
+  { gsize = sizeStack(global);
+  }
+
+  if ( !(l || g || t) )
+    return TRUE;			/* not a real request */
+
+  enterGC();				/* atom-gc synchronisation */
+  blockSignals(&mask);
+  blockGC(PASS_LD1);			/* avoid recursion due to */
+  gc_status.requested = FALSE;		/* printMessage() */
+
   if ( !fr )
     fr = environment_frame;
   if ( !ch )
     ch = LD->choicepoints;
 
-  if ( PC != NULL )			/* for now, only at the call-port */
-    fail;
-
-  if ( (l || g || t) && !shift_status.blocked )
   { TrailEntry tb = tBase;
     Word gb = gBase;
     LocalFrame lb = lBase;
-    intptr_t lsize = sizeStack(local);
-    intptr_t gsize = sizeStack(global);
-    intptr_t tsize = sizeStack(trail);
     double time = CpuTime(CPU_USER);
     int verbose = trueFeature(TRACE_GC_FEATURE);
     
     DEBUG(0, verbose = TRUE);
 
     if ( verbose )
-    { int i = 0;
-
-      Sdprintf("Expanding ");
-      if ( l ) Sdprintf("%s%s", i++ ? "and " : "", "local ");
-      if ( g ) Sdprintf("%s%s", i++ ? "and " : "", "global ");
-      if ( t ) Sdprintf("%s%s", i++ ? "and " : "", "trail ");
-      Sdprintf("stacks ");
+    { printMessage(ATOM_informational,
+		   PL_FUNCTOR_CHARS, "shift_stacks", 1,
+		     PL_FUNCTOR_CHARS, "start", 3,
+		       PL_BOOL, l,
+		       PL_BOOL, g,
+		       PL_BOOL, t);
     }
 
-    if ( !fr )
-      fr = environment_frame;
-    if ( !ch )
-      ch = LD->choicepoints;
-
     SECURE(if ( !scan_global(FALSE) ) sysError("Stack not ok at shift entry"));
     SECURE(key = checkStacks(fr, ch));
 
     if ( t )
-    { tsize = nextStackSize((Stack) &LD->stacks.trail, t);
-      tb = PL_realloc(tb, tsize);
-      shift_status.trail_shifts++;
+    { void *nw;
+
+      if ( (nw = realloc(tb, tsize)) )
+      { LD->shift_status.trail_shifts++;
+	tb = nw;
+      } else
+      { fatal = &LD->stacks.trail;
+	tsize = sizeStack(trail);
+      }
     }
 
     if ( g || l )
-    { intptr_t loffset = gsize + GL_SEPARATION;
-      assert(lb == addPointer(gb, loffset));	
+    { intptr_t ogsize = sizeStack(global); 		/* old size */
+      intptr_t olsize = sizeStack(local);
+      void *nw;
 
-      if ( g )
-      { gsize = nextStackSize((Stack) &LD->stacks.global, g);
-	shift_status.global_shifts++;
-      }
-      if ( l )
-      { lsize = nextStackSize((Stack) &LD->stacks.local, l);
-	shift_status.local_shifts++;
-      }
+      assert(lb == addPointer(gb, ogsize));	
 
-      gb = xrealloc(gb, lsize + gsize + GL_SEPARATION);
-      lb = addPointer(gb, gsize + GL_SEPARATION);
-      if ( g )				/* global enlarged; move local */
-	memmove(lb,   addPointer(gb, loffset), lsize);
-	     /* dest, src,                     size */
+      if ( gsize < ogsize )
+	memmove(addPointer(gb, gsize), lb, olsize);
+
+      if ( (nw = realloc(gb, lsize + gsize)) )
+      { if ( g )
+	  LD->shift_status.global_shifts++;
+	if ( l )
+	  LD->shift_status.local_shifts++;
+
+	gb = nw;
+	lb = addPointer(gb, gsize);
+	if ( gsize > ogsize )	
+	  memmove(lb, addPointer(gb, ogsize), olsize);
+      } else
+      { if ( g )
+	  fatal = &LD->stacks.global;
+	else
+	  fatal = &LD->stacks.local;
+
+	gsize = sizeStack(global);
+	lsize = sizeStack(local);
+      }
     }
       
-    if ( verbose )
-    { Sdprintf("to (l+g+t) = %d+%d+%d Kbytes ... ",
-	   lsize / 1024,
-	   gsize / 1024,
-	   tsize / 1024);
-    }
-
 #define PrintStackParms(stack, name, newbase, newsize) \
-	{ Sdprintf("%6s: 0x%08lx ... 0x%08lx --> 0x%08lx ... 0x%08lx\n", \
-		 name, \
-		 (uintptr_t) LD->stacks.stack.base, \
-		 (uintptr_t) LD->stacks.stack.max, \
-		 (uintptr_t) newbase, \
-		 (uintptr_t) addPointer(newbase, newsize)); \
+	{ Sdprintf("%6s: %p ... %p --> %p ... %p\n", \
+		   name, \
+		   LD->stacks.stack.base, \
+		   LD->stacks.stack.max, \
+		   newbase, \
+		   addPointer(newbase, newsize)); \
 	}
 
 
@@ -2645,19 +2767,32 @@ growStacks(LocalFrame fr, Choice ch, Code PC, intptr_t l, intptr_t g, intptr_t t
     SetHTop(LD->stacks.trail.max);
 
     time = CpuTime(CPU_USER) - time;
-    shift_status.time += time;
+    LD->shift_status.time += time;
     SECURE(if ( checkStacks(NULL, NULL) != key )
 	   { Sdprintf("Stack checksum failure\n");
 	     trap_gdb();
 	   });
     if ( verbose )
-    { Sdprintf("%.2f sec.\n", time);
+    { printMessage(ATOM_informational,
+		   PL_FUNCTOR_CHARS, "shift_stacks", 1,
+		     PL_FUNCTOR_CHARS, "done", 4,
+		       PL_DOUBLE, (double)time,
+		       PL_INTPTR, lsize,
+		       PL_INTPTR, gsize,
+		       PL_INTPTR, tsize);
     }
-
-    succeed;
   }
 
-  fail;
+  unblockGC(PASS_LD1);
+  unblockSignals(&mask);
+  leaveGC();
+
+  if ( fatal )
+  { DEBUG(1, Sdprintf("Out of %s stack due to failed rellocation\n", ((Stack)fatal)->name));
+    return outOfStack(fatal, STACK_OVERFLOW_THROW);
+  }
+
+  return TRUE;
 }
 
 #endif /*O_SHIFT_STACKS*/
@@ -2980,3 +3115,9 @@ markPredicatesInEnvironments(PL_local_data_t *ld)
 
 
 #endif /*O_CLAUSEGC*/
+
+BeginPredDefs(gc)
+#if O_SECURE || O_DEBUG || defined(O_MAINTENANCE)
+  PRED_DEF("$check_stacks", 1, check_stacks, 0)
+#endif
+EndPredDefs
