@@ -1,4 +1,4 @@
-/*  $Id: pl-gmp.c,v 1.15 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-gmp.c,v 1.22 2007/05/08 15:40:48 jan Exp $
 
     Part of SWI-Prolog
 
@@ -31,9 +31,12 @@
 
 static mpz_t MPZ_MIN_TAGGED;		/* Prolog tagged integers */
 static mpz_t MPZ_MAX_TAGGED;
-static mpz_t MPZ_MIN_INT;		/* Prolog int64_t integers */
-static mpz_t MPZ_MAX_INT;
-
+static mpz_t MPZ_MIN_PLINT;		/* Prolog int64_t integers */
+static mpz_t MPZ_MAX_PLINT;
+#if SIZEOF_LONG	< SIZEOF_VOIDP
+static mpz_t MPZ_MIN_LONG;		/* Prolog int64_t integers */
+static mpz_t MPZ_MAX_LONG;
+#endif
 
 #if 0
 		 /*******************************
@@ -122,7 +125,7 @@ get_integer(word w, Number n)
   } else
   { GET_LD
     Word p = addressIndirect(w);
-    int wsize = wsizeofInd(*p);
+    size_t wsize = wsizeofInd(*p);
 
     p++;
     if ( wsize == WORDS_PER_INT64 )
@@ -153,20 +156,20 @@ addMPZToBuffer(Buffer b, mpz_t mpz)
 void
 addMPZToBuffer(Buffer b, mpz_t mpz)
 { size_t size = (mpz_sizeinbase(mpz, 2)+7)/8;
-  int hdrsize;
+  long hdrsize;
   size_t count;
 
   growBuffer(b, size+4);
   if ( mpz_sgn(mpz) < 0 )
-    hdrsize = -size;
+    hdrsize = -(long)size;
   else
-    hdrsize = size;
+    hdrsize = (long)size;
   DEBUG(1, Sdprintf("addMPZToBuffer(): Added %d bytes\n", size));
 
-  *b->top++ = (hdrsize>>24)&0xff;
-  *b->top++ = (hdrsize>>16)&0xff;
-  *b->top++ = (hdrsize>> 8)&0xff;
-  *b->top++ = (hdrsize    )&0xff;
+  *b->top++ = (char)((hdrsize>>24)&0xff);
+  *b->top++ = (char)((hdrsize>>16)&0xff);
+  *b->top++ = (char)((hdrsize>> 8)&0xff);
+  *b->top++ = (char)((hdrsize    )&0xff);
 
   mpz_export(b->top, &count, 1, 1, 1, 0, mpz);
   assert(count == size);
@@ -264,7 +267,7 @@ mpz_init_set_si64(mpz_t mpz, int64_t i)
 { DEBUG(2, Sdprintf("Converting " INT64_FORMAT " to MPZ\n", i));
 
   if ( i >= LONG_MIN && i <= LONG_MAX )
-  { mpz_init_set_si(mpz, (intptr_t)i);
+  { mpz_init_set_si(mpz, (long)i);
   } else
   { mpz_init(mpz);
     if ( i >= 0 )
@@ -372,10 +375,14 @@ clearNumber(Number n)
 
 void
 initGMP()
-{ mpz_init_set_si(MPZ_MIN_TAGGED, PLMINTAGGEDINT);
-  mpz_init_set_si(MPZ_MAX_TAGGED, PLMAXTAGGEDINT);
-  mpz_init_set_si64(MPZ_MIN_INT, PLMININT);
-  mpz_init_set_si64(MPZ_MAX_INT, PLMAXINT);
+{ mpz_init_set_si64(MPZ_MIN_TAGGED, PLMINTAGGEDINT);
+  mpz_init_set_si64(MPZ_MAX_TAGGED, PLMAXTAGGEDINT);
+  mpz_init_set_si64(MPZ_MIN_PLINT, PLMININT);
+  mpz_init_set_si64(MPZ_MAX_PLINT, PLMAXINT);
+#if SIZEOF_LONG < SIZEOF_VOIDP
+  mpz_init_set_si64(MPZ_MIN_LONG, LONG_MIN);
+  mpz_init_set_si64(MPZ_MAX_LONG, LONG_MAX);
+#endif
 }
 
 		 /*******************************
@@ -384,11 +391,17 @@ initGMP()
 
 int
 mpz_to_int64(mpz_t mpz, int64_t *i)
-{ if ( mpz_cmp(mpz, MPZ_MIN_INT) >= 0 &&
-       mpz_cmp(mpz, MPZ_MAX_INT) <= 0 )
+{ if ( mpz_cmp(mpz, MPZ_MIN_PLINT) >= 0 &&
+       mpz_cmp(mpz, MPZ_MAX_PLINT) <= 0 )
   { int64_t v;
 
     mpz_export(&v, NULL, ORDER, sizeof(v), 0, 0, mpz);
+    DEBUG(2,
+	  { char buf[256];
+	    Sdprintf("Convert %s --> %I64d\n",
+		     mpz_get_str(buf, 10, mpz), v);
+	  });
+
     if ( mpz_sgn(mpz) < 0 )
       v = -v;
 
@@ -411,15 +424,30 @@ static word
 put_mpz(mpz_t mpz)
 { int64_t v;
 
+    DEBUG(2,
+	  { char buf[256];
+	    Sdprintf("put_mpz(%s)\n",
+		     mpz_get_str(buf, 10, mpz));
+	  });
+
+#if SIZEOF_LONG < SIZEOF_VOIDP
+  if ( mpz_cmp(mpz, MPZ_MIN_LONG) >= 0 &&
+       mpz_cmp(mpz, MPZ_MAX_LONG) <= 0 )
+#else
   if ( mpz_cmp(mpz, MPZ_MIN_TAGGED) >= 0 &&
        mpz_cmp(mpz, MPZ_MAX_TAGGED) <= 0 )
-  { intptr_t v = mpz_get_si(mpz);
+#endif
+  { long v = mpz_get_si(mpz);
 
     return consInt(v);
   } else if ( mpz_to_int64(mpz, &v) )
   { GET_LD
     
+#if SIZEOF_LONG < SIZEOF_VOIDP
+    return makeNum(v);
+#else
     return globalLong(v PASS_LD);
+#endif
   } else
   { return globalMPZ(mpz);
   }
@@ -476,6 +504,11 @@ PL_unify_number(term_t t, Number n)
   if ( canBind(*p) )
   { word w = put_number(n);
 
+#ifdef O_SHIFT_STACKS
+    p = valTermRef(t);			/* put_number can shift the stacks */
+    deRef(p);
+#endif
+
     bindConst(p, w);
     succeed;
   }
@@ -487,10 +520,13 @@ PL_unify_number(term_t t, Number n)
 #endif
       if ( isInteger(*p) )
       { number n2;
+	int rc;
 
 	get_integer(*p, &n2);
-	
-	return ar_compare(n, &n2, EQ);
+	rc = ar_compare(n, &n2, EQ);
+	clearNumber(&n2);
+
+	return rc;
       }
       break;
 #ifdef O_GMP
