@@ -1,11 +1,11 @@
-/*  $Id: pl-incl.h,v 1.267 2007/01/23 19:22:19 jan Exp $
+/*  $Id: pl-incl.h,v 1.279 2007/06/19 09:05:31 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -452,7 +452,7 @@ them.  Descriptions:
 #define SMALLSTACK		200 * 1024 /* GC policy */
 
 #define WORDBITSIZE		(8 * sizeof(word))
-#define LONGBITSIZE		(8 * sizeof(intptr_t))
+#define LONGBITSIZE		(8 * sizeof(long))
 #define INTBITSIZE		(8 * sizeof(int))
 #define INT64BITSIZE		(8 * sizeof(int64_t))
 #define WORDS_PER_DOUBLE        ((sizeof(double)+sizeof(word)-1)/sizeof(word))
@@ -874,7 +874,7 @@ typedef struct queryFrame *	QueryFrame;     /* toplevel query frame */
 typedef struct fliFrame *	FliFrame; 	/* FLI interface frame */
 typedef struct trail_entry *	TrailEntry;	/* Entry of trail stack */
 typedef struct gc_trail_entry *	GCTrailEntry;	/* Entry of trail stack (GC) */
-typedef struct data_mark	mark;		/* backtrack mark */
+typedef struct mark		mark;		/* backtrack mark */
 typedef struct index *		Index;		/* clause indexing */
 typedef struct stack *		Stack;		/* machine stack */
 typedef struct arithFunction * 	ArithFunction;  /* arithmetic function */
@@ -888,6 +888,10 @@ typedef struct on_halt *	OnHalt;		/* pl-os.c */
 typedef struct find_data_tag *	FindData; 	/* pl-trace.c */
 typedef struct feature *	Feature; 	/* pl-prims.c */
 
+typedef uintptr_t qid_t;		/* external query-id */
+typedef uintptr_t PL_fid_t;		/* external foreign context-id */
+
+#define fid_t PL_fid_t			/* avoid AIX name-clash */
 
 		 /*******************************
 		 *	    ARITHMETIC		*
@@ -917,6 +921,9 @@ typedef struct
 	} value;
 } number, *Number;
 
+#define TOINT_CONVERT_FLOAT	0x1	/* toIntegerNumber() */
+#define TOINT_TRUNCATE		0x2
+
 #ifdef O_GMP
 #define intNumber(n)	((n)->type <=  V_MPZ)
 #else
@@ -1124,7 +1131,7 @@ it mean anything?
 
 #define startCritical (void)(LD->critical++)
 #define endCritical   if ( --(LD->critical) == 0 && LD->aborted ) \
-			pl_abort(ABORT_NORMAL)
+			endCritical__LD(PASS_LD1)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 LIST processing macros.
@@ -1213,8 +1220,8 @@ extern Atom _PL_debug_atom_value(atom_t a);
 #endif
 
 struct index
-{ uintptr_t key;		/* key of index */
-  uintptr_t varmask;	/* variable field mask */
+{ word		key;		/* key of index */
+  word		varmask;	/* variable field mask */
 };
 
 struct functorDef
@@ -1299,11 +1306,17 @@ typedef struct
   char		argtype;	/* # `external' arguments code takes */
 } code_info;
 
-struct data_mark
+struct mark
 { TrailEntry	trailtop;	/* top of the trail stack */
   Word		globaltop;	/* top of the global stack */
 };
 
+typedef struct tmp_mark
+{ TrailEntry	trailtop;	/* top of the trail stack */
+  Word		globaltop;	/* top of the global stack */
+  Word		saved_mark;	/* saved LD->mark_bar */
+} tmp_mark;
+
 struct functor
 { word		definition;	/* Tagged definition pointer */
   word		arguments[1];	/* arguments vector */
@@ -1441,6 +1454,32 @@ struct choice
 };
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+EXCEPTION_GUARDED(code, cleanup) must be used  in environments that need
+cleanup  should  a  PL_throw()  happen.  The   most  commpn  reason  for
+PL_throw() instead of the nicely   synchronous PL_raise_exception() is a
+stack overflow.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#define EXCEPTION_GUARDED(code, cleanup) \
+	{ exception_frame __throw_env; \
+	  __throw_env.parent = LD->exception.throw_environment; \
+	  if ( setjmp(__throw_env.exception_jmp_env) != 0 ) \
+	  { LD->exception.throw_environment = __throw_env.parent; \
+	    cleanup; \
+	  } else \
+	  { LD->exception.throw_environment = &__throw_env; \
+	    code; \
+	    LD->exception.throw_environment = __throw_env.parent; \
+	  } \
+	}
+
+typedef struct exception_frame		/* PL_throw exception environments */
+{ struct exception_frame *parent;	/* parent frame */
+  jmp_buf	exception_jmp_env;	/* longjmp environment */
+} exception_frame;
+
+
 #define QF_NODEBUG		0x0001	/* debug-able query */
 #define QF_DETERMINISTIC	0x0002	/* deterministic success */
 #define	QF_INTERACTIVE		0x0004	/* interactive goal (prolog()) */
@@ -1459,8 +1498,10 @@ struct queryFrame
 #endif
 #if O_CATCHTHROW
   term_t	exception;		/* Exception term */
-  jmp_buf	exception_jmp_env;	/* longjmp() buffer for exception */
+  exception_frame exception_env;	/* Top PL_throw() environment */
+  exception_frame *saved_throw_env;	/* Throw environment of parent query */
 #endif
+  fid_t		foreign_frame;		/* Frame after PL_next_solution() */
   unsigned long	flags;
   debug_type	debugSave;		/* saved debugstatus.debugging */
   Word	       *aSave;			/* saved argument-stack */
@@ -1620,29 +1661,39 @@ struct alloc_pool
 		 *	     MARK/UNDO		*
 		 *******************************/
 
-#define setVar(w)		((w) = (word) 0)
+#define setVar(w)	((w) = (word) 0)
 
 #ifdef O_DESTRUCTIVE_ASSIGNMENT
 
-#define Undo(b)			do_undo(&b)
+#define Undo(b)		do_undo(&b)
+#define TmpUndo(b)	do_undo((mark*)&b)
 
 #else /*O_DESTRUCTIVE_ASSIGNMENT*/
 
-#define Undo(b)		{ TrailEntry tt = tTop; \
-			  TrailEntry mt = (b).trailtop; \
-			  while(tt > mt) \
-			  { tt--; \
-			    setVar(*tt->address); \
-			  } \
-			  tTop = tt; \
-			  gTop = (LD->frozen_bar > (b).globaltop ? \
-				  LD->frozen_bar : (b).globaltop); \
-			}
+#define Undo(b)		do { TrailEntry tt = tTop; \
+			     TrailEntry mt = (b).trailtop; \
+			     while(tt > mt) \
+			     { tt--; \
+			       setVar(*tt->address); \
+			     } \
+			     tTop = tt; \
+			     gTop = (LD->frozen_bar > (b).globaltop ? \
+			             LD->frozen_bar : (b).globaltop); \
+			    } while(0)
 #endif /*O_DESTRUCTIVE_ASSIGNMENT*/
 
-#define Mark(b)		{ (b).trailtop  = tTop; \
-			  LD->mark_bar = (b).globaltop = gTop; \
-			}
+#define Mark(b)		do { (b).trailtop  = tTop; \
+			     LD->mark_bar = (b).globaltop = gTop; \
+			   } while(0)
+
+#define TmpMark(b)	do { (b).trailtop  = tTop; \
+			     (b).saved_mark = LD->mark_bar; \
+			     LD->mark_bar = (b).globaltop = gTop; \
+			   } while(0)
+
+#define EndTmpMark(b)	do { LD->mark_bar = (b).saved_mark; \
+			   } while(0)
+
 
 		 /*******************************
 		 *	     TRAILING		*
@@ -1686,11 +1737,6 @@ value need not be trailed.
 		 *	   FLI INTERNALS	*
 		 *******************************/
 
-typedef uintptr_t qid_t;		/* external query-id */
-typedef uintptr_t PL_fid_t;		/* external foreign context-id */
-
-#define fid_t PL_fid_t			/* avoid AIX name-clash */
-
 #define consTermRef(p)	 ((Word)(p) - (Word)(lBase))
 #define valTermRef(r)	 (&((Word)(lBase))[r])
 
@@ -1847,17 +1893,14 @@ typedef struct
 #define lBase	(LD->stacks.local.base)
 #define lTop	(LD->stacks.local.top)
 #define lMax	(LD->stacks.local.max)
-#define lLimit	(LD->stacks.local.limit)
 
 #define gBase	(LD->stacks.global.base)
 #define gTop	(LD->stacks.global.top)
 #define gMax	(LD->stacks.global.max)
-#define gLimit	(LD->stacks.global.limit)
 
 #define aBase	(LD->stacks.argument.base)
 #define aTop	(LD->stacks.argument.top)
 #define aMax	(LD->stacks.argument.max)
-#define aLimit	(LD->stacks.argument.limit)
 
 #define SetHTop(val)	{ if ( (char *)(val) > hTop  ) hTop  = (char *)(val); }
 #define SetHBase(val)	{ if ( (char *)(val) < hBase ) hBase = (char *)(val); }
@@ -1874,17 +1917,19 @@ typedef struct
 	((char *)(addr) >= (char *)LD->stacks.name.base && \
 	 (char *)(addr) <  (char *)LD->stacks.name.limit)
 #endif
-#define usedStack(name) ((char *)LD->stacks.name.top - \
-			 (char *)LD->stacks.name.base)
-#define sizeStack(name) ((char *)LD->stacks.name.max - \
-			 (char *)LD->stacks.name.base)
-#define roomStack(name) ((char *)LD->stacks.name.max - \
-			 (char *)LD->stacks.name.top)
-#define spaceStack(name) ((char *)LD->stacks.name.limit - \
-			 (char *)LD->stacks.name.top)
-#define limitStack(name) ((char *)LD->stacks.name.limit - \
-			 (char *)LD->stacks.name.base)
-#define narrowStack(name) (roomStack(name) < LD->stacks.name.minfree)
+#define usedStackP(s) ((char *)(s)->top - (char *)(s)->base)
+#define sizeStackP(s) ((char *)(s)->max - (char *)(s)->base)
+#define roomStackP(s) ((char *)(s)->max - (char *)(s)->top)
+#define spaceStackP(s) ((char *)(s)->limit - (char *)(s)->top)
+#define limitStackP(s) ((char *)(s)->limit - (char *)(s)->base)
+#define narrowStackP(s) (roomStackP(s) < (s)->minfree)
+
+#define usedStack(name) usedStackP(&LD->stacks.name)
+#define sizeStack(name) sizeStackP(&LD->stacks.name)
+#define roomStack(name) roomStackP(&LD->stacks.name)
+#define spaceStack(name) spaceStackP(&LD->stacks.name)
+#define limitStack(name) limitStackP(&LD->stacks.name)
+#define narrowStack(name) narrowStackP(&LD->stacks.name)
 
 typedef enum
 { STACK_OVERFLOW_SIGNAL,
@@ -1894,14 +1939,14 @@ typedef enum
 } stack_overflow_action;
 
 #if O_DYNAMIC_STACKS
-#ifdef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
+#define requireStack(s, n)
+#else
 #define requireStack(s, n) \
 	{ if ( roomStack(s) < (intptr_t)(n) ) \
  	    ensureRoomStack((void*)&LD->stacks.s, n); \
 	}
-#else /*NO_SEGV_HANDLING*/
-#define requireStack(s, n)
-#endif /*NO_SEGV_HANDLING*/
+#endif /*O_SEGV_HANDLING*/
 #else
 #define requireStack(s, n) \
 	{ if ( roomStack(s) < (intptr_t)(n) ) \
@@ -2169,10 +2214,13 @@ decrease).
 #include SYSLIB_H
 #endif
 
+#define COMMON(type) SO_LOCAL type
+
 #include "pl-main.h"			/* Declarations needed by pl-main.c */
 #include "pl-error.h"			/* Exception generation */
 #include "pl-thread.h"			/* thread manipulation */
 #include "pl-data.h"			/* Access Prolog data */
+#include "pl-segstack.h"		/* Segmented stacks */
 #include "pl-global.h"			/* global data */
 #include "pl-funcs.h"			/* global functions */
 #include "pl-text.h"			/* text manipulation */
@@ -2185,7 +2233,7 @@ decrease).
 #endif
 
 #define NULL_ATOM ((atom_t)0)
-#define MK_ATOM(n)    		((n)<<7|TAG_ATOM|STG_STATIC)
+#define MK_ATOM(n)    		((atom_t)((n)<<7|TAG_ATOM|STG_STATIC))
 #include "pl-atom.ih"
 #include "pl-funct.ih"
 
