@@ -1,11 +1,11 @@
-/*  $Id: pl-itf.h,v 1.329 2006/08/31 09:44:01 jan Exp $
+/*  $Id: pl-itf.h,v 1.361 2007/07/09 15:27:09 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -27,9 +27,14 @@
 
 #include <stdarg.h>
 #include <stdlib.h>			/* get size_t */
-#ifdef WIN32
+#include <stddef.h>
+#ifdef __WINDOWS__
 typedef __int64 int64_t;
 typedef unsigned __int64 uint64_t;
+#if (_MSC_VER < 1300)
+typedef long intptr_t;
+typedef unsigned long uintptr_t;
+#endif
 #else
 #include <inttypes.h>			/* more portable than stdint.h */
 #endif
@@ -50,7 +55,7 @@ before loading this file.  See end of this file.
 /* PLVERSION: 10000 * <Major> + 100 * <Minor> + <Patch> */
 
 #ifndef PLVERSION
-#define PLVERSION 50618
+#define PLVERSION 50637
 #endif
 
 		 /*******************************
@@ -75,11 +80,8 @@ duplicated this stuff.
 
 #ifndef _PL_EXPORT_DONE
 #define _PL_EXPORT_DONE
-#if defined(WIN32) && !defined(__WIN32__)
-#define __WIN32__
-#endif
 
-#if (defined(__WIN32__) || defined(__CYGWIN__)) && !defined(__LCC__)
+#if (defined(_MSC_VER) || defined(__CYGWIN__)) && !defined(__LCC__)
 #define HAVE_DECLSPEC
 #endif
 
@@ -116,19 +118,19 @@ typedef Procedure	predicate_t;	/* a predicate handle */
 typedef Record		record_t;	/* handle to a recorded term */
 typedef struct PL_local_data *PL_engine_t; /* handle to a engine */
 #else
-typedef	unsigned long	atom_t;		/* Prolog atom */
+typedef	uintptr_t	atom_t;		/* Prolog atom */
 typedef void *		module_t;	/* Prolog module */
 typedef void *		predicate_t;	/* Prolog procedure */
 typedef void *		record_t;	/* Prolog recorded term */
-typedef unsigned long	term_t;		/* opaque term handle */
-typedef unsigned long	qid_t;		/* opaque query handle */
-typedef unsigned long	PL_fid_t;	/* opaque foreign context handle */
+typedef uintptr_t	term_t;		/* opaque term handle */
+typedef uintptr_t	qid_t;		/* opaque query handle */
+typedef uintptr_t	PL_fid_t;	/* opaque foreign context handle */
 typedef void *		control_t;	/* non-deterministic control arg */
 typedef void *		PL_engine_t;	/* opaque engine handle */
 #endif
-typedef unsigned long	functor_t;	/* Name/arity pair */
-typedef unsigned long	PL_atomic_t;	/* same a word */
-typedef unsigned long	foreign_t;	/* return type of foreign functions */
+typedef uintptr_t	functor_t;	/* Name/arity pair */
+typedef uintptr_t	PL_atomic_t;	/* same a word */
+typedef uintptr_t	foreign_t;	/* return type of foreign functions */
 typedef wchar_t	        pl_wchar_t;	/* Prolog wide character */
 #ifdef __cplusplus
 typedef void * 		pl_function_t;	/* can only pass function as void * */
@@ -185,19 +187,20 @@ typedef union
 #define PL_INT		 (20)		/* int */
 #define PL_LONG		 (21)		/* long */
 #define PL_DOUBLE	 (22)		/* double */
-#define PL_NCHARS	 (23)		/* unsigned, const char * */
+#define PL_NCHARS	 (23)		/* size_t, const char * */
 #define PL_UTF8_CHARS	 (24)		/* const char * */
 #define PL_UTF8_STRING	 (25)		/* const char * */
 #define PL_INT64	 (26)		/* int64_t */
-#define PL_NUTF8_CHARS	 (27)		/* unsigned, const char * */
-#define PL_NUTF8_CODES	 (29)		/* unsigned, const char * */
-#define PL_NUTF8_STRING	 (30)		/* unsigned, const char * */
-#define PL_NWCHARS	 (31)		/* unsigned, const wchar_t * */
-#define PL_NWCODES	 (32)		/* unsigned, const wchar_t * */
-#define PL_NWSTRING	 (33)		/* unsigned, const wchar_t * */
+#define PL_NUTF8_CHARS	 (27)		/* size_t, const char * */
+#define PL_NUTF8_CODES	 (29)		/* size_t, const char * */
+#define PL_NUTF8_STRING	 (30)		/* size_t, const char * */
+#define PL_NWCHARS	 (31)		/* size_t, const wchar_t * */
+#define PL_NWCODES	 (32)		/* size_t, const wchar_t * */
+#define PL_NWSTRING	 (33)		/* size_t, const wchar_t * */
 #define PL_MBCHARS	 (34)		/* const char * */
 #define PL_MBCODES	 (35)		/* const char * */
 #define PL_MBSTRING	 (36)		/* const char * */
+#define PL_INTPTR	 (37)		/* intptr_t */
 
 
 		/********************************
@@ -215,7 +218,7 @@ typedef union
 /*  Note 1: Non-deterministic foreign functions may also use the deterministic
     return methods PL_succeed and PL_fail.
 
-    Note 2: The argument to PL_retry is a 30 bits signed integer (long).
+    Note 2: The argument to PL_retry is a 30 bits signed integer (intptr_t).
 */
 
 #define PL_FIRST_CALL		(0)
@@ -225,10 +228,10 @@ typedef union
 #define PL_retry(n)		return _PL_retry(n)
 #define PL_retry_address(a)	return _PL_retry_address(a)
 
-PL_EXPORT(foreign_t)	_PL_retry(long);
+PL_EXPORT(foreign_t)	_PL_retry(intptr_t);
 PL_EXPORT(foreign_t)	_PL_retry_address(void *);
 PL_EXPORT(int)	 	PL_foreign_control(control_t);
-PL_EXPORT(long)	 	PL_foreign_context(control_t);
+PL_EXPORT(intptr_t)	 	PL_foreign_context(control_t);
 PL_EXPORT(void *)	PL_foreign_context_address(control_t);
 
 
@@ -334,11 +337,11 @@ PL_EXPORT(void)		PL_reset_term_refs(term_t r);
 
 			/* Constants */
 PL_EXPORT(atom_t)	PL_new_atom(const char *s);
-PL_EXPORT(atom_t)	PL_new_atom_nchars(unsigned int len, const char *s);
-PL_EXPORT(atom_t)	PL_new_atom_wchars(unsigned int len, const wchar_t *s);
+PL_EXPORT(atom_t)	PL_new_atom_nchars(size_t len, const char *s);
+PL_EXPORT(atom_t)	PL_new_atom_wchars(size_t len, const wchar_t *s);
 PL_EXPORT(const char *)	PL_atom_chars(atom_t a);
-PL_EXPORT(const char *) PL_atom_nchars(atom_t a, unsigned int *len);
-PL_EXPORT(const wchar_t *) PL_atom_wchars(atom_t a, unsigned int *len);
+PL_EXPORT(const char *) PL_atom_nchars(atom_t a, size_t *len);
+PL_EXPORT(const wchar_t *) PL_atom_wchars(atom_t a, size_t *len);
 #ifndef O_DEBUG_ATOMGC
 PL_EXPORT(void)		PL_register_atom(atom_t a);
 PL_EXPORT(void)		PL_unregister_atom(atom_t a);
@@ -352,18 +355,17 @@ PL_EXPORT(int)		PL_get_atom(term_t t, atom_t *a);
 PL_EXPORT(int)		PL_get_bool(term_t t, int *value);
 PL_EXPORT(int)		PL_get_atom_chars(term_t t, char **a);
 #define PL_get_string_chars(t, s, l) PL_get_string(t,s,l)
-					/* PL_get_string() is depreciated */
-PL_EXPORT(int)		PL_get_string(term_t t, char **s, unsigned int *len);
+					/* PL_get_string() is depricated */
+PL_EXPORT(int)		PL_get_string(term_t t, char **s, size_t *len);
 PL_EXPORT(int)		PL_get_chars(term_t t, char **s, unsigned int flags);
 PL_EXPORT(int)		PL_get_list_chars(term_t l, char **s,
 					  unsigned int flags);
-PL_EXPORT(int)		PL_get_atom_nchars(term_t t,
-					   unsigned int *length, char **a);
+PL_EXPORT(int)		PL_get_atom_nchars(term_t t, size_t *len, char **a);
 PL_EXPORT(int)		PL_get_list_nchars(term_t l,
-					   unsigned int *length, char **s,
+					   size_t *len, char **s,
 					   unsigned int flags);
 PL_EXPORT(int)		PL_get_nchars(term_t t,
-				      unsigned int *length, char **s,
+				      size_t *len, char **s,
 				      unsigned int flags);
 PL_EXPORT(int)		PL_get_integer(term_t t, int *i);
 PL_EXPORT(int)		PL_get_long(term_t t, long *i);
@@ -390,6 +392,7 @@ PL_EXPORT(int)		PL_is_string(term_t t);
 PL_EXPORT(int)		PL_is_float(term_t t);
 PL_EXPORT(int)		PL_is_rational(term_t t);
 PL_EXPORT(int)		PL_is_compound(term_t t);
+PL_EXPORT(int)		PL_is_callable(term_t t);
 PL_EXPORT(int)		PL_is_functor(term_t t, functor_t f);
 PL_EXPORT(int)		PL_is_list(term_t t);
 PL_EXPORT(int)		PL_is_atomic(term_t t);
@@ -402,15 +405,10 @@ PL_EXPORT(void)		PL_put_atom_chars(term_t t, const char *chars);
 PL_EXPORT(void)		PL_put_string_chars(term_t t, const char *chars);
 PL_EXPORT(void)		PL_put_list_chars(term_t t, const char *chars);
 PL_EXPORT(void)		PL_put_list_codes(term_t t, const char *chars);
-PL_EXPORT(void)		PL_put_atom_nchars(term_t t,
-					   unsigned int l, const char *chars);
-PL_EXPORT(void)		PL_put_string_nchars(term_t t,
-					     unsigned int len,
-					     const char *chars);
-PL_EXPORT(void)		PL_put_list_nchars(term_t t,
-					   unsigned int l, const char *chars);
-PL_EXPORT(void)		PL_put_list_ncodes(term_t t,
-					   unsigned int l, const char *chars);
+PL_EXPORT(void)		PL_put_atom_nchars(term_t t, size_t l, const char *chars);
+PL_EXPORT(void)		PL_put_string_nchars(term_t t, size_t len, const char *chars);
+PL_EXPORT(void)		PL_put_list_nchars(term_t t, size_t l, const char *chars);
+PL_EXPORT(void)		PL_put_list_ncodes(term_t t, size_t l, const char *chars);
 PL_EXPORT(void)		PL_put_integer(term_t t, long i);
 PL_EXPORT(void)		PL_put_pointer(term_t t, void *ptr);
 PL_EXPORT(void)		PL_put_float(term_t t, double f);
@@ -431,16 +429,13 @@ PL_EXPORT(int)		PL_unify_atom_chars(term_t t, const char *chars);
 PL_EXPORT(int)		PL_unify_list_chars(term_t t, const char *chars);
 PL_EXPORT(int)		PL_unify_list_codes(term_t t, const char *chars);
 PL_EXPORT(int)		PL_unify_string_chars(term_t t, const char *chars);
-PL_EXPORT(int)		PL_unify_atom_nchars(term_t t,
-					     unsigned int l, const char *s);
-PL_EXPORT(int)		PL_unify_list_ncodes(term_t t,
-					     unsigned int l, const char *s);
-PL_EXPORT(int)		PL_unify_list_nchars(term_t t,
-					     unsigned int l, const char *s);
+PL_EXPORT(int)		PL_unify_atom_nchars(term_t t, size_t l, const char *s);
+PL_EXPORT(int)		PL_unify_list_ncodes(term_t t, size_t l, const char *s);
+PL_EXPORT(int)		PL_unify_list_nchars(term_t t, size_t l, const char *s);
 PL_EXPORT(int)		PL_unify_string_nchars(term_t t,
-					       unsigned int len,
+					       size_t len,
 					       const char *chars);
-PL_EXPORT(int)		PL_unify_integer(term_t t, long n);
+PL_EXPORT(int)		PL_unify_integer(term_t t, intptr_t n);
 PL_EXPORT(int)		PL_unify_float(term_t t, double f);
 PL_EXPORT(int)		PL_unify_pointer(term_t t, void *ptr);
 PL_EXPORT(int)		PL_unify_functor(term_t t, functor_t f);
@@ -449,7 +444,7 @@ PL_EXPORT(int)		PL_unify_nil(term_t l);
 PL_EXPORT(int)		PL_unify_arg(int index, term_t t, term_t a);
 PL_EXPORT(int)		PL_unify_term(term_t t, ...);
 PL_EXPORT(int)		PL_unify_chars(term_t t, int flags,
-				       unsigned int len, const char *s);
+				       size_t len, const char *s);
 
 
 		 /*******************************
@@ -457,13 +452,13 @@ PL_EXPORT(int)		PL_unify_chars(term_t t, int flags,
 		 *******************************/
 
 PL_EXPORT(int)		PL_unify_wchars(term_t t, int type,
-					unsigned int len, const pl_wchar_t *s);
+					size_t len, const pl_wchar_t *s);
 PL_EXPORT(int)		PL_unify_wchars_diff(term_t t, term_t tail, int type,
-					unsigned int len, const pl_wchar_t *s);
+					size_t len, const pl_wchar_t *s);
 PL_EXPORT(int)		PL_get_wchars(term_t l,
-				      unsigned int *length, pl_wchar_t **s,
+				      size_t *length, pl_wchar_t **s,
 				      unsigned flags);
-PL_EXPORT(unsigned int)	PL_utf8_strlen(const char *s, unsigned int len);
+PL_EXPORT(size_t)	PL_utf8_strlen(const char *s, size_t len);
 
 
 		 /*******************************
@@ -498,8 +493,8 @@ PL_EXPORT(int)		PL_get_attr(term_t v, term_t a);
 #define PL_BLOB_WCHAR	0x08		/* wide character string */
 
 typedef struct PL_blob_t
-{ unsigned long		magic;		/* PL_BLOB_MAGIC */
-  unsigned long		flags;		/* PL_BLOB_* */
+{ uintptr_t		magic;		/* PL_BLOB_MAGIC */
+  uintptr_t		flags;		/* PL_BLOB_* */
   char *		name;		/* name of the type */
   int			(*release)(atom_t a);
   int			(*compare)(atom_t a, atom_t b);
@@ -525,15 +520,15 @@ typedef struct PL_blob_t
 } PL_blob_t;
 
 PL_EXPORT(int)		PL_is_blob(term_t t, PL_blob_t **type);
-PL_EXPORT(int)		PL_unify_blob(term_t t, void *blob, unsigned int len,
+PL_EXPORT(int)		PL_unify_blob(term_t t, void *blob, size_t len,
 				      PL_blob_t *type);
-PL_EXPORT(int)		PL_put_blob(term_t t, void *blob, unsigned int len,
+PL_EXPORT(int)		PL_put_blob(term_t t, void *blob, size_t len,
 				    PL_blob_t *type);
-PL_EXPORT(int)		PL_get_blob(term_t t, void **blob, unsigned int *len,
+PL_EXPORT(int)		PL_get_blob(term_t t, void **blob, size_t *len,
 				    PL_blob_t **type);
 
 PL_EXPORT(void*)	PL_blob_data(atom_t a,
-				     unsigned int *len,
+				     size_t *len,
 				     struct PL_blob_t **type);
 
 PL_EXPORT(void)		PL_register_blob_type(PL_blob_t *type);
@@ -581,7 +576,7 @@ PL_EXPORT(int)		PL_cvt_i_float(term_t p, double *c);
 PL_EXPORT(int)		PL_cvt_i_single(term_t p, float *c);
 PL_EXPORT(int)		PL_cvt_i_string(term_t p, char **c);
 PL_EXPORT(int)		PL_cvt_i_atom(term_t p, atom_t *c);
-PL_EXPORT(int)		PL_cvt_o_integer(long c, term_t p);
+PL_EXPORT(int)		PL_cvt_o_integer(intptr_t c, term_t p);
 PL_EXPORT(int)		PL_cvt_o_float(double c, term_t p);
 PL_EXPORT(int)		PL_cvt_o_single(float c, term_t p);
 PL_EXPORT(int)		PL_cvt_o_string(const char *c, term_t p);
@@ -610,7 +605,7 @@ PL_EXPORT(record_t)	PL_record(term_t term);
 PL_EXPORT(void)		PL_recorded(record_t record, term_t term);
 PL_EXPORT(void)		PL_erase(record_t record);
 
-PL_EXPORT(char *)	PL_record_external(term_t t, unsigned int *size);
+PL_EXPORT(char *)	PL_record_external(term_t t, size_t *size);
 PL_EXPORT(int)		PL_recorded_external(const char *rec, term_t term);
 PL_EXPORT(int)		PL_erase_external(char *rec);
 
@@ -663,7 +658,7 @@ UNICODE file functions.
 #define REP_ISO_LATIN_1 0x0000		/* output representation */
 #define REP_UTF8 	0x1000
 #define REP_MB		0x2000
-#ifdef WIN32
+#ifdef __WINDOWS__
 #define REP_FN		REP_UTF8
 #else
 #define REP_FN		REP_MB
@@ -678,7 +673,7 @@ UNICODE file functions.
 PL_EXPORT(int)  	PL_open_stream(term_t t, IOSTREAM *s); /* compat */
 PL_EXPORT(int)  	PL_unify_stream(term_t t, IOSTREAM *s);
 PL_EXPORT(int)  	PL_get_stream_handle(term_t t, IOSTREAM **s);
-PL_EXPORT(void) 	PL_release_stream(IOSTREAM *s);
+PL_EXPORT(int) 		PL_release_stream(IOSTREAM *s);
 PL_EXPORT(IOSTREAM *)	PL_open_resource(module_t m,
 					 const char *name,
 					 const char *rc_class,
@@ -758,16 +753,16 @@ readline overhead.
 #define PL_DISPATCH_INSTALLED 2		/* dispatch function installed? */
 
 PL_EXPORT(int)		PL_dispatch(int fd, int wait);
-PL_EXPORT(void)		PL_add_to_protocol(const char *buf, int count);
+PL_EXPORT(void)		PL_add_to_protocol(const char *buf, size_t count);
 PL_EXPORT(char *)	PL_prompt_string(int fd);
 PL_EXPORT(void)		PL_write_prompt(int dowrite);
 PL_EXPORT(void)		PL_prompt_next(int fd);
 PL_EXPORT(char *)	PL_atom_generator(const char *prefix, int state);
 PL_EXPORT(pl_wchar_t*)  PL_atom_generator_w(const pl_wchar_t *pref,
 					    pl_wchar_t *buffer,
-					    unsigned int buflen,
+					    size_t buflen,
 					    int state);
-PL_EXPORT(void)		PL_clock_wait_ticks(long waited);
+PL_EXPORT(void)		PL_clock_wait_ticks(intptr_t waited);
 
 
 		 /*******************************
@@ -847,7 +842,7 @@ PL_EXPORT(void) PL_on_halt(void (*)(int, void *), void *);
 #define PL_QUERY_ENCODING	12	/* I/O encoding */
 #define PL_QUERY_USER_CPU	13	/* User CPU in milliseconds */
 
-PL_EXPORT(long)	PL_query(int);	/* get information from Prolog */
+PL_EXPORT(intptr_t)	PL_query(int);	/* get information from Prolog */
 
 
 		 /*******************************
@@ -855,13 +850,13 @@ PL_EXPORT(long)	PL_query(int);	/* get information from Prolog */
 		 *******************************/
 
 typedef struct
-{ unsigned long	    local_size;		/* Stack sizes */
-  unsigned long	    global_size;
-  unsigned long	    trail_size;
-  unsigned long	    argument_size;
-  char *	    alias;		/* alias name */
-  int		  (*cancel)(int id);	/* cancel function */
-  void *	    reserved[5];	/* reserved for extensions */
+{ long	    local_size;			/* Stack sizes (Kbytes) */
+  long	    global_size;
+  long	    trail_size;
+  long	    argument_size;
+  char *    alias;			/* alias name */
+  int	  (*cancel)(int id);		/* cancel function */
+  void *    reserved[5];		/* reserved for extensions */
 } PL_thread_attr_t;
 
 
@@ -899,7 +894,7 @@ typedef struct
 { int	(*unify)(term_t t, void *handle); 	/* implementation --> Prolog */
   int   (*get)(term_t t, void **handle);	/* Prolog --> implementation */
   void	(*activate)(int active);		/* (de)activate */
-  long	magic;					/* PROFTYPE_MAGIC */
+  intptr_t	magic;					/* PROFTYPE_MAGIC */
 } PL_prof_type_t;
 
 PL_EXPORT(int)		PL_register_profile_type(PL_prof_type_t *type);
@@ -916,10 +911,10 @@ PL_EXPORT(void)		PL_prof_exit(void *node);
 #define PL_MSG_IGNORED 0
 #define PL_MSG_HANDLED 1
 
-PL_EXPORT(int)		PL_win_message_proc(HWND hwnd,
+PL_EXPORT(LRESULT)	PL_win_message_proc(HWND hwnd,
 					    UINT message,
-					    UINT wParam,
-					    LONG lParam);
+					    WPARAM wParam,
+					    LPARAM lParam);
 #endif /*_WINDOWS_*/
 
 
@@ -930,14 +925,14 @@ PL_EXPORT(int)		PL_win_message_proc(HWND hwnd,
 typedef struct
 { int type;				/* PL_INTEGER or PL_ATOM */
   union
-  { unsigned long i;			/* integer reference value */
+  { uintptr_t i;			/* integer reference value */
     atom_t	  a;			/* atom reference value */
   } value;
 } xpceref_t;
 
 PL_EXPORT(int)	_PL_get_xpce_reference(term_t t, xpceref_t *ref);
 PL_EXPORT(int)  _PL_unify_xpce_reference(term_t t, xpceref_t *ref);
-PL_EXPORT(void) _PL_put_xpce_reference_i(term_t t, unsigned long r);
+PL_EXPORT(void) _PL_put_xpce_reference_i(term_t t, uintptr_t r);
 PL_EXPORT(void) _PL_put_xpce_reference_a(term_t t, atom_t name);
 
 		 /*******************************
@@ -998,7 +993,7 @@ PL_EXPORT(term_t)	_PL_arg(term_t t, int n);
 		 *******************************/
 
 PL_EXPORT(term_t)	_PL_new_term(void);
-PL_EXPORT(PL_atomic_t)	_PL_new_integer(long i);
+PL_EXPORT(PL_atomic_t)	_PL_new_integer(intptr_t i);
 PL_EXPORT(PL_atomic_t)	_PL_new_float(double f);
 PL_EXPORT(PL_atomic_t)	_PL_new_string(const char *s);
 PL_EXPORT(PL_atomic_t)	_PL_new_var(void);
