@@ -1,4 +1,4 @@
-/*  $Id: pl-main.c,v 1.170 2007/01/22 11:52:23 jan Exp $
+/*  $Id: pl-main.c,v 1.172 2007/05/11 09:20:43 jan Exp $
 
     Part of SWI-Prolog
 
@@ -47,6 +47,10 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #include <crtdbg.h>
 #endif
 
+#if defined(HAVE_MTRACE) && defined(O_MAINTENANCE)
+#include <mcheck.h>
+#endif
+
 static int	usage(void);
 static int	giveVersionInfo(const char *a);
 static bool	vsysError(const char *fm, va_list args);
@@ -470,9 +474,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 
 	  switch(*s)
 	  { case 'L':	GD->options.localSize    = size; goto next;
-	    case 'G':	GD->options.globalSize   = size;
-	    		GD->options.argumentSize = size/4;
-							 goto next;
+	    case 'G':	GD->options.globalSize   = size; goto next;
 	    case 'T':	GD->options.trailSize    = size; goto next;
 	    case 'A':	GD->options.argumentSize = size; goto next;
 	    case 'H':	GD->options.heapSize     = size; goto next;
