@@ -1,4 +1,4 @@
-/*  $Id: pl-main.h,v 1.29 2007/01/09 14:09:00 jan Exp $
+/*  $Id: pl-main.h,v 1.31 2007/05/10 09:41:28 jan Exp $
 
     Part of SWI-Prolog
 
@@ -79,19 +79,11 @@ typedef struct
 #define OS "unknown"
 #endif
 
-#if O_DYNAMIC_STACKS || O_SHIFT_STACKS
-#define DEF_DEFLOCAL	(500*SIZEOF_VOIDP)
-#define DEF_DEFGLOBAL	(1000*SIZEOF_VOIDP)
-#define DEF_DEFTRAIL	(1000*SIZEOF_VOIDP)
+#define DEF_DEFLOCAL	(4096*SIZEOF_VOIDP)
+#define DEF_DEFGLOBAL	(8192*SIZEOF_VOIDP) /* 32MB on 32-bit hardware */
+#define DEF_DEFTRAIL	(8192*SIZEOF_VOIDP)
 #define DEF_DEFARGUMENT (250*SIZEOF_VOIDP)
-#define DEF_DEFHEAP        0		/* unlimited */
-#else					/* static stack areas */
-#define DEF_DEFLOCAL	(50*SIZEOF_VOIDP)
-#define DEF_DEFGLOBAL	(80*SIZEOF_VOIDP)
-#define DEF_DEFTRAIL	(80*SIZEOF_VOIDP)
-#define DEF_DEFARGUMENT	(20*SIZEOF_VOIDP)
-#define DEF_DEFHEAP       0		/* unlimited */
-#endif
+#define DEF_DEFHEAP     0		/* unlimited */
 
 #ifndef DEFLOCAL
 #define DEFLOCAL    DEF_DEFLOCAL
