@@ -1,4 +1,4 @@
-/*  $Id: pl-nt.c,v 1.67 2007/01/13 12:49:07 jan Exp $
+/*  $Id: pl-nt.c,v 1.69 2007/06/26 08:55:07 jan Exp $
 
     Part of SWI-Prolog
 
@@ -227,6 +227,22 @@ CpuTime(cputime_kind which)
 }
 
 
+static int
+CpuCount()
+{ SYSTEM_INFO si;
+
+  GetSystemInfo(&si);
+
+  return si.dwNumberOfProcessors;
+}
+
+
+void
+setOSFeatures()
+{ defFeature("cpu_count", FT_INTEGER, CpuCount());
+}
+
+
 char *
 findExecutable(const char *module, char *exe)
 { int n;
@@ -384,7 +400,7 @@ System(char *command)			/* command is a UTF-8 string */
 		      NULL,			/* Security stuff */
 		      NULL,			/* Thread security stuff */
 		      FALSE,			/* Inherit handles */
-		      NORMAL_PRIORITY_CLASS,	/* flags */
+		      CREATE_NO_WINDOW,		/* flags */
 		      NULL,			/* environment */
 		      NULL,			/* CWD */
 		      &sinfo,			/* startup info */
