@@ -1,4 +1,4 @@
-/*  $Id: pl-ntmain.c,v 1.50 2007/01/21 17:01:19 jan Exp $
+/*  $Id: pl-ntmain.c,v 1.51 2007/02/27 10:21:53 jan Exp $
 
     Part of SWI-Prolog
 
@@ -910,7 +910,7 @@ set_window_title(rlc_console c)
 #ifdef WIN64
   TCHAR *w64 = _T("AMD64, ");			/* TBD: IA64 */
 #else
-  TCHAR *w64 = _T("xx");
+  TCHAR *w64 = _T("");
 #endif
 
   _stprintf(title, _T("SWI-Prolog (%s%sversion %d.%d.%d)"),
