@@ -1,11 +1,11 @@
-/*  $Id: pl-os.c,v 1.199 2007/01/13 12:49:07 jan Exp $
+/*  $Id: pl-os.c,v 1.204 2007/06/26 09:41:58 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -327,11 +327,69 @@ WallTime(void)
   return stime;
 }
 
+		 /*******************************
+		 *	      FEATURES		*
+		 *******************************/
+
+#ifndef __WINDOWS__			/* Windows version in pl-nt.c */
+
+#ifdef PROCFS_CPUINFO
+static int
+CpuCount()
+{ FILE *fd = fopen("/proc/cpuinfo", "r");
+
+  if ( fd )
+  { char buf[256];
+    int count = 0;
+    
+    while(fgets(buf, sizeof(buf)-1, fd))
+    { int cpu;
+      char *vp;
+
+      if ( (vp = strchr(buf, ':')) )
+      { char *en;
+
+	for(en=vp; en > buf && en[-1] <= ' '; en--)
+	  ;
+	*en = EOS;
+	DEBUG(2, Sdprintf("Got %s = %s\n", buf, vp+2));
+	if ( streq("processor", buf) && isDigit(vp[2]) )
+	{ int cpu = atoi(vp+2);
+
+	  if ( cpu+1 > count )
+	    count = cpu+1;
+	}
+      }
+    }
+
+    fclose(fd);
+    return count;
+  }
+
+  return 0;
+}
+
+#else /*PROCFS_CPUINFO*/
+
+#define CpuCount() 0
+
+#endif /*PROCFS_CPUINFO*/
+
+
+void
+setOSFeatures(void)
+{ int cpu_count = CpuCount();
+
+  if ( cpu_count > 0 )
+    defFeature("cpu_count", FT_INTEGER, cpu_count);
+}
+#endif
+
 		 /*******************************
 		 *	       MEMORY		*
 		 *******************************/
 
-ulong
+uintptr_t
 UsedMemory(void)
 {
 #if defined(HAVE_GETRUSAGE) && defined(HAVE_RU_IDRSS)
@@ -350,9 +408,9 @@ UsedMemory(void)
 }
 
 
-ulong
+uintptr_t
 FreeMemory(void)
-{ ulong used = UsedMemory();
+{ uintptr_t used = UsedMemory();
 
 #if defined(HAVE_GETRLIMIT) && defined(RLIMIT_DATA)
   struct rlimit limit;
@@ -792,8 +850,7 @@ ExistsFile(const char *path)
 #if defined(HAVE_STAT) || defined(__unix__)
   struct stat buf;
 
-  if ( statfunc(OsPath(path, tmp), &buf) == -1 ||
-       (buf.st_mode & S_IFMT) != S_IFREG )
+  if ( statfunc(OsPath(path, tmp), &buf) == -1 || !S_ISREG(buf.st_mode) )
   { DEBUG(2, perror(tmp));
     fail;
   }
@@ -861,7 +918,7 @@ ExistsDirectory(const char *path)
   if ( statfunc(ospath, &buf) < 0 )
     fail;
 
-  if ( (buf.st_mode & S_IFMT) == S_IFDIR )
+  if ( S_ISDIR(buf.st_mode) )
     succeed;
 
   fail;
@@ -1735,10 +1792,14 @@ is slow and appears not to cooperate with profile/3.  getwd() is supposed
 to be implemented directly.  What about other Unixes?
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if defined(HAVE_GETWD) && (defined(__sun__) || !defined(HAVE_GETCWD))
+#if defined(HAVE_GETWD) && (defined(__sun__) || defined(__sun))
+#undef HAVE_GETCWD
+#endif
+
+#if defined(HAVE_GETWD) && !defined(HAVE_GETCWD)
     rval = getwd(buf);
 #else
-    rval = getcwd(buf, MAXPATHLEN);
+    rval = getcwd(buf, sizeof(buf));
 #endif
     if ( !rval )
     { term_t tmp = PL_new_term_ref();
@@ -2784,9 +2845,9 @@ static char *
 okToExec(const char *s)
 { struct stat stbuff;
 
-  if (statfunc(s, &stbuff) == 0 &&			/* stat it */
-     (stbuff.st_mode & S_IFMT) == S_IFREG &&	/* check for file */
-     access(s, X_OK) == 0)			/* can be executed? */
+  if (statfunc(s, &stbuff) == 0 &&	/* stat it */
+     S_ISREG(stbuff.st_mode) &&		/* check for file */
+     access(s, X_OK) == 0)		/* can be executed? */
     return (char *)s;
   else
     return (char *) NULL;
