@@ -1,11 +1,11 @@
-/*  $Id: pl-prims.c,v 1.191 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-prims.c,v 1.208 2007/07/05 15:49:35 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -25,13 +25,347 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-ctype.h"
-#undef ulong
-#define ulong uintptr_t
 
 #undef LD
 #define LD LOCAL_LD
 
-static void	duplicate_term(term_t in, term_t copy ARG_LD);
+static int	duplicate_term(term_t in, term_t copy ARG_LD);
+
+
+		 /*******************************
+		 *	   CYCLIC TERMS		*
+		 *******************************/
+
+#if O_CYCLIC
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Cyclic term unification. The algorithm has been  described to me by Bart
+Demoen. Here it is (translated from dutch):
+
+I created my own variation. You only need it during general unification.
+Here is a short description:  suppose  you   unify  2  terms  f(...) and
+f(...), which are represented on the heap (=global stack) as:
+
+     +-----+          and     +-----+
+     | f/3 |                  | f/3 |
+     +-----+                  +-----+
+      args                     args'
+
+Before working on args and args', change  this into the structure below,
+using a reference pointer pointing from functor  of the one to the other
+term.
+
+     +-----+          and      +-----+
+     | ----+----------------->| f/3 |
+     +-----+                  +-----+
+      args                     args'
+
+If, during this unification you  find  a   compound  whose  functor is a
+reference to the term at the right hand you know you hit a cycle and the
+terms are the same.
+
+Of course functor_t must be different from ref. Overwritten functors are
+collected in a stack and  reset   regardless  of whether the unification
+succeeded or failed. In SWI-Prolog we use   the  argument stack for this
+purpose.
+
+Initial measurements show a performance degradation for deep unification
+of approx. 30%. On the other hand,  if subterms appear multiple times in
+a term unification can be much faster. As only a small percentage of the
+unifications of a realistic program are   covered by unify() and involve
+deep unification the overall impact of performance is small (< 3%).
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+initvisited(ARG1_LD)
+{ LD->cycle.stack.unit_size = sizeof(Word);
+}
+
+
+static int
+empty_visited(ARG1_LD)
+{ return LD->cycle.stack.count == 0;
+}
+
+
+static inline int
+visitedWord(Word p ARG_LD)
+{ if ( is_marked(p) )
+    succeed;
+  set_marked(p);
+  pushSegStack(&LD->cycle.stack, &p);
+  fail;
+}
+
+
+static inline int
+visited(Functor f ARG_LD)
+{ Word p = &f->definition;
+
+  return visitedWord(p PASS_LD);
+}
+
+
+static void
+unvisit(ARG1_LD)
+{ Word p;
+
+  while( popSegStack(&LD->cycle.stack, &p) )
+  { clear_marked(p);
+  }
+}
+
+
+static void
+popVisited(ARG1_LD)
+{ Word p;
+
+  popSegStack(&LD->cycle.stack, &p);
+  clear_marked(p);
+}
+
+
+static inline void
+initCyclic(ARG1_LD)
+{ LD->cycle.stack.unit_size = sizeof(Word);
+}
+
+
+static inline void
+linkTermsCyclic(Functor f1, Functor f2 ARG_LD)
+{ Word p1 = (Word)&f1->definition;
+  Word p2 = (Word)&f2->definition;
+
+  *p1 = makeRefG(p2);
+  pushSegStack(&LD->cycle.stack, &p1);
+}
+
+
+static inline void
+exitCyclic(ARG1_LD)
+{ Word p;
+
+  while( popSegStack(&LD->cycle.stack, &p) )
+  { *p = *unRef(*p);
+  }
+}
+
+#else /*O_CYCLIC*/
+
+static inline visited(Functor f ARG_LD) { fail; }
+static inline unvisit(Word *base ARG_LD) { }
+static inline void initCyclic(ARG1_LD) {}
+static inline void exitCyclic(ARG1_LD) {}
+static inline void linkTermsCyclic(Functor f1, Functor f2 ARG_LD) {}
+
+#endif /*O_CYCLIC*/
+
+
+		/********************************
+		*          UNIFICATION          *
+		*********************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Unify is the general unification procedure. This raw routine should only
+be called by interpret as it  does   not  undo  bindings made during the
+unification in case the unification fails. pl_unify() (implementing =/2)
+does undo bindings and should be used   by  foreign predicates. See also
+unify_ptrs().
+
+Unification depends on the datatypes available in the system and will in
+general need updating if new types are added.  It should be  noted  that
+unify()  is  not  the only place were unification happens.  Other points
+are:
+
+  - various of the virtual machine instructions
+  - various macros, for example APPENDLIST and CLOSELIST
+  - unifyAtomic(): unification of atomic data.
+  - various builtin predicates. They should be flagged some way.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static bool
+do_unify(Word t1, Word t2 ARG_LD)
+{ 
+  word w1;
+  word w2;
+
+right_recursion:
+  w1 = *t1;
+  w2 = *t2;
+
+  while(isRef(w1))			/* this is deRef() */
+  { t1 = unRef(w1);
+    w1 = *t1;
+  }
+  while(isRef(w2))
+  { t2 = unRef(w2);
+    w2 = *t2;
+  }
+
+  if ( isVar(w1) )
+  { if ( isVar(w2) )
+    { if ( t1 < t2 )			/* always point downwards */
+      { *t2 = makeRef(t1);
+	Trail(t2);
+	succeed;
+      }
+      if ( t1 == t2 )
+	succeed;
+      *t1 = makeRef(t2);
+      Trail(t1);
+      succeed;
+    }
+#ifdef O_ATTVAR
+    *t1 = isAttVar(w2) ? makeRef(t2) : w2;
+#else
+    *t1 = w2;
+#endif
+    Trail(t1);
+    succeed;
+  }
+  if ( isVar(w2) )
+  {
+#ifdef O_ATTVAR
+    *t2 = isAttVar(w1) ? makeRef(t1) : w1;
+#else
+    *t2 = w1;
+#endif
+    Trail(t2);
+    succeed;
+  }
+
+#ifdef O_ATTVAR
+  if ( isAttVar(w1) )
+    return assignAttVar(t1, t2 PASS_LD);
+  if ( isAttVar(w2) )
+    return assignAttVar(t2, t1 PASS_LD);
+#endif
+
+  if ( w1 == w2 )
+    succeed;
+  if ( tag(w1) != tag(w2) )
+    fail;
+
+  switch(tag(w1))
+  { case TAG_ATOM:
+      fail;
+    case TAG_INTEGER:
+      if ( storage(w1) == STG_INLINE ||
+	   storage(w2) == STG_INLINE )
+	fail;
+    case TAG_STRING:
+    case TAG_FLOAT:
+      return equalIndirect(w1, w2);
+    case TAG_COMPOUND:
+    { Functor f1 = valueTerm(w1);
+      Functor f2 = valueTerm(w2);
+      Word e;
+
+#if O_CYCLIC
+      while ( isRef(f1->definition) )
+	f1 = (Functor)unRef(f1->definition);
+      while ( isRef(f2->definition) )
+	f2 = (Functor)unRef(f2->definition);
+      if ( f1 == f2 )
+	succeed;
+#endif
+
+      if ( f1->definition != f2->definition )
+	fail;
+
+      t1 = f1->arguments;
+      t2 = f2->arguments;
+      e  = t1+arityFunctor(f1->definition)-1; /* right-recurse on last */
+      linkTermsCyclic(f1, f2 PASS_LD);
+
+      for(; t1 < e; t1++, t2++)
+      { if ( !do_unify(t1, t2 PASS_LD) )
+	  fail;
+      }
+      goto right_recursion;
+    }
+  }
+
+  succeed;
+}
+
+
+bool
+raw_unify_ptrs(Word t1, Word t2 ARG_LD)
+{ bool rc;
+
+  initCyclic(PASS_LD1);
+  rc = do_unify(t1, t2 PASS_LD);
+  exitCyclic(PASS_LD1);
+
+  return rc;
+}
+
+
+static
+PRED_IMPL("=", 2, unify, 0)
+{ PRED_LD
+  Word p0 = valTermRef(A1);
+  tmp_mark m;
+  int rval;
+
+  TmpMark(m);
+  if ( !(rval = raw_unify_ptrs(p0, p0+1 PASS_LD)) )
+    TmpUndo(m);
+  EndTmpMark(m);
+
+  return rval;  
+}
+
+
+static
+PRED_IMPL("\\=", 2, not_unify, 0)
+{ PRED_LD
+  Word p0 = valTermRef(A1);
+
+  return can_unify(p0, p0+1) ? FALSE : TRUE;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Public unification procedure for  `raw'  data.   See  also  unify()  and
+PL_unify().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+bool
+unify_ptrs(Word t1, Word t2 ARG_LD)
+{ tmp_mark m;
+  bool rval;
+
+  TmpMark(m);
+  if ( !(rval = raw_unify_ptrs(t1, t2 PASS_LD)) )
+    TmpUndo(m);
+  EndTmpMark(m);
+
+  return rval;  
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+can_unify(t1, t2) succeeds if  two  terms   *can*  be  unified,  without
+actually doing so. This  is  basically   a  stripped  version of unify()
+above. See this function for comments.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+bool
+can_unify(Word t1, Word t2)
+{ GET_LD
+  tmp_mark m;
+  bool rval;
+
+  TmpMark(m);
+  if ( (rval = raw_unify_ptrs(t1, t2 PASS_LD)) )
+    rval = foreignWakeup(PASS_LD1);
+  TmpUndo(m);
+  EndTmpMark(m);
+
+  return rval;  
+}
 
 
 		/********************************
@@ -92,86 +426,6 @@ PRED_IMPL("atomic", 1, atomic, 0)
 }
 
 
-#if O_CYCLIC
-
-static inline int
-visitedWord(Word p ARG_LD)
-{ if ( is_marked(p) )
-    succeed;
-  set_marked(p);
-  requireStack(argument, sizeof(Word));
-  *aTop++ = p;
-  fail;
-}
-
-
-static inline int
-visited(Functor f ARG_LD)
-{ Word p = &f->definition;
-
-  return visitedWord(p PASS_LD);
-}
-
-
-static void
-unvisit(Word *base ARG_LD)
-{ Word *p = aTop;
-
-  while(p>base)
-  { p--;
-    clear_marked(*p);
-  }
-
-  aTop = base;
-}
-
-
-static void
-popVisited(ARG1_LD)
-{ Word *p = aTop;
-
-  p--;
-  clear_marked(*p);
-  aTop = p;
-}
-
-
-					/* see also pl-wam.c, unify() */
-static inline void
-linkTermsCyclic(Functor f1, Functor f2 ARG_LD)
-{ Word p1 = (Word)&f1->definition;
-  Word p2 = (Word)&f2->definition;
-
-  *p1 = makeRefG(p2);
-  requireStack(argument, sizeof(Word));
-  *aTop++ = p1;
-}
-
-
-static inline void
-exitCyclic(Word *base ARG_LD)
-{ Word *sp = aTop;
-
-  while(sp>base)
-  { Word p;
-
-    sp--;
-    p = *sp;
-    *p = *unRef(*p);
-  }
-
-  aTop = base;
-}
-
-#else
-
-static inline visited(Functor f ARG_LD) { fail; }
-static inline unvisit(Word *base ARG_LD) { }
-static inline void exitCyclic(ARG1_LD) {}
-static inline void linkTermsCyclic(Functor f1, Functor f2 ARG_LD) {}
-
-#endif /*O_CYCLIC*/
-
 static int
 ground(Word p ARG_LD)
 { int arity;
@@ -203,12 +457,11 @@ last:
 int
 PL_is_ground(term_t t)
 { GET_LD
-
-  Word *m = aTop;
   int rc;
 
+  initvisited(PASS_LD1);
   rc = ground(valTermRef(t) PASS_LD);
-  unvisit(m PASS_LD);
+  unvisit(PASS_LD1);
 
   return rc;
 }
@@ -217,11 +470,11 @@ PL_is_ground(term_t t)
 static
 PRED_IMPL("ground", 1, ground, 0)
 { PRED_LD
-  Word *m = aTop;
   int rc;
 
+  initvisited(PASS_LD1);
   rc = ground(valTermRef(A1) PASS_LD);
-  unvisit(m PASS_LD);
+  unvisit(PASS_LD1);
 
   return rc;
 }
@@ -235,11 +488,7 @@ PRED_IMPL("compound", 1, compound, 0)
 
 static
 PRED_IMPL("callable", 1, callable, 0)
-{ PRED_LD
-  if ( PL_is_atom(A1) || PL_is_compound(A1) )
-    succeed;
-
-  fail;
+{ return PL_is_callable(A1);
 }
 
 
@@ -417,13 +666,12 @@ PRED_IMPL("hash_term", 2, hash_term, 0)
 { PRED_LD
   Word p = valTermRef(A1);
   intptr_t hraw = 0L;
-  Word *m = aTop;
   int rc;
 
   deRef(p);
-
+  initvisited(PASS_LD1);
   rc = termHashValue(*p, &hraw PASS_LD);
-  assert(aTop == m);
+  assert(empty_visited(PASS_LD1));
 
   if ( rc )
   { hraw = hraw & PLMAXTAGGEDINT;	/* ensure tagged */
@@ -628,11 +876,11 @@ tail_recursion:
 
 int
 compareStandard(Word p1, Word p2, int eq ARG_LD)
-{ Word *m = aTop;
-  int rc;
+{ int rc;
 
+  initCyclic(PASS_LD1);
   rc = do_compare(p1, p2, eq PASS_LD);
-  exitCyclic(m PASS_LD);
+  exitCyclic(PASS_LD1);
 
   return rc;
 }
@@ -869,7 +1117,6 @@ PRED_IMPL("=@=", 2, structural_eq, 0)
   Reset r;
   Word p1 = valTermRef(A1);
   Word p2 = p1+1;
-  Word *m = aTop;
 
   deRef(p1);
   deRef(p2);
@@ -877,6 +1124,7 @@ PRED_IMPL("=@=", 2, structural_eq, 0)
   if ( *p1 == *p2 )
     succeed;
 
+  initCyclic(PASS_LD1);
   initBuffer(&buf);			/* can be faster! */
   rval = structeql(p1, p2, &buf PASS_LD);
   for(r = baseBuffer(&buf, reset); r < topBuffer(&buf, reset); r++)
@@ -884,7 +1132,7 @@ PRED_IMPL("=@=", 2, structural_eq, 0)
     setVar(*r->v2);
   }
   discardBuffer(&buf);
-  exitCyclic(m PASS_LD);
+  exitCyclic(PASS_LD1);
 
   return rval;
 }
@@ -904,17 +1152,18 @@ either equal or non-equal. I.e. X and Y are equal or they cannot unify.
 static
 PRED_IMPL("?=", 2, can_compare, 0)
 { PRED_LD
-  mark m;
+  tmp_mark m;
   bool rval;
 
-  Mark(m);
+  TmpMark(m);
   rval = PL_unify(A1, A2);
   if ( rval )
   { if ( m.trailtop != tTop )
       rval = FALSE;			/* can be equal after substitution */
   } else
     rval = TRUE;			/* cannot unify */
-  Undo(m);
+  TmpUndo(m);
+  EndTmpMark(m);
 
   return rval;
 }
@@ -1086,7 +1335,8 @@ setarg(term_t n, term_t term, term_t value, int flags)
     { if ( !(flags & SETARG_LINK) )
       { term_t copy = PL_new_term_ref();
 
-	duplicate_term(value, copy PASS_LD);
+	if ( !duplicate_term(value, copy PASS_LD) )
+	  fail;
 	value = copy;
       }
 
@@ -1271,8 +1521,9 @@ start:
     }
 
 #ifdef O_SHIFT_STACKS
-    if ( roomStack(global) < 2 * (intptr_t)sizeof(word) )
-    { growStacks(environment_frame, NULL, NULL, FALSE, TRUE, FALSE);
+    if ( roomStack(global) < (intptr_t)(2 * sizeof(word)) )
+    { if ( !growStacks(environment_frame, NULL, NULL, 0, 2 * sizeof(word), 0) )
+	return -1;
       p = valTermRef(t);
       deRef(p);
     }
@@ -1342,9 +1593,11 @@ start:
 int
 numberVars(term_t t, nv_options *options, int n ARG_LD)
 { term_t h2 = PL_copy_term_ref(t);
-  Word *m = aTop;
-  int rval = do_number_vars(h2, options, n PASS_LD);
-  unvisit(m PASS_LD);
+  int rval;
+
+  initvisited(PASS_LD1);
+  rval = do_number_vars(h2, options, n PASS_LD);
+  unvisit(PASS_LD1);
 
   PL_reset_term_refs(h2);
 
@@ -1454,12 +1707,12 @@ static int
 term_variables(term_t t, term_t vars, term_t tail ARG_LD)
 { term_t head = PL_new_term_ref();
   term_t v0   = PL_new_term_refs(0);
-  Word *m     = aTop;
   int i, n;
 
   startCritical;
+  initvisited(PASS_LD1);
   n = term_variables_loop(valTermRef(t), v0, 0 PASS_LD);
-  unvisit(m PASS_LD);
+  unvisit(PASS_LD1);
   endCritical;
 
   for(i=0; i<n; i++)
@@ -1546,14 +1799,14 @@ right_recursion:
 word
 pl_e_free_variables(term_t t, term_t vars)
 { GET_LD
-  Word *vm = aTop;
   Word t2 = valTermRef(t);
   term_t v0 = PL_new_term_refs(0);
   int i, n;
 
   startCritical;
+  initvisited(PASS_LD1);
   n = free_variables_loop(t2, v0, 0, FALSE PASS_LD);
-  unvisit(vm PASS_LD);
+  unvisit(PASS_LD1);
   endCritical;
 
   if ( PL_unify_functor(vars, PL_new_functor(ATOM_v, n)) )
@@ -1588,41 +1841,49 @@ the trouble. Note that unifying attributed   variables  is no problem as
 these always live on the global stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static
-PRED_IMPL("unifiable", 3, unifiable, 0)
-{ PRED_LD
-  mark m;
+static intptr_t
+unifiable(term_t t1, term_t t2, term_t subst ARG_LD)
+{ fid_t fid;
+  intptr_t tmark;
 
-  if ( PL_is_variable(A1) )
-  { if ( PL_compare(A1, A2) == 0 )
-      return PL_unify_atom(A3, ATOM_nil);
+  if ( PL_is_variable(t1) )
+  { if ( PL_compare(t1, t2) == 0 )
+      return PL_unify_atom(subst, ATOM_nil);
     else
-      return PL_unify_term(A3,
+      return PL_unify_term(subst,
 			   PL_FUNCTOR, FUNCTOR_dot2,
 			     PL_FUNCTOR, FUNCTOR_equals2,
-			       PL_TERM, A1,
-			       PL_TERM, A2,
+			       PL_TERM, t1,
+			       PL_TERM, t2,
 			     PL_ATOM, ATOM_nil);
   }
-  if ( PL_is_variable(A2) )
-  { return PL_unify_term(A3,
+  if ( PL_is_variable(t2) )
+  { return PL_unify_term(subst,
 			 PL_FUNCTOR, FUNCTOR_dot2,
 			   PL_FUNCTOR, FUNCTOR_equals2,
-			     PL_TERM, A2,
-			     PL_TERM, A1,
+			     PL_TERM, t2,
+			     PL_TERM, t1,
 			   PL_ATOM, ATOM_nil);
   }
 
-  Mark(m);
-  if ( PL_unify(A1, A2) )
+  fid = PL_open_foreign_frame();
+  tmark = (char*)tTop - (char*)tBase;	/* prepare for a shift */
+
+  if ( PL_unify(t1, t2) )
   { TrailEntry tt = tTop;
-    TrailEntry mt = m.trailtop;
+    TrailEntry mt = addPointer(tBase, tmark);
 
     if ( tt > mt )
-    { Word list = allocGlobalNoShift((tt-mt)*6+1);
+    { intptr_t needed = (tt-mt)*6+1;
+      Word list = allocGlobalNoShift(needed);
       Word gp = list+1;
       Word tail = list;
 
+      if ( !list )
+      { PL_rewind_foreign_frame(fid);
+	return -needed;
+      }
+
       *list = ATOM_nil;
       while(--tt >= mt)
       { Word p = tt->address;
@@ -1670,15 +1931,35 @@ PRED_IMPL("unifiable", 3, unifiable, 0)
 	}
       }
       gTop = gp;			/* may not have used all space */
-      tTop = m.trailtop;
+      tTop = addPointer(tBase, tmark);
 
-      return PL_unify(wordToTermRef(list), A3);
+      return PL_unify(wordToTermRef(list), subst);
     } else
-      return PL_unify_atom(A3, ATOM_nil);
+      return PL_unify_atom(subst, ATOM_nil);
   } else
     fail;
 }
 
+
+static
+PRED_IMPL("unifiable", 3, unifiable, 0)
+{ PRED_LD
+
+#ifdef O_SHIFT_STACKS
+  for(;;)
+  { intptr_t rc = unifiable(A1, A2, A3 PASS_LD);
+
+    if ( rc < 0 )			/* not enough space */
+    { if ( !growStacks(NULL, NULL, NULL, 0, (-rc)*sizeof(word), 0) )
+	return outOfStack(&LD->stacks.global, STACK_OVERFLOW_SIGNAL);
+    } else
+      return rc;
+  } 
+#else
+  return unifiable(A1, A2, A3 PASS_LD);
+#endif
+}
+
 #if O_CYCLIC
 
 		 /*******************************
@@ -1710,7 +1991,7 @@ attributed copied, so futher modification of the attribute list does not
 affect the copy  and  visa  versa.   The  algorithm  deals  with  shared
 variables as well as cyclic  terms.  It   works,  like  unify for cyclic
 terms, by creating references from  the   original  to the reference and
-restoring the references using the argument stack.
+restoring the references using a stack.
 
 There are three types of references between the original and the copy:
 
@@ -1720,7 +2001,7 @@ There are three types of references between the original and the copy:
 	  to the same address.
 	* For attributed variables we create an TAG_ATTVAR link to the
 	  copy.  If we find a TAG_ATTVAR pointing to a TAG_ATTVAR we
-	  no we found a copy.  Unfortunately just trailing the old
+	  know we found a copy.  Unfortunately just trailing the old
 	  location doesn't suffice as we must recreate the link to
 	  the old address, so we push this one first.
 	* Compounds use the old trick to make the functor a reference
@@ -1746,18 +2027,23 @@ shifted rather than doing it.
 
 
 static inline void
-TrailCyclic(Word p ARG_LD)
-{ requireStack(argument, sizeof(Word));
-  *aTop++ = p;
+initCyclicCopy(ARG1_LD)
+{ LD->cycle.stack.unit_size = sizeof(Word);
 }
 
 
 static inline void
-exitCyclicCopy(Word *m, int flags ARG_LD)
-{ Word *sp;
+TrailCyclic(Word p ARG_LD)
+{ pushSegStack(&LD->cycle.stack, &p);
+}
 
-  for(sp = aTop; sp > m; )
-  { Word p = *--sp;
+
+static inline void
+exitCyclicCopy(size_t count, int flags ARG_LD)
+{ while(LD->cycle.stack.count > count)
+  { Word p;
+
+    popSegStack(&LD->cycle.stack, &p);
 
     if ( isRef(*p) )
     { Word p2 = unRef(*p);
@@ -1769,7 +2055,9 @@ exitCyclicCopy(Word *m, int flags ARG_LD)
       { *p = *p2;			/* cyclic terms */
       }
     } else
-    { Word old = *--sp;
+    { Word old;
+
+      popSegStack(&LD->cycle.stack, &old);
 
       if ( !(flags&COPY_ATTRS) )
       { Word p2 = valPAttVar(*p);
@@ -1781,13 +2069,13 @@ exitCyclicCopy(Word *m, int flags ARG_LD)
       *p = consPtr(old, STG_GLOBAL|TAG_ATTVAR);
     }
   }
-  aTop = sp;
 }
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 FALSE: term cannot be shared
 TRUE:  term can be shared (ground)
+-1:    not enough space on the stack
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
@@ -1835,19 +2123,25 @@ again:
       { Word attr;			/* the new attributes */
 
 	if ( !onStackArea(global, to) )
-	{ Word t = allocGlobalNoShift(1);
+	{ Word t;
+
+	  if ( !(t = allocGlobalNoShift(1)) )
+	    return -1;
 	  
 	  *to = makeRefG(t);
 	  to = t;
 	}
-	attr = allocGlobalNoShift(1);
+	if ( !(attr = allocGlobalNoShift(1)) )
+	  return -1;
 	TrailCyclic(p PASS_LD);
 	TrailCyclic(from PASS_LD);
 	*from = consPtr(to, STG_GLOBAL|TAG_ATTVAR);
 	*to = consPtr(attr, STG_GLOBAL|TAG_ATTVAR);
 
+					/* copy attribute value */
 	flags &= ~COPY_SHARE;
-	do_copy_term(p, attr, flags PASS_LD);	/* copy attribute value */
+	if ( do_copy_term(p, attr, flags PASS_LD) < 0 )
+	  return -1;
 	return FALSE;
       }
     }
@@ -1868,9 +2162,12 @@ again:
 	Word oldtop = gTop;
 	Word to0 = to;
 	Word from0 = from;
-	Functor f2 = (Functor)allocGlobalNoShift(arity+1);
+	Functor f2;
 	int ground = TRUE;
-	Word *am = aTop;
+	size_t count = LD->cycle.stack.count;
+
+	if ( !(f2 = (Functor)allocGlobalNoShift(arity+1)) )
+	  return -1;
 
 	f2->definition = f1->definition;
 	f1->definition = makeRefG((Word)f2);
@@ -1880,11 +2177,21 @@ again:
 	from = &f1->arguments[0];
 	to   = &f2->arguments[0];
 	while(--arity > 0)
-	  ground &= do_copy_term(from++, to++, flags PASS_LD);
+	{ int rc = do_copy_term(from++, to++, flags PASS_LD);
+	  if ( rc < 0 )
+	    return rc;
+	  ground &= rc;
+	}
+
 	if ( (flags & COPY_SHARE) )
-	{ ground &= do_copy_term(from, to, flags PASS_LD);
+	{ int rc = do_copy_term(from, to, flags PASS_LD);
+
+	  if ( rc < 0 )
+	    return rc;
+
+	  ground &= rc;
 	  if ( ground )
-	  { exitCyclicCopy(am, flags PASS_LD);
+	  { exitCyclicCopy(count, flags PASS_LD);
 	    gTop = oldtop;
 	    *to0 = *from0;
 	    DEBUG(2, Sdprintf("Shared\n"));
@@ -1902,27 +2209,58 @@ again:
 }
 
 
+static int
+copy_term_refs(term_t from, term_t to, int flags ARG_LD)
+{
+#ifdef O_SHIFT_STACKS
+  intptr_t grow = sizeStack(global)/2;
+
+  for(;; grow *= 2)
+  { int rc;
+    Word gsave = gTop;
+
+    initCyclicCopy(PASS_LD1);
+    rc = do_copy_term(valTermRef(from), valTermRef(to), flags PASS_LD);
+    exitCyclicCopy(0, flags PASS_LD);
+
+    if ( rc == -1 )
+    { gTop = gsave;
+      if ( !growStacks(NULL, NULL, NULL, 0, grow, 0) )
+	return outOfStack(&LD->stacks.global, STACK_OVERFLOW_SIGNAL);
+    } else
+    { succeed;		/* do_copy_term returning FALSE just means not-ground */
+    }
+  }
+
+#else
+  int rc;
+
+  initCyclicCopy(PASS_LD1);
+  rc = do_copy_term(valTermRef(from), valTermRef(to), flags PASS_LD);
+  exitCyclicCopy(0, flags PASS_LD);
+  if ( rc == -1 )
+    return outOfStack(&LD->stacks.global, STACK_OVERFLOW_SIGNAL);
+
+  succeed;
+#endif
+}
+
+
 static
 PRED_IMPL("copy_term", 2, copy_term, 0)
 { PRED_LD
   term_t copy = PL_new_term_ref();
-  Word *m = aTop;
-  int flags = COPY_SHARE|COPY_ATTRS;
 
-  do_copy_term(valTermRef(A1), valTermRef(copy), flags PASS_LD);
-  exitCyclicCopy(m, flags PASS_LD);
-  
-  return PL_unify(copy, A2);
+  if ( copy_term_refs(A1, copy, COPY_SHARE|COPY_ATTRS PASS_LD) )
+    return PL_unify(copy, A2);
+
+  fail;
 }
 
 
-static void
+static int
 duplicate_term(term_t in, term_t copy ARG_LD)
-{ Word *m = aTop;
-  int flags = COPY_ATTRS;
-
-  do_copy_term(valTermRef(in), valTermRef(copy), flags PASS_LD);
-  exitCyclicCopy(m, flags PASS_LD);
+{ return copy_term_refs(in, copy, COPY_ATTRS PASS_LD);
 }
 
 
@@ -1931,9 +2269,10 @@ PRED_IMPL("duplicate_term", 2, duplicate_term, 0)
 { PRED_LD
   term_t copy = PL_new_term_ref();
 
-  duplicate_term(A1, copy PASS_LD);
-  
-  return PL_unify(copy, A2);
+  if ( duplicate_term(A1, copy PASS_LD) )
+    return PL_unify(copy, A2);
+
+  fail;
 }
 
 
@@ -1941,13 +2280,11 @@ static
 PRED_IMPL("copy_term_nat", 2, copy_term_nat, 0)
 { PRED_LD
   term_t copy = PL_new_term_ref();
-  Word *m = aTop;
-  int flags = COPY_SHARE;
 
-  do_copy_term(valTermRef(A1), valTermRef(copy), flags PASS_LD);
-  exitCyclicCopy(m, flags PASS_LD);
-  
-  return PL_unify(copy, A2);
+  if ( copy_term_refs(A1, copy, COPY_SHARE PASS_LD) )
+    return PL_unify(copy, A2);
+
+  fail;
 }
 
 #undef LD
@@ -2138,7 +2475,9 @@ PRED_IMPL("atom_number", 2, atom_number, 0)
 
 static
 PRED_IMPL("collation_key", 2, collation_key, 0)
-{ wchar_t *s;
+{
+#ifdef HAVE_WCSXFRM
+  wchar_t *s;
   size_t len;
   wchar_t buf[256];
   size_t buflen = sizeof(buf)/sizeof(wchar_t);
@@ -2161,6 +2500,9 @@ PRED_IMPL("collation_key", 2, collation_key, 0)
       o = PL_malloc(buflen*sizeof(wchar_t));
     }
   }
+#else
+  return PL_unify(A1, A2);
+#endif
 }
 
 
@@ -2228,7 +2570,6 @@ concat(const char *pred,
     fail;
   } else				/* -, -, + */
   { size_t at_n;
-    mark m;
 
     switch ( ForeignControl(ctx) )
     { case FRG_FIRST_CALL:
@@ -2245,20 +2586,12 @@ concat(const char *pred,
 	succeed;
     }
 
-    Mark(m);
-    for(; at_n <= L3; at_n++)
-    { if ( PL_unify_text_range(a2, &t3, at_n, L3-at_n, otype) &&
-	   PL_unify_text_range(a1, &t3, 0,    at_n, otype) )
-      { if ( at_n < L3 )
-	  ForeignRedoInt(at_n+1);
-	else
-	{ rc = TRUE;
-	  goto out;
-	}
-      }
-    }
-    rc = FALSE;
-    goto out;
+    PL_unify_text_range(a2, &t3, at_n, L3-at_n, otype);
+    PL_unify_text_range(a1, &t3, 0,    at_n, otype);
+    if ( at_n < L3 )
+      ForeignRedoInt(at_n+1);
+
+    rc = TRUE;
   }    
 
 out:
@@ -2542,7 +2875,7 @@ sub_text(term_t atom,
   sub_state *state;			/* non-deterministic state */
   atom_t expected = (type == PL_STRING ? ATOM_string : ATOM_atom);
   int match;
-  mark mrk;
+  fid_t fid;
 
 #define la ta.length
 #define ls ts.length
@@ -2674,7 +3007,7 @@ sub_text(term_t atom,
       fail;
   }
 
-  Mark(mrk);
+  fid = PL_open_foreign_frame();
 again:
   switch(state->type)
   { case SUB_SEARCH:
@@ -2767,7 +3100,7 @@ next:
   if ( match )
   { ForeignRedoPtr(state);
   } else
-  { Undo(mrk);
+  { PL_rewind_foreign_frame(fid);
     goto again;
   }
 
@@ -3137,9 +3470,9 @@ qp_statistics__LD(atom_t key, int64_t v[], PL_local_data_t *ld)
   } else if ( key == ATOM_stack_shifts )
   {
 #ifdef O_SHIFT_STACKS
-    v[0] = shift_status.global_shifts;
-    v[1] = shift_status.local_shifts;
-    v[2] = (int64_t)(shift_status.time * 1000.0);
+    v[0] = LD->shift_status.global_shifts;
+    v[1] = LD->shift_status.local_shifts;
+    v[2] = (int64_t)(LD->shift_status.time * 1000.0);
     vn = 3;
 #else
     fail;
@@ -3235,11 +3568,11 @@ swi_statistics__LD(atom_t key, Number v, PL_local_data_t *ld)
 #endif
 #if O_SHIFT_STACKS
   else if (key == ATOM_global_shifts)
-    v->value.i = shift_status.global_shifts;
+    v->value.i = LD->shift_status.global_shifts;
   else if (key == ATOM_local_shifts)
-    v->value.i = shift_status.local_shifts;
+    v->value.i = LD->shift_status.local_shifts;
   else if (key == ATOM_trail_shifts)
-    v->value.i = shift_status.trail_shifts;
+    v->value.i = LD->shift_status.trail_shifts;
 #else
   else if ( key == ATOM_global_shifts ||
 	    key == ATOM_local_shifts ||
@@ -3523,6 +3856,8 @@ PRED_IMPL("$style_check", 2, style_check, 0)
 		 *******************************/
 
 BeginPredDefs(prims)
+  PRED_DEF("=", 2, unify, 0)
+  PRED_DEF("\\=", 2, not_unify, 0)
   PRED_DEF("nonvar", 1, nonvar, 0)
   PRED_DEF("var", 1, var, 0)
   PRED_DEF("integer", 1, integer, 0)
