@@ -1,11 +1,11 @@
-/*  $Id: pl-proc.c,v 1.156 2007/01/13 12:49:07 jan Exp $
+/*  $Id: pl-proc.c,v 1.162 2007/05/16 09:42:27 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -38,7 +38,7 @@ finding source files, etc.
 static void	resetReferencesModule(Module);
 static void	resetProcedure(Procedure proc, bool isnew);
 static void	removeClausesProcedure(Procedure proc, int sfindex, int file);
-static atom_t	autoLoader(LocalFrame fr, Code PC, Definition def);
+static atom_t	autoLoader(LocalFrame *frp, Code PC, Definition def);
 static void	registerDirtyDefinition(Definition def);
 static Procedure visibleProcedure(functor_t f, Module m);
 
@@ -639,7 +639,7 @@ pl_current_predicate1(term_t spec, control_t ctx)
 	    else
 	    { Procedure proc = lookupProcedure(e->functor, e->module);
 
-	      if ( autoLoader(environment_frame, NULL,
+	      if ( autoLoader(&environment_frame, NULL,
 			      proc->definition) != ATOM_retry )
 		break;
 	    }
@@ -1441,7 +1441,7 @@ the garbage collector to scan this frame.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static atom_t
-autoLoader(LocalFrame fr, Code PC, Definition def)
+autoLoader(LocalFrame *frp, Code PC, Definition def)
 { GET_LD
   fid_t  cid  = PL_open_foreign_frame();
   term_t argv = PL_new_term_refs(4);
@@ -1460,7 +1460,9 @@ autoLoader(LocalFrame fr, Code PC, Definition def)
   
   LD->autoload_nesting++;
   if ( PC )
-  { fr->parent = environment_frame;
+  { LocalFrame fr = *frp;
+
+    fr->parent = environment_frame;
     fr->flags = fr->parent->flags;
     fr->predicate = def;
     fr->programPointer = PC;
@@ -1476,7 +1478,11 @@ autoLoader(LocalFrame fr, Code PC, Definition def)
     PL_get_atom(argv+3, &answer);
   PL_close_query(qid);
   if ( PC )
+  { LocalFrame fr = environment_frame;		/* may be shifted */
+
+    *frp = fr;
     environment_frame = fr->parent;
+  }
   LD->autoload_nesting--;
   source_file_name = sfn;
   source_line_no   = sln;
@@ -1492,7 +1498,7 @@ discontiguous should not cause an undefined predicate warning.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static Definition
-trapUndefined_unlocked(LocalFrame fr, Code PC, Procedure proc ARG_LD)
+trapUndefined_unlocked(LocalFrame *frp, Code PC, Procedure proc ARG_LD)
 { int retry_times = 0;
   Definition newdef;
   Definition def = proc->definition;
@@ -1518,7 +1524,7 @@ trapUndefined_unlocked(LocalFrame fr, Code PC, Procedure proc ARG_LD)
 
       return def;
     } else
-    { atom_t answer = autoLoader(fr, PC, def);
+    { atom_t answer = autoLoader(frp, PC, def);
 
       def = lookupProcedure(functor->functor, module)->definition;
 
@@ -1542,8 +1548,8 @@ error:
     sysError("Undefined predicate: %s", predicateName(def));
   else if ( true(module, UNKNOWN_ERROR) )
   { Definition caller;
-    if ( fr->parent )
-      caller = fr->parent->predicate;
+    if ( (*frp)->parent )
+      caller = (*frp)->parent->predicate;
     else
       caller = NULL;
     PL_error(NULL, 0, NULL, ERR_UNDEFINED_PROC, def, caller);
@@ -1575,20 +1581,23 @@ that should be considered.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Definition
-trapUndefined(LocalFrame fr, Code PC, Procedure proc ARG_LD)
-{ LocalFrame lSafe = lTop;
+trapUndefined(LocalFrame *frp, Code PC, Procedure proc ARG_LD)
+{ intptr_t lSafe = (char*)lTop - (char*)lBase;
+  LocalFrame fr = *frp;
   Definition def;
 
   lTop = (LocalFrame)argFrameP(fr, proc->definition->functor->arity);
 #ifdef O_PLMT
   PL_mutex_lock(GD->thread.MUTEX_load);
 #endif
-  def = trapUndefined_unlocked(fr, PC, proc PASS_LD);
+  def = trapUndefined_unlocked(&fr, PC, proc PASS_LD);
 #ifdef O_PLMT
   PL_mutex_unlock(GD->thread.MUTEX_load);
 #endif
-  lTop = lSafe;
 
+  lTop = (LocalFrame)((char*)lBase+lSafe);
+  if ( frp != &lTop )
+    *frp = fr;
 
   return def;
 }
@@ -1638,8 +1647,7 @@ pl_retract(term_t term, control_t h)
     Word argv;
     ClauseRef next;
     atom_t b;
-    mark mrk;
-    term_t r0;
+    fid_t fid;
 
     PL_strip_module(term, &m, cl);
     get_head_and_body_clause(cl, head, body, NULL PASS_LD);
@@ -1696,14 +1704,12 @@ pl_retract(term_t term, control_t h)
       startCritical;
     }
 
-    Mark(mrk);
-    r0 = PL_new_term_refs(0);
+    fid = PL_open_foreign_frame();
     while( cref )
     { if ( decompile(cref->clause, cl, 0) )
       { retractClauseDefinition(getProcDefinition(proc), cref->clause PASS_LD);
 	if ( !next )
-	{ PL_reset_term_refs(r0);
-	  leaveDefinition(def);
+	{ leaveDefinition(def);
 	  endCritical;
 	  succeed;
 	}
@@ -1712,8 +1718,7 @@ pl_retract(term_t term, control_t h)
 	ForeignRedoPtr(next);
       }
 
-      PL_reset_term_refs(r0);
-      Undo(mrk);
+      PL_rewind_foreign_frame(fid);
 
       cref = findClause(next, argv, environment_frame, def, &next PASS_LD);
     }
@@ -1735,8 +1740,7 @@ pl_retractall(term_t head)
   ClauseRef next;
   Word argv;
   LocalFrame fr = environment_frame;
-  mark m;
-  term_t r0;
+  fid_t fid;
 
   if ( !get_procedure(head, &proc, thehead, GP_FINDHERE) )
     succeed;
@@ -1759,9 +1763,9 @@ pl_retractall(term_t head)
   else
     argv = NULL;			/* retract(foobar) */
 
-  Mark(m);
   startCritical;
   enterDefinition(def);
+  fid = PL_open_foreign_frame();
 
   if ( !(cref = firstClause(argv, fr, def, &next PASS_LD)) )
   { endCritical;
@@ -1769,13 +1773,11 @@ pl_retractall(term_t head)
     succeed;
   }
 
-  r0 = PL_new_term_refs(0);
   while( cref )
   { if ( decompileHead(cref->clause, thehead) )
       retractClauseDefinition(def, cref->clause PASS_LD);
 
-    PL_reset_term_refs(r0);
-    Undo(m);
+    PL_rewind_foreign_frame(fid);
 
     if ( !next )
     { leaveDefinition(def);
@@ -2041,9 +2043,11 @@ setDynamicProcedure(Procedure proc, bool isdyn)
     attachMutexAndUnlock(def);
   } else				/* dynamic --> static */
   { clear(def, DYNAMIC);
-    if ( def->references && true(def, NEEDSCLAUSEGC|NEEDSREHASH) )
-      registerDirtyDefinition(def);
-    def->references = 0;
+    if ( def->references )
+    { if ( true(def, NEEDSCLAUSEGC|NEEDSREHASH) )
+	registerDirtyDefinition(def);
+      def->references = 0;
+    }
 
     detachMutexAndUnlock(def);
   }
@@ -2240,11 +2244,14 @@ reindexDefinition(Definition def)
 }
 
 
+/* MT: Definition is locked by caller
+*/
+
 void
 indexDefinition(Definition def, long pattern)
 { clear(def, AUTOINDEX);
 
-  if ( (def->indexPattern & ~NEED_REINDEX) != 0x1L &&
+  if ( pattern != 0x1L &&
        true(def, DYNAMIC) && def->references == 0 )
   { if ( def->hash_info )
     { unallocClauseIndexTable(def->hash_info);
@@ -2774,7 +2781,7 @@ pl_clause_from_source(term_t file, term_t line, term_t clause)
 static void
 listGenerations(Definition def)
 { GET_LD
-  ulong gen = environment_frame->generation;
+  uintptr_t gen = environment_frame->generation;
   ClauseRef cl;
 
   Sdprintf("%s has %d clauses at generation %ld (%s)\n",
@@ -2787,7 +2794,7 @@ listGenerations(Definition def)
   { Clause clause = cl->clause;
 
     Sdprintf("%8u: %8u-%10u %s\n",
-	     ((ulong)clause - heap_base)>>2,
+	     ((uintptr_t)clause - heap_base)>>2,
 	     clause->generation.created,
 	     clause->generation.erased,
 	     visibleClause(clause, gen) ? "ok" : "erased");
@@ -2812,7 +2819,7 @@ listGenerations(Definition def)
       { Clause clause = cl->clause;
 
 	Sdprintf("%8u: %8u-%10u %s\n",
-		 ((ulong)clause - heap_base)>>2,
+		 ((uintptr_t)clause - heap_base)>>2,
 		 clause->generation.created,
 		 clause->generation.erased,
 		 visibleClause(clause, gen) ? "ok" : "erased");
