@@ -1,4 +1,4 @@
-/*  $Id: pl-prof.c,v 1.57 2007/01/13 13:17:46 jan Exp $
+/*  $Id: pl-prof.c,v 1.59 2007/06/28 15:05:06 jan Exp $
 
     Part of SWI-Prolog
 
@@ -57,12 +57,12 @@ typedef struct call_node
   struct call_node *parent;
   void *            handle;		/* handle to procedure-id */
   PL_prof_type_t   *type;
-  ulong		    calls;		/* Calls from the parent */
-  ulong		    redos;		/* redos while here */
-  ulong		    exits;		/* exits to the parent */
-  ulong		    recur;		/* recursive calls */
-  ulong		    ticks;		/* time-statistics */
-  ulong		    sibling_ticks;	/* ticks in a siblings */
+  uintptr_t	    calls;		/* Calls from the parent */
+  uintptr_t	    redos;		/* redos while here */
+  uintptr_t	    exits;		/* exits to the parent */
+  uintptr_t	    recur;		/* recursive calls */
+  uintptr_t	    ticks;		/* time-statistics */
+  uintptr_t	    sibling_ticks;	/* ticks in a siblings */
   struct call_node *next;		/* next in siblings chain */
   struct call_node *siblings;		/* my offspring */
 } call_node;
@@ -229,6 +229,7 @@ startProfiler(void)
   if (setitimer(ITIMER_PROF, &value, &ovalue) != 0)
     return PL_error(NULL, 0, MSG_ERRNO, ERR_SYSCALL, setitimer);
   activateProfiler(TRUE PASS_LD);
+  LD->profile.time_at_start = CpuTime(CPU_USER); /* consider using thread-time? */
 
   succeed;
 }
@@ -257,6 +258,8 @@ stopProfiler(void)
   if ( !LD->profile.active )
     succeed;
 
+  LD->profile.time += CpuTime(CPU_USER) - LD->profile.time_at_start;
+
   stopItimer();
   activateProfiler(FALSE PASS_LD);
 #ifndef __WINDOWS__
@@ -427,20 +430,20 @@ typedef struct prof_ref
   void * handle;			/* Procedure handle */
   PL_prof_type_t   *type;
   int   cycle;
-  ulong ticks;
-  ulong sibling_ticks;
-  ulong calls;				/* calls to/from this predicate */
-  ulong redos;				/* redos to/from this predicate */
+  uintptr_t ticks;
+  uintptr_t sibling_ticks;
+  uintptr_t calls;			/* calls to/from this predicate */
+  uintptr_t redos;			/* redos to/from this predicate */
 } prof_ref;
 
 
 typedef struct
 { Definition def;
-  ulong ticks;
-  ulong sibling_ticks;
-  ulong calls;
-  ulong redos;
-  ulong recur;
+  uintptr_t ticks;
+  uintptr_t sibling_ticks;
+  uintptr_t calls;
+  uintptr_t redos;
+  uintptr_t recur;
   prof_ref *callers;
   prof_ref *callees;
 } node_sum;
@@ -496,7 +499,7 @@ add_parent_ref(node_sum *sum,
 
 
 static void
-add_recursive_ref(node_sum *sum, call_node *self, ulong count,
+add_recursive_ref(node_sum *sum, call_node *self, uintptr_t count,
 		  int cycle ARG_LD)
 { prof_ref *r;
 
@@ -745,14 +748,11 @@ PRED_IMPL("reset_profiler", 0, reset_profiler, 0)
 
 static
 PRED_IMPL("$profile", 1, profile, PL_FA_TRANSPARENT)
-{ PRED_LD
-  int rc;
+{ int rc;
 
   resetProfiler();
   startProfiler();
-  LD->profile.time = CpuTime(CPU_USER);
   rc = callProlog(NULL, A1, PL_Q_PASS_EXCEPTION, NULL);
-  LD->profile.time = CpuTime(CPU_USER) - LD->profile.time;
   stopProfiler();
 
   DEBUG(0,
@@ -993,10 +993,10 @@ PL_prof_exit(void *node)
 		 *******************************/
 
 
-static ulong
+static uintptr_t
 collectSiblingsNode(call_node *n)
 { call_node *s;
-  ulong count = 0;
+  uintptr_t count = 0;
 
   for(s=n->siblings; s; s=s->next)
   { count += collectSiblingsNode(s);
