@@ -1,4 +1,4 @@
-/*  $Id: pl-rc.c,v 1.14 2007/01/08 13:29:29 jan Exp $
+/*  $Id: pl-rc.c,v 1.15 2007/01/30 15:48:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -169,7 +169,7 @@ foreign_t
 pl_rc_open_archive(term_t file, term_t handle)
 { char *name;
 
-  if ( PL_get_chars_ex(file, &name, CVT_ALL) )
+  if ( PL_get_file_name(file, &name, 0) )
   { RcArchive a = rc_open_archive(name, RC_RDWR|RC_CREATE);
 
     if ( a )
