@@ -1,11 +1,11 @@
-/*  $Id: pl-rec.c,v 1.97 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-rec.c,v 1.101 2007/06/01 14:24:32 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -320,6 +320,12 @@ addFunctor(CompileInfo info, functor_t f)
 }
 
 
+typedef struct
+{ Functor	term;
+  functor_t	fdef;
+} cycle_mark;
+
+
 #define mkAttVarP(p)  ((Word)((word)(p) | 0x1L))
 #define isAttVarP(p)  ((word)(p) & 0x1)
 #define valAttVarP(p) ((Word)((word)(p) & ~0x1L))
@@ -442,12 +448,15 @@ right_recursion:
 
 	return;
       } else
-      { arity   = arityFunctor(f->definition);
+      { cycle_mark mark;
+
+	arity   = arityFunctor(f->definition);
 	functor = f->definition;
 
 	requireStack(argument, sizeof(Word)*2);
-	*aTop++ = (Word)f;
-	*aTop++ = (Word)f->definition;
+	mark.term = f;
+	mark.fdef = f->definition;
+	pushSegStack(&LD->cycle.stack, &mark);
 	f->definition = (functor_t)consInt(info->size);
 	assert(valInt(f->definition) == (intptr_t)info->size); /* overflow test */
       }
@@ -475,27 +484,26 @@ right_recursion:
 
 #if O_CYCLIC
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Argument stack contains pairs of pointer to   term and first word of the
-term (functor) before entering. The first  word itself contains a Prolog
-integer with the current stack-offset for the new term.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+static void
+init_cycle(ARG1_LD)
+{ LD->cycle.stack.unit_size = sizeof(cycle_mark);
+}
+
 
 static void
-unvisit(Word *base ARG_LD)
-{ Word *p = aTop;
-  Word gp;
+unvisit(ARG1_LD)
+{ cycle_mark mark;
 
-  while(p>base)
-  { p -= 2;
-
-    gp  = p[0];
-    *gp = (word)p[1];
+  while( popSegStack(&LD->cycle.stack, &mark) )
+  { mark.term->definition = mark.fdef;
   }
-
-  aTop = base;
 }
 
+#else
+
+static void init_cycle(ARG1_LD) {}
+static void unvisit(ARG1_LD) {}
+
 #endif
 
 
@@ -506,9 +514,7 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
   Word *p;
   size_t size;
   size_t rsize = SIZERECORD(flags);
-#if O_CYCLIC
-  Word *m = aTop;
-#endif
+  init_cycle(PASS_LD1);
 
   SECURE(checkData(valTermRef(t)));
 
@@ -530,9 +536,7 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
   }
   discardBuffer(&info.vars);
   
-#if O_CYCLIC
-  unvisit(m PASS_LD);
-#endif
+  unvisit(PASS_LD1);
   
   size = rsize + sizeOfBuffer(&info.code);
   record = allocHeap(size);
@@ -585,14 +589,12 @@ PL_record_external(term_t t, size_t *len)
   int scode, shdr;
   char *rec;
   int first = REC_HDR;
-#if O_CYCLIC
-  Word *m = aTop;
-#endif
 
   SECURE(checkData(valTermRef(t)));
   p = valTermRef(t);
   deRef(p);
 
+  init_cycle(PASS_LD1);
   initBuffer(&info.code);
 
   if ( isInteger(*p) )			/* integer-only record */
@@ -636,9 +638,7 @@ PL_record_external(term_t t, size_t *len)
   if ( info.nvars == 0 )
     first |= REC_GROUND;
 
-#if O_CYCLIC
-  unvisit(m PASS_LD);
-#endif
+  unvisit(PASS_LD1);
 
   initBuffer(&hdr);
   addBuffer(&hdr, first, uchar);		/* magic code */
@@ -1482,10 +1482,9 @@ getKeyEx(term_t key, word *w ARG_LD)
 
 word
 pl_current_key(term_t k, control_t h)
-{ GET_LD
-  TableEnum e;
+{ TableEnum e;
   Symbol s;
-  mark m;
+  fid_t fid;
 
   switch( ForeignControl(h) )
   { case FRG_FIRST_CALL:
@@ -1501,14 +1500,14 @@ pl_current_key(term_t k, control_t h)
       succeed;
   }
 
-  Mark(m);
+  fid = PL_open_foreign_frame();
   while( (s=advanceTableEnum(e)) )
   { RecordList l = s->value;
 
     if ( l->firstRecord && unifyKey(k, l->key) )
       ForeignRedoPtr(e);
 
-    Undo(m);
+    PL_rewind_foreign_frame(fid);
   }
 
   freeTableEnum(e);
@@ -1636,17 +1635,20 @@ pl_recorded(term_t key, term_t term, term_t ref, control_t h)
   }
 
 { GET_LD
-  copy = PL_new_term_ref();
-  for( ; record; record = record->next )
-  { mark m;
+  fid_t fid;
 
-    if ( true(record->record, ERASED) )
+  copy = PL_new_term_ref();
+  fid = PL_open_foreign_frame();
+
+  for( ; record; record = record->next )
+  { if ( true(record->record, ERASED) )
       continue;
 
-    Mark(m);
     copyRecordToGlobal(copy, record->record PASS_LD);
     if ( PL_unify(term, copy) && PL_unify_pointer(ref, record) )
-    { if ( !record->next )
+    { PL_close_foreign_frame(fid);
+
+      if ( !record->next )
       { if ( --rl->references == 0 && true(rl, R_DIRTY) )
 	  cleanRecordList(rl);
 	UNLOCK();
@@ -1656,8 +1658,11 @@ pl_recorded(term_t key, term_t term, term_t ref, control_t h)
 	ForeignRedoPtr(record->next);
       }
     }
-    Undo(m);
+
+    PL_rewind_foreign_frame(fid);
   }
+
+  PL_close_foreign_frame(fid);
 }
 
   if ( --rl->references == 0 && true(rl, R_DIRTY) )
@@ -1825,13 +1830,11 @@ undo_while_saving_term(mark *m, Word term)
   copy_info b;
   uint n;
   Word *p;
-#if O_CYCLIC
-  Word *cycle_mark = aTop;
-#endif
 
   SECURE(checkData(term));
   assert(onStack(local, term));
 
+  init_cycle(PASS_LD1);
   initBuffer(&info.code);
   initBuffer(&info.vars);
   info.size = 0;
@@ -1849,10 +1852,7 @@ undo_while_saving_term(mark *m, Word term)
       setVar(**p);
   }
 
-#if O_CYCLIC
-  unvisit(cycle_mark PASS_LD);
-#endif
-
+  unvisit(PASS_LD1);
   Undo(*m);
   
   b.data = info.code.base;
