@@ -1,11 +1,11 @@
-/*  $Id: pl-setup.c,v 1.156 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-setup.c,v 1.171 2007/06/18 17:44:32 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -31,8 +31,6 @@
 #include <unistd.h>
 #endif
 
-#undef ulong
-#define ulong uintptr_t
 #undef max
 #define max(a,b) ((a) > (b) ? (a) : (b))
 
@@ -347,7 +345,7 @@ dispatch_signal(int sig, int sync)
 { GET_LD
   SigHandler sh = &GD->sig_handlers[sig];
   fid_t fid;
-  LocalFrame lTopSave;
+  uintptr_t lTopSave;
   int saved_current_signal;
 
 #ifdef O_PLMT
@@ -364,7 +362,7 @@ dispatch_signal(int sig, int sync)
 		    pthread_self(),
 		    sync ? " (sync)" : " (async)"));
 	
-  lTopSave = lTop;
+  lTopSave = (char*)lTop - (char*)lBase;
   saved_current_signal = LD->current_signal;
 
   switch(sig)
@@ -440,7 +438,7 @@ dispatch_signal(int sig, int sync)
 
   LD->current_signal = saved_current_signal;
   PL_discard_foreign_frame(fid);
-  lTop = lTopSave;
+  lTop = addPointer(lBase, lTopSave);
 
   unblockGC(PASS_LD1);
 }
@@ -482,7 +480,7 @@ set_sighandler(int sig, handler_t func)
 #endif
 
 
-#if !defined(NO_SEGV_HANDLING) && defined(O_DYNAMIC_STACKS)
+#if defined(O_SEGV_HANDLING) && defined(O_DYNAMIC_STACKS)
 static handler_t
 set_stack_guard_handler(int sig, void *func)
 { 
@@ -514,7 +512,7 @@ set_stack_guard_handler(int sig, void *func)
   return signal(sig, func);
 #endif
 }
-#endif /*NO_SEGV_HANDLING*/
+#endif /*O_SEGV_HANDLING*/
 
 
 static SigHandler
@@ -731,11 +729,11 @@ PL_handle_signals()
 { int done = 0;
   PL_local_data_t *ld = LD;
 
-  if ( !ld )
+  if ( !ld || ld->critical )
     return 0;
 
-  while(!ld->critical && ld->pending_signals)
-  { ulong mask = 1L;
+  while(ld->pending_signals)
+  { uintptr_t mask = 1;
     int sig = 1;
 
     for( ; mask ; mask <<= 1, sig++ )
@@ -751,8 +749,10 @@ PL_handle_signals()
 #endif
 #ifdef O_ATOMGC
 	if ( sig == SIG_ATOM_GC )
-	  pl_garbage_collect_atoms();
-	else
+	{ if ( GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin &&
+	       !gc_status.blocked )
+	    pl_garbage_collect_atoms();
+	} else
 #endif
         if ( sig == SIG_EXCEPTION && ld->pending_exception )
 	{ record_t ex = ld->pending_exception;
@@ -780,6 +780,13 @@ PL_handle_signals()
 }
 
 
+void
+endCritical__LD(ARG1_LD)
+{ if ( LD->aborted )
+    pl_abort(ABORT_NORMAL);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 signal(+Signal, :Old, :Handler)
 	Assign Handler to be called if signal arrises.  Example:
@@ -891,9 +898,18 @@ enforce_limit(intptr_t *size, intptr_t maxarea, const char *name)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+initPrologStacks() creates the stacks for the calling thread. It is used
+both at system startup to create the stack   for the main thread as from
+pl-thread.c to create stacks for Prolog threads.
+
+allocStacks() does the  real  work   and  has  several  implementations,
+depending on the OS features.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 initPrologStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
-{ ulong maxarea;
+{ uintptr_t maxarea;
 
   maxarea = MAXTAGGEDPTR+1;		/* MAXTAGGEDPTR = 0x..fff.. */
   if ( maxarea > 1024 MB )		/* 64-bit machines */
@@ -959,16 +975,16 @@ emptyStacks()
   destroyGlobalVars();
 #endif
 
-  LD->mark_bar          = gLimit;
+  LD->mark_bar          = (Word)(~(uintptr_t)0);
 }
 
 #if O_DYNAMIC_STACKS
 
 static void init_stack(Stack s, char *name,
-		       caddress base, intptr_t limit, intptr_t minsize);
+		       void* base, intptr_t limit, intptr_t minsize);
 static void gcPolicy(Stack s, int policy);
 
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
 #ifdef USE_SIGINFO
 RETSIGTYPE _PL_segv_handler(int sig, siginfo_t *info, void *);
 #else
@@ -1108,15 +1124,15 @@ get_map_fd()
 
 static void
 mapOrOutOf(Stack s)
-{ ulong incr;
+{ uintptr_t incr;
   intptr_t  newroom;
 
   if ( s->top > s->max )
-    incr = ROUND(((ulong)s->top - (ulong)s->max), size_alignment);
+    incr = ROUND(((uintptr_t)s->top - (uintptr_t)s->max), size_alignment);
   else
     incr = size_alignment;
 
-  newroom = (ulong)s->limit - ((ulong)s->max + incr);
+  newroom = (uintptr_t)s->limit - ((uintptr_t)s->max + incr);
   if ( newroom < 0 )
     outOfStack(s, STACK_OVERFLOW_FATAL);
 
@@ -1134,7 +1150,7 @@ mapOrOutOf(Stack s)
   if ( newroom < STACK_SIGNAL )
   { if ( newroom < STACK_RESERVE )
     {
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
       LD->current_signal = SIGSEGV;
 #endif
       outOfStack(s, STACK_OVERFLOW_THROW);
@@ -1146,7 +1162,7 @@ mapOrOutOf(Stack s)
 }
 
 
-#ifdef NO_SEGV_HANDLING
+#ifndef O_SEGV_HANDLING
 void
 ensureRoomStack(Stack s, size_t bytes)
 { while((char *)s->max <= &((char *)s->top)[bytes])
@@ -1165,8 +1181,7 @@ unmap(Stack s)
 
 #ifdef MAP_FIXED
     munmap(addr, len);
-    if ( mmap(addr, len, MAPPROTECT, MAP_FIXED|MAP_FLAGS, mapfd, 0L) !=
-	 addr )
+    if ( mmap(addr, len, MAPPROTECT, MAP_FIXED|MAP_FLAGS, mapfd, 0L) != addr )
       fatalError("Failed to remap 0x%x bytes at %p: %s",
 		 len, addr, OsError());
 #else
@@ -1185,7 +1200,6 @@ static int
 allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
 { caddress lbase, gbase, tbase, abase;
   intptr_t glsize;
-  intptr_t lsep, tsep;
   intptr_t minglobal   = 4*SIZEOF_VOIDP K;
   intptr_t minlocal    = 2*SIZEOF_VOIDP K;
   intptr_t mintrail    = 2*SIZEOF_VOIDP K;
@@ -1202,13 +1216,6 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   PL_UNLOCK(L_MISC);
 #endif
 
-#ifdef NO_SEGV_HANDLING
-  lsep = tsep = 0;
-#else
-  lsep = STACK_SEPARATION;
-  tsep = size_alignment;
-#endif
-
   local    = max(local,    minlocal + STACK_SIGNAL);
   global   = max(global,   minglobal + STACK_SIGNAL);
   trail    = max(trail,    mintrail + STACK_SIGNAL);
@@ -1218,32 +1225,24 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   global   = (intptr_t) align_size(global);
   trail    = (intptr_t) align_size(trail);
   argument = (intptr_t) align_size(argument);
-  glsize   = global+tsep+local+lsep;
+  glsize   = global+local;
 
-  tbase = mmap(NULL, trail+tsep,    MAPPROTECT, MAP_FLAGS, mapfd, 0L);
-  abase = mmap(NULL, argument+tsep, MAPPROTECT, MAP_FLAGS, mapfd, 0L);
-  gbase = mmap(NULL, glsize,        MAPPROTECT, MAP_FLAGS, mapfd, 0L);
-  lbase = addPointer(gbase, global + tsep);
+  tbase = mmap(NULL, trail,    MAPPROTECT, MAP_FLAGS, mapfd, 0L);
+  abase = mmap(NULL, argument, MAPPROTECT, MAP_FLAGS, mapfd, 0L);
+  gbase = mmap(NULL, glsize,   MAPPROTECT, MAP_FLAGS, mapfd, 0L);
+  lbase = addPointer(gbase, global);
 
   if ( tbase == MAP_FAILED || abase == MAP_FAILED || gbase == MAP_FAILED )
   { if ( tbase != MAP_FAILED )
-      munmap(tbase, trail+tsep);
+      munmap(tbase, trail);
     if ( abase != MAP_FAILED )
-      munmap(abase, argument+tsep);
+      munmap(abase, argument);
     if ( gbase != MAP_FAILED )
       munmap(gbase, glsize);
 
     fail;
   }
 
-  /*
-  Sdprintf("Mapped:\n"
-	   "\t%p %ld bytes\n"
-	   "\t%p %ld bytes\n"
-	   "\t%p %ld bytes\n",
-	   tbase, trail+tsep, abase, argument+tsep, gbase, glsize);
-  */
-
 #define INIT_STACK(name, print, base, limit, minsize) \
   DEBUG(1, Sdprintf("%s stack at 0x%x; size = %ld\n", print, base, limit)); \
   init_stack((Stack) &LD->stacks.name, print, base, limit, minsize);
@@ -1254,7 +1253,7 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   INIT_STACK(trail,    "trail",    tbase, trail,    mintrail);
   INIT_STACK(argument, "argument", abase, argument, minargument);
 
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
   set_stack_guard_handler(SIGSEGV, _PL_segv_handler);
 #endif
   
@@ -1266,35 +1265,20 @@ Free stacks for the current Prolog thread
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-freeStacks(PL_local_data_t *ld)
-{ intptr_t lsep, tsep;
-  intptr_t tlen, alen, glen;
+freeStacks(ARG1_LD)
+{ intptr_t tlen, alen, gllen;
 
-#ifdef NO_SEGV_HANDLING
-  lsep = tsep = 0;
-#else
-  lsep = STACK_SEPARATION;
-  tsep = size_alignment;
-#endif
+#undef LD
+#define LD LOCAL_LD
+  tlen  = limitStack(trail);
+  alen  = limitStack(argument);
+  gllen = limitStack(global) + limitStack(local);
 
-  tlen = (char *)ld->stacks.trail.limit - (char *)ld->stacks.trail.base;
-  alen = (char *)ld->stacks.argument.limit - (char *)ld->stacks.argument.base;
-  glen = ((char *)ld->stacks.global.limit - (char *)ld->stacks.global.base) +
-         ((char *)ld->stacks.local.limit - (char *)ld->stacks.local.base);
-
-  /*
-  Sdprintf("UnMapped:\n"
-	   "\t%p %ld bytes\n"
-	   "\t%p %ld bytes\n"
-	   "\t%p %ld bytes\n",
-	   ld->stacks.trail.base,    tlen+tsep,
-	   ld->stacks.argument.base, alen+tsep,
-	   ld->stacks.global.base,   glen+tsep+lsep);
-  */
-
-  munmap((char *)ld->stacks.trail.base,    tlen+tsep);
-  munmap((char *)ld->stacks.argument.base, alen+tsep);
-  munmap((char *)ld->stacks.global.base,   glen+tsep+lsep);
+  munmap(tBase, tlen);
+  munmap(aBase, alen);
+  munmap(gBase, gllen);
+#undef LD
+#define LD GLOBAL_LD
 }
 
 
@@ -1311,15 +1295,15 @@ freeStacks(PL_local_data_t *ld)
 
 static void
 mapOrOutOf(Stack s)
-{ ulong incr;
+{ uintptr_t incr;
   intptr_t  newroom;
 
   if ( s->top > s->max )
-    incr = ROUND(((ulong)s->top - (ulong)s->max), size_alignment);
+    incr = ROUND(((uintptr_t)s->top - (uintptr_t)s->max), size_alignment);
   else
     incr = size_alignment;
 
-  newroom = (ulong)s->limit - ((ulong)s->max + incr);
+  newroom = (uintptr_t)s->limit - ((uintptr_t)s->max + incr);
   if ( newroom < 0 )
     outOfStack(s, STACK_OVERFLOW_FATAL);
 
@@ -1330,14 +1314,14 @@ mapOrOutOf(Stack s)
 
   DEBUG(1, Sdprintf("mapped %d bytes from 0x%x to 0x%x\n",
 		    incr, (unsigned) s->max,
-		    (ulong) s->max + size_alignment));
+		    (uintptr_t) s->max + size_alignment));
 
   s->max = addPointer(s->max, incr);
 
   if ( newroom <= STACK_SIGNAL )
   { if ( newroom <= STACK_RESERVE )
     {
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
       LD->current_signal = SIGSEGV;
 #endif
       outOfStack(s, STACK_OVERFLOW_THROW);
@@ -1349,7 +1333,7 @@ mapOrOutOf(Stack s)
 }
 
 
-#ifdef NO_SEGV_HANDLING
+#ifndef O_SEGV_HANDLING
 void
 ensureRoomStack(Stack s, size_t bytes)
 { while((char *)s->max <= &((char *)s->top)[bytes])
@@ -1363,7 +1347,7 @@ unmap(Stack s)
   caddress addr = (caddress) align_size((intptr_t) top + size_alignment);
 
   if ( addr < s->max )
-  { if ( !VirtualFree(addr, (ulong)s->max - (ulong)addr, MEM_DECOMMIT) )
+  { if ( !VirtualFree(addr, (uintptr_t)s->max - (uintptr_t)addr, MEM_DECOMMIT) )
       fatalError("Failed to unmap memory: %d", GetLastError());
     s->max = addr;
   }
@@ -1376,7 +1360,6 @@ static int
 allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
 { caddress lbase, gbase, tbase, abase;
   intptr_t glsize;
-  intptr_t lsep, tsep;
   SYSTEM_INFO info;
   intptr_t minglobal   = 4*SIZEOF_VOIDP K;
   intptr_t minlocal    = 2*SIZEOF_VOIDP K;
@@ -1386,13 +1369,6 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   GetSystemInfo(&info);
   size_alignment = info.dwPageSize;
 
-#ifdef NO_SEGV_HANDLING
-  lsep = tsep = 0;
-#else
-  lsep = STACK_SEPARATION;
-  tsep = size_alignment;
-#endif
-
   local    = max(local,    minlocal + STACK_SIGNAL);
   global   = max(global,   minglobal + STACK_SIGNAL);
   trail    = max(trail,    mintrail + STACK_SIGNAL);
@@ -1402,12 +1378,12 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   global   = (intptr_t) align_size(global);
   trail    = (intptr_t) align_size(trail);
   argument = (intptr_t) align_size(argument);
-  glsize   = global+tsep+local+lsep;
+  glsize   = global+local;
 
-  tbase = VirtualAlloc(NULL, trail+tsep,    MEM_RESERVE, PAGE_READWRITE);
-  abase = VirtualAlloc(NULL, argument+tsep, MEM_RESERVE, PAGE_READWRITE);
-  gbase = VirtualAlloc(NULL, glsize,        MEM_RESERVE, PAGE_READWRITE);
-  lbase = addPointer(gbase, global + tsep);
+  tbase = VirtualAlloc(NULL, trail,    MEM_RESERVE, PAGE_READWRITE);
+  abase = VirtualAlloc(NULL, argument, MEM_RESERVE, PAGE_READWRITE);
+  gbase = VirtualAlloc(NULL, glsize,   MEM_RESERVE, PAGE_READWRITE);
+  lbase = addPointer(gbase, global);
 
   if ( !tbase || !abase || !gbase )
   { if ( !tbase )
@@ -1430,7 +1406,7 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   INIT_STACK(trail,    "trail",    tbase, trail,    mintrail);
   INIT_STACK(argument, "argument", abase, argument, minargument);
 
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
   set_stack_guard_handler(SIGSEGV, _PL_segv_handler);
 #endif
 
@@ -1440,7 +1416,7 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
 
 static void
 freeStack(Stack s)
-{ VirtualFree(s->base, (ulong)s->max-(ulong)s->base, MEM_DECOMMIT);
+{ VirtualFree(s->base, (uintptr_t)s->max-(uintptr_t)s->base, MEM_DECOMMIT);
 }
 
 
@@ -1486,7 +1462,15 @@ expandStack(Stack s, void *addr)
 }
 #endif /*USE_SIGINFO*/
 
-#ifndef NO_SEGV_HANDLING
+#ifdef O_SEGV_HANDLING
+#warning "SEGV guarded stacks are no longer supported"
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This code is left here for future use,   but  it no longer works. If you
+want to make it working, please  remember   that  the  stacks may now be
+allocated without a free page at the end, so make sure to leave the last
+page blank in the signal handler.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 RETSIGTYPE
 #ifdef USE_SIGINFO
 _PL_segv_handler(int sig, siginfo_t *info, void *extra)
@@ -1518,7 +1502,7 @@ _PL_segv_handler(int sig)
 		    roomStack(global), roomStack(local), roomStack(trail)));
 
   for(i=0; i<N_STACKS; i++)
-  { intptr_t r = (ulong)stacka[i].max - (ulong)stacka[i].top;
+  { intptr_t r = (uintptr_t)stacka[i].max - (uintptr_t)stacka[i].top;
 
     if ( r < size_alignment )
     { DEBUG(1, Sdprintf("Mapped %s stack (free was %d)\n", stacka[i].name, r));
@@ -1533,19 +1517,19 @@ _PL_segv_handler(int sig)
   pl_signal_handler(sig);
 }
 
-#endif /*NO_SEGV_HANDLING*/
+#endif /*O_SEGV_HANDLING*/
 
 static void
-init_stack(Stack s, char *name, caddress base, intptr_t limit, intptr_t minsize)
+init_stack(Stack s, char *name, void *base, intptr_t limit, intptr_t minsize)
 { s->name       = name;
   s->base       = s->max = s->top = base;
   s->limit	= addPointer(base, limit);
-  s->min        = (caddress)((ulong)s->base + minsize);
+  s->min        = addPointer(base, minsize);
   s->gced_size  = 0L;			/* size after last gc */
   gcPolicy(s, GC_FAST_POLICY);
 
   DEBUG(1, Sdprintf("%-8s stack from 0x%08x to 0x%08x\n",
-		    s->name, (ulong)s->base, (ulong)s->limit));
+		    s->name, (uintptr_t)s->base, (uintptr_t)s->limit));
 
   while(s->max < s->min)
     mapOrOutOf(s);
@@ -1579,25 +1563,13 @@ gcPolicy(Stack s, int policy)
 }
 
 
-word
-pl_trim_stacks()
-{ trimStacks(PASS_LD1);
-
-  gcPolicy((Stack) &LD->stacks.global, GC_FAST_POLICY);
-  gcPolicy((Stack) &LD->stacks.trail,  GC_FAST_POLICY);
-
-  succeed;
-}
-
-
 #else /* O_DYNAMIC_STACKS */
 
+
 		/********************************
 		*    SIMPLE STACK ALLOCATION    *
 		*********************************/
 
-forwards void init_stack(Stack, char *, intptr_t, intptr_t, intptr_t);
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 On systems that do not allow us to get access to the MMU (or that do not
 have an MMU)  the  stacks  have  fixed  size  and  overflow  checks  are
@@ -1608,12 +1580,6 @@ why does IBM bring computers on the marked that are 10 years out-of-date
 at the moment of announcement?).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
-pl_trim_stacks()
-{ succeed;
-}
-
-
 word
 pl_stack_parameter(term_t name, term_t key, term_t old, term_t new)
 { atom_t a, k;
@@ -1650,36 +1616,35 @@ pl_stack_parameter(term_t name, term_t key, term_t old, term_t new)
 
 static void
 init_stack(Stack s, char *name, intptr_t size, intptr_t limit, intptr_t minfree)
-{ if ( s->base == NULL )
-  { fatalError("Not enough core to allocate stacks");
-    return;
-  }
-
-  s->name 	= name;
+{ s->name 	= name;
   s->top	= s->base;
   s->limit	= addPointer(s->base, limit);
+  s->max	= addPointer(s->base, size);
   s->minfree	= minfree;
-  s->max	= (char *)s->base + size;
-  s->gced_size = 0L;			/* size after last gc */
-  s->gc	       = ((s == (Stack) &LD->stacks.global ||
-		   s == (Stack) &LD->stacks.trail) ? TRUE : FALSE);
-  s->small     = (s->gc ? SMALLSTACK : 0);
+  s->gced_size  = 0L;			/* size after last gc */
+  s->gc	        = ((s == (Stack) &LD->stacks.global ||
+		    s == (Stack) &LD->stacks.trail) ? TRUE : FALSE);
+  s->small      = (s->gc ? SMALLSTACK : 0);
 }
 
-/* malloc() version */
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Malloc'ed version of the stacks. Our initial  stack size is a bit bigger
+than the minimum to avoid a stack-shift right at the start, We must find
+a better way to specify the minimum.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
 allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
-{ intptr_t old_heap = GD->statistics.heap;
-  intptr_t minglobal   = 25*SIZEOF_VOIDP K;
+{ intptr_t minglobal   = 25*SIZEOF_VOIDP K;
   intptr_t minlocal    = 4*SIZEOF_VOIDP K;
   intptr_t mintrail    = 4*SIZEOF_VOIDP K;
   intptr_t minargument = 1*SIZEOF_VOIDP K;
 
 #if O_SHIFT_STACKS
-  intptr_t itrail  = 8*SIZEOF_VOIDP K;
-  intptr_t iglobal = 50*SIZEOF_VOIDP K;
-  intptr_t ilocal  = 8*SIZEOF_VOIDP K;
+  intptr_t itrail  = nextStackSizeAbove(mintrail);
+  intptr_t iglobal = nextStackSizeAbove(minglobal);
+  intptr_t ilocal  = nextStackSizeAbove(minlocal);
 #else
   intptr_t itrail  = trail;
   intptr_t iglobal = global;
@@ -1691,12 +1656,15 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   trail    = max(trail,    mintrail);
   argument = max(argument, minargument);
 
-  gBase = (Word) PL_malloc(iglobal + sizeof(word) +
-			ilocal + sizeof(struct localFrame) +
-			MAXARITY * sizeof(word));
-  lBase = (LocalFrame)	addPointer(gBase, iglobal+sizeof(word));
-  tBase = (TrailEntry)	PL_malloc(itrail);
-  aBase = (Word *)	PL_malloc(argument);
+  gBase = (Word) malloc(iglobal + ilocal);
+  tBase = (TrailEntry) malloc(itrail);
+  aBase = (Word *)     malloc(argument);
+  if ( !gBase || !tBase || !aBase )
+  { freeStacks(LD);
+    fail;
+  }
+
+  lBase = (LocalFrame) addPointer(gBase, iglobal);
 
   init_stack((Stack)&LD->stacks.global,
 	     "global",   iglobal, global,  minglobal);
@@ -1707,17 +1675,20 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   init_stack((Stack)&LD->stacks.argument,
 	     "argument", argument, argument, minargument);
 
-  GD->statistics.heap = old_heap;
-
   succeed;
 }
 
 
 void
-freeStacks(PL_local_data_t *ld)
-{ PL_free(ld->stacks.global.base);
-  PL_free(ld->stacks.trail.base);
-  PL_free(ld->stacks.argument.base);
+freeStacks(ARG1_LD)
+{
+#undef LD
+#define LD LOCAL_LD
+  if ( gBase ) { free(gBase); gBase = NULL; lBase = NULL; }
+  if ( tBase ) { free(tBase); tBase = NULL; }
+  if ( aBase ) { free(aBase); aBase = NULL; }
+#undef LD
+#define LD GLOBAL_LD
 }
 
 
@@ -1741,16 +1712,24 @@ result from a stack-overflow.
 void
 trimStacks(ARG1_LD)
 {
-#ifdef O_DYNAMIC_STACKS
 #undef LD
 #define LD LOCAL_LD
   TrailEntry te;
   Word dummy = NULL;
 
+  LD->trim_stack_requested = FALSE;
+
+#ifdef O_SHIFT_STACKS
+  if ( !growStacks(NULL, NULL, NULL, -1, -1, -1) )
+    return;
+#else
+#ifdef O_DYNAMIC_STACKS
   unmap((Stack) &LD->stacks.local);
   unmap((Stack) &LD->stacks.global);
   unmap((Stack) &LD->stacks.trail);
   unmap((Stack) &LD->stacks.argument);
+#endif /*O_DYNAMIC_STACKS*/
+#endif /*O_SHIFT_STACKS*/
 
   LD->stacks.global.gced_size = usedStack(global);
   LD->stacks.trail.gced_size  = usedStack(trail);
@@ -1772,9 +1751,22 @@ trimStacks(ARG1_LD)
   }
 #undef LD
 #define LD GLOBAL_LD
-#endif /*O_DYNAMIC_STACKS*/
 }
 
+
+word
+pl_trim_stacks()
+{ trimStacks(PASS_LD1);
+
+#ifdef O_DYNAMIC_STACKS
+  gcPolicy((Stack) &LD->stacks.global, GC_FAST_POLICY);
+  gcPolicy((Stack) &LD->stacks.trail,  GC_FAST_POLICY);
+#endif
+
+  succeed;
+}
+
+
 		 /*******************************
 		 *	    LOCAL DATA		*
 		 *******************************/
@@ -1804,4 +1796,11 @@ freeLocalData(PL_local_data_t *ld)
   if ( ld->gvar.nb_vars )
     destroyHTable(ld->gvar.nb_vars);
 #endif
+
+#ifdef O_CYCLIC
+  clearSegStack(&ld->cycle.stack);
+#endif
+
+  if ( ld->arith.stack.base )
+    PL_free(ld->arith.stack.base);
 }
