@@ -1,11 +1,11 @@
-/*  $Id: pl-stream.c,v 1.143 2007/01/23 14:22:00 jan Exp $
+/*  $Id: pl-stream.c,v 1.163 2007/07/02 20:33:54 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -101,6 +101,9 @@ typedef wchar_t pl_wchar_t;
 #define MB_LEN_MAX 6
 #endif
 
+#define ROUND(p, n) ((((p) + (n) - 1) & ~((n) - 1)))
+#define UNDO_SIZE ROUND(MB_LEN_MAX, sizeof(wchar_t))
+
 #ifndef FALSE
 #define FALSE 0
 #endif
@@ -117,7 +120,7 @@ int Slinesize = SIO_LINESIZE;		/* Sgets() buffer size */
 static ssize_t	S__flushbuf(IOSTREAM *s);
 static void	run_close_hooks(IOSTREAM *s);
 static int	S__removebuf(IOSTREAM *s);
-
+static int	S__seterror(IOSTREAM *s);
 
 #ifdef O_PLMT
 #define SLOCK(s)    if ( s->mutex ) recursiveMutexLock(s->mutex)
@@ -149,10 +152,10 @@ extern IOENC			initEncoding(void);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Note that the  buffer  is  allocated   from  s->unbuffer,  which  starts
-MB_LEN_MAX before s->buffer, so we can always push-back a wide character
-into a multibyte stream. We do  not   do  this for SIO_USERBUF case, but
-this is only used by  the  output   stream  Svfprintf()  where it is not
-needed.
+UNDO_SIZE before s->buffer, so we can always push-back a wide
+character into a multibyte stream. We do not do this for SIO_USERBUF
+case, but this is only used by the output stream Svfprintf() where it is
+not needed.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
@@ -167,12 +170,12 @@ S__setbuf(IOSTREAM *s, char *buffer, int size)
   { s->unbuffer = s->buffer = buffer;
     s->flags |= SIO_USERBUF;
   } else
-  { if ( !(s->unbuffer = malloc(s->bufsize+MB_LEN_MAX)) )
+  { if ( !(s->unbuffer = malloc(s->bufsize+UNDO_SIZE)) )
     { errno = ENOMEM;
       return -1;
     }
     s->flags &= ~SIO_USERBUF;
-    s->buffer = s->unbuffer + MB_LEN_MAX;
+    s->buffer = s->unbuffer + UNDO_SIZE;
   }
 
   s->limitp = &s->buffer[s->bufsize];
@@ -182,6 +185,13 @@ S__setbuf(IOSTREAM *s, char *buffer, int size)
 }
 
 
+void
+Ssetbuffer(IOSTREAM *s, char *buffer, size_t size)
+{ S__setbuf(s, buffer, size);
+  s->flags &= ~SIO_USERBUF;
+}
+
+
 static int
 S__removebuf(IOSTREAM *s)
 { if ( s->buffer && s->unbuffer )
@@ -259,7 +269,7 @@ StryLock(IOSTREAM *s)
 { if ( !STRYLOCK(s) )
     return -1;
 
-  if ( --s->locks == 0 )
+  if ( !s->locks++ )
   { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
       return S__setbuf(s, NULL, TMPBUFSIZE);
   }
@@ -282,6 +292,8 @@ Sunlock(IOSTREAM *s)
     { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
 	rval = S__removebuf(s);
     }
+  } else
+  { assert(0);
   }
 
   SUNLOCK(s);
@@ -298,37 +310,42 @@ Sunlock(IOSTREAM *s)
 			 
 static ssize_t
 S__flushbuf(IOSTREAM *s)
-{ ssize_t size;
-  char *from = s->buffer;
+{ char *from, *to;
+  ssize_t rc;
 
-  while ( (size = s->bufp - from) > 0 )
-  { ssize_t n = (*s->functions->write)(s->handle, from, size);
+  SLOCK(s);
+  from = s->buffer;
+  to   = s->bufp;
+
+  while ( from < to )
+  { size_t size = (size_t)(to - from);
+    ssize_t n = (*s->functions->write)(s->handle, from, size);
 
     if ( n > 0 )			/* wrote some */
     { from += n;
     } else if ( n < 0 )			/* error */
-    { s->flags |= SIO_FERR;
-      return -1;
+    { S__seterror(s);
+      rc = -1;
+      goto out;
     } else				/* wrote nothing? */
     { break;
     }
   }
 
-  if ( s->bufp - from == 0 )		/* full flush */
-  { ssize_t rc = s->bufp - s->buffer;
-    
+  if ( to == from )			/* full flush */
+  { rc = s->bufp - s->buffer;
     s->bufp = s->buffer;
-
-    return rc;
   } else				/* partial flush */
-  { ssize_t rc = from - s->buffer;
-    size_t left = s->bufp - from;
+  { size_t left = to - from;
 
+    rc = from - s->buffer;
     memmove(s->buffer, from, left);
     s->bufp = s->buffer + left;
-
-    return rc;
   }
+
+out:
+  SUNLOCK(s);
+  return rc;
 }
 
 
@@ -371,7 +388,7 @@ S__fillbuf(IOSTREAM *s)
 
   if ( s->flags & (SIO_FEOF|SIO_FERR) )
   { s->flags |= SIO_FEOF2;		/* reading past eof */
-    goto error;
+    return -1;
   }
 
 #ifdef HAVE_SELECT
@@ -400,7 +417,7 @@ S__fillbuf(IOSTREAM *s)
 	if ( rc < 0 && errno == EINTR )
 	{ if ( PL_handle_signals() < 0 )
 	  { errno = EPLEXCEPTION;
-	    goto error;
+	    return -1;
 	  }
 
 	  continue;
@@ -411,12 +428,12 @@ S__fillbuf(IOSTREAM *s)
 
       if ( rc == 0 )
       { s->flags |= (SIO_TIMEOUT|SIO_FERR);
-	goto error;
+	return -1;
       }
     } else
     { errno = EPERM;			/* no permission to select */
       s->flags |= SIO_FERR;
-      goto error;
+      return -1;
     }
   }
 #endif
@@ -428,21 +445,22 @@ S__fillbuf(IOSTREAM *s)
 
     if ( (n=(*s->functions->read)(s->handle, &chr, 1)) == 1 )
     { c = char_to_int(chr);
-      goto ok;
+      return c;
     } else if ( n == 0 )
     { if ( !(s->flags & SIO_NOFEOF) )
 	s->flags |= SIO_FEOF;
-      goto error;
+      return -1;
     } else
-    { s->flags |= SIO_FERR;
-      goto error;			/* error */
+    { S__seterror(s);
+      return -1;
     }
   } else
-  { size_t n, len;
+  { ssize_t n;
+    size_t len;
 
     if ( !s->buffer )
     { if ( S__setbuf(s, NULL, 0) < 0 )
-	goto error;
+	return -1;
       s->bufp = s->limitp = s->buffer;
       len = s->bufsize;
     } else if ( s->bufp < s->limitp )
@@ -459,29 +477,24 @@ S__fillbuf(IOSTREAM *s)
     if ( (n=(*s->functions->read)(s->handle, s->limitp, len)) > 0 )
     { s->limitp += n;
       c = char_to_int(*s->bufp++);
-      goto ok;
+      return c;
     } else
     { if ( n == 0 )
       { if ( !(s->flags & SIO_NOFEOF) )
 	  s->flags |= SIO_FEOF;
-	goto error;
+	return -1;
 #ifdef EWOULDBLOCK
       } else if ( errno == EWOULDBLOCK )
       { s->bufp = s->buffer;
 	s->limitp = s->buffer;
-	goto error;
+	return -1;
 #endif
       } else
-      { s->flags |= SIO_FERR;
-	goto error;
+      { S__seterror(s);
+	return -1;
       }
     }
   }
-
-error:
-  c = -1;
-ok:
-  return c;
 }
 
 		 /*******************************
@@ -493,6 +506,11 @@ static inline void
 update_linepos(IOSTREAM *s, int c)
 { IOPOS *p = s->position;
 
+  if ( c > '\r' )			/* speedup the 99% case a bit */
+  { p->linepos++;
+    return;
+  }
+
   switch(c)
   { case '\n':
       p->lineno++;
@@ -870,12 +888,18 @@ retry:
       { int extra = UTF8_FBN(c);
 	int code;
 
+	if ( extra < 0 )
+	{ Sseterr(s, SIO_WARN, "Illegal UTF-8 start");
+	  c = UTF8_MALFORMED_REPLACEMENT;
+	  goto out;
+	}
+
 	code = UTF8_FBV(c,extra);
 	for( ; extra > 0; extra-- )
 	{ int c2 = get_byte(s);
 	  
 	  if ( !ISUTF8_CB(c2) )
-	  { Sseterr(s, SIO_WARN, "Illegal UTF-8 Sequence");
+	  { Sseterr(s, SIO_WARN, "Illegal UTF-8 continuation");
 	    c = UTF8_MALFORMED_REPLACEMENT;
 	    Sungetc(c2, s);
 	    goto out;
@@ -1268,6 +1292,24 @@ Sfeof(IOSTREAM *s)
 }
     
 
+static int
+S__seterror(IOSTREAM *s)
+{ if ( s->functions->control )
+  { char *msg;
+
+    if ( (*s->functions->control)(s->handle,
+				  SIO_LASTERROR, 
+				  (void *)&msg) == 0 )
+    { Sseterr(s, SIO_FERR, msg);
+      return 0;
+    } 
+  }
+
+  s->flags |= SIO_FERR;
+  return 0;
+}
+
+
 int
 Sferror(IOSTREAM *s)
 { return (s->flags & SIO_FERR) != 0;
@@ -1350,7 +1392,7 @@ Sflush(IOSTREAM *s)
 		 *	      SEEK		*
 		 *******************************/
 
-static int
+int
 Sunit_size(IOSTREAM *s)
 { switch(s->encoding)
   { case ENC_UNKNOWN:
@@ -1403,6 +1445,10 @@ Ssize(IOSTREAM *s)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Sseek64(IOSTREAM *s, int64_t pos, int whence)
+
+Re-position the stream to byte-no 'pos'.
+
 Maybe we should optimise this to become block-aligned?  Or can we leave
 this to read/write?
 
@@ -1416,22 +1462,22 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
   { int64_t now = Stell64(s);
 
     if ( now != -1 )
-    { int64_t rval;
+    { int64_t newpos;
       char *nbufp = (char *)-1;
     
       if ( whence == SIO_SEEK_CUR )
-      { nbufp = s->bufp + pos*Sunit_size(s);
-	rval = now + pos;
+      { nbufp = s->bufp + pos;
+	newpos = now + pos;
       } else if ( whence == SIO_SEEK_SET )
-      { nbufp = s->bufp + (pos - now)*Sunit_size(s);
-	rval = pos;
+      { nbufp = s->bufp + (pos - now);
+	newpos = pos;
       } else
-	rval = -1;			/* should not happen */
+	newpos = -1;			/* should not happen */
 
       if ( nbufp >= s->buffer && nbufp < s->limitp )
       { s->bufp = nbufp;
 
-	pos = rval;
+	pos = newpos;
 	goto update;
       }
     }
@@ -1453,7 +1499,6 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
     whence = SIO_SEEK_SET;
   }
   
-  pos *= Sunit_size(s);
   if ( s->functions->seek64 )
     pos = (*s->functions->seek64)(s->handle, pos, whence);
   else if ( pos <= LONG_MAX )
@@ -1462,14 +1507,19 @@ Sseek64(IOSTREAM *s, int64_t pos, int whence)
   { errno = EINVAL;
     return -1;
   }
-  pos /= Sunit_size(s);
+  
+  if ( pos < 0 )
+  { errno = EINVAL;
+    return -1;
+  }
 
 update:
   s->flags &= ~SIO_FEOF;		/* not on eof of file anymore */
 
   if ( s->position )
   { s->flags |= (SIO_NOLINENO|SIO_NOLINEPOS); /* no update this */
-    s->position->charno = pos;
+    s->position->byteno = pos;
+    s->position->charno = pos/Sunit_size(s); /* compatibility */
   }
 
   return 0;
@@ -1483,10 +1533,14 @@ Sseek(IOSTREAM *s, long pos, int whence)
 
 
 
+/* Stell64(IOSTREAM *s) returns the current position in the file in
+   bytes
+*/
+
 int64_t
 Stell64(IOSTREAM *s)
 { if ( s->position )
-  { return s->position->charno;
+  { return s->position->byteno;
   } else if ( s->functions->seek || s->functions->seek64 )
   { int64_t pos;
 
@@ -1494,7 +1548,6 @@ Stell64(IOSTREAM *s)
       pos = (*s->functions->seek64)(s->handle, 0L, SIO_SEEK_CUR);
     else
       pos = (*s->functions->seek)(s->handle, 0L, SIO_SEEK_CUR);
-    pos /= Sunit_size(s);
 
     if ( s->buffer )			/* open */
     { int64_t off = s->bufp - s->buffer;
@@ -1502,7 +1555,7 @@ Stell64(IOSTREAM *s)
       if ( s->flags & SIO_INPUT )
 	off -= s->limitp - s->buffer;
 
-      pos += off/Sunit_size(s);
+      pos += off;
     }
 
     return pos;
@@ -1548,15 +1601,17 @@ Sclose(IOSTREAM *s)
   if ( (s->flags&SIO_CLOSING) )		/* recursive (*) */
     return rval;
 
-  SLOCK(s);
-  while(s->locks > 0)			/* remove buffer-locks */
-    rval = Sunlock(s);
+  if ( s->upstream )
+  { errno = EPERM;
+    return -1;
+  }
 
+  SLOCK(s);
+  s->flags |= SIO_CLOSING;
   rval = S__removebuf(s);
   if ( s->mbstate )
     free(s->mbstate);
 
-  s->flags |= SIO_CLOSING;
 #ifdef __WINDOWS__
   if ( (s->flags & SIO_ADVLOCK) )
   { OVERLAPPED ov;
@@ -1570,6 +1625,8 @@ Sclose(IOSTREAM *s)
   if ( s->functions->close && (*s->functions->close)(s->handle) < 0 )
     rval = -1;
   run_close_hooks(s);
+  while(s->locks > 0)			/* remove buffer-locks */
+    rval = Sunlock(s);
 
   SUNLOCK(s);
 
@@ -1786,8 +1843,20 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	{ case 'c':
 	    *fe++ = va_arg(args, int);
 	    break;
-	  case 'd':
 	  case 'p':
+	  { void *ptr = va_arg(args, void*);
+	    char fmbuf[8], *fp=fmbuf;
+	    *fp++ = '%';
+	    if ( modified )
+	      *fp++ = '#';
+	    *fp++ = 'p';
+	    *fp   = '\0';
+	    sprintf(fs, fmbuf, ptr);
+	    fe = &fs[strlen(fs)];
+
+	    break;
+	  }
+	  case 'd':
 	  case 'i':
 	  case 'o':
 	  case 'u':
@@ -1893,10 +1962,12 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    if ( utf8 )
 	      w = utf8_strlen(fs, w);
 
-	    w = arg1 - w;
-	    while(w > 0 )
-	    { OUTCHR(s, pad);
-	      w--;
+	    if ( (ssize_t)w < arg1 )
+	    { w = arg1 - w;
+	      while(w > 0 )
+	      { OUTCHR(s, pad);
+		w--;
+	      }
 	    }
 	    while(*fs)
 	    { int c;
@@ -1959,16 +2030,12 @@ Svsprintf(char *buf, const char *fm, va_list args)
 { IOSTREAM s;
   int rval;
 
+  memset(&s, 0, sizeof(s));
   s.bufp      = buf;
   s.limitp    = (char *)(~0L);
   s.buffer    = buf;
   s.flags     = SIO_FBUF|SIO_OUTPUT;
-  s.position  = NULL;
-  s.handle    = NULL;
-  s.functions = NULL;
-  s.mutex     = NULL;
   s.encoding  = ENC_ISO_LATIN_1;
-  s.tee	      = NULL;
   
   if ( (rval = Svfprintf(&s, fm, args)) >= 0 )
     *s.bufp = '\0';
@@ -1982,15 +2049,16 @@ Svdprintf(const char *fm, va_list args)
 { int rval;
   IOSTREAM *s = Soutput;
 
-  SLOCK(s);
+  Slock(s);
   rval = Svfprintf(s, fm, args);
 #if defined(_DEBUG) && defined(__WINDOWS__)
   Sputc('\0', s);
   s->bufp--;				/* `Unput' */
   OutputDebugString(s->buffer);
 #endif
-  Sflush(s);
-  SUNLOCK(s);
+  if ( Sflush(s) != 0 )
+    rval = -1;
+  Sunlock(s);
 
   return rval;
 }
@@ -2279,6 +2347,42 @@ out:
 #endif /*0*/
 
 
+		 /*******************************
+		 *	   FILTER STREAMS	*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Link two streams in a pipeline,  where   filter  filters data for stream
+`parent'. If parent is an output steam we have
+
+	application --> filter --> parent -->
+	  
+If parent is an input stream we have
+
+	--> parent --> filter --> application
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+Sset_filter(IOSTREAM *parent, IOSTREAM *filter)
+{ if ( !parent || parent->magic != SIO_MAGIC )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  if ( filter )
+  { if ( filter->magic != SIO_MAGIC )
+    { errno = EINVAL;
+      return -1;
+    }
+  }
+
+  parent->upstream = filter;
+  if ( filter )
+    filter->downstream = parent;
+
+  return 0;
+}
+
 
 		 /*******************************
 		 *	    FILE STREAMS	*
@@ -2599,7 +2703,7 @@ Sfdopen(int fd, const char *type)
   return Snew((void *)lfd, flags, &Sfilefunctions);
 }
 
-/* MT: as intptr_t as s is valid, this should be ok
+/* MT: as long as s is valid, this should be ok
 */
 
 int
@@ -2625,60 +2729,6 @@ Sfileno(IOSTREAM *s)
 }
 
 
-#define STDIO(n, f) { NULL, NULL, NULL, NULL, \
-		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
-		      ((void *)(n)), &Sfilefunctions, \
-		      0, NULL, \
-		      (void (*)(void *))0, NULL, \
-		      -1, \
-		      0, \
-		      ENC_ISO_LATIN_1 \
-		    }
-
-#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
-#define STDIO_STREAMS \
-  STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
-  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPL), 	/* Soutput */ \
-  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT|SIO_REPPL)	/* Serror */
-
-
-IOSTREAM S__iob[] =
-{ STDIO_STREAMS
-};
-
-
-static const IOSTREAM S__iob0[] =
-{ STDIO_STREAMS
-};
-
-
-void
-SinitStreams()
-{ static int done;
-
-  if ( !done++ )
-  { int i;
-    IOENC enc = initEncoding();
-
-    for(i=0; i<=2; i++)
-    { if ( !isatty(i) )
-	S__iob[i].flags &= ~SIO_ISATTY;
-      if ( S__iob[i].encoding == ENC_ISO_LATIN_1 )
-	S__iob[i].encoding = enc;
-#ifdef O_PLMT
-      S__iob[i].mutex = malloc(sizeof(recursiveMutex));
-      recursiveMutexInit(S__iob[i].mutex);
-#endif
-    }
-  }
-}
-
-
-IOSTREAM *
-S__getiob()
-{ return S__iob;
-}
-
 		 /*******************************
 		 *	       PIPES		*
 		 *******************************/
@@ -3070,6 +3120,68 @@ Sopen_string(IOSTREAM *s, char *buf, size_t size, const char *mode)
   return s;
 }
 
+		 /*******************************
+		 *	 STANDARD HANDLES	*
+		 *******************************/
+
+#define STDIO(n, f) { NULL, NULL, NULL, NULL, \
+		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
+		      ((void *)(n)), &Sfilefunctions, \
+		      0, NULL, \
+		      (void (*)(void *))0, NULL, \
+		      -1, \
+		      0, \
+		      ENC_ISO_LATIN_1 \
+		    }
+
+#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
+#define STDIO_STREAMS \
+  STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
+  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPL), 	/* Soutput */ \
+  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT|SIO_REPPL)	/* Serror */
+
+
+IOSTREAM S__iob[] =
+{ STDIO_STREAMS
+};
+
+
+static const IOSTREAM S__iob0[] =
+{ STDIO_STREAMS
+};
+
+
+void
+SinitStreams()
+{ static int done;
+
+  if ( !done++ )
+  { int i;
+    IOENC enc = initEncoding();
+
+    for(i=0; i<=2; i++)
+    { if ( !isatty(i) )
+	S__iob[i].flags &= ~SIO_ISATTY;
+      if ( S__iob[i].encoding == ENC_ISO_LATIN_1 )
+	S__iob[i].encoding = enc;
+#ifdef O_PLMT
+      S__iob[i].mutex = malloc(sizeof(recursiveMutex));
+      recursiveMutexInit(S__iob[i].mutex);
+#endif
+#ifdef __WINDOWS__
+      pt_init();
+#endif
+    }
+  }
+}
+
+
+IOSTREAM *
+S__getiob()
+{ return S__iob;
+}
+
+
 		 /*******************************
 		 *	       HOOKS		*
 		 *******************************/
