@@ -1,11 +1,11 @@
-/*  $Id: pl-stream.h,v 1.42 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-stream.h,v 1.77 2007/06/16 20:38:03 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        jan@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -25,7 +25,27 @@
 #ifndef _PL_STREAM_H
 #define _PL_STREAM_H
 
+/* This appears to make the wide-character support compile and work
+   on HPUX 11.23.  There really should be a cleaner way ...
+*/
+#if defined(__hpux)
+#include <sys/_mbstate_t.h>
+#endif
+
 #include <stdarg.h>
+#include <wchar.h>
+#include <stddef.h>
+#ifdef __WINDOWS__
+typedef __int64 int64_t;
+#if (_MSC_VER < 1300)
+typedef long intptr_t;
+typedef unsigned long uintptr_t;
+#endif
+typedef intptr_t ssize_t;		/* signed version of size_t */
+#else
+#include <unistd.h>
+#include <inttypes.h>			/* more portable than stdint.h */
+#endif
 
 #ifdef __cplusplus
 extern "C" {
@@ -42,11 +62,8 @@ stuff.
 
 #ifndef _PL_EXPORT_DONE
 #define _PL_EXPORT_DONE
-#if defined(WIN32) && !defined(__WIN32__)
-#define __WIN32__
-#endif
 
-#if (defined(__WIN32__) || defined(__CYGWIN32__)) && !defined(__LCC__)
+#if (defined(__WINDOWS__) || defined(__CYGWIN__)) && !defined(__LCC__)
 #define HAVE_DECLSPEC
 #endif
 
@@ -84,7 +101,7 @@ stuff.
 #define NULL ((void *)0)
 #endif
 
-#if defined(__WIN32__) && !defined(EWOULDBLOCK)
+#if defined(__WINDOWS__) && !defined(EWOULDBLOCK)
 #define EWOULDBLOCK	1000		/* Needed for socket handling */
 #endif
 #define EPLEXCEPTION	1001		/* errno: pending Prolog exception */
@@ -94,9 +111,10 @@ stuff.
 #define SIO_MAGIC	(7212676)	/* magic number */
 #define SIO_CMAGIC	(42)		/* we are close (and thus illegal!) */
 
-typedef int   (*Sread_function)(void *handle, char *buf, int bufsize);
-typedef int   (*Swrite_function)(void *handle, char*buf, int bufsize);
+typedef ssize_t (*Sread_function)(void *handle, char *buf, size_t bufsize);
+typedef ssize_t (*Swrite_function)(void *handle, char*buf, size_t bufsize);
 typedef long  (*Sseek_function)(void *handle, long pos, int whence);
+typedef int64_t (*Sseek64_function)(void *handle, int64_t pos, int whence);
 typedef int   (*Sclose_function)(void *handle);
 typedef int   (*Scontrol_function)(void *handle, int action, void *arg);
 
@@ -113,14 +131,31 @@ typedef struct io_functions
   Sseek_function	seek;		/* seek to position */
   Sclose_function	close;		/* close stream */
   Scontrol_function	control;	/* Info/control */
+  Sseek64_function	seek64;		/* seek to position (intptr_t files) */
 } IOFUNCTIONS;
 
 typedef struct io_position
-{ long			charno;		/* character position in file */
+{ int64_t		byteno;		/* byte-position in file */
+  int64_t		charno;		/* character position in file */
   int			lineno;		/* lineno in file */
   int			linepos;	/* position in line */
+  intptr_t		reserved[2];	/* future extensions */
 } IOPOS;
 
+					/* NOTE: check with encoding_names */
+					/* in pl-file.c */
+typedef enum
+{ ENC_UNKNOWN = 0,			/* invalid/unknown */
+  ENC_OCTET,				/* raw 8 bit input */
+  ENC_ASCII,				/* US-ASCII (0..127) */
+  ENC_ISO_LATIN_1,			/* ISO Latin-1 (0..256) */
+  ENC_ANSI,				/* default (multibyte) codepage */
+  ENC_UTF8,
+  ENC_UNICODE_BE,			/* big endian unicode file */
+  ENC_UNICODE_LE,			/* little endian unicode file */
+  ENC_WCHAR				/* pl_wchar_t */
+} IOENC;
+
 typedef struct io_stream
 { char		       *bufp;		/* `here' */
   char		       *limitp;		/* read/write limit */
@@ -141,8 +176,17 @@ typedef struct io_stream
   void *		closure;
 					/* SWI-Prolog 5.1.3 */
   int			timeout;	/* timeout (milliseconds) */
+					/* SWI-Prolog 5.4.4 */
+  char *		message;	/* error/warning message */
+  IOENC			encoding;	/* character encoding used */
+  struct io_stream *	tee;		/* copy data to this stream */
+  mbstate_t *		mbstate;	/* ENC_ANSI decoding */
+  struct io_stream *	upstream;	/* stream providing our input */
+  struct io_stream *	downstream;	/* stream providing our output */
+  intptr_t		reserved[4];	/* reserved for extension */
 } IOSTREAM;
 
+
 #define SmakeFlag(n)	(1<<(n-1))
 
 #define SIO_FBUF	SmakeFlag(1)	/* full buffering */
@@ -171,6 +215,11 @@ typedef struct io_stream
 #define SIO_TIMEOUT	SmakeFlag(24)	/* We had a timeout */
 #define SIO_NOMUTEX	SmakeFlag(25)	/* Do not allow multi-thread access */
 #define SIO_ADVLOCK	SmakeFlag(26)	/* File locked with advisory lock */
+#define SIO_WARN	SmakeFlag(27)	/* Pending warning */
+#define SIO_CLEARERR	SmakeFlag(28)	/* Clear error after reporting */
+#define SIO_REPXML	SmakeFlag(29)	/* Bad char --> XML entity */
+#define SIO_REPPL	SmakeFlag(30)	/* Bad char --> Prolog \hex\ */
+#define SIO_BOM		SmakeFlag(31)	/* BOM was detected/written */
 
 #define	SIO_SEEK_SET	0	/* From beginning of file.  */
 #define	SIO_SEEK_CUR	1	/* From current position.  */
@@ -180,7 +229,7 @@ PL_EXPORT(IOSTREAM *)	S__getiob(void);	/* get DLL's __iob[] address */
 
 PL_EXPORT_DATA(IOFUNCTIONS)  Sfilefunctions;	/* OS file functions */
 PL_EXPORT_DATA(int)	     Slinesize;		/* Sgets() linesize */
-#if defined(__CYGWIN32__) && !defined(PL_KERNEL)
+#if defined(__CYGWIN__) && !defined(PL_KERNEL)
 #define S__iob S__getiob()
 #else
 PL_EXPORT_DATA(IOSTREAM)    S__iob[3];		/* Libs standard streams */
@@ -193,17 +242,20 @@ PL_EXPORT_DATA(IOSTREAM)    S__iob[3];		/* Libs standard streams */
 #define Sgetchar()	Sgetc(Sinput)
 #define Sputchar(c)	Sputc((c), Soutput)
 
-#define S__updatefilepos(s, c) \
-	((s)->position ? S__fupdatefilepos((s), (c)) \
+#define S__updatefilepos_getc(s, c) \
+	((s)->position ? S__fupdatefilepos_getc((s), (c)) \
 		       : (c))
 
 #define Snpgetc(s) ((s)->bufp < (s)->limitp ? (int)(*(s)->bufp++)&0xff \
 					    : S__fillbuf(s))
-#define Sgetc(s) S__updatefilepos((s), Snpgetc(s))
+#define Sgetc(s) S__updatefilepos_getc((s), Snpgetc(s))
 
 /* Control-operations */
 #define SIO_GETSIZE	(1)		/* get size of underlying object */
 #define SIO_GETFILENO	(2)		/* get underlying file (if any) */
+#define SIO_SETENCODING	(3)		/* modify encoding of stream */
+#define SIO_FLUSHOUTPUT	(4)		/* flush output */
+#define SIO_LASTERROR	(5)		/* string holding last error */
 
 /* Sread_pending() */
 #define SIO_RP_BLOCK 0x1		/* wait for new input */
@@ -269,29 +321,42 @@ PL_EXPORT_DATA(IOSTREAM)    S__iob[3];		/* Libs standard streams */
 
 PL_EXPORT(void)		SinitStreams();
 PL_EXPORT(void)		Scleanup(void);
+PL_EXPORT(void)		Sreset(void);
 PL_EXPORT(int)		S__fupdatefilepos(IOSTREAM *s, int c);
+PL_EXPORT(int)		S__fupdatefilepos_getc(IOSTREAM *s, int c);
 PL_EXPORT(int)		S__fillbuf(IOSTREAM *s);
+PL_EXPORT(int)		Sunit_size(IOSTREAM *s);
+					/* byte I/O */
 PL_EXPORT(int)		Sputc(int c, IOSTREAM *s);
 PL_EXPORT(int)		Sfgetc(IOSTREAM *s);
 PL_EXPORT(int)		Sungetc(int c, IOSTREAM *s);
+					/* multibyte I/O */
+PL_EXPORT(int)		Scanrepresent(int c, IOSTREAM *s);
+PL_EXPORT(int)		Sputcode(int c, IOSTREAM *s);
+PL_EXPORT(int)		Sgetcode(IOSTREAM *s);
+PL_EXPORT(int)		Sungetcode(int c, IOSTREAM *s);
+					/* word I/O */
 PL_EXPORT(int)		Sputw(int w, IOSTREAM *s);
 PL_EXPORT(int)		Sgetw(IOSTREAM *s);
-PL_EXPORT(int)		Sfread(void *data, int size, int elems, IOSTREAM *s);
-PL_EXPORT(int)		Sfwrite(const void *data, int size, int elems,
+PL_EXPORT(size_t)	Sfread(void *data, size_t size, size_t elems,
+			       IOSTREAM *s);
+PL_EXPORT(size_t)	Sfwrite(const void *data, size_t size, size_t elems,
 				IOSTREAM *s);
 PL_EXPORT(int)		Sfeof(IOSTREAM *s);
 PL_EXPORT(int)		Sfpasteof(IOSTREAM *s);
 PL_EXPORT(int)		Sferror(IOSTREAM *s);
 PL_EXPORT(void)		Sclearerr(IOSTREAM *s);
+PL_EXPORT(void)		Sseterr(IOSTREAM *s, int which, const char *message);
+PL_EXPORT(int)		Ssetenc(IOSTREAM *s, IOENC new_enc, IOENC *old_enc);
 PL_EXPORT(int)		Sflush(IOSTREAM *s);
 PL_EXPORT(long)		Ssize(IOSTREAM *s);
-PL_EXPORT(long)		Sseek(IOSTREAM *s, long pos, int whence);
+PL_EXPORT(int)		Sseek(IOSTREAM *s, long pos, int whence);
 PL_EXPORT(long)		Stell(IOSTREAM *s);
 PL_EXPORT(int)		Sclose(IOSTREAM *s);
 PL_EXPORT(char *)	Sfgets(char *buf, int n, IOSTREAM *s);
 PL_EXPORT(char *)	Sgets(char *buf);
-PL_EXPORT(int)		Sread_pending(IOSTREAM *s,
-				      char *buf, int limit, int flags);
+PL_EXPORT(ssize_t)	Sread_pending(IOSTREAM *s,
+				      char *buf, size_t limit, int flags);
 PL_EXPORT(int)		Sfputs(const char *q, IOSTREAM *s);
 PL_EXPORT(int)		Sputs(const char *q);
 PL_EXPORT(int)		Sfprintf(IOSTREAM *s, const char *fm, ...);
@@ -310,10 +375,18 @@ PL_EXPORT(IOSTREAM *)	Sopen_file(const char *path, const char *how);
 PL_EXPORT(IOSTREAM *)	Sfdopen(int fd, const char *type);
 PL_EXPORT(int)	   	Sfileno(IOSTREAM *s);
 PL_EXPORT(IOSTREAM *)	Sopen_pipe(const char *command, const char *type);
-PL_EXPORT(IOSTREAM *)	Sopenmem(char **buffer, int *sizep, const char *mode);
-PL_EXPORT(IOSTREAM *)	Sopen_string(IOSTREAM *s, char *buf, int sz, const char *m);
+PL_EXPORT(IOSTREAM *)	Sopenmem(char **buffer, size_t *sizep, const char *mode);
+PL_EXPORT(IOSTREAM *)	Sopen_string(IOSTREAM *s, char *buf, size_t sz, const char *m);
 PL_EXPORT(int)		Sclosehook(void (*hook)(IOSTREAM *s));
 PL_EXPORT(void)		Sfree(void *ptr);
+PL_EXPORT(int)		Sset_filter(IOSTREAM *parent, IOSTREAM *filter);
+PL_EXPORT(void)		Ssetbuffer(IOSTREAM *s, char *buf, size_t size);
+
+PL_EXPORT(int64_t)	Stell64(IOSTREAM *s);
+PL_EXPORT(int)		Sseek64(IOSTREAM *s, int64_t pos, int whence);
+
+PL_EXPORT(int)		ScheckBOM(IOSTREAM *s);
+PL_EXPORT(int)		SwriteBOM(IOSTREAM *s);
 
 #ifdef __cplusplus
 }
