@@ -1,4 +1,4 @@
-/*  $Id: pl-table.c,v 1.25 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-table.c,v 1.27 2007/01/13 10:30:37 jan Exp $
 
     Part of SWI-Prolog
 
@@ -41,7 +41,7 @@ create, advance over and destroy enumerator   objects. These objects are
 used to enumerate the symbols of these   tables,  used primarily for the
 pl_current_* predicates.
 
-The  enumerators  cause  two  things:  (1)    as  long  enumerators  are
+The  enumerators  cause  two  things:  (1)    as  intptr_t  enumerators  are
 associated, the table will not  be  rehashed   and  (2)  if  symbols are
 deleted  that  are  referenced  by  an  enumerator,  the  enumerator  is
 automatically advanced to the next free symbol.  This, in general, makes
@@ -184,7 +184,7 @@ rehashHTable(Table ht)
   { Symbol s, n;
 
     for(s=oldtab[i]; s; s = n)
-    { int v = pointerHashValue(s->name, ht->buckets);
+    { int v = (int)pointerHashValue(s->name, ht->buckets);
 
       n = s->next;
       s->next = ht->entries[v];
@@ -203,7 +203,7 @@ addHTable(Table ht, void *name, void *value)
   int v;
 
   LOCK_TABLE(ht);
-  v = pointerHashValue(name, ht->buckets);
+  v = (int)pointerHashValue(name, ht->buckets);
   if ( lookupHTable(ht, name) )
   { UNLOCK_TABLE(ht);
     return NULL;
@@ -237,7 +237,7 @@ deleteSymbolHTable(Table ht, Symbol s)
   TableEnum e;
 
   LOCK_TABLE(ht);
-  v = pointerHashValue(s->name, ht->buckets);
+  v = (int)pointerHashValue(s->name, ht->buckets);
   h = &ht->entries[v];
 
   for( e=ht->enumerators; e; e = e->next )
