@@ -1,11 +1,11 @@
-/*  $Id: pl-tai.c,v 1.21 2007/01/13 12:49:08 jan Exp $
+/*  $Id: pl-tai.c,v 1.27 2007/07/05 12:22:32 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -27,7 +27,7 @@ Solaris has asctime_r() with 3 arguments. Using _POSIX_PTHREAD_SEMANTICS
 is supposed to give the POSIX standard one.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef __sun__
+#if defined(__sun__) || defined(__sun)
 #define _POSIX_PTHREAD_SEMANTICS 1
 #endif
 
@@ -37,7 +37,7 @@ is supposed to give the POSIX standard one.
 #include "libtai/caltime.h"
 #include <stdio.h>
 
-#ifdef __WINDOWS__
+#if defined(__WINDOWS__) || defined (__CYGWIN__)
 #define timezone _timezone
 #define HAVE_VAR_TIMEZONE
 #else
@@ -83,7 +83,7 @@ do_tzset()
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-POSIX provides the avriable  timezone,  providing   the  offset  of  the
+POSIX provides the variable  timezone,  providing   the  offset  of  the
 current timezone WEST of GMT in seconds.   Some systems (FreeBSD) do not
 provide that. Instead thet provide  tm_gmtoff   in  struct  tm, but this
 value is EAST and includes the DST offset.
@@ -135,8 +135,15 @@ tz_name_as_atom(int dst)
   dst = (dst != 0);			/* 0 or 1 */
 
   if ( !a[dst] )
-  { do_tzset();
-    a[dst] = PL_new_atom(tzname[dst]);
+  { wchar_t wbuf[256];
+    const char *str = tz_name(dst);
+    size_t n;
+
+    if ( (n = mbstowcs(wbuf, str, sizeof(wbuf)/sizeof(wbuf[0])-1)) != (size_t)-1 )
+    { a[dst] = PL_new_atom_wchars(n, wbuf);
+    } else
+    { a[dst] = PL_new_atom(str);
+    }
   }
 
   return a[dst];
@@ -481,7 +488,7 @@ fmt_not_implemented(const char *key)
 	{ Sfprintf(fd, fmt, val); \
 	}
 #define SUBFORMAT(f) \
-	{ format_time(fd, f, ftm); \
+	{ format_time(fd, f, ftm, posix); \
 	}
 #define OUTCHR(fd, c) \
  	{ Sputcode(c, fd); \
@@ -489,9 +496,44 @@ fmt_not_implemented(const char *key)
 #define OUTSTR(str) \
 	{ Sfputs(str, fd); \
 	}
+#define OUTSTRA(str) \
+	{ foutstra(str, fd); \
+	}
+#define OUTATOM(a) \
+	{ writeAtomToStream(fd, a); \
+	}
+
+static void
+foutstra(const char *str, IOSTREAM *fd)
+{ wchar_t wbuf[256];
+  size_t n;
+
+  if ( (n = mbstowcs(wbuf, str, sizeof(wbuf)/sizeof(wbuf[0])-1)) != (size_t)-1 ) 
+  { wchar_t *p;
+
+    for(p=wbuf; n-- > 0; p++)
+      Sputcode(*p, fd);
+  }
+}
+
+
+static const char *abbred_weekday[] =
+{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
+static const char *weekday[] =
+{ "Sunday", "Monday", "Tuesday", "Wednesday",
+  "Thursday", "Friday", "Satuerday" };
+static const char *abbred_month[] =
+{ "Jan", "Feb", "Mar", "Apr", "May", "Jun",
+  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
+};
+static const char *month[] =
+{ "January", "February", "March", "April", "May", "June",
+  "July", "August", "September", "October", "November", "December"
+};
+
 
 static int
-format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
+format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 { wint_t c;
 
   while((c = *format++))
@@ -502,6 +544,31 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
 	  case 'A':			/* %A: weekday */
 	  case 'b':			/* %b: abbreviated month */
 	  case 'B':			/* %B: month */
+	    if ( posix )
+	    { const char *s;
+	      cal_ftm(ftm, HAS_STAMP|HAS_WYDAY);
+
+	      switch( c )
+	      { case 'a':
+		  s = abbred_weekday[ftm->tm.tm_wday];
+		  break;
+		case 'A':
+		  s = weekday[ftm->tm.tm_wday];
+		  break;
+		case 'b':
+		  s = abbred_month[ftm->tm.tm_mon];
+		  break;
+		case 'B':
+		  s = month[ftm->tm.tm_mon];
+		  break;
+		default:
+		  s = NULL;
+		  assert(0);
+	      }
+	      OUTSTR(s);
+	      break;
+	    }
+	    /*FALLTHROUGH*/
 	  case 'c':			/* %c: default representation */
 	  case 'p':			/* %p: AM/PM (locale) */
 	  case 'P':			/* %P: am/pm (locale) */
@@ -510,7 +577,6 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
 	  case_b:
 	  { char fmt[3];
 	    char buf[256];
-	    wchar_t wbuf[256];
 	    size_t n;
   
 	    fmt[0] = '%';
@@ -520,12 +586,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
 	    cal_ftm(ftm, HAS_STAMP|HAS_WYDAY);
 					/* conversion is not thread-safe under locale switch */
 	    n = strftime(buf, sizeof(buf), fmt, &ftm->tm);
-	    if ( (n = mbstowcs(wbuf, buf, sizeof(wbuf)/sizeof(wchar_t))) != (size_t)-1 )
-	    { wchar_t *p;
-
-	      for(p=wbuf; n-- > 0; p++)
-		Sputcode(*p, fd);
-	    }
+	    OUTSTRA(buf);
 	    break;
 	  }
 	  case 'C':			/* (year/100) as a 2-digit int */
@@ -687,9 +748,9 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
 	  }
 	  case 'Z':			/* Time-zone as name */
 	    if ( ftm->tzname )
-	    { OUTSTR(PL_atom_chars(ftm->tzname));
+	    { OUTATOM(ftm->tzname);
 	    } else
-	    { OUTSTR(tz_name(ftm->tm.tm_isdst));
+	    { OUTSTRA(tz_name(ftm->tm.tm_isdst));
 	    }
 	    break;
 	  case '+':
@@ -698,7 +759,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm)
 	      cal_ftm(ftm, HAS_WYDAY);
 	      asctime_r(&ftm->tm, buf);
 	      buf[24] = EOS;
-	      OUTSTR(buf);
+	      OUTSTRA(buf);
 	    }
 	    break;
 	  case '%':
@@ -725,8 +786,8 @@ Issues:
 	* Sub-second times
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static
-PRED_IMPL("format_time", 3, format_time, 0)
+static  foreign_t
+pl_format_time(term_t out, term_t format, term_t time, int posix)
 { struct taia taia;
   struct caltime ct;
   struct ftm tb;
@@ -737,12 +798,12 @@ PRED_IMPL("format_time", 3, format_time, 0)
   size_t fmtlen;
   redir_context ctx;
 
-  if ( !PL_get_wchars(A2, &fmtlen, &fmt,
+  if ( !PL_get_wchars(format, &fmtlen, &fmt,
 		      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
     fail;
 
   memset(&tb, 0, sizeof(tb));
-  if ( get_taia(A3, &taia, &tb.stamp) )
+  if ( get_taia(time, &taia, &tb.stamp) )
   { double ip;
 
     ut64 = taia.sec.x - TAI_UTC_OFFSET;
@@ -771,20 +832,38 @@ PRED_IMPL("format_time", 3, format_time, 0)
       tb.tzname     = ATOM_utc;
       tb.utcoff     = 0;
     }
-  } else if ( !get_ftm(A3, &tb) )
+  } else if ( !get_ftm(time, &tb) )
   { fail;
   }
 
-  if ( !setupOutputRedirect(A1, &ctx, FALSE) )
+  if ( !setupOutputRedirect(out, &ctx, FALSE) )
     fail;
-  if ( format_time(ctx.stream, fmt, &tb) )
+  if ( format_time(ctx.stream, fmt, &tb, posix) )
     return closeOutputRedirect(&ctx);	/* takes care of I/O errors */
 
   discardOutputRedirect(&ctx);
   fail;
 }
 
+static
+PRED_IMPL("format_time", 3, format_time3, 0)
+{ return pl_format_time(A1, A2, A3, FALSE);
+}
 
+static
+PRED_IMPL("format_time", 4, format_time4, 0)
+{ int posix = FALSE;
+  atom_t locale;
+
+  if ( !PL_get_atom_ex(A4, &locale) )
+    return FALSE;
+  if ( locale == ATOM_posix )
+    posix = TRUE;
+  else
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_locale, A4);
+
+  return pl_format_time(A1, A2, A3, posix);
+}
 
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
@@ -794,5 +873,6 @@ PRED_IMPL("format_time", 3, format_time, 0)
 BeginPredDefs(tai)
   PRED_DEF("stamp_date_time", 3, stamp_date_time, 0)
   PRED_DEF("date_time_stamp", 2, date_time_stamp, 0)
-  PRED_DEF("format_time",     3, format_time,     0)
+  PRED_DEF("format_time",     3, format_time3,    0)
+  PRED_DEF("format_time",     4, format_time4,    0)
 EndPredDefs
