@@ -1,11 +1,11 @@
-/*  $Id: pl-thread.c,v 1.189 2007/01/12 21:12:29 jan Exp $
+/*  $Id: pl-thread.c,v 1.198 2007/06/25 12:51:23 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -73,10 +73,12 @@ APPROACH
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #include <errno.h>
-#if defined(__linux__) && defined(HAVE_GETTID)
+#if defined(__linux__)
 #include <linux/unistd.h>
+#ifdef HAVE_GETTID_MACRO
 _syscall0(pid_t,gettid)
 #endif
+#endif
 
 #ifdef HAVE_SYS_SYSCALL_H
 #include <sys/syscall.h>
@@ -606,12 +608,14 @@ exitPrologThreads()
 #else
   	  if ( t->tid )
 	  { int rc;
+	    int oldstat = t->status;
 	    
+	    t->status = PL_THREAD_CANCELED;
 	    if ( (rc=pthread_cancel(t->tid)) == 0 )
-	    { t->status = PL_THREAD_CANCELED;
-	      canceled++;
+	    { canceled++;
 	    } else
-	    { Sdprintf("Failed to cancel thread %d: %s\n", i, ThError(rc));
+	    { t->status = oldstat;
+	      Sdprintf("Failed to cancel thread %d: %s\n", i, ThError(rc));
 	    }
 	  } else
 	  { DEBUG(1, Sdprintf("Destroying engine %d\n", i));
@@ -847,13 +851,17 @@ set_system_thread_id(PL_thread_info_t *info)
 { 
   info->tid = pthread_self();
   info->has_tid = TRUE;
-#ifdef HAVE_GETTID
+#ifdef HAVE_GETTID_SYSCALL
+  info->pid = syscall(__NR_gettid);
+#else
+#ifdef HAVE_GETTID_MACRO
   info->pid = gettid();
 #else
 #ifdef __WINDOWS__
   info->w32id = GetCurrentThreadId();
 #endif
 #endif
+#endif
 }
 
 
@@ -946,7 +954,7 @@ pl_thread_create(term_t goal, term_t id, term_t options)
   intptr_t stack = 0;
   int rc;
 
-  if ( !(PL_is_compound(goal) || PL_is_atom(goal)) )
+  if ( !PL_is_callable(goal) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_callable, goal);
 
   LOCK();
@@ -1259,6 +1267,7 @@ PRED_IMPL("thread_detach", 1, thread_detach, 0)
 word
 pl_current_thread(term_t id, term_t status, control_t h)
 { int current;
+  fid_t fid;
 
   switch(ForeignControl(h))
   { case FRG_FIRST_CALL:
@@ -1275,18 +1284,18 @@ pl_current_thread(term_t id, term_t status, control_t h)
     }
     case FRG_REDO:
       current = (int)ForeignContextInt(h);
-    redo:
-      for( ; current < MAX_THREADS; current++ )
-      { mark m;
 
-	if ( !threads[current].has_tid )
+    redo:
+      fid = PL_open_foreign_frame();
+      for( ; current < MAX_THREADS; current++ )
+      { if ( threads[current].status == PL_THREAD_UNUSED )
 	   continue;
 
-	Mark(m);
 	if ( unify_thread_id(id, &threads[current]) &&
 	     unify_thread_status(status, &threads[current]) )
 	  ForeignRedoInt(current+1);
-	Undo(m);
+
+	PL_rewind_foreign_frame(fid);
       }
       fail;
     case FRG_CUTTED:
@@ -1556,6 +1565,7 @@ executeThreadSignals(int sig)
     next = sg->next;
     PL_recorded(sg->goal, goal);
     PL_erase(sg->goal);
+    DEBUG(1, Sdprintf("[%d] Executing thread signal\n", PL_thread_self()));
     rval = callProlog(sg->module, goal, PL_Q_CATCH_EXCEPTION, &ex);
     freeHeap(sg, sizeof(*sg));
 
@@ -1881,9 +1891,7 @@ get_message(message_queue *queue, term_t msg)
   int isvar = PL_is_variable(msg) ? 1 : 0;
   word key = (isvar ? 0L : getIndexOfTerm(msg));
   int rval = TRUE;
-  mark m;
-
-  Mark(m);
+  fid_t fid = PL_open_foreign_frame();
 
   ctx.queue = queue;
   ctx.isvar = isvar;
@@ -1915,7 +1923,7 @@ get_message(message_queue *queue, term_t msg)
 	goto out;
       }
 
-      Undo(m);			/* reclaim term */
+      PL_rewind_foreign_frame(fid);
     }
 				/* linux docs say it may return EINTR */
 				/* does it re-lock in that case? */
@@ -1955,9 +1963,7 @@ peek_message(message_queue *queue, term_t msg)
 { thread_message *msgp;
   term_t tmp = PL_new_term_ref();
   word key = getIndexOfTerm(msg);
-  mark m;
-
-  Mark(m);
+  fid_t fid = PL_open_foreign_frame();
 
   simpleMutexLock(&queue->mutex);
   msgp = queue->head;
@@ -1972,7 +1978,7 @@ peek_message(message_queue *queue, term_t msg)
       succeed;
     }
 
-    Undo(m);
+    PL_rewind_foreign_frame(fid);
   }
      
   simpleMutexUnlock(&queue->mutex);
@@ -2713,7 +2719,7 @@ foreign_t
 pl_current_mutex(term_t mutex, term_t owner, term_t count, control_t h)
 { TableEnum e;
   Symbol s;
-  mark mrk;
+  fid_t fid;
 
   switch(ForeignControl(h))
   { case FRG_FIRST_CALL:
@@ -2741,7 +2747,7 @@ pl_current_mutex(term_t mutex, term_t owner, term_t count, control_t h)
       succeed;
   }
 
-  Mark(mrk);
+  fid = PL_open_foreign_frame();
   while ( (s = advanceTableEnum(e)) )
   { pl_mutex *m = s->value;
 
@@ -2750,7 +2756,7 @@ pl_current_mutex(term_t mutex, term_t owner, term_t count, control_t h)
 	 PL_unify_integer(count, m->count) )
     { ForeignRedoPtr(e);
     }
-    Undo(mrk);
+    PL_rewind_foreign_frame(fid);
   }
 
   freeTableEnum(e);
@@ -2896,6 +2902,7 @@ detach_engine(PL_engine_t e)
   info->w32id = 0;
 #endif
   memset(&info->tid, 0, sizeof(info->tid));
+  info->has_tid = FALSE;
 
   TLD_set(PL_ldata, NULL);
 }
@@ -3077,13 +3084,17 @@ ThreadCPUTime(PL_thread_info_t *info, int which)
 
 static double
 ThreadCPUTime(PL_thread_info_t *info, int which)
-{ if ( info->tid )
+{ if ( info->has_tid )
   { clockid_t clock_id;
     struct timespec ts;
+    int rc;
 
-    pthread_getcpuclockid(info->tid, &clock_id);
-    if (clock_gettime(clock_id, &ts) == 0)
-      return timespec_to_double(ts);
+    if ( (rc=pthread_getcpuclockid(info->tid, &clock_id)) == 0 )
+    { if (clock_gettime(clock_id, &ts) == 0)
+	return timespec_to_double(ts);
+    } else
+    { DEBUG(1, Sdprintf("Could not get thread time: %s\n", strerror(rc)));
+    }
   }
 
   return 0.0;
