@@ -1,4 +1,4 @@
-/*  $Id: pl-trace.c,v 1.114 2007/01/13 10:30:37 jan Exp $
+/*  $Id: pl-trace.c,v 1.116 2007/06/01 14:24:32 jan Exp $
 
     Part of SWI-Prolog
 
@@ -69,7 +69,7 @@ PL_put_frame(term_t t, LocalFrame fr)
 { if ( fr )
   { assert(fr >= lBase && fr < lTop);
 
-    PL_put_integer(t, (Word)fr - (Word)lBase);
+    PL_put_intptr(t, (Word)fr - (Word)lBase);
   } else
     PL_put_atom(t, ATOM_none);
 }
@@ -103,7 +103,7 @@ PL_put_choice(term_t t, Choice ch)
 { if ( ch )
   { assert(ch >= (Choice)lBase && ch < (Choice)lTop);
 
-    PL_put_integer(t, (Word)ch - (Word)lBase);
+    PL_put_intptr(t, (Word)ch - (Word)lBase);
   } else
     PL_put_atom(t, ATOM_none);
 }
@@ -785,7 +785,7 @@ writeFrameGoal(LocalFrame frame, Code PC, unsigned int flags)
     if ( true(def, FOREIGN) )
       PL_put_atom(pc, ATOM_foreign);
     else if ( PC && frame->clause )
-      PL_put_integer(pc, PC-frame->clause->clause->codes);
+      PL_put_intptr(pc, PC-frame->clause->clause->codes);
     else
       PL_put_nil(pc);
     
@@ -1635,6 +1635,8 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
 	  *argp = (needsRef(*a) ? makeRef(a) : *a);
       }
     }
+  } else if ( key == ATOM_predicate_indicator )
+  { unify_definition(result, fr->predicate, 0, GP_NAMEARITY);
   } else if ( key == ATOM_parent_goal )
   { Procedure proc;
     term_t head = PL_new_term_ref();
@@ -1672,7 +1674,7 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
       if ( pc < 0 )
 	trap_gdb();
 
-      PL_put_integer(result, pc);
+      PL_put_intptr(result, pc);
     } else
     { fail;
     }
