@@ -1,4 +1,4 @@
-/*  $Id: pl-wic.c,v 1.137 2007/01/13 10:08:41 jan Exp $
+/*  $Id: pl-wic.c,v 1.141 2007/07/09 13:21:18 jan Exp $
 
     Part of SWI-Prolog
 
@@ -173,8 +173,8 @@ between  16  and  32  bits  machines (arities on 16 bits machines are 16
 bits) as well as machines with different byte order.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define LOADVERSION 43			/* load all versions later >= X */
-#define VERSION 43			/* save version number */
+#define LOADVERSION 44			/* load all versions later >= X */
+#define VERSION 44			/* save version number */
 #define QLFMAGICNUM 0x716c7374		/* "qlst" on little-endian machine */
 
 #define XR_REF     0			/* reference to previous */
@@ -929,7 +929,7 @@ loadPredicateFlags(Definition def, IOSTREAM *fd, int skip ARG_LD)
   { int	flags = getInt(fd);
 
     if ( !skip )
-    { uintptr_t lflags = 0L;
+    { unsigned long lflags = 0L;
 
       if ( flags & PRED_SYSTEM )
 	lflags |= SYSTEM;
@@ -1411,8 +1411,6 @@ First byte:  bits 8&7  bits 1-6 (low order)
 		3      number of bytes following
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define PLMINLONG   ((intptr_t)(-1L<<(LONGBITSIZE-1)))
-
 static void
 putNum(int64_t n, IOSTREAM *fd)
 { int m;
@@ -2015,6 +2013,7 @@ writeSourceMarks(IOSTREAM *s ARG_LD)
     freeHeap(pm, sizeof(*pm));
     n++;
   }
+  source_mark_head = source_mark_tail = NULL;
   
   DEBUG(1, Sdprintf("Written %d marks\n", n));
   putLong(n, s);
