@@ -1,4 +1,4 @@
-/*  $Id: pl-write.c,v 1.95 2007/01/12 21:12:30 jan Exp $
+/*  $Id: pl-write.c,v 1.97 2007/02/20 11:21:42 jan Exp $
 
     Part of SWI-Prolog
 
@@ -669,6 +669,13 @@ writePrimitive(term_t t, write_options *options)
       case FP_PINF:
 	s = (true(options, PL_WRT_QUOTED) ? "'$Infinity'" : "Infinity");
         break;
+      case FP_NDENORM:			/* pos/neg denormalized non-zero */
+      case FP_PDENORM:
+      case FP_NNORM:			/* pos/neg normalized non-zero */
+      case FP_PNORM:
+      case FP_NZERO:			/* pos/neg zero */
+      case FP_PZERO:
+	break;
     }
 #else
 #ifdef HAVE__FPCLASS
