@@ -1,4 +1,4 @@
-/*  $Id: plld.c,v 1.78 2007/01/13 12:49:08 jan Exp $
+/*  $Id: plld.c,v 1.83 2007/07/05 12:22:32 jan Exp $
 
     Part of SWI-Prolog
 
@@ -204,6 +204,7 @@ static char *plsysinit;			/* -F file */
 static char *ctmp;			/* base executable */
 static char *pltmp;			/* base saved state */
 static char *out;			/* final output */
+static int  opt_o=FALSE;		/* -o out given */
 
 static int nostate = FALSE;		/* do not make a state */
 static int nolink = FALSE;		/* do not link */
@@ -727,6 +728,7 @@ parseOptions(int argc, char **argv)
     } else if ( streq(opt, "-o") ) 		/* -o out */
     { if ( argc > 1 )
       { out = argv[1];
+	opt_o = TRUE;
 	argc--, argv++;
       } else
 	usage();
@@ -878,7 +880,7 @@ fillDefaultOptions()
   if ( strcmp(LIB_PL_DEBUG,pllib) == 0 )
     ensureOption(&coptions, "/MDd");
   else ensureOption(&coptions, "/MD");
-  ensureOption(&coptions, "/DWIN32");
+  ensureOption(&coptions, "/D__WINDOWS__");
   ensureOption(&coptions, "/nologo");
   ensureOption(&ldoptions, "/nologo");
 #endif
@@ -903,7 +905,7 @@ fillDefaultOptions()
   }
 #if defined(__CYGWIN__)
   /* Tack correct suffix to out, otherwise cat in createOutput fails. */
-  if ( out )
+  if ( out && !nolink )
   { replaceExtension(out, shared || embed_shared ? "dll" : "exe", tmp);
     /* Dont call free(out), it could be an argv member. */
     out = strdup(tmp);
@@ -1086,13 +1088,17 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
   char *ext;
   arglist *args = copyArgList(options);
 
-  strcpy(ofile, cfile);
-  if ( (ext = (char *)file_name_extension(ofile)) )
-    strcpy(ext, EXT_OBJ);
+  if ( opt_o && nolink )
+  { strcpy(ofile, out);
+  } else
+  { strcpy(ofile, cfile);
+    if ( (ext = (char *)file_name_extension(ofile)) )
+      strcpy(ext, EXT_OBJ);
+  }
 
   prependArgList(args, "-c");
 #ifdef __WINDOWS__
-  appendArgList(args, "-DWIN32");
+  appendArgList(args, "-D__WINDOWS__");
   appendArgList(args, "-D_WINDOWS");
 #endif
   appendArgList(args, "-D__SWI_PROLOG__");
