@@ -1,4 +1,4 @@
-/*  $Id: test.pl,v 1.141 2007/01/14 21:02:58 jan Exp $
+/*  $Id: test.pl,v 1.150 2007/06/19 09:05:32 jan Exp $
 
     Part of SWI-Prolog
 
@@ -22,8 +22,12 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-:- set_prolog_flag(optimise, true).
+%:- set_prolog_flag(optimise, true).
 %:- set_prolog_flag(trace_gc, true).
+:- asserta(user:file_search_path(library, '../packages/plunit')).
+:- [library(plunit)].
+:- set_test_options([load(always), silent(true)]).
+
 :- use_module(library(lists)).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -61,27 +65,28 @@ available test sets. The public goals are:
 		 *******************************/
 
 :- dynamic
-	include_code/1.
+	user:include_code/1.
 
 term_expansion((:- if(G)), []) :-
-	(   catch(G, E, (print_message(error, E), fail))
-	->  asserta(include_code(true))
-	;   asserta(include_code(false))
+	prolog_load_context(module, M),
+	(   catch(M:G, E, (print_message(error, E), fail))
+	->  asserta(user:include_code(true))
+	;   asserta(user:include_code(false))
 	).
 term_expansion((:- else), []) :-
-	(   retract(include_code(X))
+	(   retract(user:include_code(X))
 	->  (   X == true
 	    ->  X2 = false 
 	    ;   X2 = true
 	    ),
-	    asserta(include_code(X2))
+	    asserta(user:include_code(X2))
 	;   throw(error(context_error(no_if), _))
 	).
 term_expansion((:- endif), []) :-
-	retract(include_code(_)), !.
+	retract(user:include_code(_)), !.
 
 term_expansion(_, []) :-
-	include_code(X), !,
+	user:include_code(X), !,
 	X == false.
 	    
 
@@ -249,6 +254,18 @@ cyclic(test-4) :-
 	X = f(a), acyclic_term(a(X, X)).
 
 
+		 /*******************************
+		 *	    UNIFIABLE		*
+		 *******************************/
+
+unifiable(unifiable-1) :-
+	unifiable(X, 1, S),
+	S == [X=1].
+unifiable(unifiable-2) :-
+	unifiable(a(X,X), a(Y,Z), S),
+	S == [Z=X, Y=X].
+
+
 		 /*******************************
 		 *      INTEGER ARITHMETIC	*
 		 *******************************/
@@ -562,6 +579,12 @@ gmp(pow-5) :-
 	A is -100**5,
 	A < 0,
 	abs(A) =:= 100**5.
+gmp(powm-1) :-
+	A is powm(2,4,2),
+	A == 0.
+gmp(powm-2) :-
+	A is powm(2,4,3),
+	A == 1.
 gmp(integer-1) :-			% rounding integer conversion
 	0 =:= integer(1 rdiv 3),
 	1 =:= integer(2 rdiv 3),
@@ -728,7 +751,7 @@ meta(call-2) :-
 meta(call-3) :-
 	\+ call((between(0,3,X), !, X = 2)).
 meta(call-4) :-
-	length(X, 100000), call((is_list(X) -> true ; fail)).
+	length(X, 50000), call((is_list(X) -> true ; fail)).
 meta(call-5) :-
 	call((X=a;X=b)), X = b.
 meta(call-6) :-
@@ -877,6 +900,8 @@ type_test(type-1) :-
 	var(_), X = Y, var(X), Y = a, nonvar(X).
 type_test(type-2) :-
 	atom(hello), \+ atom(10), \+ atom("hello").
+type_test(type-3) :-
+	callable(atom), callable(term(a)), \+ callable(_Var).
 
 
 		 /*******************************
@@ -1482,9 +1507,13 @@ local_overflow :-
 	choice,
 	local_overflow.
 
-global_overflow(X) :-
+global_overflow(X) :-			% Causes gracefully signalled overflow
 	global_overflow(s(X)).
 
+string_overflow([H|T]) :-		% Causes PL_throw() overflow
+	format(string(H), '~txx~1000000|', []),
+	string_overflow(T).
+		
 
 resource(stack-1) :-
 	catch(local_overflow, E, true),
@@ -1492,6 +1521,16 @@ resource(stack-1) :-
 resource(stack-2) :-			% VERY slow with -DO_SECURE
 	catch(global_overflow(0), E, true),
 	E = error(resource_error(stack), global).
+resource(stack-3) :-
+	catch(string_overflow(_), E1, true),
+	E1 = error(resource_error(stack), global),
+	catch(string_overflow(_), E2, true),
+	E2 = error(resource_error(stack), global).
+resource(stack-4) :-
+	catch(string_overflow(_), E1, true),
+	E1 = error(resource_error(stack), global),
+	catch(global_overflow(_), E2, true),
+	E2 = error(resource_error(stack), global).
 
 
 		 /*******************************
@@ -1937,7 +1976,7 @@ io(tell-2) :-
 
 popen(pwd-1) :-
 	(   current_prolog_flag(windows, true)
-	->  Command = cd
+	->  Command = 'cmd /c cd'
 	;   Command = pwd
 	),
 	open(pipe(Command), read, Fd),
@@ -1963,9 +2002,13 @@ popen(cat-1) :-
 	    A == Text
 	).
 popen(cat-2) :-
+	(   current_prolog_flag(windows, true)
+	->  Cmd = 'cmd /c rem true'
+	;   Cmd = true
+	),
 	absolute_file_name(swi('library/MANUAL'), Manual),
 	open(Manual, read, Fd),
-	open(pipe(true), write, Pipe),
+	open(pipe(Cmd), write, Pipe),
 	catch(copy_stream_data(Fd, Pipe),
 	      E,
 	      true),
@@ -2266,6 +2309,14 @@ thread(at_exit-1) :-
         thread_create(at_exit_work, Id, []),
         thread_join(Id, exited(true)),
 	retract(at_exit_called).
+thread(status-1) :-
+	thread_create(true, Id, []),
+	between(0, 10, _),
+	sleep(0.01),
+	current_thread(Id2, Status),
+	Id2 == Id,
+	Status == true, !,
+	thread_join(Id2, _).
 	
 
 		 /*******************************
@@ -2374,6 +2425,7 @@ testset(write_test).
 testset(format_test).
 testset(unify).
 testset(occurs_check).
+testset(unifiable).
 testset(arithmetic).
 testset(arithmetic_functions).
 testset(floattest).
