@@ -1,4 +1,4 @@
-/*  $Id: console.c,v 1.45 2007/01/10 12:39:51 jan Exp $
+/*  $Id: console.c,v 1.47 2007/02/23 12:58:39 jan Exp $
 
     Part of SWI-Prolog
 
@@ -71,7 +71,7 @@ Menus
 The current console provides a menu that can be extended from Prolog.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if defined(_DEBUG) && !defined(WIN64)
+#ifdef O_DEBUG_HEAP
 static void initHeapDebug(void);
 #include <crtdbg.h>
 #else
@@ -3624,6 +3624,9 @@ initHeapDebug(void)
   }
 }
 
+#endif /*initHeapDebug*/
+
+#ifdef _DEBUG
 
 static void
 Dprintf(const TCHAR *fmt, ...)
@@ -3653,4 +3656,4 @@ Dprint_lines(RlcData b, int from, int to)
   }
 }
 
-#endif /*initHeapDebug*/
+#endif
